﻿* Encoding: UTF-8.
*MATRIX - END MATRIX functions.
*SPSS macros by Kirill Orlov
*kior@akado.ru, ttnphns@gmail.com
*https://www.spsstools.net/en/KO-spssmacros
*All rights reserved.

********************* FUNCTIONS FOR MATRIX - END MATRIX *********************.

*1) Arguments in functions are separated by character %, not comma.

*2) Functions should be written and run like commands, starting a line – not as “right hand side”,
*and the name of the result to return should be written among the arguments as the last one of them.
*I.e., correct: !function(argument % result); not correct: result = !function(argument).
*Some functions return more than one result.

*3) Any argument except the result may be expression, not just name or value.
*For example, !function(1.5*name(2,4)+3 % result) is allowed. The expression however cannot contain
*the functions (the macros) described in this documents.

*4) The result name may coinside with any of the arguments’ name. I.e., it is allowed,
*for example: !function(name1 % name2 % name1). In this case the result will be name1,
*and the object which formerly carried this name will be replaced.

*5) Do not use in your syntax within MATRIX - END MATRIX names starting by character @ because
*this character is reserved for the functions here, and so using it as the 1st character
*is fraught with loss of data.

*6) Functions do not check and do not issue special notifications whether your input data meet
*the requirements of the given function (so was made for speed performance). Therefore, to minimize
*chance of error or erroneous result do attentively read the requirements of a specific function.

*7) Functions may be put inside LOOP - END LOOP and DO IF - END IF.

*8) To disable a function in syntax use not * but /*, for example:
/*func(arg%result).
*or
/*!func(arg%result).*/.

*9) These matrix functions were designed for numeric input data. Some of the functions, not tied
*with numeric analysis, might work with text (string) data (such as counting of values, recoding,
*restructuring), but it is not guaranteed since the code was written in expectation of numbers.
*So, try and check with string data yourself.

*10) Many of these functions require to set in advance, prior the first MATRIX command in a session,
*the maximum number of cycles to a sufficiently big value, in order to cope with potentially large data.
*I recommend to do beforehand the setting on big number of cycles just always when you are about
*to work with these functions – to save from worrying about that in the sequel. For example:
*set mxloops 100000000.  /*or 1E8
*matrix.
*etc.

*11) Variance/covariance and correlation, when computed within the functions, not always use one
*and the same calculation formula. For example, /*!KO_variance */ or /*!KO_corr*/ use one formula
*(giving more numerically stable result) while other functions may calculate by another formula
*(giving less numerically stable result).

*----------------------------------------------------------------------------------------------.


******** CUMULATIVE SUMS IN COLUMNS ********.
*/*!KO_ccum(mat%dir%name)*/*.
*Version 1.
*Takes matrix or column MAT with number of rows at least 2 and returns matrix NAME of the same size;
*elements in each column of NAME are cumulative sums in that column by the direction of summation
*downward (if DIR positive) or upward (if DIR nonpositive).
*DIR - scalar.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_ccum(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @dir= !3.
do if @dir.
-loop @i= 2 to nrow(!4).
- comp !4(@i,:)= !4(@i-1,:)+!4(@i,:).
-end loop.
else.
-loop @i= nrow(!4)-1 to 1 by -1.
- comp !4(@i,:)= !4(@i+1,:)+!4(@i,:).
-end loop.
end if.
release @dir,@i.
!enddefine.


******** PRODUCT OF ELEMENTS OF MATRIX ********.
*/*!KO_prod(mat%name)*/*.
*Version 2.
*Computes product of elements of MAT.
*Returns scalar NAME.

define !KO_prod(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
do if not all(@mat).
-comp !3= 0.
else.
-comp !3= exp(msum(ln(abs(@mat)))).
-do if mod(msum(@mat<0),2)).
- comp !3= -!3.
-end if.
end if.
release @mat.
!enddefine.


******** SUM OF ABOVE-DIAGONAL OR BELOW-DIAGONAL ELEMENTS OF MATRIX ********.
*/*!KO_trsum(mat%tr%name)*/*.
*Version 1.
*Sums up elements of the upper or the lower triangle of MAT.
*Returns scalar NAME.
*Matrix MAT needs not be square, but must be at least size 2x2.
*TR - scalar; if positive, elements above the main diagonal are summed;
*if nonpositive, elements below the main diagonal are summed.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_trsum(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @tr= !3.
do if @tr.
-comp @mat= t(@mat).
end if.
comp !4= 0.
loop @i= 2 to nrow(@mat).
-comp !4= !4+rsum(@mat(@i,1:rmin({@i-1,ncol(@mat)}))).
end loop.
release @mat,@tr,@i.
!enddefine.


******** CHECK TRIANGULAR INEQUALITY IN DISSIMILARITY MATRIX ********.
*/*!KO_trineq(dis%ineq%stop%name1%name2)*/*.
*Version 1.
*Takes square matrix of distances (dissimilarities) DIS and verifies if there are distances
*violating "axiom of triangular inequality". If there are such, distances in matrix are not metric.
*The function checks only above-diagonal elements of DIS; if you want to check below-diagonal
*elements of DIS - transpose the matrix.
*DIS - square, with nonnegative values, sized at least 3.
*INEQ - digit 0 or 1 (not name or expression; optionally may cover the digit by quotes or apostrophes).
*If 0, slack triangular inequality will be verified: Dab <= Dac+Dbc (where a b c are objects, i.e.
*matrix rows/columns). If 1, strict triangular inequality will be verified: Dab < Dac+Dbc (Dab = Dac+Dbc
*doesn't count as violation).
*STOP - digit 0 or 1 (not name or expression; optionally may cover the digit by quotes or apostrophes).
*If 0, checked will be all above-diagonal elements: the function works till the end. If 1, the function
*breaks upon the first encounter of violation of triangular inequality, if there are any.
*Results:
*NAME1 - row vector of length 2: first value is the number of triplets (combination of elements by 3)
*checked above the matrix diagonal; second value - number of triplets of them with triangle inequality
*violated. (With STOP=1 this value is always =1.)
*NAME2 - row vector of length 4; it contains maximal found magnitude of triangle inequality violation
*(1st value in the vector) and the three objects' indices (matrix rows/columns) making that triplet.
*If no violations were found, NAME2 contains all -1.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_trineq(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @d= !2.
!let !ie= !unquote(!3)
!let !stop= !unquote(!4)
!if (!stop) !then !let !ifstop= 'if @stop' !ifend
comp !5= 0.
comp !6= make(1,4,-1).
comp @stop= 0.
comp @cnt= 0.
loop @x= 1 to ncol(@d)-2.
-loop @y= @x+1 to ncol(@d)-1.
- comp @dxy= @d(@x,@y).
- loop @z= @y+1 to ncol(@d).
-  comp @cnt= @cnt+1.
-  comp @dxz= @d(@x,@z).
-  comp @dyz= @d(@y,@z).
-  comp @viol= rmax({@dxy-@dxz-@dyz,@dxz-@dyz-@dxy,@dyz-@dxy-@dxz}).
-  do if @viol and @viol<1E-14.
-   comp @viol= 0.
-  end if.
-  do if @viol !if (!ie) !then >=0 !ifend.
-   comp !5= !5+1.
-   do if @viol>!6(1).
-    comp !6= {@viol,@x,@y,@z}.
-   end if.
    !if (!stop) !then
-    comp @stop= 1.
    !ifend
-  end if.
- end loop !ifstop.
-end loop !ifstop.
end loop !ifstop.
comp !5= {@cnt,!5}.
release @d,@stop,@cnt,@viol,@x,@y,@z,@dxy,@dxz,@dyz.
!enddefine.


******** LOWER INCOMLETE GAMMA FUNCTION ********.
*/*!KO_ligamma(arg%ubound%name)*/*.
*Version 1.
*Returns value of lower incomplete gamma function for argument ARG and upper bound
*of integration UBOUND.
*ARG – positive number or numbers: scalar, vector, or matrix.
*UBOUND - positive number or numbers: matrix sized as ARG.
*Returns matrix NAME sized as ARG.
*This function requires prior setting of limit for number of cycles to 100 (or more)
*by command SET MXLOOPS.

define !KO_ligamma(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @arg= !2.
comp @ubound= !3.
comp @den= 1.
comp !4= 1.
loop @it= 1 to 100.
-comp @den= @den&*(@arg+@it).
-comp !4= !4+@ubound&**@it/@den.
end loop.
comp !4= @ubound&**@arg/exp(@ubound)/@arg&*!4.
release @arg,@ubound,@den,@it.
!enddefine.


******** QR DECOMPOSITION BY HAUSHOLDER REFLECTIONS METHOD ********.
*/*!KO_qrdc(mat%name1%name2)*/*.
*Version 1.
*Does QR decomposition of matrix MAT into upper triangular matrix R (NAME2) and orthonormal
*matrix Q (NAME1) by Housholder reflections method. Q*R=MAT.
*MAT - matrix with number of columns not less than the number of rows.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*In-built function GSCH (Gram-Schmidt orthonormalization) is also a QR decomposition, using another
*method and giving different result.

define !KO_qrdc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @n= nrow(!4).
comp !3= ident(@n).
loop @i= 1 to @n.
-comp @ind= @i:@n.
-comp @v= !4(@i:@n,@i).
-comp @v(1)= @v(1)+((@v(1)>=0)-(@v(1)<0))*sqrt(mssq(@v)).
-comp @h= mssq(@v).
-comp @h= ident(nrow(@v))-2*sscp(t(@v))/(@h+(@h=0)).
-comp !4(@ind,@i:ncol(!4))= @h*!4(@ind,@i:ncol(!4)).
-comp !3(:,@ind)= !3(:,@ind)*@h.
end loop.
release @n,@i,@ind,@v,@h.
!enddefine.


******** HESSENBERG MATRIX / TRIDIAGONAL MATRIX ********.
*/*!KO_hess(mat%name1%name2)*/*.
*Version 1.
*Turns square matrix MAT (of size not less than 3) into upper Hessenberg matrix NAME2.
*That matrix is upper triangular with the upper sub-diagonal filled also, and its eigenvalues
*are the same as of MAT. There returned is also an orthonormal matrix NAME1, such
*that NAME1*MAT*t(NAME1)=NAME2 and t(NAME1)*NAME2*NAME1=MAT.
*If MAT is a symmetric matrix, then NAME2 comes out tridiagonal.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_hess(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @n= ncol(!4).
comp !3= ident(@n,@n).
loop @i= 1 to @n-2.
-comp @ind= (@i+1):@n.
-comp @v= !4(@ind,@i).
-comp @v1= ((@v(1)<0)-(@v(1)>=0))*sqrt(mssq(@v)).
-comp @u1= sqrt((@v1-@v(1))/(2*@v1+(@v1=0))).
-comp @u= -@v/(2*@u1*@v1+(@v1=0)).
-comp @u(1)= @u1.
-comp @h= ident(nrow(@u))-2*sscp(t(@u)).
-comp !4(@ind,@i:@n)= @h*!4(@ind,@i:@n).
-comp !4(:,@ind)= !4(:,@ind)*@h.
-comp !3(@ind,:)= @h*!3(@ind,:).
end loop.
release @n,@i,@ind,@v,@v1,@u,@u1,@h.
!enddefine.


******** EIGENDECOMPOSITION OF 2x2 MATRIX ********.
*/*!KO_eig2x2(mat%name1%name2)*/*.
*Version 1.
*Returns (real) eigenvalues (NAME2) and eigenvectors (columns of NAME1) of a square
*matrix MAT sized 2x2. MAT may be symmetric or asymmetric; in the latter case
*the eigenvectors may be nonorthogonal (use function /*!KO_schur2x2*/ to obtain
*orthogonal eigenvectors).
*Restoration of the matrix: MAT = NAME1*mdiag(NAME2)*inv(NAME1).
*If eigenvalues of MAT are not real, NAME1 and NAME2 are returned as zero scalars.

define !KO_eig2x2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @t= @mat(1,1)+@mat(2,2).
comp @d= @mat(1,1)*@mat(2,2)-@mat(1,2)*@mat(2,1).
comp @r= @t**2-4*@d.
do if @r>=0.
-comp !4= make(2,1,0).
-comp @r= sqrt(@r).
-comp !4(1)= (@t+@r)/2.
-comp !4(2)= (@t-@r)/2.
-comp !3= make(2,2,1).
-do if @mat(1,2)=0 and @mat(2,1)=0.
- comp !3= ident(2).
- do if @mat(1,1)<@mat(2,2).
-  comp !3= !3(:,2:1).
- end if.
-else if @mat(1,2)=0.
- comp !3(1,1)= (!4(1)-@mat(2,2))/@mat(2,1).
- comp !3(1,2)= (!4(2)-@mat(2,2))/@mat(2,1).
-else.
- comp !3(2,1)= (!4(1)-@mat(1,1))/@mat(1,2).
- comp !3(2,2)= (!4(2)-@mat(1,1))/@mat(1,2).
-end if.
-comp !3= !3/sqrt(cssq(!3))({1,1},:).
else.
-comp !4= 0.
-comp !3= 0.
end if.
release @mat,@t,@d,@r.
!enddefine.


******** EIGENDECOMPOSITION OF 2x2 MATRIX (SCHUR VARIANT) ********.
*/*!KO_schur2x2(mat%name1%name2)*/*.
*Version 1.
*Returns (real) eigenvalues and eigenvectors of a square matrix MAT sized 2x2.
*Eigenvalues are placed on the diagonal of the upper triangular matrix NAME2, and eigenvectors
*are the columns of NAME1. These are orthogonal eigenvectors and NAME1 is a unitary matrix.
*Restoration of the matrix: MAT = NAME1*NAME2*t(NAME1).
*If MAT is symmetric, then NAME2 is a diagonal matrix.
*If eigenvalues of MAT are not real, NAME1 and NAME2 are returned as zero scalars.
*For Schur decomposition of a matrix of greater size use /*!KO_qreig*/.

define !KO_schur2x2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
do if @mat(2,1)=0.
-comp !4= @mat.
-comp !3= ident(2).
else.
-comp @t= @mat(1,1)+@mat(2,2).
-comp @d= @mat(1,1)*@mat(2,2)-@mat(1,2)*@mat(2,1).
-comp @r= @t**2-4*@d.
-do if @r>=0.
- comp !4= (@t+sqrt(@r))/2.
- comp !3= make(2,1,1).
- do if @mat(1,2)=0.
-  comp !3(1,1)= (!4-@mat(2,2))/@mat(2,1).
- else.
-  comp !3(2,1)= (!4-@mat(1,1))/@mat(1,2).
- end if.
- comp !3= !3/sqrt(cssq(!3)).
- comp !3= {!3,{-!3(2);!3(1)}}.
- comp !4= t(!3)*@mat*!3.
-else.
- comp !4= 0.
- comp !3= 0.
-end if.
-release @t,@d,@r.
end if.
release @mat.
!enddefine.


******** EIGENDECOMPOSITION (SCHUR VARIANT) ********.
*/*!KO_qreig(mat%needvec%conv%name1%name2%name3)*/*.
*Version 1.
*Does eigendecomposition of a square matrix MAT (symmetric or asymmetric).
*Algorithm used is Implicit QR iterations on recursively splitted Hessenberg matrix with
*double (Francis) shifts. The function returns upper triangular matrix NAME2 (with the eigenvalues
*on its diagonal) and unitary matrix NAME1, which columns are the mutually orthogonal eigenvectors.
*The order of eigenvectors corresponds to the order of eigenvalues but the eigenvalues are not
*necessarily sorted by magnitude. If MAT is symmetric NAME2 is diagonal.
*Restoration of the matrix: MAT = NAME1*NAME2*t(NAME1).
*Argument NEEDVEC - digit (not name or expression) 1 or 0 (the digit may optionally be in quotes or
*apostrophes). If "1", the function will return eigenvalues NAME2 and eigenvectors NAME1. If "0",
*the function will return only eigenvalues NAME2, not spending time on computation of the
*eigenvectors (NAME1 will come out as scalar 1).
*Argument CONV - scalar; this is the being set precision of convergence at computation of
*eigenvalues; recommended value: from 1E-8 to 1E-12.
*NAME3 - this output scalar is the maximal required number of iterations to compute a single
*eigenvalue.
*Unconvergence/fail. Asymmetric matrices MAT sometimes have complex (not real) eigenvalues tied
*in pairs. If some subdiagonal element in NAME2 appeared nonzero while NAME3 is not above 100
*(the limiting number of iterations in the function), a pair of complex eigenvalues is evident.
*The real parts of these two values are found above and to the right from the unconverged-to-zero
*subdiagonal element, however they are not necessarily calculated precisely. If a subdiagonal
*element in NAME2 appeared nonzero but NAME3=101 (i.e. all iterations exhausted), a fail of the
*algorithm is evident - this is relatively rare case, associated with a special defectness of
*the given MAT.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*If MAT is size 2x2, function /*!KO_schur2x2*/ does the same decomposition as this function.
*If MAT is symmetric, the in-built function CALL EIGEN does the same decomposition as this function.

define !KO_qreig(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
             /!pos= !charend('%') /!pos= !charend(')'))
comp !6= !2.
!let !q= !unquote(!3)
comp @conv= !4.
comp @n= ncol(!6).
!if (!q) !then
-comp !5= ident(@n).
!else
-comp !5= 1.
!ifend
do if @n>2.
-loop @i= 1 to @n-2.
- comp @ind= (@i+1):@n.
- comp @v= !6(@ind,@i).
- comp @v1= ((@v(1)<0)-(@v(1)>=0))*sqrt(mssq(@v)).
- comp @u1= sqrt((@v1-@v(1))/(2*@v1+(@v1=0))).
- comp @u= -@v/(2*@u1*@v1+(@v1=0)).
- comp @u(1)= @u1.
- comp @h= ident(nrow(@u))-2*sscp(t(@u)).
- comp !6(@ind,@i:@n)= @h*!6(@ind,@i:@n).
- comp !6(:,@ind)= !6(:,@ind)*@h.
  !if (!q) !then
-  comp !5(@ind,:)= @h*!5(@ind,:).
  !ifend
-end loop.
-release @v,@v1,@u,@u1,@h.
else.
-comp @v= 0.
-comp @l= 0.
end if.
!if (!q) !then
-comp !5= t(!5).
!ifend
comp @stv= make(1,@n,0).
comp @env= make(1,@n,0).
comp @stv(1)= 1.
comp @env(1)= @n.
comp !7= 1.
comp @wr= 1.
loop @re= 1 to @n if @stv(@re).
-comp @st= @stv(@re).
-comp @en= @env(@re).
-loop @m= 1 to @n if @st<@en.
- do if @en-@st=1.
-  do if abs(!6(@en,@st))<=@conv*rmax(abs({!6(@st,@st),!6(@en,@en)})).
-   comp !6(@en,@st)= 0.
-  else.
-   comp @l= !6(@st,@st)+!6(@en,@en).
-   comp @v= !6(@st,@st)*!6(@en,@en)-!6(@st,@en)*!6(@en,@st).
-   comp @v= @l**2-4*@v.
-   do if @v>=0.
-    comp @l= (@l+sqrt(@v))/2.
-    comp @v= make(2,1,1).
-    do if !6(@st,@en)=0.
-     comp @v(1,1)= (@l-!6(@en,@en))/!6(@en,@st).
-    else.
-     comp @v(2,1)= (@l-!6(@st,@st))/!6(@st,@en).
-    end if.
-    comp @v= @v/sqrt(mssq(@v)).
-    comp @v= {@v,{-@v(2);@v(1)}}.
-    comp !6(@st:@en,:)= t(@v)*!6(@st:@en,:).
-    comp !6(:,@st:@en)= !6(:,@st:@en)*@v.
     !if (!q) !then
-     comp !5(:,@st:@en)= !5(:,@st:@en)*@v.
     !ifend
-   end if.
-  end if.
-  break.
- else.
-  loop @it= 1 to 100.
-   comp @en1= @en-1.
-   comp @st1= @st+1.
-   loop @i= @st to @en-2.
-    comp @ind= @i:(@i+2).
-    do if @i=@st.
-     comp @k1= !6(@en1,@en1)+!6(@en,@en).
-     comp @k2= !6(@en1,@en1)*!6(@en,@en)-!6(@en1,@en)*!6(@en,@en1).
-     comp @v= {!6(@st,@st)**2-!6(@st,@st)*@k1+@k2+!6(@st,@st1)*!6(@st1,@st);!6(@st1,@st)*(!6(@st,@st)+!6(@st1,@st1)-@k1);!6(@st1,@st)*!6(@st+2,@st1)}.
-    else.
-     comp @v= !6(@ind,@i-1).
-    end if.
-    comp @s= ((@v(1)>=0)-(@v(1)<0))*sqrt(mssq(@v)).
-    do if @s<>0.
-     comp @v(1)= @v(1)+@s.
-     comp @v= @v*(t(@v)/(@v(1)*@s)).
-     comp @v= ident(3)-@v.
-     comp !6(@ind,:)= @v*!6(@ind,:).
-     comp !6(:,@ind)= !6(:,@ind)*@v.
      !if (!q) !then
-      comp !5(:,@ind)= !5(:,@ind)*@v.
      !ifend
-    end if.
-   end loop.
-   comp @b= !6(@en,@en-2).
-   comp @r= !6(@en1,@en-2).
-   comp @h= sqrt(@r**2+@b**2).
-   do if @h.
-    comp @c= abs(@r)/@h.
-    comp @s= ((@r>=0)-(@r<0))*@b/@h.
-    comp @v= {@c,-@s;@s,@c}.
-    comp !6(@en1:@en,:)= t(@v)*!6(@en1:@en,:).
-    comp !6(:,@en1:@en)= !6(:,@en1:@en)*@v.
     !if (!q) !then
-     comp !5(:,@en1:@en)= !5(:,@en1:@en)*@v.
     !ifend
-   end if.
-   comp @dg= abs(diag(!6)).
-   comp @az= abs(diag(!6(@st1:@en,@st:@en1)))<=@conv*rmax({@dg(@st1:@en),@dg(@st:@en1)}).
-   do if any(@az).
-    comp @p= cmax(@az&*t(@st:@en1)).
-    comp !6(@p+1,@p)= 0.
-    break.
-   end if.
-  end loop.
-  comp @wr= @wr+1.
-  comp @stv(@wr)= @st.
-  comp @env(@wr)= @p.
-  comp @st= @p+1.
-  comp !7= rmax({!7,@it}).
- end if.
-end loop.
end loop.
do if @n>2.
-release @i,@ind,@h,@it,@en1,@st1,@k1,@k2,@s,@b,@r,@c,@dg,@az,@p.
end if.
release @conv,@n,@v,@l,@stv,@env,@wr,@re,@st,@en,@m.
!enddefine.


******** VECTOR OF ARITHMETIC MEANS ********.
*/*!KO_mean(data%name)*/*.
*Version 1.
*Computes means in columns of DATA.
*Returns row vector NAME.

define !KO_mean(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= csum(!3)/nrow(!3).
!enddefine.


******** VECTOR OF VARIANCES ********.
*/*!KO_variance(data%name)*/*.
*Version 1.
*Computes variances in columns of DATA.
*Returns row vector NAME.

define !KO_variance(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
comp !3= cssq(!3)/(nrow(!3)-1).
!enddefine.


******** VECTOR OF VARIANCES (DF=N) ********.
*/*!KO_variance2(data%name)*/*.
*Version 1.
*Computes biased (computed on "df=n", not "df=n-1") variances in columns of DATA.
*Returns row vector NAME.

define !KO_variance2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
comp !3= cssq(!3)/nrow(!3).
!enddefine.


******** VECTOR OF SKEWNESSES ********.
*/*!KO_skewness(data%name)*/*.
*Version 1.
*Computes skewnesses in columns of DATA.
*Returns row vector NAME.

define !KO_skewness(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp @n= nrow(!3).
comp !3= !3-(csum(!3)/@n)(make(1,@n,1),:).
comp @m2= cssq(!3).
comp @sd= sqrt(@m2/(@n-1)).
comp @m3= csum(!3&*!3&*!3).
comp !3= @n*@m3/((@n-1)*(@n-2)*@sd&**3).
release @n,@m2,@m3,@sd.
!enddefine.


******** VECTOR OF KURTOSES ********.
*/*!KO_kurtosis(data%name)*/*.
*Version 1.
*Computes kurtoses (tailedness) in columns of DATA. DATA must have at least three rows.
*Returns row vector NAME.

define !KO_kurtosis(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp @n= nrow(!3).
comp !3= !3-(csum(!3)/@n)(make(1,@n,1),:).
comp @m2= cssq(!3).
comp @sd= sqrt(@m2/(@n-1)).
comp @m4= csum(!3&*!3&*!3&*!3).
comp !3= (@n*(@n+1)*@m4-3*@m2&**2*(@n-1))/((@n-1)*(@n-2)*(@n-3)*@sd&**4).
release @n,@m2,@m4,@sd.
!enddefine.


******** VECTOR OF MEDIANS ********.
*/*!KO_median(data%name)*/*.
*Version 1.
*Computes medians in columns of DATA.
*Returns row vector NAME.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_median(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
loop @i= 1 to ncol(@data).
-comp @data(grade(@data(:,@i)),@i)= @data(:,@i).
end loop.
do if mod(nrow(@data),2).
-comp !3= @data(nrow(@data)/2+.5,:).
else.
-comp !3= (@data(nrow(@data)/2,:)+@data(nrow(@data)/2+1,:))/2.
end if.
release @data,@i.
!enddefine.


******** PERCENTILES ********.
*/*!KO_ptile(data%sort%p%method%name)*/*.
*Version 1.
*Computes requested percentiles in columns of DATA.
*Returns matrix NAME with rows corresponding to the percentile levels.
*SORT - digit (not name or expression) 1 or 0 (optionally in quotes or apostrophes). Use 0 only
*in the special case when values in each column of DATA are already sorted ascendingly; this
*option allows to spare time doing the sorting.
*P - scalar or vector (row or column) containing percents (percent levels, percentile ranks) for
*which you need to determine percentile values in the variables (each column of DATA).
*Values in P must be greater than 0 and lesser than 100. They need not to go in ascending order.
*METHOD - method to compute percentiles, capitalized keyword (optionally in quotes or apostrophes):
*"WAVER" - Weighted Average Definition 2
*"HAVER" - Weighted Average Definition 1
*"HAVER2" - Weighted Average Definition 1
*"EMPIR" - Empirical Distribution Function
*"AEMPIR" - Empirical Distribution Function with Averaging
*"ROUND" - Observation closest aka Rounded
*These are the methods found in SPSS procedure EXAMINE (Explore). HAVER and HAVER2 are the same,
*but HAVER2 does not compute percentiles that are "too high" for the input data (replacing them
*in NAME with arbitrary number -999), like SPSS procedures Explore and Frequencies do for the
*default method "HAVERAGE". HAVER puts, instead of "-999", the maximal value in the variable.
*Most popular or recommended methods are HAVER/HAVER2 and AEMPIR. 50th percentile of these methods
*coincides with the median after function /*!KO_median*/.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_ptile(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !meth= !unquote(!5) !let !sort= !unquote(!3)
comp @data= !2.
comp @p= !4/100.
do if nrow(@p)=1.
-comp @p= t(@p).
end if.
comp !6= make(nrow(@p),ncol(@data),0).
loop @v= 1 to ncol(@data).
-comp @var= @data(:,@v).
 !if (!sort) !then
- comp @var(grade(@var))= @var.
 !ifend
 !if (!meth='ROUND') !then
- comp @t= nrow(@var)*@p+.5.
- comp @i= trunc(@t).
- comp @i= @i+(not @i).
- comp @f= 0.
- comp !6(:,@v)= @var(@i).
 !else !if (!meth='EMPIR' !or !meth='AEMPIR') !then
- comp @t= nrow(@var)*@p.
- comp @i= trunc(@t).
- comp @f= @t<>@i.
  !if (!meth='EMPIR') !then
-  comp !6(:,@v)= @var(@i+@f).
  !else
-  comp !6(:,@v)= (@var(@i+@f)+@var(@i+@f+(not @f)))/2.
  !ifend
 !else
  !if (!meth<>'WAVER') !then !let !p_= '+1' !let !n_= '-(@i=nrow(@var))' !ifend
- comp @t= (nrow(@var) !p_)*@p.
- comp @i= trunc(@t).
- comp @f= mod(@t,1).
- comp !6(:,@v)= (1-@f)&*@var(@i +(@i=0) )+@f&*@var(@i+1 !n_).
  !if (!meth='HAVER2') !then
-  comp @f= @i=nrow(@var) and @f.
-  comp !6(:,@v)= !6(:,@v)&*(not @f)-999&*@f.
  !ifend
 !ifend !ifend
end loop.
release @data,@var,@p,@v,@t,@i,@f.
!enddefine.


******** MINIMUM, MAXIMUM, MEAN, AND VARIANCE, BY GROUPS ********.
*/*!KO_gdescr(col%bin%name1%name2%name3%name4)*/*.
*Version 2.
*Takes column COL (one analyzed variable) and one or more binary (1 vs 0)
*variables constituting columns of matrix BIN with number of rows as in COL.
*Each column of BIN represents a group, and value 1 in it means that the case (row) belongs to it.
*Groups may intersect or be disjoint by the composition of cases (in the latter case
*BIN appears to be a set of dummy variables). Each column of BIN must contain at least one one.
*A column may, as a specific case, contain no zeros, in which case the group corresponds to the
*whole dataset, all its rows: statistic "in total" will be computed.
*The function returns 4 columns of by-group univariate statistics, with rows corresponding to the groups, columns of BIN:
*minimal value (NAME1), maximal value (NAME2), arithmetic mean (NAME3),
*variance (on "df=n-1", NAME4). NAME4 won't be computed (issuing the error of division by zero)
*if any column in BIN contains less than 2 ones.
*If your grouping is defined by a categorical variable, create dummy variables BIN out of it
*with the help of function /*!KO_freq*/.

define !KO_gdescr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @var= !2.
comp @bin= !3.
comp @n= t(csum(@bin)).
comp @vars= @var*make(1,ncol(@bin),1).
comp !4= t(cmin(@vars&*@bin+(not @bin)*mmax(@var))).
comp !5= t(cmax(@vars&*@bin+(not @bin)*mmin(@var))).
comp @bin= t(@bin).
comp @sum= @bin*@var.
comp !6= @sum/@n.
comp !7= (@bin*(@var&*@var)-@sum&*!6)/(@n-1).
release @var,@bin,@n,@vars,@sum.
!enddefine.


******** VARIOUS UNIVARIATE STATISTICS, BY GROUPS ********.
*/*!KO_aggr(data%bin%stat%name)*/*.
*Version 1.
*Computes the required descriptive statistic in the columns of DATA, in total and/or by groups
*of cases (the rows).
*This function resembles out-of-MATRIX command AGGREGATE.
*Takes data DATA (matrix with any number of columns, variables) and one or more binary (1 vs 0)
*variables constituting columns of matrix BIN with number of rows as in DATA.
*Each column of BIN represents a group, and value 1 in it means that the case (row) belongs to it.
*Groups may intersect or be disjoint by the composition of cases (in the latter case
*BIN appears to be a set of dummy variables). Each column of BIN must contain at least one one.
*A column may, as a specific case, contain no zeros, in which case the group corresponds to the
*whole dataset, all its rows: statistic "in total" will be computed.
*Result NAME - matrix with column corresponding to columns of DATA, i.e. the variables, and with rows
*corresponding to columns of BIN, i.e. the groups. NAME contains values of the computed statistic for
*each variable in each group.
*STAT - the requested statistic. It is a keyword in capital letters. Quotes or apostrophes around the
*keyword are optional. The following statistics are available:
*"SUM" - sum
*"MEAN" - arithmetic mean
*"SSDEV" - sum of squared deviations from the mean
*"VARIANCE" - variance [sum in every column of BIN must be >1]
*"VARIANCE2" - variance, computed on "df=n", not "df=n-1"
*"SEMEAN" - st. error of the arithmetic mean [sum in every column of BIN must be >1]
*"MIN" - minimal value
*"MAX" - maximal value
*"FIRST" - first value
*"LAST" - last value
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*If your grouping is defined by a categorical variable, create dummy variables BIN out of it
*with the help of function /*!KO_freq*/.
*If you want to replace, by the computed statistic, the values of DATA, use function /*!KO_aggrv*/.

define !KO_aggr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @bin= !3.
!let !stat= !unquote(!4)
!if (!index('SUM FIRST LAST MIN MAX',!stat)=0) !then !let !n= ',@n'
-comp @n= csum(@bin).
!ifend
!if (!stat='SUM') !then
-comp !5= t(@bin)*@data.
!else !if (!stat='MEAN') !then
-comp !5= t(@bin)*@data/(t(@n)*make(1,ncol(@data),1)).
!else !if (!stat='SSDEV' !or !stat='VARIANCE' !or !stat='VARIANCE2' !or !stat='SEMEAN') !then
-comp @sum= t(@bin)*@data.
-comp @cssq= t(@bin)*(@data&*@data).
-comp @n= t(@n)*make(1,ncol(@data),1).
-comp !5= @cssq-@sum&*@sum/@n.
 !if (!stat<>'SSDEV') !then
- comp !5= !5/(@n !if (!stat<>'VARIANCE2') !then -1 !ifend).
  !if (!stat='SEMEAN') !then
-  comp !5= sqrt(!5/@n).
  !ifend
 !ifend
-release @sum,@cssq.
!else !if (!stat='MIN') !then
-comp @ones= make(1,ncol(@data),1).
-comp !5= make(ncol(@bin),ncol(@data),0).
-comp @bigv= 1E300.
-loop @i= 1 to ncol(@bin).
- comp @bini= @bin(:,@i)*@ones.
- comp !5(@i,:)= cmin(@data&*@bini+(not @bini)*@bigv).
-end loop.
-release @ones,@bigv,@i,@bini.
!else !if (!stat='MAX') !then
-comp @ones= make(1,ncol(@data),1).
-comp !5= make(ncol(@bin),ncol(@data),0).
-comp @smallv= mmin(@data).
-loop @i= 1 to ncol(@bin).
- comp @bini= @bin(:,@i)*@ones.
- comp !5(@i,:)= cmax(@data&*@bini+(not @bini)*@smallv).
-end loop.
-release @ones,@smallv,@i,@bini.
!else !if (!stat='LAST') !then
-comp !5= make(ncol(@bin),ncol(@data),0).
-comp @rnums= t(1:nrow(@data)).
-loop @i= 1 to ncol(@bin).
- comp @ind= cmax(@bin(:,@i)&*@rnums).
- comp !5(@i,:)= @data(@ind,:).
-end loop.
-release @rnums,@i,@ind.
!else !if (!stat='FIRST') !then
-comp !5= make(ncol(@bin),ncol(@data),0).
-comp @rnums= t(nrow(@data):1).
-loop @i= 1 to ncol(@bin).
- comp @ind= nrow(@data)-cmax(@bin(:,@i)&*@rnums)+1.
- comp !5(@i,:)= @data(@ind,:).
-end loop.
-release @rnums,@i,@ind.
!ifend !ifend !ifend !ifend !ifend !ifend !ifend
release @data,@bin !n.
!enddefine.


******** VALUE CLOSEST TO A SPECFIED ONE, FROM BELOW OR ABOVE ********.
*/*!KO_closest(data%ineq%val%name)*/*.
*Version 1.
*In each column of data DATA, finds out the observed value closest to value VAL by size;
*and condition INEQ stipulates whether it must be lesser or greater than VAL.
*INEQ is inequality operator, keyword (may put in quotes or apostrophes)
*to choose: >, >=, <, <= (or letters: GT, GE, LT, LE). For instance, <= means to identify the greatest
*observed value which is <=VAL, i.e., it adjoins it from "below" and is either less or equals it.
*VAL - scalar or row vector lengthed as the number of columns of DATA. Vector means specifying
*a separate value VAL for each column of DATA.
*Result is row NAME containing the sought-for values. Value 1E+300 or 1E-300 in NAME says that
*in the given column of DATA there was found no value satisfying condition INEQ VAL.

define !KO_closest(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !ineq= !upcase(!unquote(!3))
comp @data= !2.
comp @val= !4.
!if (!index(!ineq,'>')>0 !or !index(!ineq,'G')>0) !then
 !let !plus= '+' !let !max= 'cmin' !else !let !plus= '-' !let !max= 'cmax'
!ifend
do if ncol(@val)>1.
-comp @val= @val(make(1,nrow(@data),1),:).
end if.
comp @flag= @data !ineq @val.
comp !5= !max(@flag&*@data !plus (not @flag)*1E300).
release @data,@val,@flag.
!enddefine.


******** COVARIANCE MATRIX ********.
*/*!KO_cov(data%name)*/*.
*Version 1.
*Computes covariances between colums of DATA.
*Returns square symmetric matrix NAME.

define !KO_cov(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
comp !3= sscp(!3)/(nrow(!3)-1).
!enddefine.


******** COVARIANCE MATRIX (DF=N) ********.
*/*!KO_cov2(data%name)*/*.
*Version 1.
*Computes covariances between colums of DATA. The values are computed on "df=n" (not on "df=n-1").
*Returns square symmetric matrix NAME.

define !KO_cov2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
comp !3= sscp(!3)/nrow(!3).
!enddefine.


******** CORRELATION MATRIX ********.
*/*!KO_corr(data%name)*/*.
*Version 1.
*Computes Pearson correlations between colums of DATA.
*Returns square symmetric matrix NAME.

define !KO_corr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= sscp(!3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:)).
comp @rssdev= sqrt(diag(!3)).
comp !3= !3/(@rssdev*t(@rssdev)).
call setdiag(!3,1).
release @rssdev.
!enddefine.


******** COSINE MATRIX ********.
*/*!KO_cosine(data%name)*/*.
*Version 1.
*Computes cosine similarities (coefficient of proportionality or Tucker's coefficient
*of congruence) between columns of DATA.
*Returns square symmetric matrix NAME.

define !KO_cosine(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= sscp(!3).
comp @rss= sqrt(diag(!3)).
comp !3= !3/(@rss*t(@rss)).
call setdiag(!3,1).
release @rss.
!enddefine.


******** MATRIX OF COEFFICIENTS OF IDENTITY ********.
*/*!KO_idc(data%name)*/*.
*Version 1.
*Computes coefficients of identity [Zegers, ten Berge, 1985]
*between columns of DATA.
*Returns square symmetric matrix NAME.

define !KO_idc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= sscp(!3).
comp @ss2= (diag(!3)/2)*make(1,ncol(!3),1).
comp !3= !3/(@ss2+t(@ss2)).
release @ss2.
!enddefine.


******** MATRIX OF SIMILARITY RATIOS ********.
*/*!KO_simr(data%name)*/*.
*Version 1.
*Computes similarity ratio coefficients, also known as Kohonen similarity,
*between columns of DATA.
*Returns square symmetric matrix NAME.

define !KO_simr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= sscp(!3).
comp @ss= diag(!3)*make(1,ncol(!3),1).
comp !3= !3/(@ss+t(@ss)-!3).
release @ss.
!enddefine.


******** COVARIANCE, CORRELATION, COSINE, COEFFICIENT OF IDENTITY, SIMILARITY RATIO ********.
*/*!KO_biv(col1%col2%name)*/*.
*Version 2.
*For two column vectors COL1 and COL2 (equal length) computes row of five statistics of similarity:
*coefficient of covariation (1st element of NAME), Pearson correlation (2nd element), coefficient of
*proportionality (cosine, 3rd element), coefficient of identity (4th element),
*similarity ratio (5th element).

define !KO_biv(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @var1= !2.
comp @var2= !3.
comp @vars= {@var1,@var2}.
comp !4= make(1,5,0).
comp @ss= cssq(@vars).
comp @scp= csum(@var1&*@var2).
comp !4(3)= @scp/sqrt(@ss(1)*@ss(2)).
comp !4(4)= 2*@scp/(@ss(1)+@ss(2)).
comp !4(5)= @scp/(@ss(1)+@ss(2)-@scp).
comp @vars= @vars-(csum(@vars)/nrow(@vars))(make(1,nrow(@vars),1),:).
comp @ss= cssq(@vars).
comp @scp= csum(@vars(:,1)&*@vars(:,2)).
comp !4(1)= @scp/(nrow(@vars)-1)).
comp !4(2)= @scp/sqrt(@ss(1)*@ss(2)).
release @var1,@var2,@vars,@ss,@scp.
!enddefine.


******** RECTANGULAR MATRIX OF COVARIANCES/CORRELATIONS/COSINES ********.
*/*!KO_rect(vars1%vars2%type%name)*/*.
*Version 1.
*Computes requested association coefficent between columns of VARS1 and columns of VARS2.
*These two matrices must have the same number of rows (rows are cases, and columns are variables).
*Returns rectangular matrix NAME sized num_col_VARS1 x num_col_VARS2.
*TYPE - requested association measure, capitalized keyword (optionally may put in quotes or
*apostrophes):
*"COV" - coefficient of covariance (on "df=n-1")
*"COV2" - coefficient of covariance (on "df=n")
*"CORR" - Pearson correlation coefficient
*"COSINE" - cosine similarity (Tucker's coefficient of congruence).

define !KO_rect(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !type= !unquote(!4)
comp @v1= !2.
comp @v2= !3.
comp @n= nrow(@v1).
!if (!type<>'COSINE') !then
-comp @v1= @v1-(csum(@v1)/@n)(make(1,@n,1),:).
-comp @v2= @v2-(csum(@v2)/@n)(make(1,@n,1),:).
!ifend
comp !5= t(@v1)*@v2.
!if (!type='COV') !then
-comp !5= !5/(@n-1).
!else !if (!type='COV2') !then
-comp !5= !5/@n.
!else
-comp @v1= sqrt(cssq(@v1)).
-comp @v2= sqrt(cssq(@v2)).
-comp !5= !5/(t(@v1)*(@v2)).
!ifend !ifend
release @v1,@v2,@n.
!enddefine.


******** COVARIANCE/CORRELATION MATRIX, BY GROUPS AND POOLED ********.
*/*!KO_gcov(data%bin%type%out%name)*/*.
*Version 2.
*Computes scatter, covariance, or correlation matrix of variables by groups
*of cases, and also the pooled (across the groups) matrix.
*Takes data DATA (matrix with columns-variables) and one or more binary (1 vs 0)
*variables constituting columns of matrix BIN with number of rows as in DATA.
*Each column of BIN represents a group, and value 1 in it means that the case (row) belongs to it.
*Groups may intersect or be disjoint by the composition of cases (in the latter case
*BIN appears to be a set of dummy variables). Each column of BIN must contain at least one one.
*A column may, as a specific case, contain no zeros, in which case the group corresponds to the
*whole dataset, all its rows: matrix for the total sample will be computed.
*Arguments TYPE and OUT is a keyword in capital letters. Quotes or apostrophes around
*the keyword are optional.
*TYPE - the requested type of matrix. The following types are available:
*"SCAT" - scatter matrix
*"COV" - covariance matrix [warning: sum in every column of BIN must be >1]
*"COV2" - covariance matrix, computed on "df=n", not "df=n-1"
*"CORR" - correlation matrix [warning: sum in every column of BIN must be >1]
*OUT - this argument determines the composition of the returned result NAME:
*"BYGR" - matrices, own for each group, column of BIN; the matrices are stacked, one below another,
*their order follows the order of BIN's columns
*"POOL" - pooled matrix of the groups
*"BOTH" - both this and that (the pooled matrix will be the upper one in the stack).
*What the pooled matrix is in case of different matrix type:
*- the pooled scatter matrix is the summed scatter matrix of the groups;
*- the pooled covariance matrix is the weighted averaged group covariance matrix, where weights are the
*number of the degrees of freedom in a group ("n-1" with COV or "n" with COV2, where n is the frequency
*in a group);
*- the pooled correlation matrix is obtained from the pooled covariance matrix
*(or the pooled scatter matrix), not by averaging the group correlation matrices.
*When computing a pooled matrix of any type the function takes for the total sample frequency
*the sum of frequencies in groups, i.e. number of ones in BIN, - and not the number of rows in BIN
*(for in BIN some rows might be empty or might count in more than one group).
*Because, if DATA has only one column, COV (or COV2) matrix turns in a scalar, the variance,
*you may use the current function also to compute the pooled group variance.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*If your grouping is defined by a categorical variable, create dummy variables BIN out of it
*with the help of function /*!KO_freq*/.

define !KO_gcov(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @bin= !3.
!let !type= !unquote(!4)
!let !out= !unquote(!5)
comp @n= csum(@bin).
comp !6= make(!if (!out='BYGR') !then 1 !else ncol(@data) !ifend,ncol(@data),0).
comp @m= @bin.
loop @i= 2 to nrow(@m).
-comp @m(@i,:)= @m(@i-1,:)+@m(@i,:).
end loop.
comp @m= @m&*@bin.
comp @inds= make(nrow(@m)+1,ncol(@m),0).
comp @nums= t(1:nrow(@m)).
comp @m= @m+(not @m)*nrow(@inds).
loop @i= 1 to ncol(@m).
-comp @inds(@m(:,@i),@i)= @nums.
end loop.
comp @inds= @inds(1:(nrow(@inds)-1),:).
release @m,@nums.
loop @i= 1 to ncol(@bin).
-comp @ni= @n(@i).
-comp @datai= @data(@inds(1:@ni,@i),:).
-comp @sumi= csum(@datai).
-comp @datai= (sscp(@datai)-t(@sumi)*(@sumi/@ni)).
 !if (!out<>'BYGR') !then
- comp !6(1:ncol(@data),:)= !6(1:ncol(@data),:)+@datai.
 !ifend
 !if (!out<>'POOL') !then
  !if (!type='COV') !then
-  comp @datai= @datai/(@ni-1).
  !else !if (!type='COV2') !then
-  comp @datai= @datai/@ni.
  !else !if (!type='CORR') !then
-  comp @sumi= diag(@datai).
-  comp @datai= @datai/sqrt(@sumi*t(@sumi)).
  !ifend !ifend !ifend
- comp !6= {!6;@datai}.
 !ifend
end loop.
!if (!out<>'BYGR' !and !type<>'SCAT') !then
 !if (!type='COV') !then
- comp !6(1:ncol(@data),:)= !6(1:ncol(@data),:)/(rsum(@n)-ncol(@bin)).
 !else !if (!type='COV2') !then
- comp !6(1:ncol(@data),:)= !6(1:ncol(@data),:)/rsum(@n).
 !else !if (!type='CORR') !then
- comp @sumi= diag(!6).
- comp !6(1:ncol(@data),:)= !6(1:ncol(@data),:)/sqrt(@sumi*t(@sumi)).
 !ifend !ifend !ifend
!ifend
!if (!out='BYGR') !then
-comp !6= !6(2:nrow(!6),:).
!ifend
release @data,@bin,@n,@i,@inds,@ni,@sumi,@datai.
!enddefine.


******** BETWEEN-GROUP AND POOLED WITHIN-GROUP SCATTER MATRICES ********.
*/*!KO_bwscat(data%dummy%name1%name2)*/*.
*Version 1.
*Takes data DATA and groups DUMMY and returns between-group scatter matrix NAME1 and
*pooled (summed) within-group scatter matrix NAME2.
*(SSbetween = trace(NAME1) и SSwithin = trace(NAME1).)
*DATA - data with any number of columns (variables).
*NAME1 (attention!) depends on whether DATA has centered
*columns or not, but NAME2 does not depend (except for numerical stability issue, which is
*better under centered data).
*DUMMY - binary dummy (indicator) variables marking groups (each variable corresponds to a group,
*with value 1 = the case belongs to it, 0 = doesn't belong to it).
*Each column in DUMMY must contain at least one 1.
*The groups must be disjoint by membership: sum in each row in DUMMY must not exceed 1.
*If DUMMY has rows complete of zeros (cases out of the groups) then NAME2 won't be correct.
*The returned matrices sum into the sscp matrix of variables of the whole dataset: NAME1+NAME2=sscp(DATA).
*You can create dummy variables DUMMY out of a categorical grouping variable
*with the help of function /*!KO_freq*/.
*If you need a pooled within-group scatter matrix from groups which intersect by case membership,
*use function /*!KO_gcov*/.

define !KO_bwscat(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @des= !3.
comp @n= csum(@des).
comp !4= t(@des)*@data/(t(@n)*make(1,ncol(@data),1)).
comp !4= sscp(@des*!4).
comp !5= sscp(@data)-!4.
release @data,@des,@n.
!enddefine.


******** IMAGE OR ANTI-IMAGE MATRIX ********.
*/*!KO_image(cov%type%name)*/*.
*Version 1.
*Takes covariance or correlation matrix COV and returns as NAME the image
*(if TYPE is positive scalar) or anti-image (if TYPE is nonpositive scalar) covariance matrix.
*COV must be square symmetric nonsingular, normally - positive definite.
*TYPE is scalar.
*"Images" stand on the diagomal of image matrix, equal to Rsq*Ssq, where Rsq is the squared multiple
*correlation for the variable as dependent on all the other, аnd Ssq is the diagonal element of COV.
*Off-diagonal elements of anti-image matrix are the (negated) coefficients of partial covariance.
*If you negate the sign of off-diagonal elements of the anti-image matrix and add it with the
*image matrix, the initial matrix COV would be restored.
*This function might be used for any sscp-type matrix: correlation, covariance, cosine, raw sscp.

define !KO_image(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @cov= !2.
comp @type= !3.
comp @icov= inv(@cov).
comp @diag= diag(@icov).
comp !4= @icov/(@diag*t(@diag)).
comp !4= (!4+t(!4))/2.
do if @type.
-call setdiag(!4,-diag(!4)).
-comp !4= !4+@cov.
end if.
release @cov,@type,@icov,@diag.
!enddefine.


******** INFLUENCE MATRIX (HAT MATRIX) ********.
*/*!KO_hat(data%name)*/*.
*Version 1.
*Takes data matrix DATA which must be with noncollinear columns.
*If necessary do prior centration or standardization of DATA columns.
*Returns influence matrix (hat matrix, projection matrix) NAME.
*Values on its diagonal are known as leverages.
*If DATA (having n rows) has centered columns then NAME*(n-1) is the double centration
*(see function  /*!KO_dcenter*/) of matrix of squared Mahalanobis distances between rows of DATA,
*and leverages*(n-1) are squared Mahalanobis distances to centroid of DATA.

define !KO_hat(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3*solve(sscp(!3),t(!3)).
comp !3= (!3+t(!3))/2.
!enddefine.


******** CO-COUNTS MATRICES FOR BINARY DATA ********.
*/*!KO_bincnt(data%name1%name2)*/*.
*Version 1.
*Takes matrix of binary (values 0 and 1) data DATA and returns two square matrices with co-counts
*between columns of DATA:
*NAME1 - symmetric matrix which element ij shows the number of rows in DATA where column i
*contains 1 and column j contains 1. These counts are known as counts A (characteristic is in both columns).
*NAME2 - asymmetric matrix which element ij shows the number of rows in DATA where column i
*contains 0 and column j contains 1. These counts are known as counts B (characteristic present in one
*and is absent in the other column).
*Counts C matrix is transposed B.
*Counts D (characteristic is absent on both columns) = number_of_rows_in_DATA-A-B-C.
*Counts A, B, C, D are useful for computation of various proximity measures invented for binary data.

define !KO_bincnt(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp !3= sscp(!4).
comp !4= csum(!4)(make(1,ncol(!4),1),:)-!3.
!enddefine.


******** MATRIX OF SQUARED EUCLIDEAN DISTANCES ********.
*/*!KO_seuclid(data%name)*/*.
*Version 2.
*Computes square euclidean distances between rows of DATA.
*Returns square symmetric matrix NAME.
*If the number of columns of DATA >1, the function computes the distances via scalar products;
*this is more fast, but calculatory potentially less precise mode than the "direct" mode
*via raising differences to square (if you need "direct mode", use more general
*function /*!KO_pwmink*/).

define !KO_seuclid(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
do if ncol(!3)=1.
-comp !3= t(!3)(make(1,nrow(!3),1),:).
-comp !3= !3-t(!3).
-comp !3= !3&*!3.
else.
-comp !3= sscp(t(!3)).
-comp @hsq= t(diag(!3))(make(1,ncol(!3),1),:).
-comp !3= @hsq+t(@hsq)-2*!3.
-comp !3= !3&*(!3>0).
-release @hsq.
end if.
!enddefine.


******** RECTANGULAR MATRIX OF SQUARED EUCLIDEAN DISTANCES ********.
*/*!KO_seuclidr(cases1%cases2%name)*/*.
*Version 2.
*Computes square euclidean distances between rows of CASES1 and rows of CASES2.
*These two matrices must have the same number of columns (columns are data dimensions).
*Returns rectangular matrix NAME sized num_rows_CASES1 x num_rows_CASES2.
*If the number of columns of data >1, the function computes the distances via scalar products;
*this is more fast, but calculatory potentially less precise mode than the "direct" mode
*via raising differences to square.

define !KO_seuclidr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @g1= !2.
comp @g2= !3.
do if ncol(@g1)=1.
-comp !4= (@g1*make(1,nrow(@g2),1))-t(@g2)(make(1,nrow(@g1),1),:).
-comp !4= !4&*!4.
else.
-comp @g2= t(@g2).
-comp !4= @g1*@g2.
-comp !4= rssq(@g1)*make(1,ncol(@g2),1)+cssq(@g2)(make(1,nrow(@g1),1),:)-2*!4.
-comp !4= !4&*(!4>0).
end if.
release @g1,@g2.
!enddefine.


******** MATRIX OF POWERED WEIGHTED MINKOWSKI DISTANCES ********.
*/*!KO_pwmink(data%power%weights%name)*/*.
*Version 1.
*Computes Minkowski distances (raised to power POWER) between rows of DATA.
*Returns square symmetric matrix NAME.
*POWER - scalar (typically >=1); this is the power for the distance.
*WEIGHTS - vector (row or column) lengthed as the number of columns of DATA and containing positive numbers;
*these are weights for columns of DATA. Only ratio between the weights size matters because the function
*normalizes them. If all weights are equal there will come out unweighted distance.
*The distance between rows a and b =  Sum(w_i*|a_i-b_i|^p), where a_i, b_i are values in column i,
*p is power, w_i is weight of column i, Sum is sum across all columns.
*If POWER=2 the function will return (weighted) squared euclidean distances.
*If POWER=1 the function will return (weighted) Manhattan distances.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_pwmink(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @p= !3.
comp @w= !4.
comp @w= @w/msum(@w)*rmax({nrow(@w),ncol(@w)}).
comp !5= make(nrow(@data),nrow(@data),0).
comp @ones= make(1,nrow(@data),1).
loop @i= 1 to ncol(@data).
-comp @kroni= @data(:,@i)*@ones.
-comp !5= !5+@w(@i)&*abs(@kroni-t(@kroni))&**@p.
end loop.
release @data,@p,@w,@kroni,@i,@ones.
!enddefine.


******** RECTANGULAR MATRIX OF POWERED WEIGHTED MINKOWSKI DISTANCES ********.
*/*!KO_pwminkr(cases1%cases2%power%weights%name)*/*.
*Version 1.
*Computes Minkowski distances (raised to power POWER) between rows of CASES1 and rows of CASES2.
*These two matrices must have the same number of columns (columns are data dimensions).
*Returns rectangular matrix NAME sized num_rows_CASES1 x num_rows_CASES2.
*Arguments POWER and WEIGHTS are the same as in function /*!KO_pwmink*/.
*If POWER=2 the function will return (weighted) squared euclidean distances.
*If POWER=1 the function will return (weighted) Manhattan distances.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_pwminkr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @g1= !2.
comp @g2= !3.
comp @p= !4.
comp @w= !5.
comp @w= @w/msum(@w)*rmax({nrow(@w),ncol(@w)}).
comp !6= make(nrow(@g1),nrow(@g2),0).
comp @ones1= make(1,nrow(@g1),1).
comp @ones2= make(1,nrow(@g2),1).
loop @i= 1 to ncol(@g1).
-comp @kroni1= @g1(:,@i)*@ones2.
-comp @kroni2= t(@g2(:,@i))(@ones1,:).
-comp !6= !6+@w(@i)&*abs(@kroni1-@kroni2)&**@p.
end loop.
release @g1,@g2,@p,@w,@kroni1,@kroni2,@i,@ones1,@ones2.
!enddefine.


******** MATRIX OF SQUARED MAHALANOBIS DISTANCES ********.
*/*!KO_smahal(data%cov%name)*/*.
*Version 1.
*Computes square Mahalanobis distances between rows of DATA (matrix n x p, p>=1).
*Returns square symmetric matrix NAME.
*Argument COV - either square p x p symmetric nonsingular matrix that is a user-specified
*covariance matrix between the variables, or scalar 0 - then matrix COV will
*be set equal to the observed covariance matrix between the p variables of DATA.
*Then DATA must have more rows than columns and the columns must be noncollinear.
*The distance formula: d^2 = (x-y)'inv(COV)(x-y).
*If the covariance matrix is identity matrix, the result are squared euclidean distances.
*If the input COV is a correlation matrix, the unscaled squared Mahalanobis distances
*are obtained.
*If you have a user-specified inv(COV) instead of COV, you can use function /*!KO_sbutler*/
*to get Mahalanobis distances.

define !KO_smahal(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @cov= !3.
do if nrow(@cov)*ncol(@cov)=1 and @cov(1,1)=0.
-comp !4= !4-(csum(!4)/nrow(!4))(make(1,nrow(!4),1),:).
-comp !4= !4*solve(sscp(!4),t(!4))*(nrow(!4)-1).
else.
-comp !4= !4*inv(@cov)*t(!4).
end if.
comp !4= (!4+t(!4))/2.
comp @hsq= diag(!4)*make(1,ncol(!4),1).
comp !4= @hsq+t(@hsq)-2*!4.
comp !4= !4&*(!4>0).
release @cov,@hsq.
!enddefine.


******** VECTOR OF SQUARED MAHALANOBIS DISTANCES TO CENTROID ********.
*/*!KO_smahalc(data%mean%cov%name)*/*.
*Version 1.
*Computes squared Mahalanobis distances between rows of DATA and a centroid.
*Returns column vector NAME with number of rows as DATA has.
*DATA must have p (p>=1) columns and any number of rows. It is coordinates of points in
*p-dimensional space.
*If you want to get distances to the centroid of the observed DATA cloud itself, specify
*argument COV as zero scalar. DATA will then need to have at least two rows and
*noncollinear columns. Argument MEAN will be ignored.
*If you want to get distances of DATA points to the centroid of some other cloud in the same space,
*then you have to specify vector of means MEAN and covariance matrix COV of that cloud:
*MEAN must have p columns and 1 row. These are coordinates of the centre of that cloud.
*COV must be p X p, symmetric nonsingular. It is the covariances in that cloud.
*If you want to get distance between two groups from the same population, enter row DATA = centroid
*of one group, row MEAN = centroid of the other group, COV = pooled within-group covariance matrix
*(you can obtain the latter by function /*!KO_gcov*/).
*If COV is identity matrix, the result are squared euclidean distances from the points to
*centroid MEAN.

define !KO_smahalc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @mean= !3.
comp @cov= !4.
do if nrow(@cov)*ncol(@cov)=1 and @cov(1,1)=0.
-comp @mean= csum(@data)/nrow(@data).
-comp @data= @data-(csum(@data)/nrow(@data))(make(1,nrow(@data),1),:).
-comp !5= @data*solve(sscp(@data)/(nrow(@data)-1),t(@data)).
else.
-comp @data= @data-@mean(make(1,nrow(@data),1),:).
-comp !5= @data*inv(@cov)*t(@data).
end if.
comp !5= diag(!5).
release @data,@mean,@cov.
!enddefine.


******** MATRIX OF SQUARED BUTLER DISTANCES ********.
*/*!KO_sbutler(data%cos%name)*/*.
*Version 1.
*Computes square Butler's distances between rows of DATA (matrix n x p, p>=1).
*Returns square symmetric matrix NAME.
*"Butler's" distances are euclidean distances not in cartesian (orthogonal) but in
*oblique system, where the angle cosines between the axes are equal to the correlations
*between the variables or are specified by the user; and input data are taken for skew
*(contravariant) coordinates in the system. The formula: d^2 = (x-y)'COS(x-y).
*Argument COS - either square p x p symmetric matrix of angle cosines between the axes
*(this is user-specification of the cosines), or a scalar 0 - then the matrix of angle
*cosines COS will be set equal to the observed correlation matrix between the p
*variables of DATA.
*If the angle cosines matrix is identity matrix, the result are squared euclidean distances.
*If as COS you input an inverted covariance matrix, the result are squared Mahalanobis distances,
*and if an inverted correlation matrix - then unscaled squared Mahalanobis distances.

define !KO_sbutler(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @cos= !3.
do if nrow(@cos)*ncol(@cos)=1 and @cos(1,1)=0.
-comp !4= !4-(csum(!4)/nrow(!4))(make(1,nrow(!4),1),:).
-comp @cos= sscp(!4).
-comp @h= sqrt(diag(@cos)).
-comp @cos= @cos/(@h*t(@h)).
end if.
comp !4= !4*@cos*t(!4).
comp !4= (!4+t(!4))/2.
comp @h= diag(!4)*make(1,ncol(!4),1).
comp !4= @h+t(@h)-2*!4.
comp !4= !4&*(!4>0).
release @cos,@h.
!enddefine.


******** LIST OF UNIQUE VALUES OF VECTOR AND POSITION OF THEIR FIRST OCCURRENCE ********.
*/*!KO_unique(vec%ignore%name1%name2)*/*.
*Version 3.
*Takes vector (column or row) VEC and returns row vector NAME1, which is
*the list of unique values in VEC; the values go in the order of their first occurrence in VEC.
*The positions of the first occurrence are returned as the corresponding row NAME2.
*Argument IGNORE - digit (not name or expression, optionally in quotes or apostrophes): if 1,
*value 0 in VEC is ignored (considered nonvalid); if 0, zero is considered valid value.
*Result NAME2 in the form of scalar 0 under IGNORE=1 means that VEC consisted of just zeros.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_unique(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @vec= !2.
!let !ignore= !unquote(!3)
!if (!ignore) !then !let !st= 1
-comp !4= 0.
-comp !5= 0.
!else !let !st= 2
-comp !4= @vec(1).
-comp !5= 1.
!ifend
loop @i= !st to rmax({nrow(@vec),ncol(@vec)}).
-do if not any(@vec(@i)=!4).
- comp !4= {!4,@vec(@i)}.
- comp !5= {!5,@i}.
-end if.
end loop.
!if (!ignore) !then
-do if ncol(!4)>1.
- comp !4= !4(2:ncol(!4)).
- comp !5= !5(2:ncol(!5)).
-end if.
!ifend
release @vec,@i.
!enddefine.


******** FREQUENCIES OF SPECIFIED VALUES IN COLUMNS ********.
*/*!KO_freqval(data%vals%name)*/*.
*Version 1.
*Takes matrix DATA consisting of one or more columns, and list of values VALS (scalar or vector),
*and returns frequency of each of these values in each column of DATA. Result NAME is matrix of the
*frequencies; its rows correspond to the values VALS and its columns correspond to the columns of DATA.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_freqval(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @vals= !3.
comp !4= make(nrow(@vals)*ncol(@vals),ncol(@data),0).
loop @i= 1 to nrow(@vals)*ncol(@vals).
-comp !4(@i,:)= csum(@data=@vals(@i)).
end loop.
release @data,@vals,@i.
!enddefine.


******** TWO-WAY FREQUENCY CROSSTABULATION FOR SPECIFIED VALUES ********.
*/*!KO_ctabval(rcol%ccol%rvals%cvals%name)*/*.
*Version 1.
*Takes column vectors RCOL and CCOL which must be of equal length, and lists of values: RVALS (for RCOL)
*and CVALS (for CCOL). RVALS and CVALS are vectors (or scalars), their length may be different.
*Returns frequency cross-table NAME of size r x c with rows defined by the r values RVALS and columns
*defined by the c values CVALS.
*Counts in the table are the number of cases with corresponding combination of values between RCOL and CCOL.

define !KO_ctabval(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @rcol= !2.
comp @ccol= !3.
comp @rvals= !4.
comp @cvals= !5.
do if ncol(@rvals)>1.
-comp @rvals= t(@rvals).
end if.
do if nrow(@cvals)>1.
-comp @cvals= t(@cvals).
end if.
comp !6= make(nrow(@rcol),1,1).
comp !6= (t(@rcol)(make(nrow(@rvals),1,1),:)=@rvals*t(!6))*(@ccol*make(1,ncol(@cvals),1)=@cvals(!6,:)).
release @rcol,@ccol,@rvals,@cvals.
!enddefine.


******** DUMMY VARIABLES AND FREQUENCIES OF VALUES OF A COLUMN ********.
*/*!KO_freq(col%sort%name1%name2%name3)*/*.
*Version 2.
*Takes column vector COL which must not be a constant and returns:
*matrix NAME1 of dummy variables, one column for each unique value of COL;
*column vector NAME2 with frequencies of unique values of COL (it's sums within columns of NAME1);
*column vector NAME3 - list of unique values of COL.
*Argument SORT (scalar) - whether to sort the results ascendingly by the values: if positive argument,
*columns in NAME1 and rows in NAME2 and NAME3 will be sorted in ascending order of the unique values
*of COL; but if the argument is nonpositive then the order of rows and columns in the results is
*determined by the sequence of encounter of the unique values of COL (screening COL top-down).
*This function may require computer to have a lot of RAM memory in case of massive data.

define !KO_freq(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @var= !2.
comp @sort= !3.
comp !4= design(@var).
comp !5= csum(!4).
comp !6= @var*make(1,ncol(!4),1)&*!4.
comp !6= cmax(!6)+cmin(!6).
do if @sort.
-comp @grd= grade(!6).
-comp !4(:,@grd)= !4.
-comp !5(@grd)= !5.
-comp !6(@grd)= !6.
-release @grd.
end if.
comp !5= t(!5).
comp !6= t(!6).
release @var,@sort.
!enddefine.


******** TWO-WAY FREQUENCY CROSSTABULATION ********.
*/*!KO_crosstab(rcol%ccol%sort%name1%name2%name3)*/*.
*Version 2.
*Takes column vectors RCOL and CCOL which must not be constants and must be of the same length.
*Returns frequency cross-table NAME1 in which the rows are defined by unique values of RCOL
*and the columns are defined by unique values of CCOL. The lists of the unique values themselves
*are returned as column vector NAME2 (values of RCOL) and row vector NAME3 (values of CCOL).
*Argument SORT (scalar) -  whether to sort the results ascendingly by the values: if positive argument,
*rows in NAME1 and NAME2 will be sorted in ascending order of the unique values of RCOL and columns in
*NAME1 and NAME3 will be sorted in ascending order of the unique values of CCOL;
*but if the argument if nonpositive then the order of rows and columns in the results is determined
*by the sequence of encounter of the unique values in RCOL and CCOL, respectively
*(screening these columns top-down).
*This function may require computer to have a lot of RAM memory in case of massive data.

define !KO_crosstab(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @rcol= !2.
comp @ccol= !3.
comp @sort= !4.
comp !6= design(@rcol).
comp !7= design(@ccol).
comp !5= t(!6)*!7.
comp !6= @rcol*make(1,ncol(!6),1)&*!6.
comp !6= t(cmax(!6)+cmin(!6)).
comp !7= @ccol*make(1,ncol(!7),1)&*!7.
comp !7= cmax(!7)+cmin(!7).
do if @sort.
-comp @rgrd= grade(!6).
-comp @cgrd= grade(!7).
-comp !5(@rgrd,@cgrd)= !5.
-comp !6(@rgrd)= !6.
-comp !7(@cgrd)= !7.
-release @rgrd,@cgrd.
end if.
release @rcol,@ccol,@sort.
!enddefine.


******** INTERACTION DUMMY VARIABLES AND MULTI-WAY FREQUENCY AGGREGATION ********.
*/*!KO_aggrtab(cols%sort%empty%name1%name2%name3)*/*.
*Version 1.
*Does frequency crosstabulation for any number of variables.
*Returns results in the form of aggregated set (not in the form of cross-tables).
*(Same way out-of-matrix command AGGREGATE returns frequencies.)
*The function also returns binary dataset of dummy variables corresponding to the interaction of
*the highest order.
*COLS - categorical data where variables are the columns; there must be at least two columns and they
*must not be constants.
*The function returns frequency column vector NAME2 where each element corresponds to one combination
*of values from all the variables, i.e. it is the frequency in a cell of multi-way table defined by
*the variables.
*Variable values' combinations themselves are contained in NAME3 where rows correspond to rows of NAME2
*and columns correspond to variables, columns of COLS.
*Vector NAME2 is the sums in columns of the returned dataset NAME1, which is the dummy variables: they
*correspond to the above combinations of unique values. I.e. columns of NAME1 correspond to rows
*of NAME2 and NAME3.
*Argument SORT (scalar) - whether to sort results ascendingly by the values: if positive, then
*columns in NAME1 and rows in NAME2 and NAME3 will be sorted in ascending order of the unique values
*of the variables; but if nonpositive, then the mentioned order in the results is determined by the
*sequence of encounter of the unique values in data (screening columns COLS top-down).
*Argument EMPTY (scalar) - if positive, columns (in NAME1) and rows (in NAME2 and NAME3) with frequency 0
*in NAME2 will be omitted in the results. If the argument is nonpositive, results will be returned in full.
*This function may require computer to have a lot of RAM memory in case of massive data.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_aggrtab(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @cols= !2.
comp @sort= !3.
comp @empty= !4.
comp @grd= 0.
comp !5= design(@cols(:,ncol(@cols))).
comp !7= @cols(:,ncol(@cols))*make(1,ncol(!5),1)&*!5.
comp !7= cmax(!7)+cmin(!7).
do if @sort.
-comp @grd= grade(!7).
-comp !5(:,@grd)= !5.
-comp !7(@grd)= !7.
end if.
loop @i= ncol(@cols)-1 to 1 by -1.
-comp @desi= design(@cols(:,@i)).
-comp @codesi= @cols(:,@i)*make(1,ncol(@desi),1)&*@desi.
-comp @codesi= cmax(@codesi)+cmin(@codesi).
-do if @sort.
- comp @grd= grade(@codesi).
- comp @desi(:,@grd)= @desi.
- comp @codesi(@grd)= @codesi.
-end if.
-comp !5= kroneker(@desi,make(1,ncol(!5),1))&*kroneker(make(1,ncol(@desi),1),!5).
-comp !7= {kroneker(@codesi,make(1,ncol(!7),1));kroneker(make(1,ncol(@codesi),1),!7)}.
end loop.
comp !6= t(csum(!5)).
comp !7= t(!7).
do if @empty.
-comp @nz= !6>0.
-comp @nnz= csum(@nz).
-do if @nnz<nrow(@nz).
- do if @nnz/nrow(@nz)>.8.
-  comp @grd= make(1,@nnz+1,0).
-  comp @c= 1.
-  loop @i= 1 to nrow(@nz).
-   comp @grd(@c)= @i.
-   comp @c= @c+@nz(@i).
-  end loop.
-  comp @grd= @grd(1:@nnz).
-  release @c.
- else if @nnz/nrow(@nz)<.08.
-  comp @grd= make(1,@nnz,0).
-  comp @s= 1.
-  loop @c= 1 to @nnz.
-   loop @i= @s to nrow(@nz).
-   end loop if @nz(@i).
-   comp @grd(@c)= @i.
-   comp @s= @i+1.
-  end loop.
-  release @s,@c.
- else.
-  comp @grd= make(1,@nnz,0).
-  comp @c= 0.
-  loop @i= 1 to nrow(@nz).
-   do if @nz(@i).
-    comp @c= @c+1.
-    comp @grd(@c)= @i.
-   end if.
-  end loop.
-  release @c.
- end if.
- comp !5= !5(:,@grd).
- comp !6= !6(@grd).
- comp !7= !7(@grd,:).
-end if.
-release @nz,@nnz.
end if.
release @cols,@sort,@empty,@grd,@desi,@codesi,@i.
!enddefine.


******** INTERACTION DUMMY VARIABLES AND MULTI-WAY FREQUENCY AGGREGATION (DICHOTOMOUS DATA) ********.
*/*!KO_baggrtab(cols%empty%name1%name2%name3)*/*.
*Version 1.
*Does frequency crosstabulation for any number of dichotomous variables.
*The variables must be coded binary (1 vs 0); if you have other values, make them binary.
*This function is equivalent to /*!KO_aggrtab*/ but it is for all variables COLS
*binary (0 vs 1). It is faster than /*!KO_aggrtab*/ and spares memory.
*Argument EMPTY and results NAME1, NAME2, NAME3 - see in the description of /*!KO_aggrtab*/.
*The order of values in results is always sorted as 0 1, so there is no argument SORT.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_baggrtab(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @cols= !2.
comp @empty= !3.
comp !4= @cols(:,ncol(@cols)).
comp !4= {not !4,!4}.
comp !6= {0,1}.
comp @codesi= {0,1}.
loop @i= ncol(@cols)-1 to 1 by -1.
-comp @desi= @cols(:,@i).
-comp @desi= {not @desi,@desi}.
-comp !4= kroneker(@desi,make(1,ncol(!4),1))&*kroneker(make(1,ncol(@desi),1),!4).
-comp !6= {kroneker(@codesi,make(1,ncol(!6),1));kroneker(make(1,ncol(@codesi),1),!6)}.
end loop.
comp !5= t(csum(!4)).
comp !6= t(!6).
do if @empty.
-comp @nz= !5>0.
-comp @nnz= csum(@nz).
-do if @nnz<nrow(@nz).
- do if @nnz/nrow(@nz)>.8.
-  comp @grd= make(1,@nnz+1,0).
-  comp @c= 1.
-  loop @i= 1 to nrow(@nz).
-   comp @grd(@c)= @i.
-   comp @c= @c+@nz(@i).
-  end loop.
-  comp @grd= @grd(1:@nnz).
-  release @c.
- else if @nnz/nrow(@nz)<.08.
-  comp @grd= make(1,@nnz,0).
-  comp @s= 1.
-  loop @c= 1 to @nnz.
-   loop @i= @s to nrow(@nz).
-   end loop if @nz(@i).
-   comp @grd(@c)= @i.
-   comp @s= @i+1.
-  end loop.
-  release @s,@c.
- else.
-  comp @grd= make(1,@nnz,0).
-  comp @c= 0.
-  loop @i= 1 to nrow(@nz).
-   do if @nz(@i).
-    comp @c= @c+1.
-    comp @grd(@c)= @i.
-   end if.
-  end loop.
-  release @c.
- end if.
- comp !4= !4(:,@grd).
- comp !5= !5(@grd).
- comp !6= !6(@grd,:).
- release @grd.
-end if.
-release @nz,@nnz.
end if.
release @cols,@empty,@desi,@codesi,@i.
!enddefine.


******** ELEMENTARY VARIABLES AND FREQUENCIES IN A CATEGORICAL MULTIPLE RESPONSE SET ********.
*/*!KO_mrfreq(mrc%dupl%sort%name1%name2%name3)*/*.
*Version 1.
*This function is like /*!KO_freq*/ - returns binary variables and frequencies of values - but as
*input it takes not one categorical variable but a set of those; those variables having common
*pool of values. It is "categorical multiple response set". If each case is a respondent then he
*might have thus selected several, not just single, responses.
*MRC - input data which columns are the categorical variables constituting the mentioned set.
*Zero in MRC is considered the "filler", i.e. empty cell. Any other value is a valid value.
*It is allowed that some rows (cases) or some columns (variables) be empty.
*Argument DUPL (scalar) - if positive, each unique value counts in a MRC row no more than once,
*i.e. repeating of the same response by the same respondent, if encounters, is ignored;
*but if the argument is nonpositive, then all repetition by a respondent get counted as valid.
*The function returns:
*NAME1 - matrix, one column for each unique valid value of MRC. Entry (i,j) in NAME1
*is how much times respondent i gave response j (if DUPL nonpositive), or 0 or 1
*(if DUPL positive); thus under positive DUPL NAME1 is the "dichotomous (binary) multuple
*response set".
*NAME2 - column vector with frequencies of unique values of MRC (it's sums within columns of NAME1);
*NAME3 - column vector, list of unique values of MRC.
*Argument SORT (scalar) - whether to sort the results ascendingly by the values: if positive argument,
*columns in NAME1 and rows in NAME2 and NAME3 will be sorted in ascending order of the unique values
*of MRC; but if the argument is nonpositive then the order of rows and columns in the results is
*determined by the sequence of encounter of the unique values of MRC (screening MRC top-down, 1st col,
*then 2nd col...).
*This function may require computer to have a lot of RAM memory in case of massive data.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_mrfreq(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @dupl= !3.
comp @sort= !4.
comp @n= nrow(!5).
comp @p= ncol(!5).
comp !5= reshape(t(!5),@n*@p,1).
comp @des= design({0;!5}).
comp @des= @des(2:nrow(@des),2:ncol(@des)).
comp !6= csum(@des).
comp !7= !5*make(1,ncol(@des),1)&*@des.
comp !7= cmax(!7)+cmin(!7).
comp !5= @des(1:@n,:).
comp @nums= 1:@n.
loop @i= 1 to @p-1.
-comp !5= !5+@des(@nums+@i*@n,:).
end loop.
do if @dupl.
-comp !5= !5>0.
-comp !6= csum(!5).
end if.
do if @sort.
-comp @grd= grade(!7).
-comp !5(:,@grd)= !5.
-comp !6(@grd)= !6.
-comp !7(@grd)= !7.
-release @grd.
end if.
comp !6= t(!6).
comp !7= t(!7).
release @dupl,@sort,@n,@p,@des,@nums,@i.
!enddefine.


******** COUNTS OF CONCORDANT, DISCORDANT, AND TIED PAIRS ********.
*/*!KO_concdisc(col1%col2%name1%name2%name3%name4%name5)*/*.
*Version 1.
*Takes column vectors COL1 and COL2, which must not be constants and must be of the same length
*(it is two scale or ordinal variables) and returns 5 scalar quantities necessary for computation
*of such coefficients of association between ordinal variables as Goodman-Kruskal Gamma,
*Somer's coefficient, Kendall correlation, etc.
*These quantities:
*NAME1 - number of concordant pairs (P);
*NAME2 - number of discordant pairs (Q);
*NAME3 - number of pairs tied by the 1st variable, COL1 (Tx);
*NAME4 - number of pairs tied by the 2nd variable, COL2 (Ty);
*NAME5 - number of pairs tied by both variables at once (Txy).
*If COL1 and COL2 are categorical (not continuous) data, another function, /*!KO_concdisct*/
*calculating the same quantities may occur more fast.

define !KO_concdisc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                   /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @x= !2.
comp @y= !3.
comp @n= nrow(@x).
comp @x= @x*make(1,@n,1).
comp @y= @y*make(1,@n,1)).
comp @x= @x-t(@x).
comp @y= @y-t(@y).
comp @xy= @x&*@y.
comp !4= msum(@xy>0)/2.
comp !5= msum(@xy<0)/2.
comp !6= msum(@x=0)-@n.
comp !7= msum(@y=0)-@n.
comp !8= (msum(@x=0 and @y=0)-@n)/2.
comp !6= !6/2-!8.
comp !7= !7/2-!8.
release @x,@y,@n,@xy.
!enddefine.


******** COUNTS OF CONCORDANT, DISCORDANT, AND TIED PAIRS (FROM TABLE) ********.
*/*!KO_concdisct(tab%name1%name2%name3%name4%name5)*/*.
*Version 2.
*Takes frequency cross-table TAB obtained for two scale or ordinal variables
*(i.e. rows and columns in the table go in ascending order of the unique values of the variables)
*and returns 5 scalar quantities necessary for computation of such coefficients of association
*between ordinal variables as Goodman-Kruskal Gamma, Somer's coefficient, Kendall correlation, etc.
*These quantities:
*NAME1 - number of concordant pairs (P);
*NAME2 - number of discordant pairs (Q);
*NAME3 - number of pairs tied by the 1st variable, forming the table rows (Tx);
*NAME4 - number of pairs tied by the 2nd variable, forming the table columns (Ty);
*NAME5 - number of pairs tied by both variables at once (Txy).
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_concdisct(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @tab= !2.
comp @c= make(nrow(@tab),ncol(@tab),0).
comp @c(1:(nrow(@tab)-1),1:(ncol(@tab)-1))= @tab(2:nrow(@tab),2:ncol(@tab)).
loop @i= nrow(@c)-1 to 1 by -1.
-comp @c(@i,:)= @c(@i,:)+@c((@i+1),:).
end loop.
comp @c= t(@c).
loop @i= nrow(@c)-1 to 1 by -1.
-comp @c(@i,:)= @c(@i,:)+@c((@i+1),:).
end loop.
comp !3= msum(@tab&*t(@c)). /*P
comp @c= make(nrow(@tab),ncol(@tab),0).
comp @c(1:(nrow(@tab)-1),2:ncol(@tab))= @tab(2:nrow(@tab),1:(ncol(@tab)-1)).
loop @i= nrow(@c)-1 to 1 by -1.
-comp @c(@i,:)= @c(@i,:)+@c((@i+1),:).
end loop.
comp @c= t(@c).
loop @i= 2 to nrow(@c).
-comp @c(@i,:)= @c(@i,:)+@c((@i-1),:).
end loop.
comp !4= msum(@tab&*t(@c)). /*Q
comp @c= make(ncol(@tab),nrow(@tab),0).
comp @c(1:(ncol(@tab)-1),:)= t(@tab(:,2:ncol(@tab))).
loop @i= nrow(@c)-1 to 1 by -1.
-comp @c(@i,:)= @c(@i,:)+@c((@i+1),:).
end loop.
comp !5= msum(@tab&*t(@c)). /*Tx
comp @c= make(nrow(@tab),ncol(@tab),0).
comp @c(1:(nrow(@tab)-1),:)= @tab(2:nrow(@tab),:).
loop @i= nrow(@c)-1 to 1 by -1.
-comp @c(@i,:)= @c(@i,:)+@c((@i+1),:).
end loop.
comp !6= msum(@tab&*@c). /*Ty
comp @n= msum(@tab).
comp !7= (@n**2-@n)/2-!3-!4-!5-!6. /*Txy
release @tab,@c,@i,@n.
!enddefine.


******** COMEMBERSHIP CONFUSION MATRIX FOR TWO PARTITIONS ********.
*/*!KO_ccm(part1%part2%name)*/*.
*Version 1.
*Comembership confusion matrix (or table) is such a 2x2 confusion matrix where the unit of count is a pair
*of objects (not an object). This matrix is the basis of assessment of similarity of two classifications
*(partitions) of the same objects.
*Let there be n objects that were classified by two methods, I and II. Then in the comembership confusion
*matrix cell (1,1) contains number of pairs in which both objects are met in a same class both in
*classification I and in classification II; cell (1,2) contains number of pairs in which both objects
*are met in a same class in classification I but not in classification II; cell (2,1) contains number of
*pairs in which both objects are met in a same class in classification II but not in classification I;
*cell (2,2) contains number of pairs in which both objects are not met in a same class neither in
*classification I nor in classification II.
*The sum in such table is equal to the number of object pairs, n(n-1)/2.
*Input PART1 and PART2 are two arrays with the same number of rows (rows are the objects), representing
*themselves as partitions/classifications I and II, respectively. The array may be either a column with
*any values or several columns with binary (1 and 0) values. When classification is specified by a
*single column the function takes it for a categorical variable: each unique value represents a class.
*While when classification is specified by more than one column the function takes it for a set of binary
*variables each representing a class, the binary values designating belonging to it
*(1=belongs, 0=doesn't belong). Binary set can be any, for example may include rows or columns full of
*zeros or of ones; i.e. a binary set can render not only classifications with classes disjoint by objects
*membership, but any kind, including classifications with empty classes or objects not belonging to classes.
*In case if your classification represents only one class, then specify it as two columns: one binary
*(belongs - doesn't belong), and the other is empty, of zeros.
*Result NAME - comembership confusion matrix (table).

define !KO_ccm(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @p1= !2.
comp @p2= !3.
do if ncol(@p1)=1.
-comp @p1= @p1*make(1,nrow(@p1),1).
-comp @p1= @p1=t(@p1).
else.
-comp @p1= sscp(t(@p1)).
-call setdiag(@p1,1).
end if.
do if ncol(@p2)=1.
-comp @p2= @p2*make(1,nrow(@p2),1).
-comp @p2= @p2=t(@p2).
else.
-comp @p2= sscp(t(@p2)).
-call setdiag(@p2,1).
end if.
comp !4= make(2,2,0).
comp !4(1,1)= msum(@p1 and @p2)-ncol(@p1).
comp @p2= not @p2.
comp !4(1,2)= msum(@p1 and @p2).
comp @p1= not @p1.
comp !4(2,2)= msum(@p1&*@p2).
comp !4(2,1)= ncol(@p1)*(ncol(@p1)-1)-msum(!4).
comp !4= !4/2.
release @p1,@p2.
!enddefine.


******** COUNT VALUES IN ROWS (EXACT MATCH) ********.
*/*!KO_count1(data%vals%name)*/*.
*Version 2.
*Counts in the rows of DATA number of elements which match any of the values of list VALS.
*Returns column NAME lengthed as the number of rows in DATA.
*This function is similar to non-MATRIX command COUNT, where the values are specified in list.
*VALS - vector (row or column) of any length. Values in the list may duplicate: it doesn't affect
*the result, but it slows work.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_count1(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @vals= !3.
comp !4= @data=@vals(1).
loop @i= 2 to nrow(@vals)*ncol(@vals).
-comp !4= !4+(@data=@vals(@i)).
end loop.
comp !4= rsum(!4>0).
release @data,@vals,@i.
!enddefine.


******** COUNT VALUES IN ROWS (HITTING INTO RANGE) ********.
*/*!KO_count2(data%low%high%name)*/*.
*Version 2.
*Counts in the rows of DATA number of elements which lie in any of the ranges from LOW to HIGH (inclusive).
*Returns column NAME lengthed as the number of rows in DATA.
*This function is similar to non-MATRIX command COUNT, where the values are specified by range(s) from...to.
*LOW and HIGH - two scalars or two vectors of equal length. Each pair of their correspondent elements
*LOW(i) and HIGH(i) define a range; HIGH(i) should normally be >= LOW(i). Ranges may intersect.
*In order to count all values <=HIGH, set LOW on a knowingly too small value.
*In order to count all values >=LOW, set HIGH on a knowingly too large value.
*If arguments LOW and HIGH are identical the function will return the same result as /*!KO_count1*/.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_count2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @lo= !3.
comp @hi= !4.
comp !5= @data>=@lo(1) and @data<=@hi(1).
loop @i= 2 to nrow(@lo)*ncol(@lo).
-comp !5= !5+(@data>=@lo(@i))&*(@data<=@hi(@i)).
end loop.
comp !5= rsum(!5>0).
release @data,@lo,@hi,@i.
!enddefine.


******** CHI-SQUARE (PEARSON AND LIKELIHOOD RATIO) OF TWO-WAY CONTINGENCY TABLE ********.
*/*!KO_chitab(tab%type%name)*/*.
*Version 1.
*Computes chi-square statistic which tells about association of rows and columns in a two-way
*contingency table. Returns scalar NAME.
*Argument TYPE (scalar) - which version of the statistic to return: Pearson chi-square (TYPE
*non-positive) or likelihood ratio chi-square (TYPE positive).
*Table TAB must contain nonnegative numbers, and sums in its rows and columns
*must be all nonzero.

define !KO_chitab(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @type= !3.
comp @n= msum(!4).
comp @exp= rsum(!4)*csum(!4)/@n.
do if @type.
-comp !4= 2*msum(!4&*ln(!4/@exp+(!4=0))).
else.
-comp !4= msum(!4&*!4/@exp)-@n.
end if.
release @type,@n,@exp.
!enddefine.


******** PROPORTIONS AND RESIDUALS IN TWO-WAY CONTINGENCY TABLE ********.
*/*!KO_cells(tab%stat%name)*/*.
*Version 1.
*Takes two-way contingency table (for example frequency crosstable) TAB with
*(normally) nonnegative elements and (necessarily) nonzero sums in rows and columns.
*Returns table NAME of the same size, containing the required statistic in its cells.
*The required statistic - argument STAT, keyword in capital letters:
*"TABPR" - proportions of the table sum
*"ROWPR" - proportions of the sum in rows
*"COLPR" - proportions of the sum in columns
*"EXPECT" - expected (under independence of rows and colunds) values 
*"RESID" - residuals
*"SRESID" - standardized residuals
*"ASRESID" - adjusted standardized residuals.
*Quotes or apostrophes around the keyword are optional.

define !KO_cells(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @tab= !2.
!let !stat= !unquote(!3)
!if (!stat='TABPR') !then
-comp !4= @tab/msum(@tab).
!else !if (!stat='ROWPR') !then
-comp !4= @tab/(rsum(@tab)*make(1,ncol(@tab),1)).
!else !if (!stat='COLPR') !then
-comp !4= @tab/(make(nrow(@tab),1,1)*csum(@tab)).
!else
-comp @rsum= rsum(@tab).
-comp @csum= csum(@tab).
-comp @msum= msum(@tab).
-comp @exp= @rsum*@csum/@msum.
 !if (!stat='EXPECT') !then
- comp !4= @exp.
 !else !if (!stat='RESID') !then
- comp !4= @tab-@exp.
 !else !if (!stat='SRESID') !then
- comp !4= (@tab-@exp)/sqrt(@exp).
 !else !if (!stat='ASRESID') !then
- comp !4= (@tab-@exp)/sqrt(@exp&*((1-@rsum/@msum)*(1-@csum/@msum))).
 !ifend !ifend !ifend !ifend
-release @rsum,@csum,@msum,@exp.
!ifend !ifend !ifend
!enddefine.


******** MATRIX OF SQUARED CHI-SQUARE DISTANCES IN CONTINGENCY TABLE ********.
*/*!KO_schitab(tab%name)*/*.
*Version 1.
*Computes squared chi-square distances between rows of contingency table TAB.
*Returns square symmetric matrix NAME.
*TAB is a matrix with nonnegative elements, and sums in its rows and columns must
*be all nonzero.
*Chi-square distance between the rows is the weighted euclidean distance between the row proportions,
*where the inverted sums in columns serve as weights. The (squared) distance between rows i и i' is:
*N * Sum_c[1/Nj * (Nij/Ni-Ni'j/Ni')^2], where Nij is the value in the cell, Ni is the sum in row i, Nj is
*the sum in column j, N is the total sum, Sum_c means sum across all columns.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_schitab(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @tab= !2.
comp @w= msum(@tab)/csum(@tab).
comp @tab= @tab/(rsum(@tab)*make(1,ncol(@tab),1)).
comp !3= make(nrow(@tab),nrow(@tab),0).
loop @i= 1 to ncol(@tab).
-comp @coli= @tab(:,@i)*make(1,nrow(@tab),1).
-comp @coli= @coli-t(@coli).
-comp !3= !3+@w(@i)&*@coli&*@coli.
end loop.
release @tab,@w,@i,@coli.
!enddefine.


******** MATRIX OF CHI-SQUARES BETWEEN ROWS OF COUNT DATA PAIRWISELY ********.
*/*!KO_schi2c(data%type%name)*/*.
*Version 1.
*Takes data DATA sized not less than 2x2 and with nonnegative values which are considered as frequencies.
*Sums in rows and columns must be nonzero.
*For each two rows of DATA pairwise, isolates table 2xNC (NC is number of cols in DATA) and computes
*chi-square (or phi-square) Pearson statistic for the table. That statistic can be thought of a measure of
*dissimilarity between the data rows: the chi-square distance (squared).
*Returned is square symmetric matrix NAME of such distances between all rows of DATA.
*These are same distances, only squared, as returned by command PROXIMITIES /MEASURE=CHISQ (or /MEASURE=PH2).
*These "distances" are not to confuse with the chi-square distances given by function /*!KO_schitab*/.
*TYPE - scalar: if nonpositive, chi-square is computed; if positive, phi-square is computed
*(chi-square is divided by the total frequency in 2xNC table), phi-square is insensitive to nonequality of
*sums in rows of DATA.
*If in a 2xNC table there encounters empty column the function ignores it and attempts to compute based on
*the remaining columns.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_schi2c(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @type= !3.
comp @ones= make(1,nrow(@data),1).
comp @r= rsum(@data)*@ones.
comp @tot= @r+t(@r).
comp !4= make(nrow(@data),nrow(@data),0).
loop @i= 1 to ncol(@data).
-comp @c= @data(:,@i)*@ones.
-comp @cc= @c+t(@c).
-comp @e= @cc&*@r/@tot.
-comp @e= @e+(@e=0).
-comp @cc= @c&*@c/@e.
-comp !4= !4+@cc+t(@cc).
end loop.
comp !4= !4-@tot.
do if @type.
-comp !4= !4/@tot.
end if.
release @data,@type,@ones,@r,@tot,@i,@c,@cc,@e.
!enddefine.


******** DISTANCES TO GROUP CENTROIDS ********.
*/*!KO_dtoc(data%bin%name)*/*.
*Version 1.
*Takes data DATA with any number of rows (points) and columns (variables).
*And takes one or more binary (1 vs 0) variables forming matrix BIN with number of rows as in DIS.
*Each column of BIN represents a group, and value 1 in it means that the point belongs to it.
*In whole, groups may intersect or be disjoint by the composition of their points (in the latter case
*BIN appears to be a set of dummy variables). Each column of BIN must contain at least one one.
*A column may, as a specific case, contain no zeros.
*Result NAME - matrix points x groups containing squared euclidean distances between the points and
*the group centroids.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_dtoc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp !4= !3.
comp @mean= t(!4)*@data/(t(csum(!4))*make(1,ncol(@data),1)).
comp @ones= make(1,nrow(@data),1).
loop @i= 1 to ncol(!4).
-comp !4(:,@i)= rssq(@data-@mean(@i,:)(@ones,:)).
end loop.
release @data,@mean,@ones,@i.
!enddefine.


******** DISTANCES TO GROUP CENTROIDS (COMPUTATION FROM DISTANCE MATRIX) ********.
*/*!KO_dtocfrd(dis%bin%self%name)*/*.
*Version 2.
*Takes square symmetric matrix of distances between points DIS. The distances must be
*squared euclidean; anyway the function will perceive them as squared euclidean.
*And takes one or more binary (1 vs 0) variables forming matrix BIN with number of rows as in DIS.
*Each column of BIN represents a group, and value 1 in it means that the point belongs to it.
*In whole, groups may intersect or be disjoint by the composition of their points (in the latter case
*BIN appears to be a set of dummy variables). Each column of BIN must contain at least one one.
*A column may, as a specific case, contain no zeros.
*Result NAME - matrix points x groups containing squared euclidean distances between the points and
*the group centroids.
*Argument SELF - digit 0 or 1 (not name or expression; you may surround the digit by quotes/apostrophes).
*If 1, then at computation of the distance between point i and the centroid of the group to which it
*belongs (and which contains other points too), point i participates in definition of the centroid.
*If 0, then such point i does not participate in definition of the centroid: the point temporarily
*is considered as an external point.

define !KO_dtocfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @bin= !3.
!let !self= !unquote(!4)
comp @n= csum(@bin).
comp @ones= make(1,nrow(@bin),1).
comp @pgs= !5*@bin.
comp !5= csum(@pgs&*@bin)/2.
!if (!self) !then
-comp !5= (@pgs-(!5/@n)(@ones,:))/@n(@ones,:).
!else
-comp !5= !5(@ones,:)-@pgs&*@bin.
-comp @n= @n(@ones,:)-@bin.
-comp @n= @n+(@n=0).
-comp !5= (@pgs-!5/@n)/@n.
!ifend
release @bin,@n,@ones,@pgs.
!enddefine.


******** DISTANCES BETWEEN GROUP CENTROIDS (COMPUTATION FROM DISTANCE MATRIX) ********.
*/*!KO_dbwcfrd(dis%bin%name)*/*.
*Version 1.
*Takes square symmetric matrix of distances between points DIS. The distances must be
*squared euclidean; anyway the function will perceive them as squared euclidean.
*And takes no less than two binary (1 vs 0) variables forming matrix BIN with number of rows as in DIS.
*Each column of BIN represents a group, and value 1 in it means that the point belongs to it.
*In whole, groups may intersect or be disjoint by the composition of their points (in the latter case
*BIN appears to be a set of dummy variables). Each column of BIN must contain at least one one.
*A column may, as a specific case, contain no zeros.
*Result NAME - square symmetric matrix of squared euclidean distances between centroids of
*the groups of the points.

define !KO_dbwcfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @bin= !3.
comp @n= csum(@bin).
comp @nsq= @n&*@n.
comp @n= sscp(@n).
comp @bgs= t(@bin)*!4*@bin.
comp !4= diag(@bgs)/2*@nsq.
comp !4= (@bgs-(!4+t(!4))/@n)/@n.
comp !4= (!4+t(!4))/2.
call setdiag(!4,0).
release @bin,@n,@nsq,@bgs.
!enddefine.


******** AVERAGE DISTANCES TO GROUPS (COMPUTATION FROM DISSIMILARITY MATRIX) ********.
*/*!KO_datofrd(dis%bin%self%name)*/*.
*Version 1.
*Average distance from each point to points of each of the offered groups.
*Takes square symmetric matrix of distances (dissimilarities) between points DIS.
*It must be nonnegative matrix with zero diagonal.
*And takes one or more binary (1 vs 0) variables forming matrix BIN with number of rows as in DIS.
*Each column of BIN represents a group, and value 1 in it means that the point belongs to it.
*In whole, groups may intersect or be disjoint by the composition of their points (in the latter case
*BIN appears to be a set of dummy variables). Each column of BIN must contain at least one one.
*A column may, as a specific case, contain no zeros.
*Result NAME - matrix points x groups: element ij in it is the averaged distance between
*point i and points of group j.
*Argument SELF - digit 0 or 1 (not name or expression; you may surround the digit by quotes/apostrophes).
*If 1, the distance of a point to itself (the zero) when the point belongs to the group - is counted, i.e.
*it enters the averaging. If 0, the distance of a point to itself is not counted, i.e. doesn't enter the
*averaging, except when the point is the only member of the group.
*If your matrix DIS are similarities convert them first to dissimilarities by any method you like.

define !KO_datofrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @bin= !3.
!let !self= !unquote(!4)
comp !5= !5*@bin.
!if (!self) !then
-comp @n= csum(@bin)(make(1,nrow(@bin),1),:).
!else
-comp @n= csum(@bin)(make(1,nrow(@bin),1),:)-@bin.
-comp @n= @n+(not @n).
!ifend
comp !5= !5/@n.
release @bin,@n.
!enddefine.


******** AVERAGE DISTANCES BETWEEN GROUPS (COMPUTATION FROM DISSIMILARITY MATRIX) ********.
*/*!KO_dabwfrd(dis%bin%self%name)*/*.
*Version 1.
*Distances of between-group average linkage, i.e. average distances between points from every two groups.
*Takes square symmetric matrix of distances (dissimilarities) between points DIS.
*It must be nonnegative matrix with zero diagonal.
*And takes no less than two binary (1 vs 0) variables forming matrix BIN with number of rows as in DIS.
*Each column of BIN represents a group, and value 1 in it means that the point belongs to it.
*In whole, groups may intersect or be disjoint by the composition of their points (in the latter case
*BIN appears to be a set of dummy variables). Each column of BIN must contain at least one one.
*A column may, as a specific case, contain no zeros.
*Result NAME - square symmetric matrix groups x groups containing averaged distances between
*points of the different groups. Averaged distances within groups stand on the diagonal.
*Argument SELF - digit 0 or 1 (not name or expression; you may surround the digit by quotes/apostrophes).
*If 1, the distance of a point to itself (the zero) when the point belongs to the group (or to the both
*groups) - is counted, i.e. it enters the averaging. If 0, the distance of a point to itself is not
*counted, i.e. doesn't enter the averaging, except when the point is the only member of the group
*(or of both groups, when the speach is about the different groups).
*If your matrix DIS are similarities convert them first to dissimilarities by any method you like.

define !KO_dabwfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @bin= !3.
!let !self= !unquote(!4)
comp !5= t(@bin)*!5*@bin.
comp @n= csum(@bin).
comp @nn= sscp(@n).
!if (!self) !then
-call setdiag(@nn,@n&*(@n+1)).
!else
-comp @nn= @nn-sscp(@bin).
-call setdiag(@nn,@n&*(@n-1)).
-comp @nn= @nn+(not @nn).
!ifend
comp !5= !5/@nn.
release @bin,@n,@nn.
!enddefine.


******** DISTANCES TO GROUP FARTHEST NEIGHBOURS (COMPUTATION FROM DISSIMILARITY MATRIX) ********.
*/*!KO_dtoffrd(dis%bin%alsoind%name)*/*.
*Version 1.
*Distance from each point to its farthest neighbour of each of the offered groups.
*Takes square symmetric matrix of distances (dissimilarities) between points DIS.
*It must be nonnegative matrix with zero diagonal.
*And takes one or more binary (1 vs 0) variables forming matrix BIN with number of rows as in DIS.
*Each column of BIN represents a group, and value 1 in it means that the point belongs to it.
*In whole, groups may intersect or be disjoint by the composition of their points (in the latter case
*BIN appears to be a set of dummy variables). Each column of BIN must contain at least one one.
*A column may, as a specific case, contain no zeros.
*Argument ALSOIND - digit 0 or 1 (not name or expression; you may surround the digit by quotes/apostrophes).
*If you need only distances to farthest neighbours, specify 0. Result NAME will be matrix
*points x groups: element ij in it is the distance (a value from DIS) between point i and its farthest
*neighbour of group j. If you want to know also the numbers of those farthest neighbours (their row indices in
*DIS and BIN), specify ALSOIND as 1. Then there will be twice more columns in NAME: to the matrix with distances
*to farthest neighbours there will be right-joined the matrix containing the numbers (indices) of those points
*which are the farthest neighbours.
*If there exist more than one farthest neighbour for point i in group j, the number of the last one of them
*(i.e. with the greatest index) is shown.
*Only if a group consists of single point the neighbour of it in that group is that point itself.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*If your matrix DIS are similarities convert them first to dissimilarities by any method you like.

define !KO_dtoffrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @d= !2.
comp @bin= !3.
!let !ind= !unquote(!4)
comp !5= make(nrow(@bin),ncol(@bin),0).
comp @ones= make(nrow(@bin),1,1).
!if (!ind) !then
-comp @ind= make(nrow(@bin),ncol(@bin),0).
-comp @nums= t(1:nrow(@bin)).
-comp @numsm= (1:nrow(@bin))(@ones,:).
-comp @n_1= csum(@bin)=1.
-comp @d= @d+1.
-call setdiag(@d,0).
!ifend
loop @i= 1 to ncol(@bin).
-comp @bini= @bin(:,@i).
-comp @dbini= @d&*t(@bini)(@ones,:).
-comp @dneii= rmax(@dbini).
-comp !5(:,@i)= @dneii.
 !if (!ind) !then
- do if @n_1(@i).
-  comp @ind(:,@i)= make(nrow(@bin),1,cmax(@bini&*@nums)).
- else.
-  comp @ind(:,@i)= rmax((@dbini=(@dneii*t(@ones)))&*@numsm).
- end if.
 !ifend
end loop.
!if (!ind) !then
-comp !5= {!5-(!5>0),@ind}.
-release @ind,@nums,@numsm,@n_1.
!ifend
release @d,@bin,@ones,@i,@dbini,@dneii.
!enddefine.


******** DISTANCES BETWEEN GROUP FARTHEST NEIGHBOURS (COMPUTATION FROM DISSIMILARITY MATRIX) ********.
*/*!KO_dbwffrd(dis%bin%alsoind%name)*/*.
*Version 1.
*Complete-linkage distances between groups, i.e. maximal distance between points of every two groups.
*Takes square symmetric matrix of distances (dissimilarities) between points DIS.
*It must be nonnegative matrix with zero diagonal.
*And takes no less than two binary (1 vs 0) variables forming matrix BIN with number of rows as in DIS.
*Each column of BIN represents a group, and value 1 in it means that the point belongs to it.
*In whole, groups may intersect or be disjoint by the composition of their points (in the latter case
*BIN appears to be a set of dummy variables). Each column of BIN must contain at least one one.
*A column may, as a specific case, contain no zeros.
*Argument ALSOIND - digit 0 or 1 (not name or expression; you may surround the digit by quotes/apostrophes).
*If you need only distances between farthest neighbours, specify 0. Result NAME will be square symmetric
*matrix groups x groups: element ij in it is the distance (a value from DIS) between most apart from each
*other points, one of group i, the other of group j. On the diagonal are the distances between two most
*apart points inside each group. If you want to know also the numbers of those farthest neighbours (their
*row indices in DIS and BIN), specify ALSOIND as 1. Then there will be twice plus one more columns in NAME:
*to the matrix with distances between farthest neighbours there will be right-joined the matrix containing
*the numbers (indices) of those points which are the farthest neighbours. In this right-joined matrix and
*which will be asymmetric, element ij is the number of the point which is the farthest neighbour from the
*group i side (i.e. the row group). For example, elements (2,4)=10 and (4,2)=6 mean that the maximal
*distance between groups 2 and 4 is the distance between point 10 (it is from group 2) and point 6
*(it is from group 4). As for the numbers of most far apart points inside each group, the index of one of
*the two is on the diagonal of that right-joined matrix, while the index of the other is carried into the
*separate column, which is joined to the right of all. Thus, if k is the number of columns of BIN
*(number of groups) then with ALSOIND=1 NAME has 2*k+1 columns.
*A point may come to be its own neighbour only in case it is the only group member in both groups being
*currently compared.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*If your matrix DIS are similarities convert them first to dissimilarities by any method you like.

define !KO_dbwffrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @d= !2.
comp @bin= !3.
comp @k= ncol(@bin).
!let !ind= !unquote(!4)
comp @dtog= make(nrow(@bin),ncol(@bin),0).
comp @ones= make(nrow(@bin),1,1).
!if (!ind) !then
-comp @ind= make(nrow(@bin),ncol(@bin),0).
-comp @nums= t(1:nrow(@bin)).
-comp @numsm= (1:nrow(@bin))(@ones,:).
-comp @n_1= csum(@bin)=1.
-comp @d= @d+1.
-call setdiag(@d,0).
!ifend
loop @i= 1 to ncol(@bin).
-comp @bini= @bin(:,@i).
-comp @dbini= @d&*t(@bini)(@ones,:).
-comp @dneii= rmax(@dbini).
-comp @dtog(:,@i)= @dneii.
 !if (!ind) !then
- do if @n_1(@i).
-  comp @ind(:,@i)= make(nrow(@bin),1,cmax(@bini&*@nums)).
- else.
-  comp @ind(:,@i)= rmax((@dbini=(@dneii*t(@ones)))&*@numsm).
- end if.
 !ifend
end loop.
comp !5= make(@k,@k,0).
!if (!ind) !then
-comp @numsm= make(@k,@k,0).
-comp @2nd= make(@k,1,0).
-comp @dtog= @dtog+1.
!ifend
loop @i= 1 to @k.
-comp @d= make(1,@k-@i+1,1).
-comp @dbini= @bin(:,@i)*@d&*@dtog(:,@i:@k).
-comp @dneii= cmax(@dbini).
-comp !5(@i,@i:@k)= @dneii.
 !if (!ind) !then
- comp @dbini= @dbini=@dneii(@ones,:).
- comp @n_1= cmax(@nums*@d&*@dbini).
- comp @numsm(@i,@i:@k)= @n_1.
- comp @2nd(@i)= @n_1(1).
- comp @indi= @ind(:,@i:@k).
- comp @numsm(@i:@k,@i)= t(cmax((@indi<>@n_1(@ones,:))&*@dbini&*@indi)).
 !ifend
end loop.
comp !5= !5+t(!5)-mdiag(diag(!5)).
!if (!ind) !then
-comp @numsm= @numsm+t(@numsm)&*(not @numsm).
-comp @dg= diag(@numsm).
-call setdiag(@numsm,@dg+@2nd&*(not @dg)).
-comp !5= {!5-1-(!5>1),@numsm,@2nd}.
-release @ind,@indi,@nums,@numsm,@n_1,@2nd,@dg.
!ifend
release @d,@bin,@k,@dtog,@ones,@i,@bini,@dbini,@dneii.
!enddefine.


******** DISTANCES TO GROUP NEAREST NEIGHBOURS (COMPUTATION FROM DISSIMILARITY MATRIX) ********.
*/*!KO_dtonfrd(dis%bin%alsoind%name)*/*.
*Version 1.
*Distance of every point to its nearest neighbour of each of the offered groups.
*This function is same as function /*!KO_dtoffrd*/ but it outputs distances to nearest, not farthest,
*neighbours.

define !KO_dtonfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @d= !2.
comp @bin= !3.
!let !ind= !unquote(!4)
comp @maxd1= mmax(@d)+1.
comp @d= @maxd1-@d.
call setdiag(@d,0).
comp !5= make(nrow(@bin),ncol(@bin),0).
comp @ones= make(nrow(@bin),1,1).
!if (!ind) !then
-comp @ind= make(nrow(@bin),ncol(@bin),0).
-comp @nums= t(1:nrow(@bin)).
-comp @numsm= (1:nrow(@bin))(@ones,:).
-comp @n_1= csum(@bin)=1.
!ifend
loop @i= 1 to ncol(@bin).
-comp @bini= @bin(:,@i).
-comp @dbini= @d&*t(@bini)(@ones,:).
-comp @dneii= rmax(@dbini).
-comp !5(:,@i)= @dneii.
 !if (!ind) !then
- do if @n_1(@i).
-  comp @ind(:,@i)= make(nrow(@bin),1,cmax(@bini&*@nums)).
- else.
-  comp @ind(:,@i)= rmax((@dbini=(@dneii*t(@ones)))&*@numsm).
- end if.
 !ifend
end loop.
comp !5= @maxd1&*(!5>0)-!5.
!if (!ind) !then
-comp !5= {!5,@ind}.
-release @ind,@nums,@numsm,@n_1.
!ifend
release @d,@bin,@ones,@i,@dbini,@dneii,@maxd1.
!enddefine.


******** DISTANCES BETWEEN GROUP NEAREST NEIGHBOURS (COMPUTATION FROM DISSIMILARITY MATRIX) ********.
*/*!KO_dbwnfrd(dis%bin%alsoind%name)*/*.
*Version 1.
*Single-linkage distances between groups, i.e. minimal distance between points of every two groups.
*This function is same as function /*!KO_dbwffrd*/ but it outputs distances between nearest, not farthest,
*neighbours.

define !KO_dbwnfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @d= !2.
comp @bin= !3.
comp @k= ncol(@bin).
!let !ind= !unquote(!4)
comp @maxd1= mmax(@d)+1.
comp @d= @maxd1-@d.
call setdiag(@d,0).
comp @dtog= make(nrow(@bin),ncol(@bin),0).
comp @ones= make(nrow(@bin),1,1).
!if (!ind) !then
-comp @ind= make(nrow(@bin),ncol(@bin),0).
-comp @nums= t(1:nrow(@bin)).
-comp @numsm= (1:nrow(@bin))(@ones,:).
-comp @n_1= csum(@bin)=1.
!ifend
loop @i= 1 to ncol(@bin).
-comp @bini= @bin(:,@i).
-comp @dbini= @d&*t(@bini)(@ones,:).
-comp @dneii= rmax(@dbini).
-comp @dtog(:,@i)= @dneii.
 !if (!ind) !then
- do if @n_1(@i).
-  comp @ind(:,@i)= make(nrow(@bin),1,cmax(@bini&*@nums)).
- else.
-  comp @ind(:,@i)= rmax((@dbini=(@dneii*t(@ones)))&*@numsm).
- end if.
 !ifend
end loop.
comp !5= make(@k,@k,0).
!if (!ind) !then
-comp @numsm= make(@k,@k,0).
-comp @2nd= make(@k,1,0).
-comp @dtog= @dtog+1.
!ifend
loop @i= 1 to @k.
-comp @d= make(1,@k-@i+1,1).
-comp @dbini= @bin(:,@i)*@d&*@dtog(:,@i:@k).
-comp @dneii= cmax(@dbini).
-comp !5(@i,@i:@k)= @dneii.
 !if (!ind) !then
- comp @dbini= @dbini=@dneii(@ones,:).
- comp @n_1= cmax(@nums*@d&*@dbini).
- comp @numsm(@i,@i:@k)= @n_1.
- comp @2nd(@i)= @n_1(1).
- comp @indi= @ind(:,@i:@k).
- comp @numsm(@i:@k,@i)= t(cmax((@indi<>@n_1(@ones,:))&*@dbini&*@indi)).
 !ifend
end loop.
comp !5= !5+t(!5)-mdiag(diag(!5))-!ind.
comp !5= @maxd1&*(!5>0)-!5.
!if (!ind) !then
-comp @numsm= @numsm+t(@numsm)&*(not @numsm).
-comp @dg= diag(@numsm).
-call setdiag(@numsm,@dg+@2nd&*(not @dg)).
-comp !5= {!5,@numsm,@2nd}.
-release @ind,@indi,@nums,@numsm,@n_1,@2nd,@dg.
!ifend
release @d,@bin,@k,@dtog,@ones,@i,@bini,@dbini,@dneii,@maxd1.
!enddefine.


******** DISTANCES TO GROUP MEDOIDS (COMPUTATION FROM DISSIMILARITY MATRIX) ********.
*/*!KO_dtomfrd(dis%bin%name1%name2)*/*.
*Version 1.
*Distance of every point to the medoid of each of the offered groups.
*(Medoid is the point in a group of points whose sum of distances to the group's points is minimal.)
*Takes square symmetric matrix of distances (dissimilarities) between points DIS.
*It must be nonnegative matrix with zero diagonal.
*And takes no less than two binary (1 vs 0) variables forming matrix BIN with number of rows as in DIS.
*Each column of BIN represents a group, and value 1 in it means that the point belongs to it.
*In whole, groups may intersect or be disjoint by the composition of their points (in the latter case
*BIN appears to be a set of dummy variables). Each column of BIN must contain at least one one.
*A column may, as a specific case, contain no zeros.
*Result NAME1 - matrix points x groups containing distances between the points and the group medoids.
*NAME2 is a row showing indices of points that are the medoids in their groups.
*If more than one point in a group appear to be medoids (such as, for example, in a group of 2 points),
*the function appoints the point with the greater index to be the group's medoid.
*The matrix of distances between group medoids is DIS(NAME2,NAME2).
*If your matrix DIS are similarities convert them first to dissimilarities by any method you like.

define !KO_dtomfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @d= !2.
comp @bin= !3.
comp !4= @d*@bin.
comp !4= !4+(not @bin)*1E300.
comp !4= !4=cmin(!4)(make(nrow(@bin),1,1),:).
comp !5= cmax(t(1:nrow(!4))*make(1,ncol(!4),1)&*!4).
comp !4= @d(:,!5).
release @d,@bin.
!enddefine.


******** WITHIN-GROUP SUMS OF SQUARES OF DEVIATIONS (COMPUTATION FROM DISTANCE MATRIX) ********.
*/*!KO_sswfrd(dis%bin%name)*/*.
*Version 1.
*Takes square symmetric matrix of distances between points DIS. The distances must be
*squared euclidean; anyway the function will perceive them as squared euclidean.
*And takes one or more binary (1 vs 0) variables forming matrix BIN with number of rows as in DIS.
*Each column of BIN represents a group, and value 1 in it means that the point belongs to it.
*In whole, groups may intersect or be disjoint by the composition of their points (in the latter case
*BIN appears to be a set of dummy variables). Each column of BIN must contain at least one one.
*A column may, as a specific case, contain no zeros.
*Result NAME - column lenghted as the number of columns in BIN, showing sum of squares of deviations
*from centroid (SSW) in each of these groups.

define !KO_sswfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @m= !2.
comp @bin= !3.
comp !4= diag(t(@bin)*@m*@bin)/(2*t(csum(@bin))).
release @m,@bin.
!enddefine.


******** INTRACLASS CORRELATION COEFFICIENT (ONE-WAY) ********.
*/*!KO_iccow(col%dummy%type%name)*/*.
*Version1.
*Computes intraclass correlation (ICC) in situation of one-way random design: scalar NAME.
*Classes may be unequal-size (unbalanced).
*COL - column of quantitative data.
*DUMMY - binary dummy (indicator) variables (at least two) marking classes (groups): each variable
*corresponds to a group, with value 1 = the case belongs to it, 0 = doesn't belong to it.
*Each row of DUMMY must contain exactly one 1.
*Each column in DUMMY must contain at least two 1s (with TYPE=PAIR2 or PAIR3), or at least one 1
*and at least one column more than one 1s (with other TYPE).
*TYPE - ICC estimator version, keyword in capital letters (optionally in quotes or apostrophes)
*[see Donner, 1986: A review of inference procedures...; Shieh, 2012: A comparison of two indices...]:
*"ANOVA" - ANOVA estimator.
*"CETASQ" - Corrected Eta-square estimator.
*"PAIR1" - Combinatorial pairing r Pearson estimator, without class weighting, classes of greater
*size will informationally over-prevail.
*"PAIR2" - as above, but with weighted classes: classes are given equal weight of importance irrespective
*of their unequal size.
*"PAIR3" - as above with intermediate, compromise variant of weighting: large classes will moderately
*prevail.
*If all classes (groups) are of same size (i.e. balanced) then PAIR1=PAIR2=PAIR3=CETASQ.
*NAME from data may sometimes be negative whereas ICC in population is nonnegative; you may follow
*the advice to zero-off negative values.
*You can create dummy variables DUMMY out of a categorical grouping variable
*with the help of function /*!KO_freq*/.

define !KO_iccow(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @col= !2.
comp @des= !3.
!let !type= !unquote(!4)
comp @n= csum(@des).
!if (!type='ANOVA' !or !type='CETASQ') !then
-comp @sum= t(@col)*@des.
-comp @ssq= t(@col&*@col)*@des.
-comp @ssw= rsum(@ssq-@sum&*@sum/@n).
-comp @sst= cssq(@col)-csum(@col)&**2/nrow(@des).
-comp @ssb= @sst-@ssw.
-comp @n_= (nrow(@des)-rssq(@n)/nrow(@des))/(ncol(@des)-1).
 !if (!type='ANOVA') !then
- comp @ssw= @ssw/(nrow(@des)-ncol(@des)).
- comp @ssb= @ssb/(ncol(@des)-1).
- comp !5= (@ssb-@ssw)/(@ssb+(@n_-1)*@ssw).
 !else
- comp !5= (@ssb/@sst-1/@n_)/(1-1/@n_).
 !ifend
-release @sum,@ssq,@ssw,@ssb,@sst,@n_.
!else
 !if (!type='PAIR1') !then
- comp @w= 1/rsum(@n&*(@n-1)).
 !else !if (!type='PAIR2') !then
- comp @w= 1/(ncol(@des)*@n&*(@n-1)).
 !else
- comp @w= 1/(nrow(@des)*(@n-1)).
 !ifend !ifend
-comp @gsum= t(@col)*@des.
-comp @gmean= @gsum/@n.
-comp @mean= rsum(@gsum&*(@n-1)&*@w).
-comp @col= @col-@mean.
-comp @wss= t(@col&*@col)*@des&*@w.
-comp !5= (rsum((@gmean-@mean)&**2&*@n&**2&*@w)-rsum(@wss))/rsum(@wss&*(@n-1)).
-release @w,@gsum,@gmean,@mean,@wss.
!ifend
release @col,@des,@n.
!enddefine.


******** INTRACLASS CORRELATION COEFFICIENT (ONE-WAY, BALANCED CLASSES) ********.
*/*!KO_iccowb(data%type%name)*/*.
*Version1.
*This function computes same ICC as function /*!KO_iccow*/, but it is for particular case of
*equal-size (balanced) classes only. The input here must be shaped in the "wide" format:
*DATA is a data matrix (not less than 2x2) classes x measures, i.e. each class is a row, all rows
*containing the same number of values. One shouldn't apply to DATA columns the meaning of regular
*raters because in the current case one-way random design is assumed.
*TYPE - ICC estimator version, keyword in capital letters (optionally in quotes or apostrophes)
*[see Donner, 1986: A review of inference procedures...]
*"ANOVA" - ANOVA estimator.
*"PAIR" - Combinatorial pairing r Pearson estimator.
*Because groups are of equal size, pairing estimator is identical to the Corrected Eta-square and
*to maximum likelihood estimators, and it is unbiased, in contrast to the ANOVA estimator.
*NAME from data may sometimes be negative whereas ICC in population is nonnegative; you may follow
*the advice to zero-off negative values.

define !KO_iccowb(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
!let !type= !unquote(!3)
comp @n= ncol(@data).
comp @sum= msum(@data).
comp @mean= @sum/(nrow(@data)*@n).
comp @sst= (mssq(@data)-@sum&*@mean).
comp @gmean= rsum(@data)/@n.
comp @ssb= cssq(@gmean-@mean)*@n.
!if (!type='PAIR') !then
-comp !4= 1/(@n-1)*(@n*@ssb/@sst-1).
!else
-comp @ssw= @sst-@ssb.
-comp @ssw= @ssw/(nrow(@data)*@n-nrow(@data)).
-comp @ssb= @ssb/(nrow(@data)-1).
-comp !4= (@ssb-@ssw)/(@ssb+(@n-1)*@ssw).
-release @ssw.
!ifend
release @data,@n,@sum,@mean,@gmean,@sst,@ssb.
!enddefine.


******** COEFFICIENT OF DIAGONALNESS OF MATRIX ********.
*/*!KO_diagns(mat%name)*/*.
*Version 1.
*One of possible coefficients measuring the degree of concentration of large (by absolute magnitude)
*values near its main diagonal. Ranges between 0 and 1.
*Let MAT be an arbitrary matrix, then the coefficient is 1-D/Dmax, where D is the weighted Manhattan
*physical distance of elements from the main diagonal, equal Sum[abs(MAT)&*OFF], where OFF is
*matrix sized as MAT, wherein the main diagonal comprised 0s, the pair of diagonals next to it
*(i.e. apart of it by order 1) are comprised of 1s, the pair of diagonals apart of it by order 2
*are comprised of 2s, and so forth. (&* is multiplication of correspondent elements, and Sum is the sum
*of elements.)
*Dmax is the maximally possible distance for the given MAT, i.e. observed when all the large elements
*of MAT are concentrated at bottom-left and top-right corners of the matrix; Dmax is computed as D,
*only after sorting all MAT values by absolute magnitude and sorting OFF values by their magnitude.
*If MAT is not square then not one but several diagonals in it pretend to be called "main diagonal".
*These are all diagonals (going top-left to bottom-right) of the same, maximal length. For example,
*in a matrix sized 4x6 it is three diagonals starting respectively from positions (1,1), (1,2) and (1,3).
*The function will return, respectively, three coefficients, one pertaining to each diagonal: result NAME
*is the column vector containing as much coefficients as there "main diagonals" embedding in the matrix,
*and the 1st coefficient pertains to the diagonal starting from position (1,1).
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_diagns(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= abs(!2).
do if nrow(@mat)>ncol(@mat).
-comp @mat= t(@mat).
end if.
comp @matv= reshape(@mat,1,nrow(@mat)*ncol(@mat)).
comp @matv(grade(@matv))= @matv.
comp @off= make(nrow(@mat),1,1)*(1:ncol(@mat))-t(1:nrow(@mat))*make(1,ncol(@mat),1).
comp !3= make(abs(nrow(@mat)-ncol(@mat))+1,1,0).
loop @i= 1 to nrow(!3).
-comp @offi= abs(@off-@i+1).
-comp @offiv= reshape(@offi,1,nrow(@off)*ncol(@off)).
-comp @offiv(grade(@offiv))= @offiv.
-comp !3(@i)= msum(@mat&*@offi)/rsum(@matv&*@offiv).
end loop.
comp !3= 1-!3.
release @mat,@matv,@off,@offi,@offiv,@i.
!enddefine.


******** CENTRATION OF DATA COLUMNS ********.
*/*!KO_center(data%name)*/*.
*Version 1.
*Centers columns of DATA (brings mean in them to 0).
*Returns data NAME.

define !KO_center(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
!enddefine.


******** Z-STANDARDIZATION OF DATA COLUMNS ********.
*/*!KO_zscore(data%name)*/*.
*Version 1.
*Standardizes columns of DATA (brings mean in them to 0 and variance to 1).
*Returns data NAME.

define !KO_zscore(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
comp !3= !3/sqrt(cssq(!3)/(nrow(!3)-1))(make(1,nrow(!3),1),:).
!enddefine.


******** Z-STANDARDIZATION OF DATA COLUMNS (DF=N) ********.
*/*!KO_zscore2(data%name)*/*.
*Version 1.
*Standardizes columns of DATA (brings mean in them to 0 and variance to 1).
*Standard deviation standardizing the data is computed on "df=n" (not "df=n-1").
*Returns data NAME.

define !KO_zscore2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
comp !3= !3/sqrt(cssq(!3)/nrow(!3))(make(1,nrow(!3),1),:).
!enddefine.


******** SCALING OF DATA COLUMNS TO SS=1 ********.
*/*!KO_scale(data%name)*/*.
*Version 1.
*Scales columns of DATA (brings sum of squares of values, SS, to 1).
*Returns data NAME.
*If need to make SS=num_rows DATA, multiply the result by sqrt(num_rows).

define !KO_scale(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3/sqrt(cssq(!3))(make(1,nrow(!3),1),:).
!enddefine.


******** CENTRATION OF DATA COLUMNS, BY GROUPS ********.
*/*!KO_gcenter(data%dummy%name)*/*.
*Version 1.
*Centers columns of DATA (brings mean in them to 0), by groups of cases.
*Returns data NAME.
*DUMMY - binary dummy (indicator) variables marking groups (each variable corresponds to a group,
*with value 1 = the case belongs to it, 0 = doesn't belong to it).
*Each column in DUMMY must contain at least one 1.
*The groups must be disjoint by membership: sum in each row in DUMMY must not exceed 1.
*If DUMMY has rows complete of zeros, i.e. cases out of the groups, the data for the cases remain intact.
*You can create dummy variables DUMMY out of a categorical grouping variable
*with the help of function /*!KO_freq*/.

define !KO_gcenter(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @des= !3.
comp @m= t(@des)*!4/(t(csum(@des))*make(1,ncol(!4),1)).
comp !4= !4-@des*@m.
release @des,@m.
!enddefine.


******** Z-STANDARDIZATION OF DATA COLUMNS, BY GROUPS ********.
*/*!KO_gzscore(data%dummy%df%name)*/*.
*Version 2.
*Standardizes columns of DATA (brings mean in them to 0 and variance to 1), by groups of cases.
*Returns data NAME.
*DUMMY - binary dummy (indicator) variables marking groups (each variable corresponds to a group,
*with value 1 = the case belongs to it, 0 = doesn't belong to it).
*The groups must be disjoint by membership, and each case must belong to a group,
*i.e. sum in each row of DUMMY must be equal to 1.
*DF - scalar. If positive, the standardizing  st. deviation will be computed on "df=n-1"
*(then in each column of DUMMY there must be at least two ones); if non-positive, the st. deviation
*will be computed on "df=n" (in each column of DUMMY there must be at least one one).
*You can create dummy variables DUMMY out of a categorical grouping variable
*with the help of function /*!KO_freq*/.

define !KO_gzscore(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @des= !3.
comp @df= !4.
comp @n= t(csum(@des))*make(1,ncol(!5),1).
comp @m= t(@des)*!5/@n.
comp !5= !5-@des*@m.
do if @df.
-comp @sd= sqrt(t(@des)*(!5&*!5)/(@n-1)).
else.
-comp @sd= sqrt(t(@des)*(!5&*!5)/@n).
end if.
comp !5= !5/(@des*@sd).
release @des,@df,@n,@m,@sd.
!enddefine.


******** SCALING OF DATA COLUMNS, BY GROUPS ********.
*/*!KO_gscale(data%dummy%scale%name)*/*.
*Version 1.
*Scales columns of DATA, by groups of cases.
*Returns data NAME.
*SCALE - scalar. If nonpositive, sum of squares of variable values in each group will be
*brought to 1; but if positive, then to the frequency in a group (i.e. mean square will be 1).
*DUMMY - binary dummy (indicator) variables marking groups (each variable corresponds to a group,
*with value 1 = the case belongs to it, 0 = doesn't belong to it).
*Each column in DUMMY must contain at least one 1.
*The groups must be disjoint by membership, and each case must belong to a group,
*i.e. sum in each row of DUMMY must be equal to 1.
*You can create dummy variables DUMMY out of a categorical grouping variable
*with the help of function /*!KO_freq*/.

define !KO_gscale(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @des= !3.
comp @sc= !4.
comp @sqrn= 1.
do if @sc.
-comp @sqrn= t(sqrt(csum(@des)))*make(1,ncol(!5),1).
end if.
comp @norm= sqrt(t(@des)*(!5&*!5)).
comp !5= !5&*(@des*(@sqrn/@norm)).
release @des,@sc,@sqrn,@norm.
!enddefine.


******** LINEAR RESCALING ********.
*/*!KO_rescale(data%type%val1%val2%name)*/*.
*Version 1.
*Linearly rescales variables in the dataset (columns of DATA) to the wanted magnitudes of statistics
*in these data such as minimum, maximum, mean, st. deviation.
*Data DATA must be of no less than two rows and the variables (columns) must be not constants.
*Each variable is being rescaled independently; result NAME will have in each column
*the linearly transformed data exactly satisfying two specified values VAL1 and VAL2 of the two
*corresponding statistics according to TYPE.
*VAL1 and VAL2 - scalars or row vectors lengthed as the number of columns of DATA; vector(s) mean
*that the variables need to be brought each to its own specified value of the statistic.
*TYPE (keyword in capital letters, optionally quoted or in apostrophes) sets which two
*statistics the data in the variables are to be brought to:
*"MIN_MAX" - to specified minimum (VAL1) and maximum (VAL2), (VAL1<VAL2)
*"MIN_MEAN" - to specified minimum (VAL1) and mean (VAL2), (VAL1<VAL2)
*"MAX_MEAN" - to specified maximum (VAL1) and mean (VAL2), (VAL1>VAL2)
*"MIN_SD" - to specified minimum (VAL1) and standard deviation (VAL2), (VAL2 positive)
*"MAX_SD" - to specified maximum (VAL1) and standard deviation (VAL2), (VAL2 positive)
*"MEAN_SD" - to specified mean (VAL1) and standard deviation (VAL2), (VAL2 positive).

define !KO_rescale(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')')) 
!let !type= !unquote(!3)
comp !6= !2.
comp @val1= !4.
comp @val2= !5.
comp @ones= make(1,nrow(!6),1).
do if ncol(@val1)=1.
-comp @val1= make(1,ncol(!6),@val1).
end if.
do if ncol(@val2)=1.
-comp @val2= make(1,ncol(!6),@val2).
end if.
!if (!type='MIN_MAX') !then
-comp @max= cmax(!6).
-comp !6= ((@val2-@val1)/(@max-cmin(!6)))(@ones,:)&*(!6-@max(@ones,:))+@val2(@ones,:).
-release @max.
!else !if (!type='MIN_MEAN') !then
-comp !6= !6-(csum(!6)/nrow(!6))(@ones,:).
-comp !6= ((@val1-@val2)/cmin(!6))(@ones,:)&*!6+@val2(@ones,:).
!else !if (!type='MAX_MEAN') !then
-comp !6= !6-(csum(!6)/nrow(!6))(@ones,:).
-comp !6= ((@val1-@val2)/cmax(!6))(@ones,:)&*!6+@val2(@ones,:).
!else
-comp !6= !6-(csum(!6)/nrow(!6))(@ones,:).
-comp !6= !6/(sqrt(cssq(!6)/(nrow(!6)-1))(@ones,:)).
 !if (!type='MIN_SD') !then
- comp !6= (!6-cmin(!6)(@ones,:))&*@val2(@ones,:)+@val1(@ones,:).
 !else !if (!type='MAX_SD') !then
- comp !6= (!6-cmax(!6)(@ones,:))&*@val2(@ones,:)+@val1(@ones,:).
 !else
- comp !6= !6&*@val2(@ones,:)+@val1(@ones,:).
 !ifend !ifend
!ifend !ifend !ifend
release @val1,@val2,@ones.
!enddefine.


******** CONVERT COVARIANCE MATRIX INTO CORRELATION ********.
*/*!KO_covcorr(cov%name)*/.
*Version 1.
*Converts covariance matrix into correlation one (or sscp matrix into cosine similarities).
*Obtained matrix NAME has ones on the diagonal.
*Input matrix COV must have positive diagonal elements.

define !KO_covcorr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp @vrnc= diag(!3).
comp !3= !3/sqrt(@vrnc*t(@vrnc)).
release @vrnc.
!enddefine.


******** CONVERT CORRELATION MATRIX INTO COVARIANCE ********.
*/*!KO_corrcov(corr%dg%name)*/.
*Version 1.
*Converts correlation matrix into covariance (or cosine similarity matrix into sscp matrix).
*Obtained matrix NAME has vector DG on the diagonal.
*Input matrix CORR must have ones on the diagonal.
*DG - column vector of positive values, lengthed as the number of rows/columns of CORR.

define !KO_corrcov(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @vrnc= !3.
comp !4= !4&*sqrt(@vrnc*t(@vrnc)).
release @vrnc.
!enddefine.


******** DOUBLE CENTERING OF MATRIX OF SQUARED DISTANCES ********.
*/*!KO_dcenter(dis%name)*/*.
*Version 1.
*Converts matrix of squared dissimilarities DIS into matrix of similarities NAME which are scalar products.
*If DIS is squared dissimilarities between points of a cloud, then on the diagonal of NAME are squared
*euclidean distances from the points to the centroid of the cloud, and off-diagonal elements are scalar
*products between vectors shed from the centroid and ending on the points.
*If the dissimilarities are squared euclidean distances then this operation is a "geometrically correct"
*transformation them into similarities. Namely, then NAME=X*t(X), where matrix X is any data (points by
*dimensions, and the dimensions, the columns, are centered) having between the points the sq. euclidean
*distances equal to DIS.
*In double-centered matrix NAME row and column sums equal 0.
*DIS must be square symmetric with non-negative elements and with 0 on the diagonal.
*For back conversion use the opposite function /*!KO_sdcosth*/.

define !KO_dcenter(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp @rmean= (rsum(!3)/ncol(!3))*make(1,ncol(!3),1).
comp !3= (@rmean+t(@rmean)-!3-msum(!3)/ncol(!3)**2)/2.
release @rmean.
!enddefine.


******** CONVERT MATRIX OF ANGULAR SIMILARITIES INTO SQUARED DISTANCES BY COSINE THEOREM ********.
*/*!KO_sdcosth(sim%name)*/*.
*Version 1.
*Converts matrix of similarities SIM, which it takes as scalar products, into matrix of squared
*dissimilarities NAME. Diagonal values of SIM are used as the squares of the distances from the points to the space
*origin. If SIM are scalar products and SIM is positive semidefinite then dissimilarities NAME are
*squared euclidean distances. If SIM is not positive semidefinite then NAME may be sometimes
*squared euclidean distances (converge in euclidean space).
*This operation is a "geometrically correct" transformation of covariances, correlations or cosine similarities into
*euclidean distances corresponding to them.
*SIM must be square symmetric with positive values on the diagonal.
*This function is opposite to functions /*!KO_dcenter*/ or /*!KO_dscosth*/.

define !KO_sdcosth(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')')) 
comp !3= !2.
comp @hsq= diag(!3)*make(1,ncol(!3),1).
comp !3= @hsq+t(@hsq)-2*!3.
release @hsq.
!enddefine.


******** SWITCHING MATRIX OF SCALAR PRODUCTS ONTO NEW DIAGONAL (PRESERVING ANGLE) ********.
*/*!KO_swdiag1(sim%diag%name)*/.
*Version 1.
*Recalculates elements of SSCP-type matrix (= matrix of scalar products; in particular,
*covariance, correlation, cosine similarity matrices belong to) onto the given diagonal values.
*At that, angles (cosines) between the vectors are kept unchanged.
*With this function one can, for example, turn a covariance matrix into a correlation or vice versa.
*SIM must be square symmetric, with positive elements on the diagonal. The matrix is treated as
*a matrix of similarities of SSCP type (a matrix of scalar products).
*DIAG - vector (row or column) lengthed as the number of rows/columns in SIM, containing positive
*elements; it is the new diagonal for SIM.
*Result NAME is SIM with diagonal DIAG and recalculated off-diagonal elements.
*If the input SIM is gramian (all eigenvalues nonnegative), NAME is also gramian.

define !KO_swdiag1(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @diag= !3.
do if nrow(@diag)=1.
-comp @diag= t(@diag).
end if.
comp @coef= sqrt(@diag/diag(!4)).
comp !4= !4&*(@coef*t(@coef)). /*Ск произведение sp = h1*h2*cos;
          /*ты хочешь новое sp' с тем же косинусом, но новыми h; поэтому sp' = (h1'h2')/(h1h2) * sp
call setdiag(!4,@diag).
release @diag,@coef.
!enddefine.


******** SWITCHING MATRIX OF SCALAR PRODUCTS ONTO NEW DIAGONAL (PRESERVING DISTANCE) ********.
*/*!KO_swdiag2(sim%diag%name)*/.
*Version 1.
*Recalculates elements of SSCP-type matrix (= matrix of scalar products; in particular,
*covariance, correlation, cosine similarity matrices belong to) onto the given diagonal values.
*At that, euclidean distances between the vectors' endpoints are kept unchanged.
*SIM must be square symmetric, with positive elements on the diagonal. The matrix is treated as
*a matrix of similarities of SSCP type (a matrix of scalar products).
*DIAG - vector (row or column) lengthed as the number of rows/columns in SIM, normally containing 
*positive elements; it is the new diagonal for SIM.
*Result NAME is SIM with diagonal DIAG and recalculated off-diagonal elements.
*Depending on DIAG (and, of course, on the input SIM), NAME may appear gramian
*(all eigenvalues nonnegative) or nongramian.

define !KO_swdiag2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @diag= !3.
do if nrow(@diag)=1.
-comp @diag= t(@diag).
end if.
comp @diff= (@diag-diag(!4))*make(1,ncol(!4),1).
comp !4= (@diff+t(@diff)+2*!4)/2. /*Теорема кос: d^2 = h1^2 + h2^2 - 2sp; d2 мы хотим не менять, поэтому:
           /*d^2 = h1^2 + h2^2 - 2sp = h1'^2 + h2'^2 - 2sp', где апострофом помечены новые значения диаг эл-тов (h^2) и внедиагональных (sp, т е ск произв)
           /*Отсюда, sp' = [(h1'^2 - h1^2) + (h2'^2 - h2^2) + 2sp] / 2
call setdiag(!4,@diag). /*Ну, а диагональ, ради точности, можно прямо подставить (иначе калькуляторно она может выйти в 14-15 дес цифре неточной)
release @diag,@diff.
!enddefine.


******** SWITCHING MSCP MATRIX ONTO NEW DATA CENTROID ********.
*/*!KO_swcentr(sim%oldmean%newmean%name)*/.
*Version 1.
*Recalculates elements of MSCP matrix defined by columns of some data onto a new centroid, that is,
*as would be if those data are shifted onto that new centroid (multivariate mean). The function does not
*need the original data. Say, a MSCP matrix (it is SSCP-матрица - i.e. matrix t(data)*data - devided by
*the number of cases in the data) was computed for some variables, as well as the vector of their means.
*Then the data were lost. What would be the new MSCP matrix after the data would have been changed their
*means arbitrarily by adding some constant to each variable? The function answers this question in the
*absence data themselves. For example this function can restore the MSCP matrix of the original data
*from the covariance matrix (which, as we know, is a MSCP matrics too) if the original means of the
*variables are known.
*SIM - square symmetric MSCP matrix between coluns of some data; means in those columns are recorded
*as vector (row or column) OLDMEAN. After shifting of this centroid onto new coordinates NEWMEAN
*the new MSCP matrix corresponding to the shifted data is just the matrix NAME which is returned.
*Be aware that the result matrix does not have to be gramian (positive [semi]-definite) -
*the gramianness will depend on the input matrix SIM and feasability of OLDMEAN and NEWMEAN. The function
*does not watch after realisticity of the inputs. When input statistics come from real variables there
*should be no problem with gramianness.

define !KO_swcentr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @omean= !3.
comp @nmean= !4.
do if ncol(@omean)=1.
-comp @omean= t(@omean).
end if.
do if ncol(@nmean)=1.
-comp @nmean= t(@nmean).
end if.
comp !5= !5+t(@nmean)*@nmean-t(@omean)*@omean.
release @omean,@nmean.
!enddefine.


******** EXPONENTIAL NORMALIZATION (SOFTMAX FUNCTION) ********.
*/*!KO_nef(data%name)*/*.
*Version 1.
*Exponentially normalizes elements in columns of DATA.
*Result NAME - in each column values are exponentially transformed so that
*they lie in the range [0,1] and their sum is equal to 1.

define !KO_nef(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp @epv= exp(!3).
comp !3= @epv/csum(@epv)(make(1,nrow(!3),1),:).
release @epv.
!enddefine.


******** CENTERED LOGRATIO TRANSFORM ********.
*/*!KO_clr(data%name)*/*.
*Version 1.
*Performs centered logratio transformation of elements in columns of DATA. Values in DATA
*must be positive.
*The transform consists first of taking column sum to 1, then division by
*geometric mean in column, and taking logarithm.
*Result NAME - values in each column have zero mean.

define !KO_clr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3/csum(!3)(make(1,nrow(!3),1),:).
comp @gmean= exp(csum(ln(!3))/nrow(!3)).
comp !3= ln(!3/@gmean(make(1,nrow(!3),1),:)).
release @gmean.
!enddefine.


******** BOX-COX TRANSFORM ********.
*/*!KO_boxcox(col%grid%name1%name2)*/*.
*Version 1.
*Transforms a quantitative variable COL monotonically by exponentiation so that its distribution
*become closer to normal. Selecting of the optimal value for the power parameter lambda is done by
*grid search the same way SPSS command ADP does it when performing Box-Cox transform.
*COL - input variable, column of positive values. If nonpositive values are present, please add
*a constant beforehand to make all values positive (as a variant, you may always, irrespective
*whether there are or there are no nonpositive data, do COMPUTE COL= COL-cmin(COL)+1, the way ADP
*command does, for example).
*GRID - vector (row or column) of three values - this is the grid to search a value of lambda:
*{min,max,step}. Max must be > Min, Step positive, and (Max-Min)/Step must give integer.
*For example, often used grid is {-3,3,0.5}.
*Result:
*NAME2 - the selected power lambda of the Box-Cox transform.
*NAME1 - the transformed by it data COL. (Command ADP additionally z-standardizes it, but this
*function omits doing it).
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_boxcox(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @p= !3.
comp @b_= (@p(2)-@p(1))/@p(3).
comp @p= (@p(2)-@p(1))/@b_*(0:@b_)+@p(1).
comp @n= nrow(!4).
comp @lny= ln(!4).
comp !4= !4*make(1,ncol(@p),1).
loop @j= 1 to ncol(@p).
-comp @pj= @p(@j).
-do if @pj=0.
- comp !4(:,@j)= @lny.
-else.
- comp !4(:,@j)= (!4(:,@j)&**@pj-1)/@pj.
-end if.
end loop.
comp @vrnc= !4-(csum(!4)/nrow(!4))(make(1,nrow(!4),1),:).
comp @vrnc= cssq(@vrnc)/(nrow(@vrnc)-1).
comp @loglik= -@n/2*ln((1-1/@n)*@vrnc)+(@p-1)&*csum(@lny).
comp @j= rmax((@loglik=rmax(@loglik))&*(ncol(@loglik):1)).
comp @j= (ncol(@loglik)+1)*(@j>0)-@j.
comp !5= @p(@j).
comp !4= !4(:,@j).
release @p,@b_,@n,@lny,@j,@pj,@vrnc,@loglik.
!enddefine.


******* WHITENING (SPHERING) MATRIX ********.
*/*!KO_white(cov%method%name)*/*.
*Version 1.
*Whitening or sphering is a name for transforming multivariate data into "spherical" cloud.
*The function takes covariance or correlation matrix COV and computes whitening (sphering) matrix 
*NAME such, so if there are some variables X corresponding to COV, then variables Z=X*NAME
*will have identity covariance matrix. Thus, whitened data are decorrelated and standardized
*variables, while the function returns the matrix to do such transformation.
*COV - square p x p symmetric matrix, covariance or correlation.
*METHOD - whitening method, keyword in capital letters (might take in quotes or apostrophes):
*"ZCA" - zero-phase component method; returns NAME (symmetric, with this method) minimizing squared
*differences between values of Z and X. Use this method if you need variables Z to depart minimally
*from their corresponding variables X.
*"PCA" - principal component method; returns NAME maximizing the account of variance of X in the first
*columns of Z; Z are actually the principal components of X, and Z values are the standardized principal
*component scores. Use this method if you may wish to reduce dimensionality (number of columns) in Z
*comparatively with X, with minimal informational loss.
*"CHOL" - Cholesky method; returns NAME (upper triangular, with this method) providing complete
*correlation of Z1 with X1 and orthogonality of Zp with all X except Xp. Use if it needs that Z1
*be associated with all Xs, especially X1; Z2 be not associated with X1; Z3 be not associated with X1
*and X2, etc.
*With methods "ZCA" and "PCA" whitened variables Z and their relation with the initial variables X will
*depend on whether COV is covariance or correlation matrix (so it matters if you input covariance or
*correlation COV into the function). Let matrix R be covariances (if COV is covariance) or correlations
*(if COV is correlation) between X variables (rows of R) and Z variables (columns of R). Then:
*-"ZCA" method maximizes diagonal elements of R (they're positive).
*-with "PCA" method, R is the matrix of principal component loadings, so the method maximizes in the
*first place the sum of squared elements of the 1st column in R (i.e. covariances/correlations of Z1 with
*all Xs), in the second place the sum of squared elements of the 2nd column in R (i.e. Z2 with
*all Xs), etc.
*With method "CHOL" whitened variables Z will be the same be COV covariance or correlation matrix
*(so it matters not if you input covariance or correlation COV into the function).
*This method leads to correlation matrix R such that it is lower triangular with positive diagonal
*and element (1,1)=1, while in its last column all elements except (p,p) are zero.
*It is important to note that for all the methods there must be followed the rule: if COV is correlation
*then at whitening Z=X*NAME columns of X must be standardized variables (not raw or just centered).
*Methods "ZCA" and "PCA" are connected together via data rotation so that ZCA whitening is equivalent
*to PCA whitening with subsequent procrustes rotation of Z back to X, and the matrix of such rotation is
*eigenvectors of COV.

define !KO_white(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @cov= !2.
!let !method= !unquote(!3)
!if (!method='CHOL') !then
-comp !4= inv(chol(@cov)).
!else
-call eigen(@cov,@eivec,@eival).
 !if (!method='PCA') !then
- comp !4= @eivec/sqrt(t(@eival))(make(1,ncol(@cov),1),:).
- comp @cov= t(diag(!4)>=0).
- comp !4= !4&*(@cov-(not @cov))(make(1,ncol(!4),1),:).
 !else
- comp !4= @eivec/sqrt(t(@eival))(make(1,ncol(@cov),1),:)*t(@eivec).
 !ifend
-release @eivec,@eival.
!ifend
release @cov.
!enddefine.


******** TRANSFORM VARIABLES EXACTLY TO HAVE SPECIFIED COVARIANCES ********.
*/*!KO_tocov(ingot%type%cov%method%name)*/*.
*Version 2.
*Transforming of multivariate quantitative data INGOT so that the observed covariances (or correlations
*or other sscp measures of association) between these variables become exactly equal to the
*user-specified ones.
*Argument INGOT - n cases x p variables (n>p) data matrix with values serving an "ingot", the stuff to
*obtain variates associations between which will satisfy matrix COV. INGOT could be randomly generated
*data that you want to take to the needed correlatedness, or INGOT could be existing, analysis data
*which you want to modify minimally and make them reproduce the specified correlatedness
*(see METHOD below). Columns in INGOT must be noncollinear.
*Argument TYPE - keyword in capital letters (optionally may put in quotes or apostrophes). Specify
*the type of the target matrix you input as argument COV: 
*"COV" - covariance matrix (on "df=n-1");
*"COV2" - covariance matrix (on "df=n") or mscp matrix;
*"CORR" - correlation matrix or cosine similarity matrix.
*(All the listed matrices are akin: they are matrices of a sscp(variables)/"df" kind.)
*Argument COV - p x p square symmetric positive definite matrix of associations (its type you marked
*in TYPE). It is this target matrix for which there must be obtained, out of INGOT, the variates
*satisfying its coefficients.
*Result NAME - n x p data where columns are such variates that they exactly reproduce the specified
*coefficients COV.
*Use the following rules to specify input:
*-If the target matrix COV is covariance one (on "df=n-1"), then type="COV", INGOT columns must be centered
* (one can center variables with function /*!KO_center*/).
*-If the target matrix COV is covariance one (on "df=n"), then type="COV2", INGOT columns must be centered.
*-If the target matrix COV is mean squares-and-crossproducts (mscp), then type="COV2", INGOT is raw data
* as is.
*-If the target matrix COV is correlation one (unit must be on the diagonal), then type="CORR", INGOT
* columns must be centered. At output, variables will retain their initial scale (variance).
*-If the target matrix COV is cosine similarity (unit must be on the diagonal), then type="CORR", INGOT is
* raw data as is. At output, variables will retain their initial scale (mean square).
*Argument METHOD - scalar, defines the fitting algorithm. If it is positive then at data transformation
*there will be applied component loadings of the target matrix procrustes-rotated towards component
*loadings of INGOT. Use this variant if you want the values of NAME to differ minimally in terms of
*overall sum of squares from the values of INGOT: mssq(INGOT-NAME)=min. In other words, if the task
*is to "correct" INGOT, taking the data to the wanted amount correlation. 
*If argument METHOD is nonpositive then at data transformation there will be applied Cholesky root of
*the target matrix. Use this more fast variant if you want simply to create data, i.e. when INGOT for
*you is just a "pig" of random values that were generated before. This variant does not seek to modify
*INGOT minimally.
*Note: if INGOT are values from normal distribution then NAME are data with also multivariate
*normal distribution; otherwise distribution in NAME does not have to relate to distribution 
*in INGOT.

define !KO_tocov(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
!let !type= !unquote(!3)
comp @mx= !4.
comp @method= !5.
comp @mscp= sscp(@data)/(nrow(@data)-!eval(!type='COV')).
comp @s= diag(@mscp).
!if (!type='CORR') !then
-comp @mx= @mx&*sqrt(@s*t(@s)).
!ifend
call eigen(@mscp,@v,@s).
comp @s= t(sqrt(@s))(make(1,ncol(@data),1),:).
comp @u= @data*(@v/@s).
do if @method.
-call eigen(@mx,@eivec,@eival).
-comp @mx= @eivec&*t(sqrt(@eival))(make(1,ncol(@data),1),:).
-release @eivec,@eival.
-call svd(t(@v&*@s)*@mx,@u_,@s_,@v_).
-comp @mx= t(@mx*@v_*t(@u_)).
-release @u_,@s_,@v_.
else.
-comp @mx= chol(@mx).
end if.
comp !6= @u*@mx.
release @data,@mx,@method,@mscp,@s,@v,@u.
!enddefine.


******** TRANSFORM A VARIABLE EXACTLY TO HAVE SPECIFIED COVARIANCES WITH OTHER VARIABLES ********.
*/*!KO_ytocov(yingot%xvars%type%cov%name)*/*.
*Version 1.
*Transforms YINGOT variable of the dataset so that its observed covariances (or correlations or
*other sscp association measures) with variables XVARS in the data will become exactly equal to
*the user-specified ones. Result NAME - the modified YINGOT, altered minimally in the sense of sums of
*squares: mssq(YINGOT-NAME)=min. XVARS will not be modified.
*Argument INGOT - column of n cases, it is the "ingot" quantitative variable for obtaining variable NAME.
*You may generate YINGOT as random values or it could be existing, analysis data.
*Argument XVARS - one or more quantitative variables in the dataset, n x p, these p columns must be
*noncollinear.
*Argument TYPE - keyword in capital letters (optionally may put in quotes or apostrophes). Specify
*the type of the target vector you input as COV argument:
*"COV" - covariances (on "df=n-1");
*"COV2" - covariances (on "df=n") or mean cross-products;
*"CORR" - correlations or cosine similarities.
*(All the listed types are akin: they are coefficients of the sscp(variables)/"df" kind.)
*Argument COV - row vector of the target coefficients themselves (their type you've marked in TYPE).
*It is the coefficients of association with XVARS which YINGOT will have to satisfy exactly when it
*is transformed to NAME. The length of COV row is p+1. First p values correspond respectively to
*the p columns (variables) XVARS, and the last additional value is the scale for variable NAME (see below).
*Use the following rules to specify input:
*-If your target coefficients COV are covariances (on "df=n-1"), then type="COV", the last element
* of COV vector must be the desired variance for NAME (positive value), YINGOT and XVARS columns must be
* centered (one can center variables with function /*!KO_center*/).
*-If your target coefficients COV are covariances (biased, on "df=n"), then type="COV2", the last element
* of COV vector must be the desired biased variance for NAME (positive value), YINGOT and XVARS columns
* must be centered.
*-If your target coefficients COV are mean crossproducts, then type="COV2", the last element
* of COV vector must be the desired mean square for NAME (positive value), YINGOT and XVARS are raw
* data as is.
*-If your target coefficients COV are correlations, then type="CORR", the last element of COV vector must
* be 1, YINGOT and XVARS columns must be centered. NAME will retain YINGOT's initial scale (variance).
*-If your target coefficients COV are cosine similarities, then type="CORR", the last element of COV vector
* must be 1, YINGOT and XVARS are raw data as is. NAME will retain YINGOT's initial scale (mean square).
*Note1: If the specified coefficients COV, their magnitudes, do not permit creation of a correct variable
*NAME out of variable YINGOT (because of non-gramian, not positive-definite configuration occuring)
*then the function will return NAME= scalar 0.
*Note2: if YINGOT are values from normal distribution then NAME are values also with normal
*distribution; otherwise distribution in NAME does not have to relate to distribution 
*in YINGOT.

define !KO_ytocov(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !type= !unquote(!4)
comp @y= !2.
comp @x= !3.
comp @mx= !5.
comp @p= ncol(@x).
comp @df= nrow(@x)-!eval(!type='COV').
comp @mscp= sscp({@x,@y})/@df.
comp @ms= diag(@mscp).
!if (!type='CORR') !then
-comp @mx= @mx&*t(sqrt(@ms*@ms(nrow(@ms)))).
!ifend
comp @res= @y-@x*solve(@mscp(1:@p,1:@p),@mscp(1:@p,ncol(@mscp))).
call eigen(@mscp(1:@p,1:@p),@v,@s).
comp @s= sqrt(@s).
comp @u= @x*(@v/t(@s)(make(1,@p,1),:)).
comp @xdual= @u/t(@s)(make(1,nrow(@x),1),:)*t(@v).
release @u,@s,@v,@x.
comp @mscp= sscp(@xdual)/@df.
comp @mxy= @mx(ncol(@mx)).
comp @mx= @mx(1:@p).
comp @s2= (@mxy-@mx*@mscp*t(@mx))/(cssq(@res)/@df).
do if @s2.
-comp !6= @xdual*t(@mx)+sqrt(@s2)*@res.
else.
-comp !6= 0.
end if.
release @y,@mx,@p,@df,@mscp,@ms,@res,@xdual,@s2.
!enddefine.


******** NORMAL CLOUD INTO UNIFORM BALL ********.
*/*!KO_unifball(data%alpha%name1%name2)*/*.
*Version 1.
*Takes uni- or- multivariate data DATA, which the function expects to be from
*the standard normal distribution. The function reduces the data kurtosis by pulling
*the periphery (tails) in stronger than the paracentral regions. Under parameter ALPHA=1
*the normal cloud turns into the uniform, into the (hyper)ball.
*DATA - data (cases x variables) from normal distribution with mean 0 and variance 1;
*if there are multiple variables, the function expects that they are independent, uncorrelated.
*(You can generate random standard orthogonal data by function /*!KO_normal*/,
*and to standardize variables, if necessary - by function /*!KO_zscore*/.)
*ALPHA - scalar, parameter of the degree of contraction of the periphery (tails) of data.
*The higher it is, the stronger will the function reduce kurtosis in the data.
*Under parameter 1 a normal cloud transforms to uniform one, and that will be uniform hyper(ball)
*if the data are multivariate. Under parameter 0 the data will remain as they were at input.
*With parameter between 0 and 1 you will obtain data with kurtosis in-between of normal and of
*uniform. Parameter greater than 1 makes data yet bimodal (concave in the middle).
*The formula of data transformation:
*Y_i = X_i * coef,
*coef = 1 + alpha * (sqrt(p+2)*cdf.chisq(d^2,p)^(1/p)/d - 1),
*where Y_i and X_i - values of data point i at output (NAME1) and at input (DATA);
*coefficient of contraction coef is returned as NAME2;
*p - dimensionality, number of columns of DATA;
*d - euclidean distance from data point i in DATA to the origin, i.e. to 0 (according to the said
*above, centroid of DATA should lie in 0 or approach 0);
*cdf.chisq(val,df) - chi-square cumulative distribution function;
*sqrt(p+2) factor preserves the same expected variability (i.e., 1 per dimension);
*and when ALPHA=1, the hyperball NAME1 expects radius sqrt(p+2).
*(See: https://stats.stackexchange.com/questions/79919).
*Correlating variables. Although the function will reduce kurtosis in such a cloud and will make
*the ellipsoid more ball-like, it will not attain uniformity of distribution under ALPHA=1. The
*function accomplishes its role only when the input variables are uncorrelated. If you want to
*convert correlated variables from normal distribution into ellipse with uniform distribution inside
*it, do first whitening of the data (sphering and standardization), for instance, by principal
*component analysis (function /*!KO_pcomp*/). Then apply the current function, and in the end
*do the transform backward of that principal component analysis - see EXAMPLE.

define !KO_unifball(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @alpha= !3.
comp @p= ncol(!4).
comp @d= rssq(!4).
comp !5= 1+@alpha*(chicdf(@d,@p)&**(1/@p)*sqrt(@p+2)/sqrt(@d)-1).
comp !4= !4&*(!5*make(1,@p,1)).
release @alpha,@p,@d.
!enddefine.


******** LINEAR REGRESSION ********.
*/*!KO_regress(dv%ivs%const%name1%name2%name3)*/*.
*Version 1.
*Linear regression OLS (without significance tests).
*Takes dependent variable DV and one or more independent variables IVS (which must be noncollinear).
*Argument CONST (scalar) - request model with constant (CONST positive) or without constant (CONST nonpositive).
*Results:
*NAME1 - SSregression, SSresiduals, observed coefficient of determination R-square.
*NAME2 - two-column matrix of regression coefficients: raw coefficients in the 1st col; standardized
*coefficients (beta) in the 2nd col. If model was with constant, number of rows in NAME2 is 1 more
*than the number of columns IVS, and the constant is in the last row.
*NAME3 - predicted values (raw) of the dependent variable.

define !KO_regress(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                  /!pos= !charend('%') /!pos= !charend(')'))
comp @dv= !2.
comp !7= !3.
comp @const= !4.
comp @p= ncol(!7).
comp @p1= @p+1.
do if @const.
-comp @mean= csum({!7,@dv})/nrow(@dv).
-comp @scp= sscp({!7,@dv}-@mean(make(1,nrow(!7),1),:)).
-comp !6= make(@p1,2,0).
else.
-comp @scp= sscp({!7,@dv}).
-comp !6= make(@p,2,0).
end if.
comp !6(1:@p,1)= solve(@scp(1:@p,1:@p),@scp(1:@p,@p1)).
comp @ss= diag(@scp).
comp !6(1:@p,2)= !6(1:@p,1)&*sqrt(@ss(1:@p)/@ss(@p1)).
comp !7= !7*!6(1:@p,1).
do if @const.
-comp !6(@p1,1)= @mean(@p1)-@mean(1:@p)*!6(1:@p,1).
-comp !7= !7+!6(@p1,1).
-comp !5= cssq(!7-@mean(@p1)(make(1,nrow(!7),1),:))/@ss(@p1).
-release @mean.
else.
-comp !5= cssq(!7)/@ss(@p1).
end if.
comp !5= {!5*@ss(@p1),@ss(@p1)-!5*@ss(@p1),!5}.
release @dv,@const,@p,@p1,@scp,@ss.
!enddefine.


******** PRINCIPAL COMPONENTS ********.
*/*!KO_pcomp(data%df%m%out%name1%name2%name3%name4%name5)*/*.
*Version 3.
*Principal component analysis.
*Takes data DATA (where rows are points, columns are space axes) and returns principal
*components of these data.
*Attention: The data are analyzed "as is", therefore if preliminary centration
*or standardization is needed then use first the corresponding functions.
*(Centration corresponds to the analysis of covariances and standardization - to the analysis of correlations.)
*Argument DF (scalar) - if positive, decomposed will be matrix sscp(DATA)/(n-1), if not positive then matrix
*sscp(DATA)/n, where n is the number of rows in DATA.
*Argument M (scalar) - how many 1st principal components to extract: specify positive integer from 1 to
*number_columns DATA. Or specify any nonpositive number: then all nonzero principal components will be extracted.
*Argument OUT (scalar) controls the form of results returned.
*Results:
*NAME1 contains in the 1st col eigenvalues of the decomposed matrix,
*and in the 2nd col - proportions of explained variance (or scale), i.e. proportions of the trace of the matrix.
*NAME2 - eigenvectors (if OUT nonpositive) or loadings (if OUT positive); sum-of-squares
*in columns of loadings = eigenvalues. Loadings are the coefficients of prediction of variables by scaled components.
*Eigenvectors are the coefficients of prediction of variables by raw components, or the cosines of angles of rotation.
*NAME3 - rescaled loadings (if OUT positive) or eigenvectors NAME2 (if OUT nonpositive).
*Rescaled or standardized loading is a loading divided by the scale of the variable - i.e. by the root of the corresponding
*diagonal element of the decomposed matrix. By this division the effect of uniqual scales (variances) of variables
*on the loadings is taken off. Rescaled loading is the cosine similarity or correlation between a component and a variable.
*Squared elements of eigenvectors can be understood as the magnitude of contributions of variables into components.
*Squared elements of rescaled loadings can be understood as the magnitude of contributions of components into variables.
*NAME4 - regressional coefficients of computation of component scores by variables. If OUT is nonpositive
*this matrix is identical to eigenvectors NAME2. If OUT is positive this matrix is computed from loadings NAME2.
*You can use matrix NAME4 to compute principal component scores of the data points not participated
*in the analysis.
*NAME5 - principal component scores. They are DATA*NAME4. If OUT is nonpositive these are raw scores:
*sums-of-squares in the columns of NAME5 = eigenvalues of matrix sscp(DATA)). If OUT is positive these
*are scaled scores: sums-of-squares in the columns of NAME5 = n-1 or n, depending on argument DF.

define !KO_pcomp(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
             /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !10= !2.
comp @df= !3.
comp @m= !4.
comp @out= !5.
call eigen(sscp(!10)/(nrow(!10)-(@df>0)),!7,!6).
comp !6= {!6,!6/csum(!6)}.
do if not @m.
-comp @m= csum(!6(:,1)>0).
end if.
comp !6= !6(1:@m,:).
comp !7= !7(:,1:@m).
comp @eipos= !6(:,1)>0.
comp !6(:,1)= !6(:,1)&*@eipos.
comp @ones= make(1,nrow(!7),1).
comp !7= !7&*t(@eipos)(@ones,:).
comp @sign= csum(!7)>=0.
comp !7= !7&*(@sign-(not @sign))(@ones,:).
do if @out.
-comp !7= !7&*t(sqrt(!6(:,1)))(@ones,:).
-comp !8= !7/(t(sqrt(cssq(!10)/(nrow(!10)-(@df>0))))*make(1,@m,1)).
-comp !9= !7/t(!6(:,1)+(not @eipos))(@ones,:).
else.
-comp !8= !7.
-comp !9= !7.
end if.
comp !10= !10*!9.
release @df,@m,@out,@eipos,@sign,@ones.
!enddefine.


******** PRINCIPAL COMPONENTS (FOR N<P) ********.
*/*!KO_pcomp2(data%df%m%out%name1%name2%name3%name4%name5)*/*.
*Version 1.
*Principal component analysis.
*This function is identical with function /*!KO_pcomp*/ but is faster when number of columns in data
*is greater than number of rows. And is slower than it when number of columns in data is less than number of rows.

define !KO_pcomp2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
             /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @df= !3.
comp @m= !4.
comp @out= !5.
comp @df= nrow(@data)-(@df>0).
call eigen(sscp(t(@data))/@df,@u,!6).
comp !6= {!6,!6/csum(!6)}.
do if not @m.
-comp @m= csum(!6(:,1)>0).
end if.
comp !6= !6(1:@m,:).
comp @u= @u(:,1:@m).
comp @eipos= !6(:,1)>0.
comp !6(:,1)= !6(:,1)&*@eipos.
comp @u= @u&*t(@eipos)(make(1,nrow(@u),1),:).
comp !7= t(@data)*@u/sqrt(@df).
do if @out.
-comp !10= @u*sqrt(@df).
-comp !8= !7/(t(sqrt(cssq(@data)/@df))*make(1,@m,1)).
-comp !9= !7/t(!6(:,1)+(not @eipos))(make(1,nrow(!7),1),:).
else.
-comp !10= @u*sqrt(@df)&*t(sqrt(!6(:,1)))(make(1,nrow(@data),1),:).
-comp !7= !7/t(sqrt(!6(:,1)))(make(1,nrow(!7),1),:).
-comp !8= !7.
-comp !9= !7.
end if.
comp @sign= csum(!7)>=0.
comp @sign= @sign-(not @sign).
comp !7= !7&*@sign(make(1,nrow(!7),1),:).
comp !8= !8&*@sign(make(1,nrow(!8),1),:).
comp !10= !10&*@sign(make(1,nrow(!10),1),:).
comp !9= !9&*@sign(make(1,nrow(!9),1),:).
release @data,@df,@m,@out,@u,@eipos,@sign.
!enddefine.


******** PRINCIPAL COMPONENTS (MATRIX INPUT) ********.
*/*!KO_pca(cov%w%m%out%name1%name2%name3%name4)*/*.
*Version 1.
*Principal component analysis (with option to weight variables).
*This function, unlike /*!KO_pcomp*/, takes as input not data but ready matrix
*(covariance or like), and it does not compute itself principal component scores.
*Present is an option to supply nonequal weights (importancies) to variables, i.e. perform weighted PCA.
*Input:
*COV - square symmetric matrix of coefficients of sscp type - i.e. correlation, covariance,
*cosine similarity or raw sscp matrix of association coefficients between variables.
*W - vector of length as the number of rows/columns of COV, containing weights for the variables; or specify
*a nonpositive scalar (to mean - no weights supplied, results will be identical with /*!KO_pcomp*/). Weights must
*be positive values in ratios you need (only ratios between weights matter). Weight exactly 0
*is not allowed, however you may give very close to zero positive weight if you aim to make some variable
*"passive", supplementary. 
*Arguments M and OUT - same as in function /*!KO_pcomp*/.
*Results:
*NAME1, NAME2, NAME3, NAME4 - analogous to those in function /*!KO_pcomp*/.
*Component score coefficient matrix NAME4 you can use to compute principal component scores,
*which would be equal to DATA*NAME4, where DATA is data cases x variables (if COV is covariance
*matrix the variables should be centered, if COV is correlation matrix the variables should be
*z-standardized, etc, i.e. there must be correspondence with the matrix type).
*Nongramian COV input:
*If among the M extracted components there occurs such with negative eigenvalue in NAME1, i.e. imaginary
*one(s), then its corresponding results in NAME2, NAME3, NAME4 will be computed as if that component
*is real (positive eigenvalue), but those its results should be considered imaginary.

define !KO_pca(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
             /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @r= !2.
comp @w= !3.
comp @m= !4.
comp @out= !5.
comp @dg= diag(@r).
do if nrow(@w)>ncol(@w).
-comp @w= t(@w).
end if.
do if ncol(@w)=1 and not @w(1).
-comp @w= make(1,ncol(@r),1).
else.
-comp @w= @w/rsum(@w)*ncol(@w).
-comp @r= @r&*sscp(sqrt(@w)).
end if.
call eigen(@r,!7,!6).
comp @einzero= abs(!6)>1E-13.
comp !6= !6&*@einzero.
comp !6= {!6,!6/csum(!6)}.
do if not @m.
-comp @m= csum(!6(:,1)>0).
end if.
comp !6= !6(1:@m,:).
comp @einzero= @einzero(1:@m).
comp !7= !7(:,1:@m).
comp @eival= abs(!6(1:@m,1)).
comp @ones= make(1,nrow(!7),1).
comp !7= !7&*t(@einzero)(@ones,:).
comp !7= !7/(t(sqrt(@w))*make(1,ncol(!7),1)).
comp @sign= csum(!7)>=0.
comp !7= !7&*(@sign-(not @sign))(@ones,:).
comp @w= t(@w)*make(1,ncol(!7),1).
do if @out.
-comp !7= !7&*t(sqrt(@eival))(@ones,:).
-comp !8= !7/(sqrt(@dg)*make(1,ncol(!7),1)).
-comp !9= !7&*@w/t(@eival+(not @einzero))(@ones,:).
else.
-comp !8= !7.
-comp !9= !7&*@w.
end if.
release @r,@w,@m,@out,@dg,@einzero,@eival,@sign,@ones.
!enddefine.


******** ORTHOGONAL PROCRUSTES ROTATION ********.
*/*!KO_procr(fit%targ%refl%iso%name1%name2)*/*.
*Version 1.
*Performs orthogonal procrustes rotation of fit configuration FIT into target
*configuration TARG. The fitted (rotated) fit configuration is returned as NAME2,
*and the rotation matrix as NAME1.
*FIT and TARG - matrices of equal size. Their rows are data points and columns are dimensions (axes)
*of space. The same row in the two matrices is a pair of correspondent points.
*FIT and TARG each have their own space dimensions. Procrustes rotation is the rotation of FIT axes
*into TARG axes which provides the maximal juxtaposition of the corresponding points.
*NAME2 are the coordinates of FIT points in the space of TARG axes after the rotation.
*If number of dimensions differ in FIT and TARG, add zero-filled columns to one of the matrices
*because the function requires that the number of columns be equal in both.
*In rotation matrix NAME1 rows correspond to axes of FIT and columns correspond to axes of TARG.
*REFL is a scalar; if positive, the rotation is allowed to be with reflection (if it enhances the fitting);
*if nonpositive, reflection is prohibited (fitting might be worse).
*ISO is a scalar; if positive, additional operation of isoscaling (for better fitting) is applied
*after rotation; if nonpositive, isoscaling is not used. This argument does not affect
*the rotation matrix NAME1.
*The function takes FIT and TARGET data "as is". If need translation (superposition of centroids)
*and/or scaling (equalizing magnitudes) of both configurations prior operation of procrustes rotation
*do these transformations in advance.

define !KO_procr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @a= !2.
comp @b= !3.
comp @refl= !4.
comp @iso= !5.
call svd(t(@b)*@a,@u,@s,@v).
comp @s= diag(@s).
comp !6= @v*t(@u).
do if not @refl and not det(!6).
-comp @v(:,ncol(@v))= -@v(:,ncol(@v)).
-comp @s(nrow(@s))= -@s(nrow(@s)).
-comp !6= @v*t(@u).
end if.
release @b,@u,@v.
comp !7= @a*!6.
do if @iso.
-comp !7= !7*(csum(@s)/trace(sscp(@a))).
end if.
release @a,@refl,@iso,@s.
!enddefine.


******** LINEAR DISCRIMINANTS ********.
*/*!KO_discrim(data%dummy%m%out%name1%name2%name3%name4%name5)*/*.
*Version 2.
*Performs linear discriminant analysis (without classifying and tests of significance): extracts discriminants.
*Takes data DATA (rows - points, columns - axes of space) and grouping DUMMY.
*Columns of DATA (variables) must be already centered (or standardized). Columns of DATA must be
*noncollinear.
*DUMMY - binary dummy (indicator) variables marking groups (each variable corresponds to a group,
*with value 1 = the case belongs to it, 0 = doesn't belong to it).
*Each column in DUMMY must contain at least one 1.
*The groups must be disjoint by case membership, and each case must belong to a group,
*i.e. sum in each row in DUMMY must be equal to 1.
*Argument M (scalar) - how many discriminants to extract. Specify positive integer from 1 to
*min(number_groups-1,number_variables). Or specify any nonpositive number: then all nonzero discriminants
*will be extracted.
*Argument OUT (scalar) controls the form of results returned.
*Results:
*NAME1 - two-column matrix: eigenvalues in the 1st col, canonical correlations in the 2nd.
*NAME2 - eigenvectors with normalized columns (if OUT nonpositive) or raw (unstandardized) canonical
*discriminant coefficients (if OUT positive).
*NAME3 - standardized canonical discriminant coefficients (weights of variables into the discriminants).
*NAME4 - pooled within-group correlations between variables and discriminants (structure matrix).
*NAME5 - discriminant scores, they are DATA*NAME2, i.e. computed with the help of normalized eigenvectors
*(if OUT nonpositive) or with the help of discriminant coefficients (if OUT positive).
*If you need to supply the vector of constants to the unstandardized coefficients (to obtain un-centered
*discriminants) then it is equal: -csum(mdiag(MEAN)*NAME2), where MEAN is the vector of means of the original
*variables and NAME2 is the unstandardized canonical discriminant coefficients.
*You can create dummy variables DUMMY out of a categorical grouping variable with
*the help of function /*!KO_freq*/.

define !KO_discrim(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !10= !2.
comp @des= !3.
comp @k= ncol(@des).
comp @m= !4.
comp @out= !5.
comp @b= t(@des)*!10/(t(csum(@des))*make(1,ncol(!10),1)).
comp @b= sscp(@des*@b).
comp @w= sscp(!10)-@b.
comp @invc= inv(chol(@w)). 
comp @invcb= t(@invc)*@b*@invc.
comp @invcb= (@invcb+t(@invcb))/2.
call eigen(@invcb,!7,@eival).
do if not @m.
-comp @m= rmin({@k-1,ncol(!10),csum(@eival>0)}). 
end if.
comp !6= make(@m,2,0).
comp @eival= @eival(1:@m).
comp !6(:,1)= @eival.
comp !6(:,2)= sqrt(!6(:,1)/(!6(:,1)+1)).
comp !7= (@invc*!7)(:,1:@m).
comp !7= !7&*t(@eival>0)(make(1,nrow(!7),1),:).
comp @sign= csum(!7)>=0.
comp !7= !7&*(@sign-(not @sign))(make(1,nrow(!7),1),:).
release @b,@invc,@invcb,@eival,@sign.
comp @dgw= sqrt(diag(@w))*make(1,@m,1).
comp !8= !7&*@dgw.
comp !9= @w*!7/@dgw.
do if @out.
-comp !7= sqrt(nrow(!10)-@k)*!7.
else.
-comp @dgw= cssq(!7).
-comp !7= !7/(sqrt(@dgw)+(@dgw=0))(make(1,ncol(!10),1),:).
end if.
comp !10= !10*!7.
release @k,@out,@m,@w,@dgw.
!enddefine.


******** PRINCIPAL COORDINATES ********.
*/*!KO_pcoord(dis%m%method%name1%name2)*/*.
*Version 1.
*Principal coordinate analysis (= Torgerson's metric multidimensional scaling).
*Takes matrix of dissimilarities DIS, which it treats as squared (sic!) distances, and embeds
*these objects in euclidean space, returning the coordinates as NAME2.
*Dimensions (columns of NAME2) are uncorrelated variables.
*DIS must be square symmetric, num_objects x num_objects.
*M (integer scalar) - needed dimensionality of the euclidean space: number from 1 to num_objects,
*or a nonpositive number (then M will be = all really existing nonredundant dimensions).
*Argument METHOD (scalar) is the method of decomposition of the double-centration matrix, the intermediate
*product of DIS: eigen-decomposition (with METHOD positive) or svd-decomposition (with METHOD nonpositive).
*All eigen or singular values of these decompositions are returned as vector NAME1.
*The difference between the methods is the following: possible negative eigenvalues of the double-centration
*matrix are discovered by eigen-decomposition, and the function discards them further, but corrects with
*them the sum of positive eigenvalues. Negative eigenvalues are not recognized by svd-decomposition and
*come out as junior singular values. From the point of view of accurateness of reproduction of input
*dissimilarities DIS by resultant euclidean distances (the distances between the objects in the obtained
*configuration NAME2), version with eigen-decomposition is preferable, as a rule.
*If DIS is squared euclidean distances, than both methods are same, and coordinates NAME2 are then identical
*to raw component scores which would be obtained by Principal component analysis of those data (centered)
*from which one might compute the distances DIS. (This is an evidence of the tight kindred link between
*Principal coordinate analysis and Principal component analysis.).

define !KO_pcoord(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @m= !3.
comp @method= !4.
comp @rmean= rsum(@mat)/ncol(@mat)*make(1,ncol(@mat),1).
comp @mat= (@rmean+t(@rmean)-@mat-msum(@mat)/ncol(@mat)**2)/2.
release @rmean.
do if @method.
-call eigen(@mat,@v,!5).
-comp !5= !5&*(abs(!5)>1E-13).
-comp @eival= !5&*(!5>0).
-comp @eival= @eival*(trace(@mat)/csum(@eival)).
else.
-call svd(@mat,@u,@eival,@v).
-release @u.
-comp @eival= diag(@eival).
-comp @eival= @eival&*(@eival>1E-13).
-comp !5= @eival.
end if.
comp @sign= csum(@v)>=0.
comp @v= @v&*(@sign-(not @sign))(make(1,nrow(@v),1),:).
do if not @m.
-comp @m= csum(@eival>0).
end if.
comp !6= @v(:,1:@m)*mdiag(sqrt(@eival(1:@m))).
release @mat,@m,@method,@eival,@v,@sign.
!enddefine.


******** BIPLOT ********.
*/*!KO_biplot(data%rw%cw%norm%m%name1%name2%name3%name4)*/*.
*Version 1.
*Biplot with option to weight rows and columns.
*Returns coordinates of row points and column points, and other statistics.
*This function also can be considered performing weighted Principal component analysis (see below).
*DATA - arbitrary data matrix (at least two rows and columns). The analysis takes the data as is, therefore
*if it is necessary do the standardization you need in advance.
*RW - weight for each row of DATA (column vector) or a nonpositive scalar (which means no weighting: the
*program will give equal weight to all rows).
*CW - weight for each column of DATA (row vector) or a nonpositive scalar (which means no weighting: the
*program will give equal weight to all columns).
*Specified weights can be any nonnegative numbers, only ratios between weights matter.
*Specified weight 0 for row/column means that this row/column must be passive (supplementary).
*Every positive weight designates active row/column.
*NORM - vector of two numbers from 0 to 1. These are coefficients of normalizing, or endowing with inertia, of
*coordinates of row points (first number) and column points (second number). Coefficient 0 means "standard
*normalization" when scale of coordinates is standardized to 1. Coefficient 1 means "principal normalization"
*when all the scale of eigenvalues of dimensions is given to the scale of coordinates. For example, {0,1} is
*standard normalization for row points and principal normalization for column points; {1,1} is principal
*normalization for these and those; {.5,.5} is so called "symmetric normalization".
*M - how many dimensions (principal axes) to extract: integer from 1 to min(num_active_rows,num_active_cols)
*in DATA. Or specify any nonpositive number - then all nonzero dimensions will be extracted.
*Results:
*NAME1 - two-column matrix: eigenvalues (scale, inertia) of dimensions (1st col) and proportion of explained
*overall inertia of DATA (2nd col).
*In NAME2, NAME3, NAME4 rows are points of biplot: first go row points of DATA, then, below them, column\
*points of DATA.
*NAME2 - two-column matrix: point mass (1st col) and point inertia (2nd col).
*Mass of an active row or column is proportional to its weight. Mass of a passive row or column is proportional
*to the averaged weight of active rows or columns, respectively.
*Inertia of a point is its scale in the normalized (weighted) matrix DATA. If there were no weighting or
*weights for all active rows/columns were equal then the inertia of a row/column in its sum of squares in DATA
*divided by r*c (r = number of active rows, c = number of active columns).
*NAME3 - coordinates of row points and column points in the space of M dimensions. Way of normalization by
*inertia (argument NORM) affects the coordinates.
*NAME4 - contributions of points to inertia of dimensions (first M cols) and contributions of dimensions to
*inertia of points (last M cols). Contributions of points to dimensions are the squared elements of
*eigenvectors. Points with high contribution to inertia of a dimension are the points to whom the dimension
*chiefly owes its emergence. Contribution of passive points is 0 because they do not influence the formation
*of dimensions. Contribution of dimension to inertia of point is the index of completeness of explanation
*of the point by the given dimension; high contribution means that the point is well represented via this
*dimension and thus it is suffice to "explain" the point.
*RELATION to Principal component analysis (PCA). Biplot and PCA are almost identical: the only difference is
*in that biplot brings the sum of column weights of input data to 1, and PCA - to the number of (active)
*columns. This makes results of the two methods proportionally identical, and the current function is usable to
*execute the principal component analysis, including PCA with weighting of importancies of data rows and/or
*columns. Specifically, let C be the number of active columns, then the following relationships are true, to
*convert results:
*1) eigenvalues of PCA = eigenvalues of biplot * C;
*2) loadings = column coordinates under "principal normalization" of columns;
*3) standardized component scores = row coordinates under "standard normalization" of rows;
*4) eigenvectors of PCA = column coordinates under "standard normalization" of columns / sqrt(C);
*5) raw component scores = row coordinates under "principal normalization" of rows * sqrt(C).

define !KO_biplot(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
              /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @rw= !3.
comp @cw= !4.
comp @norm= !5.
comp @m= !6.
do if nrow(@rw)=1 and not @rw(1).
-comp @rw= make(nrow(@data),1,1).
end if.
do if ncol(@cw)=1 and not @cw(1).
-comp @cw= make(1,ncol(@data),1).
end if.
comp @rw= @rw/csum(@rw).
comp @cw= @cw/rsum(@cw).
comp @posrw= @rw>0.
comp @poscw= @cw>0.
comp @rm= @rw+(not @posrw)/csum(@posrw).
comp @cm= @cw+(not @poscw)/rsum(@poscw).
comp @dat= @data&*sqrt(@rm*@cm).
comp @ri= rssq(@dat&*@poscw(make(1,nrow(@dat),1),:)).
comp @ci= cssq(@dat&*(@posrw*make(1,ncol(@dat),1))).
comp !8= {@rm,@ri;t(@cm),t(@ci)}.
comp @dat= @data&*sqrt(@rw*@cw).
call svd(@dat,@u,@s,@v).
release @dat.
comp @s= diag(@s)(1:rmin({csum(@posrw),rsum(@poscw)})).
comp @s= @s&*(@s>1E-13).
comp !7= @s&**2.
comp !7= {!7,!7/csum(!7)}.
comp @m_= csum(!7(:,1)>0).
do if not @m.
-comp @m= @m_.
end if.
comp !7= !7(1:@m,:).
comp @mm_= @m-@m_.
comp @m= rmin({@m,@m_}).
comp @s= @s(1:@m).
comp @u= @u(:,1:@m).
comp @v= @v(:,1:@m).
comp !10= {@u&**2;@v&**2}.
comp @u= @u/((sqrt(@rw)+(not @posrw))*make(1,ncol(@u),1)).
comp @v= @v/(t(sqrt(@cw)+(not @poscw))*make(1,ncol(@v),1)).
comp @sign= csum({@u;@v})>=0.
comp @sign= @sign-(not @sign).
comp @u= @u&*@sign(make(1,nrow(@u),1),:).
comp @v= @v&*@sign(make(1,nrow(@v),1),:).
comp !9= {@data&*@cw(make(1,nrow(@data),1),:)*@v&*(t(@s)&**(@norm(1)-1))(make(1,nrow(@data),1),:);
        t(@data&*(@rw*make(1,ncol(@data),1)))*@u&*(make(ncol(@data),1,1)*t(@s)&**(@norm(2)-1))}.
comp !10= {!10,{!9(1:nrow(@data),:)&**2&*(@rm/@ri*t(@s&**(2-2*@norm(1))));
                !9((nrow(@data)+1):nrow(!9),:)&**2&*t(@s&**(2-2*@norm(2))*(@cm/@ci))}}.
do if @mm_.
-comp @mm_= make(nrow(!9),@mm_,0).
-comp !9= {!9,@mm_}.
-comp !10= {!10(:,1:@m),@mm_,!10(:,(@m+1):ncol(!10)),@mm_}.
end if.
release @data,@rw,@cw,@norm,@m,@m_,@mm_,@posrw,@poscw,@rm,@cm,@ri,@ci,@s,@u,@v,@sign.
!enddefine.


******** SIMPLE CORRESPONDENCE ANALYSIS ********.
*/*!KO_corresp(tab%sta%rsup%csup%norm%m%name1%name2%name3%name4)*/*.
*Version 1.
*Simple (two-way) correspondence analysis. It is an application of the biplot idea to the analysis of
*contingency table. Returns coordinates of row points and column points, and other statistics.
*TAB - contingency table (at least two rows and columns). It is any table with nonnegative entries,
*in which a value in a cell has meaning of contingency, proximity between the row and the column.
*In TAB, row sums from active columns must be all nonzero, and column sums from active rows must be all
*nonzero (active rows/columns are not passive rows/columns).
*STA - needed standardization for the table, integer from 1 to 6:
*1 - two-way centering and weighted normalization
*2 - two-way centering and simple normalization
*3 - centering of rows and simple normalization
*4 - centering of columns and simple normalization
*5 - equalizing row sums, centering of rows and simple normalization
*6 - equalizing column sums, centering of columns and simple normalization
*Option (1) is chi-square model of analysis, it is optimal if TAB is a frequency table. The rest five
*options to choose is euclidean model of analysis, it suits if TAB is not frequencies (but, for example,
*sums or means). Chi-square model implies biplot with weighting.
*RSUP - vector of row numbers that should be treated as passive (supplementary), or any nonpositive scalar if
*no passive are needed. CSUP - analogously, vector of column numbers that should be treated as passive, or any
*nonpositive scalar. Rows and columns that are not passive are considered active.
*NORM - vector of two numbers from 0 to 1. These are coefficients of normalizing, or endowing with inertia, of
*coordinates of row points (first number) and column points (second number). Coefficient 0 means "standard
*normalization" when scale of coordinates is standardized to 1. Coefficient 1 means "principal normalization"
*when the all scale of eigenvalues of dimensions is given to the scale of coordinates. For example, {0,1} is
*standard normalization for row points and principal normalization for column points; {1,1} is principal
*normalization for these and those; {.5,.5} is so called "symmetric normalization".
*M - how many dimensions to extract: integer from 1 to min(num_active_rows-1,num_active_cols-1) in TAB. Or
*specify any nonpositive number - then all nonzero dimensions will be extracted.
*Results (they are of the same kind as in biplot /*!KO_biplot*/):
*NAME1 - two-column matrix: eigenvalues (scale, inertia) of dimensions (1st col) and proportion of explained
*overall inertia of TAB transformed by standardization (2nd col).
*In NAME2, NAME3, NAME4 rows are points of biplot: first go row points of TAB, then, below them, column points
*of TAB.
*NAME2 - two-column matrix: point mass (1st col) and point inertia (2nd col).
*Mass of an active row or column is proportional to its weight. Mass of a passive row or column is proportional
*to the averaged weight of active rows or columns, respectively.
*Inertia of a point is its scale (sum of squares) in the standardized TAB.
*NAME3 - coordinates of row points and column points in the space of M dimensions. Way of normalization by
*inertia (argument NORM) affects the coordinates.
*NAME4 - contributions of points to inertia of dimensions (first M cols) and contributions of dimensions to
*inertia of points (last M cols). Contributions of points to dimensions are the squared elements of
*eigenvectors. Points with high contribution to inertia of a dimension are the points to whom the dimension
*chiefly owes its emergence. Contribution of passive points is 0 because they do not influence the formation
*of dimensions.
*Contribution of dimension to inertia of point is the index of completeness of explanation of the point by
*the given dimension; high contribution means that the point is well represented via this dimension and thus
*it is suffice to "explain" the point.

define !KO_corresp(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
               /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @tab= !2.
comp @sta= !3.
comp @rsup= !4.
comp @csup= !5.
comp @norm= !6.
comp @m= !7.
comp @nar= nrow(@tab).
comp @nac= ncol(@tab).
comp @rn= rsum(@tab).
comp @cn= csum(@tab).
comp @n= rsum(@cn).
do if @rsup(1).
-comp @nar= @nar-nrow(@rsup)*ncol(@rsup).
-comp @cn= @cn-csum(@tab(@rsup,:)).
-comp @n= rsum(@cn).
end if.
do if @csup(1).
-comp @nac= @nac-nrow(@csup)*ncol(@csup).
-comp @rn= @rn-rsum(@tab(:,@csup)).
-comp @n= csum(@rn).
end if.
do if @rsup(1) and @csup(1).
-comp @n= @n-csum(@rn(@rsup)).
end if.
do if @sta=1.
-comp @rm= @rn/@n.
-comp @cm= @cn/@n.
-comp @tb= @tab-@rn*@cn/@n.
-comp @tab= @tb.
-comp @tb= @tb/sqrt(@rn*@cn).
else.
-comp @rm= make(nrow(@tab),1,1/@nar).
-comp @cm= make(1,ncol(@tab),1/@nac).
-do if @sta=2.
- comp @tb= @tab-@rn*@cn/@n.
-else if @sta=3.
- comp @tb= @tab-(@rn/@nac)*make(1,ncol(@tab),1).
-else if @sta=4.
- comp @tb= @tab-(@cn/@nar)(make(1,nrow(@tab),1),:).
-else if @sta=5.
- comp @tb= @tab&*((@n/(@nar*@rn))*make(1,ncol(@tab),1)).
- comp @tb= @tb-@n/@nar/@nac.
-else if @sta=6.
- comp @tb= @tab&*(@n/(@nac*@cn))(make(1,nrow(@tab),1),:).
- comp @tb= @tb-@n/@nac/@nar.
-end if.
-comp @tb= @tb*@nar*@nac/@n.
-comp @tab= @tb.
-comp @tb= @tb/sqrt(@nar*@nac).
end if.
comp @ri= rssq(@tb).
comp @ci= cssq(@tb).
do if @rsup(1).
-comp @ci= @ci-cssq(@tb(@rsup,:)).
end if.
do if @csup(1).
-comp @ri= @ri-rssq(@tb(:,@csup)).
end if.
comp !9= {@rm,@ri;t(@cm),t(@ci)}.
do if @rsup(1).
-comp @tb(@rsup,:)= make(nrow(@rsup)*ncol(@rsup),ncol(@tb),0).
end if.
do if @csup(1).
-comp @tb(:,@csup)= make(nrow(@tb),nrow(@csup)*ncol(@csup),0).
end if.
call svd(@tb,@u,@s,@v).
release @tb.
comp @s= diag(@s).
comp @s= @s(1:(rmin({@nar,@nac})-1)).
comp @s= @s&*(@s>1E-13).
comp !8= @s&**2.
comp !8= {!8,!8/csum(!8)}.
comp @m_= csum(!8(:,1)>0).
do if not @m.
-comp @m= @m_.
end if.
comp !8= !8(1:@m,:).
comp @mm_= @m-@m_.
comp @m= rmin({@m,@m_}).
comp @s= @s(1:@m).
comp @u= @u(:,1:@m).
comp @v= @v(:,1:@m).
comp !11= {@u&**2;@v&**2}.
do if @sta=1.
-comp @u= @u/(sqrt(@rn/@n)*make(1,ncol(@u),1)).
-comp @v= @v/(t(sqrt(@cn/@n))*make(1,ncol(@v),1)).
-comp @denr= @rn*make(1,@m,1).
-comp @denc= t(@cn)*make(1,@m,1).
else.
-comp @u= @u*sqrt(@nar).
-comp @v= @v*sqrt(@nac).
-comp @denr= @nac.
-comp @denc= @nar.
end if.
comp @sign= csum({@u;@v})>=0.
comp @sign= @sign-(not @sign).
comp @u= @u&*@sign(make(1,nrow(@u),1),:).
comp @v= @v&*@sign(make(1,nrow(@v),1),:).
comp !10= {@tab*@v&*(t(@s)&**(@norm(1)-1))(make(1,nrow(@tab),1),:)/@denr; 
         t(@tab)*@u&*(t(@s)&**(@norm(2)-1))(make(1,ncol(@tab),1),:)/@denc}.
comp !11= {!11,{!10(1:nrow(@tab),:)&**2&*(@rm/@ri*t(@s&**(2-2*@norm(1))));
                !10((nrow(@tab)+1):nrow(!10),:)&**2&*t(@s&**(2-2*@norm(2))*(@cm/@ci))}}.
do if @mm_.
-comp @mm_= make(nrow(!10),@mm_,0).
-comp !10= {!10,@mm_}.
-comp !11= {!11(:,1:@m),@mm_,!11(:,(@m+1):ncol(!11)),@mm_}.
end if.
release @tab,@rsup,@csup,@norm,@m,@m_,@mm_,@nar,@nac,@rm,@cm,@ri,@ci,@s,@u,@v,@sign,@denr,@denc.
!enddefine.


******** FACTORS BY PRINCIPAL AXIS METHOD ********.
*/*!KO_paf(cov%comm%m%conv%name1%name2%name3%name4%name5)*/*.
*Version 1.
*Linear factor analysis (only extraction of factor loadings from a matrix; the function does not
*do factor rotations or computation of factor scores). Principal axis extraction method is used.
*Input:
*COV - square p x p symmetric matrix of coefficients of sscp type - i.e. correlation, covariance,
*cosine similarity or raw sscp matrix of association coefficients between variables.
*COMM - column vector of initial communality estimates, or nonpositive scalar. If vector, it must be
*of length as diagonal of COV (i.e. equal the number of variables p) and contain positive values not
*exceeding the corresponding values of the diagonal (i.e. the scales, variances of the variables).
*If COMM is nonpositive scalar the function will use images of COV matrix as the initial
*communalities - as a rule, they are the best canditates for initial communalities. This option
*requires positive definite COV.
*M - how many factors to extract, specify positive integer less than the number of variables
*(you must decide in advance, using known methods of estimation, how many factors to extract).
*CONV - criteria of iterative convergence: vector of 2 numbers, for example {.001,25}. The first is the
*threshold for communality change, nonnegative number; if the change is less than it communalities will
*be considered stabilized and iterations will stop. The second number is the maximal allowed number of
*iterations - specify positive integer.
*Results:
*NAME1 - M x 4 matrix of scales (variances) of the extracted factors and the proportions of total scale
*explained by them. The 1st column is the scales (variances) of the factors as they are, equal to the M first
*eigenvalues of the reduced COV at last iteration. 2nd column is the proportion of these scales of
*the total scale equal to the trace of the input COV. 3rd column is the rescaled scales (variances), i.e.
*such from which the effect of unequal scales (variances) of variables is removed. 4th column relates to
*the 3rd like the 2nd to the 1st.
*If in NAME1 there occurs factor with negative variance, it is imaginary; often rising the number of
*iterations moves it to real; or try to lessen M.
*NAME2 - p x M matrix of factor loadings, as they are. Their column sums of squares equal the factor scales
*and their row sums of squares equal the final communalities.
*NAME3 - rescaled loadings. It is the loadings from which the effect of unequal scales (variances) of variables
*is removed; they are the correlations (or cosines) between factors and variables and are convenient in
*interpretation of factors.
*NAME4 - initial and final communality estimates, p x 4. Columns left to right: initial as they are,
*final as they are, initial rescaled, final rescaled. A rescaled communality is the communality
*divided by the scale (variance) of that variable, and is the proportion of its scale explained
*by the common factors.
*If COV is correlation or cosine matrix then "raw" and rescaled results
*in NAME1, in NAME2/NAME3, in NAME4 will be identical.
*NAME5 - report of convergence, vector of 2 values which is comparable with argument CONV:
*the 1st value is the maximal change in communalities observed on the last iteration done and
*which, as desired, should be less than the threshold in CONV. The 2nd number is the number of iterastions
*done; if this appears by 1 greater than the specified in CONV then it means that the maximal number of
*iterations was done but the requested threshold of convergernce was not reached.
*Heywood case. If on any iteration a communality value exceeds the variable's scale then iterations will 
*cut off, results are returned as existing at the moment of break. You can see Heywood case occurred by presence
*of a value greater than one in the last column of NAME4; besides, the function will - for attention -
*give the 2nd value in NAME5 negative sign. Results in Heywood case should be acknowledged abortive.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_paf(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
           /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @r= !2.
comp @inic= !3.
comp @m= !4.
comp @conv= !5.
comp @dg= diag(@r).
do if nrow(@inic)*ncol(@inic)=1 and not @inic(1).
-comp @inic= @dg-1/diag(inv(@r)).
end if.
comp !9= @inic.
loop @i= 1 to @conv(2).
-comp @rc= !9/@dg.
-call setdiag(@r,!9).
-call eigen(@r,!7,!6).
-comp @eival= abs(!6(1:@m)).
-comp !7= !7(:,1:@m)&*t(sqrt(@eival))(make(1,nrow(!7),1),:).
-comp !9= rssq(!7).
-comp !10= cmax(abs(!9/@dg-@rc)).
end loop if !10<@conv(1) or any(!9>@dg).
comp !10= {!10,@i}.
do if any(!9>@dg).
-comp !10(2)= -!10(2).
end if.
comp @sign= csum(!7)>=0.
comp !7= !7&*(@sign-(not @sign))(make(1,nrow(!7),1),:).
comp !8= !7/(sqrt(@dg)*make(1,@m,1)).
comp @ssload= {!6(1:@m),t(cssq(!8))}.
comp @ssload(:,2)= @ssload(:,2)&*((@ssload(:,1)>=0)-(@ssload(:,1)<0)).
comp !6= {@ssload(:,1),@ssload(:,1)/csum(@dg),@ssload(:,2),@ssload(:,2)/nrow(!8)}.
comp !9= {@inic,!9,@inic/@dg,!9/@dg}.
release @r,@inic,@m,@conv,@dg,@i,@rc,@eival,@sign,@ssload.
!enddefine.


******** ORTHOGONAL ANALYTIC ROTATIONS ********.
*/*!KO_ortrot(load%kaiser%method%conv%reord%name1%name2%name3)*/*.
*Version 1.
*Does different kinds of orthogonal analytic rotations combinedly known as "orthomax",
*which usually are applied in factor analysis to rotate the extracted factors.
*LOAD - input matrix of loadings of orthogonal factors (p variables x m factors, p and m >1),
*that need be rotated (the result of rotation is matrix NAME1). LOAD is not necessarily
*factor loadings, it can just be coordinates of some p points in the space of m
*perpendicular axes.
*KAISER - scalar. If positive, then, before the rotation, LOAD will undergo Kaiser normalization
*(and after the rotation the result will undergo the back de-normalizing). Kaiser normalization
*brings sums of squares in the rows of LOAD to unit. If the scalar is nonpositive, then
*no Kaiser normalization will take place.
*METHOD - method of the rotation, capitalized keyword (may put in quotes or apostrophes).
*This is the criterion being maximized with the given method:
*"QUARTIMAX" - quartimax; it equals the sum of the resultant loadings (final coordinates) taken
*to power of 4: msum(NAME1^4).
*The other methods have the formula of the criterion p*Q-c*V, where Q is the quartimax criterion
*and V is the sum of squared variances of the resultant factors, the columns of NAME1:
*rsum(cssq(NAME1)^2). The coefficient c defines the method:
*"VARIMAX" - varimax: c=1
*"EQUAMAX" - equamax: c=m/2
*"PARSIMAX" - parsimax: c=p*(m-1)/(p+m-2)
*"FACPARS" - facpars (factor parsimony): c=p
*"CUSTOM c" - arbitrary value of c: set a value for it (number or scalar name). Value can be any.
*The higher is c the farther will the configuration be from a "general factor".
*With c approaching +infinity, factor variances, i.e. cssq(NAME1), will be equal.
*With c approaching -infinity, factor variances will be that as when you subject LOAD to rotation
*into its principal components (without preliminary centering).
*CONV - criteria of iterative convergence: vector of 2 numbers, for example {.0001,50}. The first is
*the threshold for the criterion change, nonnegative number; if the change is less than it the
*criterion will be considered stabilized and iterations will stop (see NAME3). The second number
*is the maximal allowed number of iterations - specify positive integer.
*REORD - scalar. If positive, then factors after the rotation, the columns of NAME1, will be
*reordered by descend of their variances (i.e. column sums of squares), and also in each column
*the sign of loadings will be ascribed so that the sum of loadings there be positive. If the scalar
*is nonpositive, these two postrotational actions won't be done.
*Results:
*NAME1 - the resultant matrix of loadings (coordinates).
*NAME2 - orthogonal rotation matrix, so that NAME1=LOAD*NAME2.
*NAME3 - iteration history, a vector of the criterion values on leaving iterations. The 1st value of
*the vector is the magnitude of the criterion for the input LOAD; the 2nd value is the magnitude of
*the criterion after the 1st iteration, and so on. The last value in the vector is the magnitude of
*the criterion on exiting the last iteration used. Therefore, there was, in all, nrow(NAME3)-1
*iterations. A criterion's values grow from iteration to iteration until they stabilize or the limit
*of iterations is reached. Quartimax and varimax values are always positive, but the rest of the
*methods may show negative values too, what depends on the input data.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Note: If factor/principal component analysis has been done on covariances (and not correlations),
*then loadings can be of two kind - raw and rescaled (the latter are obtained by dividing the former
*by the variables's st. deviations). There's more sense to rotate towards a "simple structure" just the
*rescaled ones, because it is they which are usually interpreted. However, if you are doing the rotation
*with Kaiser normalization, results of rotating the rescaled and the raw loadings will be equivalent,
*for NAME2 is the same in this case.

define !KO_ortrot(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                 /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !method= !unquote(!4)
comp !7= !2.
comp @kais= !3.
comp @conv= !5.
comp @reord= !6.
comp @p= nrow(!7).
comp @m= ncol(!7).
comp @h= 0.
do if @kais.
-comp @h= sqrt(rssq(!7)).
-comp !7= !7/(@h*make(1,@m,1)).
end if.
comp @a2= !7&*!7.
!if (!method='QUARTIMAX') !then !let !crit= 'mssq(@a2)'
!else !if (!method='VARIMAX') !then !let !crit= '@p*mssq(@a2)-rssq(csum(@a2))' !let !c1= '2/@p' !let !c2= '1/@p'
!else !if (!method='EQUAMAX') !then !let !crit= '@p*mssq(@a2)-@m/2*rssq(csum(@a2))' !let !c1= '@m/@p' !let !c2= '@m/(2*@p)'
!else !if (!method='PARSIMAX') !then !let !crit= '@p*(mssq(@a2)-(@m-1)/(@p+@m-2)*rssq(csum(@a2)))' !let !c1= '2*(@m-1)/(@p+@m-2)' !let !c2= '(@m-1)/(@p+@m-2)'
!else !if (!method='FACPARS') !then !let !crit= '@p*(mssq(@a2)-rssq(csum(@a2)))' !let !c1= '2' !let !c2= '1'
!else !let !crit= '@p*mssq(@a2)-@c*rssq(csum(@a2))' !let !c1= '2*@c/@p' !let !c2= '@c/@p'
-comp @c= !tail(!method).
!ifend !ifend !ifend !ifend !ifend
comp !9= !crit.
comp !8= ident(@m).
loop @it= 1 to @conv(2).
-loop @i= 1 to @m-1.
- loop @j= @i+1 to @m.
-  comp @ij= {@i,@j}.
-  comp @ai= !7(:,@i).
-  comp @aj= !7(:,@j).
-  comp @u= @ai&*@ai-@aj&*@aj.
-  comp @v= 2*@ai&*@aj.
-  comp @num= 2*csum(@u&*@v).
-  comp @den= csum(@u&*@u-@v&*@v).
   !if (!method<>'QUARTIMAX') !then
-   comp @us= csum(@u).
-   comp @vs= csum(@v).
-   comp @num= @num-!c1*@us*@vs.
-   comp @den= @den-!c2*(@us*@us-@vs*@vs).
   !ifend
-  comp @den= @den+(@den=0)*1E-14.
-  comp @phi= artan(@num/@den).
-  do if @den.
-   comp @phi= @phi/4.
-  else if @num.
-   comp @phi= (3.14159265358979+@phi)/4.
-  else.
-   comp @phi= (@phi-3.14159265358979)/4.
-  end if.
-  comp @sinphi= sin(@phi).
-  do if abs(@sinphi)>1E-14.
-   comp @cosphi= cos(@phi).
-   comp @qij= {@cosphi,-@sinphi;@sinphi,@cosphi}.
-   comp !7(:,@ij)= {@ai,@aj}*@qij.
-   comp !8(:,@ij)= !8(:,@ij)*@qij.
-  end if.
- end loop.
-end loop.
-comp @a2= !7&*!7.
-comp !9= {!9;!crit}.
end loop if abs(!9(nrow(!9))-!9(@it))<=@conv(1).
do if @kais.
-comp !7= !7&*(@h*make(1,@m,1)).
end if.
do if @reord.
-comp @h= csum(!7)>=0.
-comp !7= !7&*(@h-(not @h))(make(1,@p,1),:).
-comp @h= grade(-cssq(!7)).
-comp !7(:,@h)= !7.
-comp !8(:,@h)= !8.
end if.
release @kais,@conv,@p,@m,@reord,@h,@a2,@it,@i,@j,@ij,@ai,@aj,@u,@v,@num,@den,@phi,@sinphi,@cosphi, @qij
 !if (!method<>'QUARTIMAX') !then ,@us,@vs !ifend !if (!head(!method)='CUSTOM') !then ,@c !ifend.
!enddefine.


******** OBLIQUE ROTATION PROMAX AND PROMAJ ********.
*/*!KO_promax(ort%kaiser%k%name1%name2%name3)*/*.
*Version 1.
*Does rotation promax or promaj of an orthogonal configuration ORT, which the function takes as
*"already close to simple structure". ORT is a matrix of factor loadings already rotated
*towards a "simple structure" orthogonally; in the classic variant - by varimax method (use
*function */*!KO_ortrot*/ for that). Promax considers ORT^K as the target configuration - the more
*ideal "simple structure" than attained in ORT (raising elements of ORT to power K>1 while retaining
*their sign brings small loadings to zero faster than big loadings, thus "sharpening" the pattern),
*and tries to make ORT closer to ORT^K by loosing orthogonality of the axes (factors) through
*nonorthogonal procrustes rotation, while ORT^K is taken for the target pattern (the loadings, not
*the correlations). The result is an oblique pattern NAME1 obtained from the orthogonal ORT.
*Promaj is similar to promax, but creates the target configuration for rotation not by raising
*to power K but by another method (see Mulaik, Foundations of factor analysis, 2010).
*Input:
*ORT - p x m (p>=m) matrix of loadings or some coordinates in the orthogonal (cartesian) frame.
*ORT must be full rank, i.e. t(ORT)*ORT invertible. ORT is seen as close to "simple structure"
*yet orthogonal so far.
*KAISER - scalar. If positive, then the target configuration will be based of ORT undergone Kaiser
*normalization. It is the recommended way (Kaiser normalization brings sums of squares in the rows
*to unit). If the scalar is nonpositive, the target configuration will be based on ORT as it is.
*In the latter case it is preferable that values of ORT don't exceed 1 in abs. magnitude.
*K - scalar, promax parameter, the power for elements of ORT. In promax, only power >1 makes sense.
*Most frequently, K from 2 to 4 is used.
*If you specify a nonpositive value for K, the function will realize method promaj instead of promax.
*The specific value of K doesn't play a role then, because promaj does not raise elements to power K.
*Results:
*NAME3 - matrix of promax (or promaj) transformation (axes de-orthogonalization).
*NAME1 - resultant factor pattern matrix, = ORT*NAME3.
*NAME2 - resultant factor correlation matrix (angle cosines between the nonorthogonal axes in
*NAME1 configuration); it equals inv(sscp(NAME3)).
*The factor structure matrix is NAME1*NAME2.

define !KO_promax(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @a= !2.
comp @kais= !3.
comp @k= !4.
comp @ones= make(1,ncol(@a),1)).
do if @kais.
-comp @b= @a/(sqrt(rssq(@a))*@ones).
else.
-comp @b= @a.
end if.
do if @k.
-comp @b= abs(@b)&**(@k+1)/(@b+(@b=0)).
else.
-comp @b2= @b&*@b.
-comp @flag= @b2<=(csum(@b2)/nrow(@b2))(make(1,nrow(@b2),1),:).
-comp @z= cmax(@flag&*@b2)(make(1,nrow(@b2),1),:).
-comp @b= abs(@b)/(@b+(@b=0))&*sqrt((@b2-@z)&*(not @flag)).
-release @b2,@flag,@z.
end if.
comp @q= solve(sscp(@a),t(@a))*@b.
comp @q= @q/sqrt(cssq(@q))(@ones,:).
comp @ic= inv(sscp(@q)).
comp @d= t(sqrt(diag(@ic)))).
comp !7= @q&*@d(@ones,:).
comp !5= @a*!7.
comp !6= @ic/sscp(@d).
release @a,@kais,@k,@ones,@b,@q,@ic,@d.
!enddefine.


******** FACTOR SCORE COEFFICIENTS ********.
*/*!KO_fsc(cov%load%fcorr%method%name1%name2)*/*.
*Version 1.
*Computes factor score coefficient matrix by which one can obtain factor scores (estimated factor
*values as as variables) of the factors extracted in a factor analysis. There is several methods
*of the matrix computation (several kinds of factor scores).
*Input necessary for all methods METHOD:
*COV - the analyzed in the factor analysis correlation or covariance matrix, size p x p, symmetric.
*LOAD - p x m (1<=m<p) matrix of factor loadings. If the factors underwent oblique rotation then
*this must be factor pattern matrix. If COV is covariances, not correlations, LOAD is the "raw"
*(not rescaled) loadings.
*Input necessary not for all methods METHOD:
*FCORR - m x m matrix of correlations between the factors. If factors are orthogonal or m=1 you may
*specify an arbitrary scalar in place of the matrix.
*Method of computation of factor scores METHOD (keyword in capital letters, optionally quoted
*or apostrophed):
*"REG" - regression method of Thurstone (Thompson); maximizes validity of the scores.
*"REG2" - same method, only as off-diagonal elements of COV there will be used values reproduced
*by the factors.
*"REG3" - Horst's or "idealized variables" method; factor scores are computed exactly same formula
*by which component scores are computed in principal component analysis; equivalent to method "REG"
*using, in place of COV, the reduced reproduced matrix. Does not need FCORR (specify any scalar).
*"BART" - Bartlett's method; minimizes bias of scores. Does not need FCORR (specify any scalar).
*Three next methods are called correlation-preserving, because they achieve correlations between
*factor scores that are equal to correlations between factors:
*"AR" - Anderson-Rubin method. Does not need FCORR (specify any scalar). This method always gives
*uncorrelated scores, even for oblique factors.
*"MAR" - McDonald-Anderson-Rubin method is the generalization of the previous one, and is suitable
*also for oblique factors.
*"GREEN" - Green's method. It can be seen as an approximation to "MAR" to which it becomes identical
*if communalities of all variables are equal.
*For methods REG, MAR, GREEN nonsingular COV is needed.
*Results:
*NAME1 - coefficients to compute factor scores; the latter themselves are DATA*NAME1, where DATA is
*the variables analyzed by the factor analysis (centered if factor analysis was based on covariances
*or standardized if it was based on correlations).
*NAME2 - rescaled (standardized) coefficients, i.e. such from which the effect of inequality of
*variable variances was removed. If COV is correlations NAME2=NAME1.
*If instead of factor analysis proper there principal component analysis was used, then the component
*scores, same, are returned by methods REG, REG3, GREEN.

define !KO_fsc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @r= !2.
comp @p= !3.
comp @c= !4.
!let !method= !unquote(!5)
do if nrow(@c)*ncol(@c)=1.
-comp @c= 1.
end if.
comp @u2_1= 1/diag(@r-@p*@c*t(@p)).
!if (!method='REG') !then
-comp !6= solve(@r,@p*@c).
!else !if (!method='REG2') !then
-comp @r#= @p*@c*t(@p).
-call setdiag(@r#,diag(@r)).
-comp !6= solve(@r#,@p*@c).
-release @r#.
!else !if (!method='REG3') !then
-comp !6= t(solve(sscp(@p),t(@p))).
!else !if (!method='BART') !then
-comp @tpu= t(@p&*(@u2_1*make(1,ncol(@p),1))).
-comp !6= t(solve(@tpu*@p,@tpu)).
-release @tpu.
!else !if (!method='AR') !then
-comp @tpu= t(@p&*(@u2_1*make(1,ncol(@p),1))).
-call eigen(@tpu*@r*t(@tpu),@eivec,@eival).
-do if ncol(@p)=1.
- comp !6= t(@eivec*sqrt(1/@eival)*t(@eivec)*@tpu).
-else.
- comp !6= t(@eivec*mdiag(sqrt(1/@eival))*t(@eivec)*@tpu).
-end if.
-release @tpu,@eivec,@eival.
!else !if (!method='MAR') !then
-call eigen(@r,@eivec,@eival).
-comp @r12= @eivec*mdiag(sqrt(@eival))*t(@eivec).
-comp @r_12= inv(@r12).
-call eigen(@c,@eivec,@eival).
-do if ncol(@p)=1.
- comp @c12= @eivec*sqrt(@eival)*t(@eivec).
-else.
- comp @c12= @eivec*mdiag(sqrt(@eival))*t(@eivec).
-end if.
-call svd(@r12&*t(@u2_1)(make(1,nrow(@r12),1),:)*@p*@c12,@g,@d,@m).
-comp !6= @r_12*(@g(:,1:ncol(@p))*t(@m))*@c12.
-release @r12,@r_12,@c12,@eivec,@eival,@g,@d,@m.
!else
-call eigen(@r,@eivec,@eival).
-comp @r_12= @eivec*mdiag(1/sqrt(@eival))*t(@eivec).
-call eigen(@c,@eivec,@eival).
-do if ncol(@p)=1.
- comp @c12= @eivec*sqrt(@eival)*t(@eivec).
-else.
- comp @c12= @eivec*mdiag(sqrt(@eival))*t(@eivec).
-end if.
-call svd(@r_12*@p*(@c*@c12),@g,@d,@m).
-comp !6= @r_12*(@g(:,1:ncol(@p))*t(@m))*@c12.
-release @r_12,@c12,@eivec,@eival,@g,@d,@m.
!ifend !ifend !ifend !ifend !ifend !ifend
comp !7= sqrt(diag(@r))*make(1,ncol(!6),1)&*!6.
release @r,@p,@c,@u2_1.
!enddefine.


******** K-MEANS CLUSTERING ********.
*/*!KO_kmeans(data%ini%iter%empty%name1%name2%name3%name4)*/*.
*Version 1.
*Cluster analysis by k-means method.
*Takes scale data DATA (n cases x p variables) and clusters the cases in specified number of
*clusters k (k<n). Initial centres must be offered by the user.
*INI - matrix of initial centres (means) sized k clusters x p variables (columns in DATA).
*ITER - do this number of iterations (nonnegative integer scalar, e.g. 10). ITER=0 corresponds to
*classification without iterations, i.e. the assignment of cases to initial centres without improving of
*the latter.
*EMPTY - digit (not name or expression, you may optionally quote or apostrophe the digit) 0 or 1.
*It is specification of the way the problem of empty clusters is to be solved if such clusters emerge:
*"1" - get rid of empty cluster by reducing k; k will not be able to return to the initial value.
*"0" - leave empty cluster its previous centre, in hope that later it will pick cases (which may
*sometimes not happen).
*Empty clusters take place under bad initial centres and also with strongly discrete data.
*Results:
*NAME1 - the obtained cluster membership, in the form n x k binary matrix, i.e. k dummy variables
*corresponding to the clusters. Each case is assigned to one cluster: 1 stands in that column.
*You can create single categorical variable of cluster membership as: NAME1*t(1:ncol(NAME1)).
*NAME2 - k x p matrix of final, recalculated means: these are the centroids of NAME1 clusters.
*NAME3 - k x p  matrix of pre-final means: these are the latest centres to which cases were
*classified to.
*NAME4 - n x 1 column of squared euclidean distances from cases to centres NAME3, to which cases
*were last time assigned. For example, if a case according to NAME1 appeared to be 
*put in cluster 3, then its distance NAME4 to the centre of that cluster is the distance to the centre
*which is the 3rd row in NAME3, for just based on that distance the case was assigned to it.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_kmeans(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
              /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp !7= !3.
comp @k= nrow(!7).
comp @iter= !4.
!let !empty= !unquote(!5)
comp @ind= 0.
comp @cones= make(nrow(@data),1,1).
comp @rones= make(1,ncol(@data),1).
comp !6= make(nrow(@data),@k,0).
loop @it= 0 to @iter.
-comp !8= !7.
-loop @i= 1 to @k.
- comp !6(:,@i)= rssq(@data-!8(@i,:)(@cones,:)).
-end loop.
-comp !9= rmin(!6).
-comp @ass= make(nrow(@data),1,0).
-loop @i= 1 to @k.
- comp !6(:,@i)= (!6(:,@i)=!9)&*(not @ass).
- comp @ass= @ass+!6(:,@i).
-end loop.
-comp @n= t(csum(!6)).
-do if all(@n).
- comp !7= t(!6)*@data/(@n*@rones).
-else.
  !if (!empty) !then
-  comp @ass= @n>0.
-  comp @nnz= csum(@ass).
-  comp @ind= make(1,@nnz+1,0).
-  comp @c= 1.
-  loop @i= 1 to @k.
-   comp @ind(@c)= @i.
-   comp @c= @c+@ass(@i).
-  end loop.
-  comp @ind= @ind(1:@nnz).
-  comp @n= @n(@ind).
-  comp !6= !6(:,@ind).
-  comp @k= @nnz.
-  comp !7= t(!6)*@data/(@n*@rones).
-  release @nnz,@c.
  !else
-  comp @ass= @n=0.
-  comp @ind= @ass*@rones&*!8.
-  comp @n= @n+@ass.
-  comp !7= t(!6)*@data/(@n*@rones)+@ind.
  !ifend
-end if.
end loop.
release @data,@k,@iter,@cones,@rones,@it,@i,@ass,@n,@ind.
!enddefine.


******** BLOCK-DIAGONALIZATION BY VAT/IVAT METHOD ********.
*/*!KO_vat(mat%method%priority%name1%name2)*/*.
*Version 1.
*Reorders rows/columns of square distance matrix MAT by VAT (and IVAT) method.
*It is one of the ways (Bezdek, Hathaway, "VAT: a tool for...", 2002; Havens, Bezdek, "An efficient
*formulation of...", 2012) of matrix block-diagonalization which aim is to concentrate small
*distance values near the matrix diagonal. If to render a block-diagonalized matrix on the "heatmap"
*chart, one can get an impression about the presence of cluster structure in data and to estimate
*tentatively the number of clusters.
*The matrix after block-diagonalization (rows/columns permutation) - result NAME1.
*Input:
*MAT - square symmetric nonnegative matrix of distances (or costs) with zero diagonal.
*METHOD - capitalized keyword (optionally in quotes or apostrophes):
*"VAT" - do reordering by VAT algorithm. It is closely tied with Prim's algorithm of growing
*minimal spanning tree. The obtained order will be written down to vector NAME2.
*"IVAT" - having done the VAT reordering, then do iVAT operation. It specifically replaces
*some distances in the matrix by other distances of it, reducing diversity of distances in the
*matrix. The effect of iVAT-replacement is that on the "heatmap" (1) the contrast between
*between-cluster and within-cluster distances will intensify, aiding visual detection of clusters;
*(2) detectability of clusters of chain structure (including strongly extended, tree-like, ring-like)
*will enhance. In its effect, iVAT operation is identical to processing the obtained in VAT matrix
*by the Floyd-Warshall algorithm in its "determine easiest passes" version.
*Thus, method=IVAT returns the matrix with reordered+affected (discretized) values, in comparison
*to those of MAT.
*PRIORITY - digit (not name or expression) 0 or 1 (optionally in quotes or apostrophes).
*It is a technical parameter potentially affecting only if there are equal values (ties) in MAT.
*It defines the order of search/select of an element. 0 is search "from smaller to larger": the
*element with a smaller index will be preferred. 1 is search "from larger to smaller": the element
*with a larger index will be preferred (this variant is slightly faster). PRIORITY doesn't change
*validity of the result, but under ties condition the result may differ, as a consequense of that
*the order (NAME2) may become another.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_vat(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !method= !unquote(!3)
!let !pr= !unquote(!4)
comp !5= !2.
comp @n= ncol(!5).
!if (!pr) !then !let !nums1= '1:ncol(@m)' !let !nums2= '1:nrow(@m)'
!else
 !let !nums1= 'ncol(@m):1' !let !nums2= 'nrow(@m):1'
 !let !comp1= 'comp @ci= ncol(@m)+1-@ci' !let !comp2= 'comp @ri= nrow(@m)+1-@ri'
!ifend
call setdiag(!5,1E300).
comp !6= mmax((!5=mmin(!5))&*(1:@n)(make(1,@n,1),:)).
call setdiag(!5,0).
do if !6=1.
-comp @out= 2:@n.
else if !6=@n.
-comp @out= 1:(@n-1).
else.
-comp @out= {1:(!6-1),(!6+1):@n}.
end if.
!if (!method='IVAT') !then
-comp @rilst= 0.
!ifend
loop @i= 2 to @n.
-comp @m= !5(!6,@out).
-comp @val= mmin(@m).
-comp @m= @m=@val.
-comp @ci= mmax(@m&*(!nums1)(make(1,nrow(@m),1),:)).
-!comp1.
-comp !6= {!6,@out(@ci)}.
!if (!method='IVAT') !then
-comp @ri= mmax(@m(:,@ci)&*t(!nums2)).
-!comp2.
-comp @rilst= {@rilst,@ri}.
!ifend
-do if @i<@n.
- do if @ci=1.
-  comp @out= @out(2:ncol(@out)).
- else if @ci=ncol(@out).
-  comp @out= @out(1:(ncol(@out)-1)).
- else.
-  comp @out= @out({1:(@ci-1),(@ci+1):ncol(@out)}).
- end if.
-end if.
end loop.
comp !5= !5(!6,!6).
release @out,@ci.
!if (!method='IVAT') !then
-comp @m= make(@n,@n,0).
-loop @i= 2 to @n.
- comp @j= @rilst(@i).
- comp @val= !5(@i,@j).
- comp @m(@i,1:(@i-1))= make(1,@i-1,@val).
- loop @c= 1 to @j-1.
-  comp @m(@i,@c)= mmax({@val,@m(@j,@c)}).
- end loop.
-end loop.
-comp !5= @m+t(@m).
-release @rilst,@ri,@j,@c.
!ifend
release @n,@i,@val,@m.
!enddefine.


******** RANDOM VALUES FROM STANDARD NORMAL DISTRIBUTION ********.
*/*!KO_normal(nr%nc%name)*/*.
*Version 1.
*Generates matrix NAME sized NR x NC, which elements are random
*values from standard normal distribution.
*NR, NC – positive integer scalars.
*To manage random numbers  seed use out-of-matrix generator
*of random numbers (corresponds to commands SET MTINDEX or SET SEED).
*To generate data from population with correlating normal variables use
*more general function /*!KO_mvnorm*/.

define !KO_normal(!pos= !token(1) /!pos= !charend('%')  /!pos= !charend('%') /!pos= !charend(')'))
comp @nr= !2.
comp @nc= !3.
comp !4= sqrt(-2*ln(uniform(@nr,@nc)))&*cos(6.28318530717959*uniform(@nr,@nc)).
release @nr,@nc.
!enddefine.


******** RANDOM VALUES FROM NORMAL DISTRIBUTION WITH SPECIFIED PARAMETERS ********.
*/*!KO_rvnorm(mean%sd%name)*/*.
*Version 1.
*Generates random number from normal distribution with mean MEAN and standard deviation SD.
*MEAN and SD - matrices of any, same size. Result NAME - of the same size.
*Element ij NAME contains random value from normal distribution with parameters ij MEAN and ij SD.
*To manage random numbers  seed use out-of-matrix generator
*of random numbers (corresponds to commands SET MTINDEX or SET SEED).

define !KO_rvnorm(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mean= !2.
comp @sd= !3.
comp !4= sqrt(-2*ln(uniform(nrow(@mean),ncol(@mean))))&*cos(6.28318530717959*uniform(nrow(@mean),ncol(@mean))).
comp !4= !4&*@sd+@mean.
release @mean,@sd.
!enddefine.


******** RANDOM VALUES FROM UNIFORM DISTRIBUTION WITH SPECIFIED PARAMETERS ********.
*/*!KO_rvunif(min%max%name)*/*.
*Version 1.
*Generates random number from uniform distribution with minimum MIN and maximum MAX.
*MIN and MAX - matrices of any, same size. Result NAME - of the same size.
*Element ij NAME contains random value from uniform distribution with parameters ij MIN and ij MAX.
*(ij MIN must be not greater than ij MAX).
*To manage random numbers  seed use out-of-matrix generator
*of random numbers (corresponds to commands SET MTINDEX or SET SEED).

define !KO_rvunif(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @min= !2.
comp @max= !3.
comp !4= uniform(nrow(@min),ncol(@min)).
comp !4= (@max-@min)&*!4+@min.
release @min,@max.
!enddefine.


******** RANDOM VALUES FROM CHI-SQUARE DISTRIBUTION ********.
*/*!KO_chisq(nr%nc%df%name)*/*.
*Version 1.
*Generates matrix NAME sized NR x NC, which elements are random
*values from chi-square distribution with degrees of freedom DF.
*NR, NC, DF – positive integer scalars.
*To manage random numbers  seed use out-of-matrix generator
*of random numbers (corresponds to commands SET MTINDEX or SET SEED).

define !KO_chisq(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @nr= !2.
comp @nc= !3.
comp @df= !4.
comp !5= sqrt(-2*ln(uniform(@df,@nr*@nc)))&*cos(6.28318530717959*uniform(@df,@nr*@nc)). 
comp !5= reshape(cssq(!5),@nr,@nc).
release @nr,@nc,@df.
!enddefine.


******** RANDOM VALUES FROM BINOMIAL DISTRIBUTION ********.
*/*!KO_binom(nr%ntrials%prob%name)*/*.
*Version 1.
*Generates column vector NAME of length NR, which elements are random values from
*binomial distribution with parameters number of trials NTRIALS and success (event) probability PROB.
*NTRIALS (positive integer) and PROB (between 0 and 1) - either two scalars or two vectors length NR.
*If vectors, it means different values of NAME can be generated under different values of the parameters.
*To manage random numbers  seed use out-of-matrix generator
*of random numbers (corresponds to commands SET MTINDEX or SET SEED).
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_binom(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @nr= !2.
comp @n= !3.
comp @p= !4.
do if nrow(@n)*ncol(@n)*nrow(@p)*ncol(@p)=1.
-comp !5= rsum(uniform(@nr,@n)<=@p).
else.
-comp !5= make(@nr,1,0).
-loop @i= 1 to @nr.
- comp !5(@i)= rsum(uniform(1,@n(@i))<=@p(@i)).
-end loop.
-release @i.
end if.
release @nr,@n,@p.
!enddefine.


******** RANDOM VALUES FROM CATEGORICAL DISTRIBUTION ********.
*/*!KO_categ(nr%nc%kprob%name)*/*.
*Version 2.
*Generates matrix NAME sized NR x NC, which elements are random
*values (codes 1 to k) from k-categorical distribution. (Categorical distribution is
*multinomial distribution with number of trials 1. Bernoulli distribution is 2-categorical
*distribution.)
*NR, NC – positive integer scalars.
*KPROB - either vector of length k, containing probabilities for the k categories, or positive integer
*scalar equal to k.
*If KPROB - probability vector, it must be nonnegative numbers, at least one positive. The numbers
*need not sum to 1: the function will bring their sum to 1; it is suffice that the numbers are in
*the same ratios as the probabilities.
*If KPROB - scalar, it must be positive integer (fractional value will be truncated).
*It is k, and the k propabilities are supposed to be equal. When the task is the equiprobable categorical
*distribution this way of specifying is more convenient than by probability vector, because it will be
*faster result.
*To manage random numbers  seed use out-of-matrix generator
*of random numbers (corresponds to commands SET MTINDEX or SET SEED).
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_categ(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @nr= !2.
comp @nc= !3.
comp @kprob= !4.
comp @k= nrow(@kprob)*ncol(@kprob).
do if @k=1.
-comp !5= trunc(uniform(!2,!3)*trunc(@kprob))+1.
else.
-comp @kprob= @kprob/msum(@kprob).
-comp !5= uniform(!2,!3).
-comp @cump= 0.
-loop @i= 1 to @k.
- comp @cump= @cump+@kprob(@i).
- comp !5= !5+(!5<=@cump)*@i.
-end loop.
-comp !5= trunc(!5).
-release @cump,@i.
end if.
release @nr,@nc,@kprob,@k.
!enddefine.


******** SAMPLE FROM CATEGORICAL DISTRIBUTION WITHOUT REPLACEMENT ********.
*/*!KO_catwor(n%kprob%stock%method%name1%name2)*/*.
*Version 1.
*Collects random sample of N objects from a limited set of objects belonging to different
*categories. This function returns random values from k-categorical distribution
*(k is the number of object categories), like function /*!KO_categ*/, but it does sampling without
*replacement from finite populatioin, whereas /*!KO_categ*/ does sampling with replacement.
*Because objects are drawn one by one and are not returned back to "stock" (i.e. to limited
*population) the probabilities of selection ("realization") for different categories change at
*the process flow. These probabilities are determined, from one side, by the specified probabilities
*of "demand", and, from other side, by the current probabilities of "availability in stock", so to say.
*Input:
*N - positive integer scalar: the required sample size.
*KPROB - vector of length k, containing probabilities for the k categories of objects. It must be
*nonnegative numbers, at least one positive. The numbers need not sum to 1: the function will bring
*their sum to 1; it is suffice that the numbers are in the same ratios as the probabilities.
*KPROB should be understood as the initial probabilities of "demand"; i.e. they are probabilities
*of selection if the population had been infinite.
*STOCK - vector of length k and same orientation as KPROB, containing frequencies for the k categories
*of objects - it is the observed distribution of our finite population (the "stock") from where
*objects will be drawn into the sample. The frequencies must be nonnegative integers, at least one
*positive. If you input fractional values the function will first truncate them to integers.
*While weights KPROB set "demand" probabilities for categories, frequencies STOCK set "availability"
*probabilities for them.
*METHOD - digit (not variable name or expression; you may optionally quote or apostrophe the digit).
*It is specification of the method of determination of current probabilities of selection (realization):
*"1" - method "Initial demand, if there left". Probability PROB(i) to select now object of category i
*is proportional to KPROB(i)*LEFT(i), where LEFT(i)=1, if there are objects of category i left yet in
*the population, otherwise LEFT(i)=0. If frequencies STOCK are all big enough to be not emptied prior
*the end of the sampling procedure then this method is equivalent to application of function /*!KO_categ*/.
*"2" - method "Initial demand * how many left". Probability PROB(i) to select now object of category i
*is proportional to KPROB(i)*FLEFT(i), where FLEFT(i) is the fraction of category i among the objects yet
*left in the population. It is classical random selection without replacement.
*"3" - method "Floating demand * how many left". Probability PROB(i) to select now object of category i
*is proportional to PROB_(i)*FLEFT(i), where PROB_(i) is PROB(i) computed before selection of the
*preceding object into the sample (before selection of the 1st object PROB_(i)=KPROB(i)). It is
*"sharpening" method leading to soon, during selections, shift of preference toward one category,
*typically the one with high representedness on the "stock", the limited population.
*"4" - method "Floating non-demand * how many left". Probability PROB(i) to select now object of
*category i is proportional to (1-PROB_(i))*FLEFT(i) (before selection of the 1st object
*1-PROB_(i)=KPROB(i)). This method can be interesting as a "smoothing" alternative to method "2":
*selection from low-represented categories will be somewhat enhanced and selection from high-represented
*categories will be somewhat reduced, in comparison with that method.
*Results:
*NAME1 - column vector of length N containing category codes (numbers from sequence 1,2,...,k)
*of the created random sample (and with sequence there also random). In case when there happened not
*enough objects in the population to draw N objects without relpacement, length of NAME1 will be
*shorter than N. If there failed to draw even a single object, NAME1=0.
*NAME2 - column vector of length k containing the frequencies left in STOCK.
*To manage random numbers  seed use out-of-matrix generator
*of random numbers (corresponds to commands SET MTINDEX or SET SEED).
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*If KPROB are equal probabilities and STOCK is vector of ones or is binary (1 and 0), then equivalent
*will be to use function /*!KO_sample*/ which is faster.

define !KO_catwor(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
              /!pos= !charend('%') /!pos= !charend(')'))
comp @n= !2.
comp @kprob= !3.
comp !7= trunc(!4).
!let !method= !unquote(!5)
comp @k= nrow(@kprob)*ncol(@kprob).
comp @tstock= msum(!7).
comp !6= uniform(@n,1).
!if (!method='2') !then !let !kprob= '@kprob'
!else !if (!method='3') !then !let !kprob= '@prob'
-comp @prob= @kprob.
!else !if (!method='4') !then !let !kprob= '(1-@prob)'
-comp @prob= 1-@kprob/msum(@kprob).
!ifend !ifend !ifend
loop @c= 1 to @n.
 !if (!method='1') !then
- comp @prob= @kprob&*(!7>0).
 !else
- comp @prob= !kprob&*(!7/@tstock).
 !ifend
-comp @psum= msum(@prob).
-do if @psum.
- comp @prob= @prob/@psum.
- comp @objc= !6(@c).
- comp @cump= 0.
- loop @i= 1 to @k.
-  comp @cump= @cump+@prob(@i).
-  do if @objc<=@cump.
-   comp !6(@c)= @i.
-   comp !7(@i)= !7(@i)-1.
-   comp @tstock= @tstock-1.
-   break.
-  end if.
- end loop.
-else.
- break.
-end if.
end loop if not @tstock.
do if @c<=@n.
-comp @c= @c-(not @psum).
-do if @c.
- comp !6= !6(1:@c).
-else.
- comp !6= 0.
-end if.
end if.
do if nrow(!7)=1.
-comp !7= t(!7).
end if.
release @n,@kprob,@k,@tstock,@c,@prob,@psum,@objc,@cump,@i.
!enddefine.


******** SELECT N RANDOM ELEMENTS FROM VECTOR ********.
*/*!KO_sample(n%cases%name1%name2)*/*.
*Version 1.
*Randomly selects N different elements from sequence of numbers 1,2,...,k.
*This function is convenient for forming a random subsample ("without replacement") from a dataset,
*and also for random sorting of cases in a dataset.
*Input:
*N - positive integer scalar: how many different elements to select.
*CASES - "population" of elements: either positive integer scalar or binary (0 vs 1) vector.
*If scalar, then it is k, and selection will be from numbers 1,2,...,k. If binary vector, then its
*length is k and its element numbers (indices) correspond to numbers 1,2,...,k; element's value 1
*in the vector permits to select the element's number while value 0 prohibits to select. Thus, vector
*CASES is filter. Vector consisting of just 1s is equivalent to specifying CASES as scalar.
*Vector must contain at least one 1.
*Results:
*NAME1 - column vector of length N containing the randomly selected numbers, and in random order.
*NAME2 - binary column vector of length k, where 1 flags not selected (remaining) element CASES,
*and 0 flags selected (removed) element.
*If N>=k length of NAME1 will be k, and NAME2 will consist only of zeros.
*To manage random numbers  seed use out-of-matrix generator
*of random numbers (corresponds to commands SET MTINDEX or SET SEED).
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*The function is relatively slow. If k is great but N is small, then sampling with replacement that
*you can accomplish with the help of function /*!KO_categ*/ will be quite good approximation to the
*sampling without replacement; but /*!KO_categ*/ is faster than /*!KO_sample*/.
*If your population CASES is weighted, i.e. consists of counts (rather than 0 or 1), or if its elements
*are not equiprobable for selection, use function /*!KO_catwor*/.

define !KO_sample(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')')) 
comp @n= !2.
comp !5= !3.
do if nrow(!5)*ncol(!5)=1.
-comp !5= make(!5,1,1).
else if nrow(!5)=1.
-comp !5= t(!5).
end if.
comp @ns= csum(!5).
comp @n_= rmin({@n,@ns}).
comp @lnk= ln(nrow(!5)).
comp @ns= ln(@ns/nrow(!5)).
do if nrow(!5)<15000 and ln(@n)>22.189-147.7272/@lnk+.78529*@ns+268.29/@lnk**2+.044*@ns**2+3.266*@ns/@lnk.
-comp !4= {uniform(nrow(!5),1)&*!5,t(1:nrow(!5))}.
-comp !4(grade(-!4(:,1)),:)= !4.
-comp !4= !4(1:@n_,2).
-comp !5(!4)= make(@n_,1,0).
else.
-comp @g= 0.
-comp @val= 0.
-comp !4= trunc(uniform(@n_,1)*nrow(!5))+1.
-loop @i= 1 to @n_.
- do if !5(!4(@i)).
-  comp !5(!4(@i))= 0.
- else.
-  loop @g= 1 to 1E12.
-   comp @val= trunc(uniform(1,1)*nrow(!5))+1.
-  end loop if !5(@val).
-  comp !4(@i)= @val.
-  comp !5(@val)= 0.
- end if.
-end loop.
-release @i,@g,@val.
end if.
release @n,@n_,@ns,@lnk.
!enddefine.


******** RANDOM MATRIX FROM WISHART DISTRIBUTION ********.
*/*!KO_wishart(chl%n%name)*/*.
*Version 1.
*Returns random matrix NAME from Wishart distribution with scale parameter C and degrees of freedom N.
*C is a population covariance matrix (multivariate normal distribution is assumed in the
*population). Argument CHL of the function is not C, but its Cholesky root: upper triangular
*matrix returned by the in-built matrix function chol(C).
*N (positive integer scalar) is the sample size for generating of the random matrix NAME.
*The latter is a scatter matrix; to turn it into covariance matrix divide it by N.
*If as a CHL you specify a scalar (positive) then it is counted a population standard deviation,
*and NAME/N will be a random value of variance.
*To manage random numbers  seed use out-of-matrix generator
*of random numbers (corresponds to commands SET MTINDEX or SET SEED).
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_wishart(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @n= !3.
comp @a= sqrt(-2*ln(uniform(ncol(!4),ncol(!4))))&*cos(6.28318530717959*uniform(ncol(!4),ncol(!4))).
comp @v= sqrt(-2*ln(uniform(1,@n)))&*cos(6.28318530717959*uniform(1,@n)).
comp @a(1,1)= sqrt(rssq(@v)).
loop @i= 2 to nrow(@a).
-comp @v= sqrt(-2*ln(uniform(1,@n-@i+1)))&*cos(6.28318530717959*uniform(1,@n-@i+1)).
-comp @a(@i,@i)= sqrt(rssq(@v)).
-comp @a(@i,1:(@i-1))= make(1,@i-1,0).
end loop.
comp !4= @a*!4.
comp !4= sscp(!4).
release @n,@a,@v,@i.
!enddefine.


******** RANDOM DATA FROM NORMAL POPULATION WITH SPECIFIED COVARIANCES ********.
*/*!KO_mvnorm(n%chl%mean%name)*/*.
*Version 1.
*Generates data NAME - random sample of size N from normally distributed population having
*covariances C and centroid MEAN.
*Argument CHL of the function is not covariance matrix C (size p x p) itself between variables,
*but its Cholesky root: upper triangular matrix (of same size) returned by the in-built matrix
*function chol(C).
*N - number of cases to generate, positive integer scalar.
*MEAN - scalar or row vector of length p: this is the means of the variables in the population.
*Result NAME - matrix of random data n cases x p variables from normal population of the specified
*covariances between the variables and the means of the variables.
*In a particular case, C can be of size 1 x 1; then C is the variance (and CHL is st. deviation).
*If matrix C is correlations, not covariances, keep argument MEAN as 0, the obtained variables NAME
*you can then multiply by the population standard deviations and then add the population means;
*however, more convenient way is: convert correlation matrix into covariance one by function
*/*!KO_corrcov*/, after which apply /*!KO_mvnorm*/ with specifying of the means as MEAN in it.
*To manage random numbers  seed use out-of-matrix generator
*of random numbers (corresponds to commands SET MTINDEX or SET SEED).

define !KO_mvnorm(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @n= !2.
comp !5= !3.
comp @mean= !4.
comp !5= sqrt(-2*ln(uniform(@n,ncol(!5))))&*cos(6.28318530717959*uniform(@n,ncol(!5)))*!5.
do if ncol(@mean)=1.
-comp !5= !5+@mean.
else.
-comp !5= !5+@mean(make(1,@n,1),:).
end if.
release @n,@mean.
!enddefine.


******** RECODE VALUES (EXACT MATCH) ********.
*/*!KO_recode1(data%filter%old%new%name)*/*.
*Version 1.
*In data matrix DATA of any size replaces values of list OLD by corresponding values of list NEW.
*Returns matrix NAME which is matrix DATA with the values thus recoded.
*This function is similar to the out-of-MATRIX command RECODE in which the values to recode are specified
*by listing.
*OLD and NEW - row vectors. i-th value in OLD, if found in DATA, is recoded into i-th value of NEW.
*Both these vectors are of equal length or NEW may be by one element longer; in the latter case all other
*than OLD values will be recoded into that last value of NEW.
*If some value encounters in OLD more than once it is recoded into the NEW value corresponding to the first
*(leftmost) its encounter in the vector OLD.
*Argument FILTER allows to prohibit recoding of values in some rows of DATA. FILTER is either a scalar or
*a column lengthed as the number of rows in DATA. If nonzero scalar - no ban exist; if zero scalar -
*then complete ban (function won't act, NAME will be equal to DATA). If column, then 0 in it bans
*recoding in that row of DATA, while other than 0 value - allows recoding in the row.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_recode1(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @filt= !3.
comp @old= !4.
comp @new= !5.
do if any(@filt).
-comp @nflagm= make(nrow(@data),ncol(@data),1).
-comp @newm= make(nrow(@data),ncol(@data),0).
-loop @i= 1 to ncol(@old).
- comp @flagi= (@data=@old(@i))&*@nflagm.
- comp @newm= @newm+@flagi*@new(@i).
- comp @nflagm= @nflagm-@flagi.
-end loop.
-comp !6= @data&*@nflagm+@newm.
-do if ncol(@new)>ncol(@old).
- comp !6= !6&*(not @nflagm)+@nflagm*@new(@i).
-end if.
-do if not all(@filt).
- comp @filt= (@filt<>0)*make(1,ncol(@data),1).
- comp !6= !6&*@filt+@data&*(not @filt).
-end if.
-release @nflagm,@newm,@i,@flagi.
else.
-comp !6= @data.
end if.
release @data,@filt,@old,@new.
!enddefine.


******** RECODE VALUES (HITTING INTO RANGE) ********.
*/*!KO_recode2(data%filter%low%high%new%name)*/*.
*Version 1.
*In data matrix DATA of any size replaces values falling into range from LOW to HIGH (inclusive)
*by corresponding values of list NEW.
*Returns matrix NAME which is matrix DATA with the values thus recoded.
*This function is similar to the out-of-MATRIX command RECODE in which the values to recode are specified
*by range(s) from... to...
*LOW, HIGH, and NEW - row vectors. LOW(i) and HIGH(i) is a range; HIGH(i) should normally be >= LOW(i).
*Any value in DATA lying in the range from LOW(i) to HIGH(i) will be recoded into
*value NEW(i). LOW and HIGH must be of equal length. NEW may be by one element longer; in the latter
*case all other than the recoded values will be recoded into that last value of NEW.
*If some ranges specified by vectors LOW and HIGH superimpose, their common values
*are recoded into the NEW value corresponding to the earliest (leftmost in the vectors) of these ranges.
*For example, if LOW is {6,5} and HIGH is {11,10}, i.e. intersecting ranges 6-11 and 5-10 were specified,
*while NEW is {23,37}, then "disputable" data values - from 6 to 10 - will be recoded into 23,
*not into 37, because range 6-11 is specified "earlier" (more left in the vectors) than 5-10.
*(The same way command RECODE behaves.)
*Argument FILTER allows to prohibit recoding of values in some rows of DATA. FILTER is either a scalar or
*a column lengthed as the number of rows in DATA. If nonzero scalar - no ban exist; if zero scalar -
*then complete ban (function won't act, NAME will be equal to DATA). If column, then 0 in it bans
*recoding in that row of DATA, while other than 0 value - allows recoding in the row.
*If vectors LOW and HIGH are identical the function will return the same result as /*!KO_recode1*/.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_recode2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @filt= !3.
comp @lo= !4.
comp @hi= !5.
comp @new= !6.
do if any(@filt).
-comp @nflagm= make(nrow(@data),ncol(@data),1).
-comp @newm= make(nrow(@data),ncol(@data),0).
-loop @i= 1 to ncol(@lo).
- comp @flagi= (@data>=@lo(@i) and @data<=@hi(@i))&*@nflagm.
- comp @newm= @newm+@flagi*@new(@i).
- comp @nflagm= @nflagm-@flagi.
-end loop.
-comp !7= @data&*@nflagm+@newm.
-do if ncol(@new)>ncol(@lo).
- comp !7= !7&*(not @nflagm)+@nflagm*@new(@i).
-end if.
-do if not all(@filt).
- comp @filt= (@filt<>0)*make(1,ncol(@data),1).
- comp !7= !7&*@filt+@data&*(not @filt).
-end if.
-release @nflagm,@newm,@i,@flagi.
else.
-comp !7= @data.
end if.
release @data,@filt,@lo,@hi,@new.
!enddefine.


******** VARIOUS UNIVARIATE STATISTICS, BY GROUPS: REPLACE ORIGINAL VALUES WITH THEM ********.
*/*!KO_aggrv(data%bin%stat%check%value1%value2%name)*/*.
*Version 1.
*This function computes a requested descriptive statistic by groups - exactly as function /*!KO_aggr*/ does,
*but it returns, as result NAME, not the groups X variables table but the dataset sized as DATA, in
*which the initial values of DATA are replaced with the computed summary statistic: if a row (case)
*belongs to group i and only to it then its values in NAME is the row i of the mentioned table returned
*by function /*!KO_aggr*/. Therefore, /*!KO_aggrv()*/ performs what SPSS command AGGREGATE under option
*MODE=ADDVARIABLES does.
*Arguments DATA, BIN, STAT are exactly as in function /*!KO_aggr*/ - see.
*Argument CHECK (digit 0 or 1, not name or expression):
*If you know that sums in all the rows of BIN equal 1, i.e. that each case belongs to just one group,
*so that BIN are dummy variables, set CHECK to 0. While if you admit that sums in
*some rows of BIN may not equal 1, set CHECK to 1.
*Arguments VALUE1 and VALUE2 will be ignored if СHECK is 0.
*VALUE1 (number, not name or expression): specify value which should be inserted in NAME for those cases
*which do not belong to any group (i.e. whose sum in row of BIN =0).
*VALUE2: specify value (number, not name or expression), which should be inserted in NAME for those cases
*which belong to more than one group (i.e. whose sum in row of BIN >1); or specify in capital letters
*the keyword "AVER" - which means the request, for a case belonging to more than one group
*and thus having more than one computed value of the statistic at its disposal, to average these its
*values of the statistic.
*Values of arguments CHECK, VALUE1, VALUE2 may be optionnally wrapped in quotes or apostrohes.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*If your grouping is defined by a categorical variable, create dummy variables BIN out of it
*with the help of function /*!KO_freq*/.

define !KO_aggrv(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
             /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @bin= !3.
!let !stat= !unquote(!4)
!if (!index('SUM FIRST LAST MIN MAX',!stat)=0) !then !let !n= ',@n'
-comp @n= csum(@bin).
!ifend
!if (!stat='SUM') !then
-comp !8= t(@bin)*@data.
!else !if (!stat='MEAN') !then
-comp !8= t(@bin)*@data/(t(@n)*make(1,ncol(@data),1)).
!else !if (!stat='SSDEV' !or !stat='VARIANCE' !or !stat='VARIANCE2' !or !stat='SEMEAN') !then
-comp @sum= t(@bin)*@data.
-comp @cssq= t(@bin)*(@data&*@data).
-comp @n= t(@n)*make(1,ncol(@data),1).
-comp !8= @cssq-@sum&*@sum/@n.
 !if (!stat<>'SSDEV') !then
- comp !8= !8/(@n !if (!stat<>'VARIANCE2') !then -1 !ifend).
  !if (!stat='SEMEAN') !then
-  comp !8= sqrt(!8/@n).
  !ifend
 !ifend
-release @sum,@cssq.
!else !if (!stat='MIN') !then
-comp @ones= make(1,ncol(@data),1).
-comp !8= make(ncol(@bin),ncol(@data),0).
-comp @bigv= 1E300.
-loop @i= 1 to ncol(@bin).
- comp @bini= @bin(:,@i)*@ones.
- comp !8(@i,:)= cmin(@data&*@bini+(not @bini)*@bigv).
-end loop.
-release @ones,@bigv,@i,@bini.
!else !if (!stat='MAX') !then
-comp @ones= make(1,ncol(@data),1).
-comp !8= make(ncol(@bin),ncol(@data),0).
-comp @smallv= mmin(@data).
-loop @i= 1 to ncol(@bin).
- comp @bini= @bin(:,@i)*@ones.
- comp !8(@i,:)= cmax(@data&*@bini+(not @bini)*@smallv).
-end loop.
-release @ones,@smallv,@i,@bini.
!else !if (!stat='LAST') !then
-comp !8= make(ncol(@bin),ncol(@data),0).
-comp @rnums= t(1:nrow(@data)).
-loop @i= 1 to ncol(@bin).
- comp @ind= cmax(@bin(:,@i)&*@rnums).
- comp !8(@i,:)= @data(@ind,:).
-end loop.
-release @rnums,@i,@ind.
!else !if (!stat='FIRST') !then
-comp !8= make(ncol(@bin),ncol(@data),0).
-comp @rnums= t(nrow(@data):1).
-loop @i= 1 to ncol(@bin).
- comp @ind= nrow(@data)-cmax(@bin(:,@i)&*@rnums)+1.
- comp !8(@i,:)= @data(@ind,:).
-end loop.
-release @rnums,@i,@ind.
!ifend !ifend !ifend !ifend !ifend !ifend !ifend
comp !8= @bin*!8.
!if (!unquote(!5)) !then !let !n= ',@n'
-comp @ones= make(1,ncol(@data),1).
-comp @n= rsum(@bin).
-comp @notn= not @n.
-comp !8= (not @notn)*@ones&*!8+@notn*!unquote(!6)*@ones.
 !if (!unquote(!7)='AVER') !then
- comp !8= !8/((@n+@notn)*@ones).
-!else
- comp @notn= @n>1.
- comp !8= (not @notn)*@ones&*!8+@notn*!unquote(!7)*@ones.
 !ifend
-release @notn,@ones.
!ifend
release @data,@bin !n.
!enddefine.


******** POSITIONS OF NONZERO ELEMENTS IN VECTOR ********.
*/*!KO_indices(vec%name)*/*.
*Version 2.
*Takes vector (row or column) VEC and returns row vector NAME containing the list of indices
*(sequential numbers, positions) of nonzero elements in VEC. The indices go ascendingly.
*If there is no nonzero elements in VEC the function returns scalar 0.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_indices(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @vec= !2.
comp @vec= @vec<>0.
comp @n= nrow(@vec)*ncol(@vec).
comp @nnz= msum(@vec).
do if not @nnz.
-comp !3= 0.
else if @nnz=@n.
-comp !3= 1:@n.
else if @nnz/@n>.8.
-comp !3= make(1,@nnz+1,0).
-comp @c= 1.
-loop @i= 1 to @n.
- comp !3(@c)= @i.
- comp @c= @c+@vec(@i).
-end loop.
-comp !3= !3(1:@nnz).
-release @c,@i.
else if @nnz/@n<.08.
-comp !3= make(1,@nnz,0).
-comp @s= 1.
-loop @c= 1 to @nnz.
- loop @i= @s to @n.
- end loop if @vec(@i).
- comp !3(@c)= @i.
- comp @s= @i+1.
-end loop.
-release @s,@c,@i.
else.
-comp !3= make(1,@nnz,0).
-comp @c= 0.
-loop @i= 1 to @n.
- do if @vec(@i).
-  comp @c= @c+1.
-  comp !3(@c)= @i.
- end if.
-end loop.
-release @c,@i.
end if.
release @vec,@n,@nnz.
!enddefine.


******** POSITIONS OF NONZERO AND ZERO ELEMENTS IN VECTOR ********.
*/*!KO_indices2(vec%name1%name2)*/*.
*Version 1.
*Takes vector (row or column) VEC and returns row vector NAME1 containing the list of indices
*(sequential numbers, positions) of nonzero elements in VEC, and row vector NAME2 containing
*the list of indices of zero elements in VEC. The indices go ascendingly.
*If there is no nonzero elements in VEC, NAME1 will be scalar 0. If there is no zero elements
*in VEC, NAME2 will be scalar 0.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_indices2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @vec= !2.
comp @vec= @vec<>0.
comp @n= nrow(@vec)*ncol(@vec).
comp @nnz= msum(@vec).
do if not @nnz.
-comp !3= 0.
-comp !4= 1:@n.
else if @nnz=@n.
-comp !3= 1:@n.
-comp !4= 0.
else if @nnz/@n>.3.
-comp !3= make(1,@nnz,0).
-comp !4= make(1,@n-@nnz,0).
-comp @c= 0.
-comp @c2= 0.
-loop @i= 1 to @n.
- do if @vec(@i).
-  comp @c= @c+1.
-  comp !3(@c)= @i.
- else.
-  comp @c2= @c2+1.
-  comp !4(@c2)= @i.
- end if.
-end loop.
-release @c,@c2,@i.
else.
-do if nrow(@vec)=1.
- comp @vec= {@vec,1}.
-else.
- comp @vec= {@vec;1}.
-end if.
-comp @n= @n+1.
-comp !3= make(1,@nnz+1,0).
-comp !4= 0.
-comp @s= 1.
-loop @c= 1 to @nnz+1.
- loop @i= @s to @n.
- end loop if @vec(@i).
- comp !3(@c)= @i.
- do if @i>@s.
-  comp !4= {!4,@s:(@i-1)}.
- end if.
- comp @s= @i+1.
-end loop.
-release @s,@c,@i.
-comp !3= !3(1:(ncol(!3)-1)).
-comp !4= !4(2:ncol(!4)).
end if.
release @vec,@n,@nnz.
!enddefine.


******** POSITIONS OF NONZERO ELEMENTS IN MATRIX (AND THE ELEMENTS THEMSELVES) ********.
*/*!KO_indicesm(mat%ANDVAL%name)*/*.
*Version 1.
*Takes matrix MAT of any size and returns two-column matrix NAME containing the list of indices
*(sequential numbers, positions) of nonzero elements in MAT. The indices go ascendingly.
*The row number is written in the 1st column of NAME and the column number is written in the
*2nd column of NAME.
*Argument ANDVAL allows also to write out the nonzero values themselves, as the 3rd column of NAME.
*ANDVAL is the digit (not name or expression) 0 or 1 (you may take the digit in quotes or
*apostrophes). If 1, the write-out of values will take place.
*If there is no nonzero elements in MAT, the function returns scalar 0.
*This function works somewhat faster on wide than on tall matrices.
*This function is cognate to /*!KO_nzlist*/ but returns result in somewhat different form.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_indicesm(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !andval= !unquote(!3)
!if (!andval) !then !let !nc= 3 !let !vals= ',t(@mat(@i,@rami))' !else !let !nc= 2 !ifend
comp @mat= !2.
comp @ram= @mat<>0.
comp @rnnz= rsum(@ram).
do if not any(@rnnz).
-comp !4= 0.
else if mmin(@rnnz)=ncol(@ram).
-comp @rind= t(1:nrow(@ram)).
-comp @cind= t(1:ncol(@ram)).
-comp !4= {kroneker(@rind,make(nrow(@cind),1,1)),kroneker(make(nrow(@rind),1,1),@cind)}.
 !if (!andval) !then
- comp !4= {!4,reshape(@mat,nrow(!4),1)}.
 !ifend
-release @rind,@cind.
else.
-comp @m= @ram.
-loop @i= 2 to ncol(@m).
- comp @m(:,@i)= @m(:,@i-1)+@m(:,@i).
-end loop.
-comp @m= @m&*@ram.
-comp @ram= make(nrow(@m),ncol(@m)+1,0).
-comp @nums= 1:ncol(@m).
-comp @m= @m+(not @m)*ncol(@ram).
-loop @i= 1 to nrow(@m).
- comp @ram(@i,@m(@i,:))= @nums.
-end loop.
-comp !4= make(msum(@rnnz),!nc,0).
-comp @st= 1.
-loop @i= 1 to nrow(@ram).
- comp @n= @rnnz(@i).
- do if @n.
-  comp @en= @st+@n-1.
-  comp @rami= @ram(@i,1:@n).
-  comp !4(@st:@en,:)= {make(@n,1,@i),t(@rami) !vals}.
-  comp @st= @en+1.
- end if.
-end loop.
-release @m,@i,@nums,@st,@en,@n,@rami.
end if. 
release @mat,@ram,@rnnz.
!enddefine.


******** POSITION OF FIRST ENCOUNTER OF VALUE IN ROW/ROWS ********.
*/*!KO_indx(mat%val%name)*/*.
*Version 1.
*Takes MAT - matrix or row vector. Returns index (position) of first (leftmost)
*encounter of value VAL in each row of MAT.
*VAL may be a scalar or a column vector of length as the number of rows in MAT; in the latter
*instance each row of MAT is given its corresponding value VAL.
*Result NAME - column vector with number of rows as in MAT.
*Position 0 means that VAL is not found in the row.

define !KO_indx(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @val= !3.
do if nrow(@val)>1.
-comp @val= @val*make(1,ncol(@mat),1).
end if.
comp @mat= (@mat=@val)&*(ncol(@mat):1)(make(1,nrow(@mat),1),:).
comp !4= rmax(@mat).
comp !4= (ncol(@mat)+1)*(!4>0)-!4.
release @mat,@val.
!enddefine.


********  POSITION OF LAST ENCOUNTER OF VALUE IN ROW/ROWS ********.
*/*!KO_rindx(mat%val%name)*/*.
*Version 1.
*Takes MAT - matrix or row vector. Returns index (position) of last (rightmost)
*encounter of value VAL in each row of MAT.
*VAL may be a scalar or a column vector of length as the number of rows in MAT; in the latter
*instance each row of MAT is given its corresponding value VAL.
*Result NAME - column vector with number of rows as in MAT.
*Position 0 means that VAL is not found in the row.

define !KO_rindx(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @val= !3.
do if nrow(@val)>1.
-comp @val= @val*make(1,ncol(@mat),1).
end if.
comp @mat= (@mat=@val)&*(1:ncol(@mat))(make(1,nrow(@mat),1),:).
comp !4= rmax(@mat).
release @mat,@val.
!enddefine.


******** POSITION (PAIR OF INDICES) OF ONE VALUE IN MATRIX ********.
*/*!KO_ij(mat%val%priority%name)*/*.
*Version 2.
*Finds in matrix MAT one element equal to value VAL (scalar).
*Returns, as NAME, the position (indices i,j) of such element in MAT.
*If there are more than one element equal to VAL are found, the position of one of them is
*returned according to argument PRIORITY. If none is found, {0,0} is returned.
*PRIORITY - capitalized keyword (may take in quotes or apostrophes):
*"TBLR" - search top to bottom, left to right: returned is position from top-most row,
*of left-most element.
*"LRTB" - search left to right, top to bottom: returned is position from left-most column,
*of top-most element.
*"BTRL" - search bottom to top, right to left: returned is position from bottom-most row,
*of right-most element.
*"RLBT" - search right to left, bottom to top: returned is position from right-most column,
*of bottom-most element.
*"RANDOM" - of all positions of VAL, return one randomly chosen.
*There is a tiny difference in performance speed; if you don't mind the priority itself but
*are greedy on account of speed, select RLBT as the fastest.

define !KO_ij(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !pr= !unquote(!4)
comp @mat= !2.
comp @val= !3.
comp @mat= @mat=@val.
!if (!pr='RLBT' !or !pr='BTRL' !or !pr='RANDOM') !then !let !rnums= '1:nrow(@mat)' !let !cnums= '1:ncol(@mat)'
!else !let !rnums= 'nrow(@mat):1' !let !cnums= 'ncol(@mat):1'
!ifend
!if (!pr='LRTB' !or !pr='TBLR') !then
 !let !rrevert= 'comp @r= (nrow(@mat)+1)*(@r>0)-@r' !let !crevert= 'comp @c= (ncol(@mat)+1)*(@c>0)-@c'
!ifend
!if (!pr='RANDOM') !then
-comp @val= msum(@mat).
-do if @val.
!else
-do if any(@mat).
!ifend
 !if (!pr='RANDOM') !then
- do if @val>1.
-  comp @mat= @mat&*uniform(nrow(@mat),ncol(@mat)).
-  comp @mat= @mat=mmax(@mat).
- end if. 
 !ifend
 !if (!pr='BTRL' !or !pr='TBLR') !then
- comp @r= mmax(@mat&*(t(!rnums)*make(1,ncol(@mat),1))).
- !rrevert.
- comp @c= mmax(@mat(@r,:)&*(!cnums)).
- !crevert.
 !else
- comp @c= mmax(@mat&*(!cnums)(make(1,nrow(@mat),1),:)).
- !crevert.
- comp @r= mmax(@mat(:,@c)&*t(!rnums)).
- !rrevert.
 !ifend
- comp !5= {@r,@c}.
- release @r,@c.
-else.
- comp !5= {0,0}.
-end if.
release @mat,@val.
!enddefine.


******** HIGHLIGHT IN ROW AND COLUMN NO MORE THAN ONE ELEMENT EQUAL TO VALUE ********.
*/*!KO_prime(mat%val%name)*/*.
*Version 1.
*Takes matrix or vector MAT and value VAL.
*Returns binary matrix NAME of size as MAT, where 1 marks elements which are equal to VAL, but the function
*leaves in each row and column at most one such mark. In other words, the 1 in NAME
*exclusively matches rows and columns: "one row - one column".
*This is not the optimal matching algorithm: the result may depend on on the direction of the search of the
*elements equal to VAL. This function scans rows downward and elements in a row right to left.
*If you wand another direction, simply change in accordance to it the order of rows and/or columns in MAT.
*Argument VAL may be not only scalar but also a column vector of length as the number of rows in MAT.
*In that latter case in each row of MAT there is searched and marked a different value instances, but then
*(erasing the "superfluous" marks) the function act same as with scalar value.
*The function works somewhat faster with wide matrix than with high matrix.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_prime(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @val= !3.
do if nrow(@val)>1.
-comp @val= @val*make(1,ncol(!4),1).
end if.
comp !4= (!4=@val)&*(1:ncol(!4))(make(1,nrow(!4),1),:).
comp @rzero= make(1,ncol(!4),0).
comp @czero= make(nrow(!4),1,0).
loop @i= 1 to nrow(!4).
-comp @j= rmax(!4(@i,:)).
-do if @j. 
- comp !4(@i,:)= @rzero.
- comp !4(:,@j)= @czero.
- comp !4(@i,@j)= 1.
-end if.
end loop.
release @val,@rzero,@czero,@i,@j.
!enddefine.


******** HIGHLIGHT DIRECT CHAINS ********.
*/*!KO_runs(mat%maxw%dir%hl%name)*/*.
*Version 2.
*In a binary matrix MAT highlights direct chains of 1s going in succession.
*Chains of what length will be highlighted by what "weight" value depends on argument
*"maximal weight" MAXW (nonnegative integer scalar).
*If MAXW>0, chain of length >=MAXW will be highlighted by number MAXW,
*chain of length MAXW-1 will be highlighted by number MAXW-1,
*chain of length MAXW-2 will be highlighted by number MAXW-2, and so on.
*if MAXW is specified nonpositive (any such number), chain will be highlighted by the
*number equal to its length, whatever long it comes to be. In other words, nonpositive
*MAXW is the refusal of the "maximal weight" limit.
*Argument DIR (scalar) - chains of what direction interest you: vertical, in columns
*(positive argument), or horizontal, in rows (nonpositive argument).
*Argument HL - digit 1 or 0 (not name or expression, you may take the digit in quotes
*or apostrophes). This argument is about how to highlight. If HL=1, all members of a chain
*will be highlighted by one number; if HL=0, they will be highlighted by ordinal numbers.
*For example, if MAXW=0, the chain 1 1 1 1 1 will be highlighted as 5 5 5 5 5 under HL=1
*and as 1 2 3 4 5 under HL=0. And if MAXW=3, it will be highlighted as 3 3 3 3 3 under HL=1
*and as 1 2 3 3 3 under HL=0.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Function /*!KO_runs2*/ does the same job as this function, but is realized differently.

define !KO_runs(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !6= !2.
comp @maxw= !3.
comp @dir= !4.
!let !hl= !unquote(!5)
do if not @maxw.
-comp @maxw= 1E300.
end if.
do if not @dir.
-comp !6= t(!6).
end if.
comp @nr= nrow(!6).
comp @r_= !6(1,:).
loop @i= 2 to @nr.
-comp @r= !6(@i,:).
-comp @r_= @r&*(@r_+(@r_<@maxw)).
-comp !6(@i,:)= @r_.
end loop.
!if (!hl) !then
-comp @r_= !6(@nr,:).
-loop @i= @nr-1 to 1 by -1.
- comp @r= !6(@i,:).
- comp @rr_= @r and @r_.
- comp @r_= @r_&*@rr_+@r&*(not @rr_).
- comp !6(@i,:)= @r_.
-end loop.
-release @rr_.
!ifend
do if not @dir.
-comp !6= t(!6).
end if.
release @maxw,@dir,@nr,@i,@r,@r_.
!enddefine.


******** HIGHLIGHT DIRECT CHAINS (ANOTHER ALGORITHM) ********.
*/*!KO_runs2(mat%maxw%dir%name)*/*.
*Version 1.
*In a binary matrix MAT highlights direct chains of 1s going in succession.
*Chains of what length will be highlighted by what "weight" value depends on argument
*"maximal weight" MAXW (positive integer scalar from 2 to num_rows
*or num_columns of the matrix - depending on DIR):
*ones in chains lengthed >=MAXW will turn into value MAXW; ones in chains
*lengthed MAXW-1 will turn into value MAXW-1; ones in chains lengthed
*MAXW-2 will turn ito value MAXW-2, etc.
*Argument DIR (scalar) - chains of what direction interest you: vertical, in columns
*(positive argument), or horizontal, in rows (nonpositive argument).
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Function /*!KO_runs*/ does the same job as this function, but is realized differently.
*/*!KO_runs2*/ differs from /*!KO_runs*/ by the following:
*- argument MAXW must be positive;
*- argument HL is absent;
*- speed of /*!KO_runs2*/ depends on MAXW, but of /*!KO_runs*/ - not.
*- speed of /*!KO_runs2*/ is a bit higher than of /*!KO_runs*/ if MAXW is not above 3 and MAT
*is small; in all other cases /*!KO_runs*/ is faster and therefore commonly preferable.

define !KO_runs2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @maxw= !3.
comp @dir= !4.
do if not @dir.
-comp !5= t(!5).
end if.
loop @w= 2 to @maxw.
-comp @w_= @w-1.
-comp @a= !5(1:(nrow(!5)-@w_),:).
-comp @b= @a.
-loop @i= 2 to @w.
- comp @b= @b and !5(@i:(nrow(!5)-@w+@i),:).
-end loop.
-comp !5(1:(nrow(!5)-@w_),:)= @a+@b.
-loop @i= 1 to @w_.
- comp @a= !5(2:nrow(!5),:).
- comp @b= !5(1:(nrow(!5)-1),:).
- comp !5(2:nrow(!5),:)= @a+(@a=@w_)&*(@b=@w).
-end loop.
end loop.
do if not @dir.
-comp !5= t(!5).
end if.
release @maxw,@dir,@w,@w_,@a,@b,@i.
!enddefine.


******** HIGHLIGHT SLANT CHAINS ********.
*/*!KO_slant(mat%maxw%dir%hl%name)*/*.
*Version 2.
*In a binary matrix MAT highlights slant chains of 1s going in succession.
*Chains of what length will be highlighted by what "weight" value depends on argument
*"maximal weight" MAXW (nonnegative integer scalar).
*If MAXW>0, chain of length >=MAXW will be highlighted by number MAXW,
*chain of length MAXW-1 will be highlighted by number MAXW-1,
*chain of length MAXW-2 will be highlighted by number MAXW-2, and so on.
*if MAXW is specified nonpositive (any such number), chain will be highlighted by the
*number equal to its length, whatever long it comes to be. In other words, nonpositive
*MAXW is the refusal of the "maximal weight" limit.
*Argument DIR (scalar) - chains of what direction interest you: top-left/bottom-right
*(positive argument), top-right/bottom-left (nonpositive argument).
*Argument HL - digit 1 or 0 (not name or expression, you may take the digit in quotes
*or apostrophes). This argument is about how to highlight. If HL=1, all members of a chain
*will be highlighted by one number; if HL=0, they will be highlighted by ordinal numbers.
*For example, if MAXW=0, the chain 1 1 1 1 1 will be highlighted as 5 5 5 5 5 under HL=1
*and as 1 2 3 4 5 under HL=0. And if MAXW=3, it will be highlighted as 3 3 3 3 3 under HL=1
*and as 1 2 3 3 3 under HL=0.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Function /*!KO_slant2*/ does the same job as this function, but is realized differently.

define !KO_slant(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !6= !2.
comp @maxw= !3.
comp @dir= !4.
!let !hl= !unquote(!5)
do if not @maxw.
-comp @maxw= 1E300.
end if.
comp @nr= nrow(!6).
comp @nc= ncol(!6).
do if not @dir.
-comp !6= !6(:,@nc:1).
end if.
loop @i= 2 to @nr.
-comp @r= !6(@i,2:@nc).
-comp @r_= !6(@i-1,1:(@nc-1)).
-comp !6(@i,2:@nc)= @r&*(@r_+(@r_<@maxw)).
end loop.
!if (!hl) !then
-loop @i= @nr-1 to 1 by -1.
- comp @r= !6(@i,1:(@nc-1)).
- comp @r_= !6(@i+1,2:@nc).
- comp @rr_= @r and @r_.
- comp !6(@i,1:(@nc-1))= @r_&*@rr_+@r&*(not @rr_).
-end loop.
-release @rr_.
!ifend
do if not @dir.
-comp !6= !6(:,@nc:1).
end if.
release @maxw,@dir,@nr,@nc,@i,@r,@r_.
!enddefine.


******** HIGHLIGHT SLANT CHAINS (ANOTHER ALGORITHM) ********.
*/*!KO_slant2(mat%maxw%dir%name)*/*.
*Version 1.
*In a binary matrix MAT highlights slant chains of 1s going in succession.
*Chains of what length will be highlighted by what "weight" value depends on argument
*"maximal weight" MAXW (positive integer scalar from 2 to
*min(num_rows,num_columns) of the matrix): ones in chains lengthed >=MAXW will
*turn into value MAXW; ones in chains lengthed MAXW-1 will turn into value MAXW-1;
*ones in chains lengthed MAXW-2 will turn ito value MAXW-2, etc.
*Argument DIR (scalar) - chains of what direction interest you: top-left/bottom-right
*(positive argument), or top-right/bottom-left (nonpositive argument).
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Function /*!KO_slant*/ does the same job as this function, but is realized differently.
*/*!KO_slant2*/ differs from /*!KO_slant*/ by the following:
*- argument MAXW must be positive;
*- argument HL is absent;
*- speed of /*!KO_slant2*/ depends on MAXW, but of /*!KO_slant*/ - not.
*- speed of /*!KO_slant2*/ is a bit higher than of /*!KO_slant*/ if MAXW=2, or if MAXW= 3-4 and
*MAT is small; in all other cases /*!KO_slant*/ is faster and therefore commonly preferable.

define !KO_slant2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @maxw= !3.
comp @dir= !4.
do if not @dir.
-comp !5= !5(:,ncol(!5):1).
end if.
loop @w= 2 to @maxw.
-comp @w_= @w-1.
-comp @a= !5(1:(nrow(!5)-@w_),1:(ncol(!5)-@w_)).
-comp @b= @a.
-loop @i= 2 to @w.
- comp @b= @b and !5(@i:(nrow(!5)-@w+@i),@i:(ncol(!5)-@w+@i)).
-end loop.
-comp !5(1:(nrow(!5)-@w_),1:(ncol(!5)-@w_))= @a+@b.
-loop @i= 1 to @w_.
- comp @a= !5(2:nrow(!5),2:ncol(!5)).
- comp @b= !5(1:(nrow(!5)-1),1:(ncol(!5)-1)).
- comp !5(2:nrow(!5),2:ncol(!5))= @a+(@a=@w_)&*(@b=@w).
-end loop.
end loop.
do if not @dir.
-comp !5= !5(:,ncol(!5):1).
end if.
release @maxw,@dir,@w,@w_,@a,@b,@i.
!enddefine.


******** DEPTH-FIRST SEARCH IN UNIPARTITE GRAPH ********.
*/*!KO_dfs(adj%start%stop%forest%out%name)*/*.
*Version 1.
*Performs depth-first traversal/search of connected vertices in a unipartite unweighted graph ADJ,
*starting from vertex START. Optional vertex STOP - point of stopping the traversal: coming up to it,
*visiting of vertices ceases. START and STOP are scalars, vertex numbers, and must be different.
*If you don't want to specify STOP, i.e. traversal of the graph must go maximally far, set this
*argument to value 0.
*Argument OUT - digit 1 or 0 (not name or expression; you may put the digit in quotes or apostrophes).
*If 0, result NAME will contain only one column, showing, top to bottom, the order vertices were
*visited. If 1, there will be additional computations and NAME will contain more columns:
*2nd column shows the parent of the given visited (indicated in column 1) vertex. By "parent"
*here is meant the most early visited vertex directing to (i.e., having it as adjacent) our vertex.
*3rd column shows the distance of our visited vertex from vertex START (the root vertex); this
*distance is the length of the shortest path from START to our vertex through vertices already visited
*(in DFS, this distance from START to our vertex is not necessarily the globally minimal).
*4th column marks, with unit, the beginning of a new chain in the visits. A chain is a continuous
*sequence into the graph depth, where each next visited vertex is an adjacent one to the previous and
*this edge (arc) was passed (accounted) by the algorithm. New chain begins when no further move into
*depth is possible (for there are no more unvisited adjacent vertices left); then a new chain is
*conceived from a yet not visited vertex lying "under" some "fork" of paths; the algorithm returns to
*the fork to start from that vertex a new, another path into depth. DFS algorithm goals to maximize
*the length of a chain.
*5th column is added if FOREST=1 (see). Here is being unit-marked a starting of a new tree.
*Argument FOREST - digit 1 or 0 (not name or expression; you may take the digit in quotes or
*apostrophes). If 0, the algorithm will limit itself traversing into depth starting only from one
*vertex - the one specified as START. NAME will represent a single DF tree (Depth-first tree), going
*from the root START. If argument is 1, then - in case the tree from root START could not visit all
*vertices of the graph - an additional root vertex is taken (from the vertices yet unvisited), and a
*tree of yet unvisited vertices is built from it. So is repeated until lastly all vertices of the graph
*are visited. NAME in such a case will represent not one tree but a collection of trees called
*DF forest (Depth-first forest). The start (root vertex) of a new tree of the forest is marked with
*unit in the 5th column of NAME (you must specify OUT=1, for NAME to have several columns). The
*difference of a new tree from a new chain is that a tree begins from a new start, and the distances
*in a new tree are measured from that new start, while the distances in a new chain continue to be
*measured from the old start - from the root of the tree that chain belongs to. I.e., chains are
*paths nested in trees.
*Argument ADJ - input graph. The graph can be any - undirected or directed, with or without cycles,
*with or without loops, be tree or network. It is specified via adjacency list ADJ.
*Each row in ADJ is the graph vertex, and the row contains the numbers of the vertices adjacent to it
*(with a directed graph, these are those it indicates to); the numbers must be listed left to right,
*rammed. For example, if from the 2nd vertex there go edges to vertices 1,4 и 5, the 2nd row should
*look like {1,4,5,0,...,0}, where emty spaces are padded with zeros. The order of numbers may be any:
*{4,5,1,0,...,0}, for example, is also correct. The number of columns of ADJ must be enough to carry
*all the information about the graph's edges. Adjacency list corresponds to a binary matrix in which
*unit in cell (i,j) signifies the presence of the edge (arc) from vertex i to vertex j. Adjacency list
*is another mode to represent such a matrix. If your input data are the binary matrix, you can turn it
*into the adjacency list by function /*!KO_ram*/.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_dfs(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
              /!pos= !charend('%') /!pos= !charend(')'))
!let !for= !unquote(!5) !let !out= !unquote(!6)
comp @a= !2.
comp @start= !3.
comp @g= !4.
comp @brk= 0.
comp @clst= 0.
comp @na= rsum(@a<>0).
comp @notd= make(nrow(@a),1,1).
!if (!for) !then
-comp @newt= make(nrow(@a),1,0).
-comp @newt(1)= 1.
!ifend
!if (!out) !then
-comp @p= make(1,nrow(@a),0).
-comp @d= make(1,nrow(@a),0).
-comp !7= make(nrow(@a),4,0).
-comp !7(1,4)= 1.
-comp @not= 0.
!else
-comp !7= make(nrow(@a),1,0).
!ifend
comp @cnt= 0.
!if (!for) !then
-loop.
!ifend
- comp @s= {@start,0}.
- loop if ncol(@s)>1.
-  comp @u= @s(1).
-  comp @s= @s(2:ncol(@s)).
-  do if @notd(@u).
-   comp @notd(@u)= 0.
-   comp @cnt= @cnt+1.
    !if (!out) !then
-    comp !7(@cnt,1:3)= {@u,@p(@u),@d(@u)}.
-    do if @cnt>1.
-     comp !7(@cnt,4)= not any(@a(!7(@cnt-1,1),:)=@u).
-    end if.
    !else
-    comp !7(@cnt)= @u.
    !ifend
-   do if @u=@g.
-    comp @brk= 1.
-    break.
-   end if.
-   do if @na(@u).
-    comp @clst= @a(@u,1:@na(@u)).
-    comp @s= {@clst,@s}.
     !if (!out) !then
-     comp @not= not @p(@clst).
-     comp @p(@clst)= @p(@clst)+@not*@u.
-     comp @d(@clst)= @d(@clst)+@not*(@d(@u)+1).
     !ifend
-   end if.
-  end if.
- end loop.
!if (!for) !then
- do if any(@notd).
-  comp @start= cmax(@notd&*t(1:nrow(@notd))).
-  comp @newt(@cnt+1)= 1.
- else.
-  break.
- end if.
-end loop if @brk.
 !if (!out) !then
- comp !7= {!7,@newt}.
 !ifend
!ifend
comp !7= !7(1:@cnt,:).
release @a,@start,@u,@g,@brk,@na,@notd,@cnt,@s,@clst !if (!for) !then ,@newt !ifend !if (!out) !then ,@not,@p,@d !ifend.
!enddefine.


******** BREADTH-FIRST SEARCH IN UNIPARTITE GRAPH ********.
*/*!KO_bfs(adj%start%stop%forest%out%name)*/*.
*Version 1.
*Performs breadth-first traversal/search of connected vertices in a unipartite unweighted graph ADJ,
*starting from vertex START. Optional vertex STOP - point of stopping the traversal: coming up to it,
*visiting of vertices ceases. START and STOP are scalars, vertex numbers, and must be different.
*If you don't want to specify STOP, i.e. traversal of the graph must go maximally far, set this
*argument to value 0.
*Argument OUT - digit 1 or 0 (not name or expression; you may put the digit in quotes or apostrophes).
*If 0, result NAME will contain only one column, showing, top to bottom, the order vertices were
*visited. If 1, there will be additional computations and NAME will contain more columns:
*2nd column shows the parent of the given visited (indicated in column 1) vertex. By "parent"
*here is meant the most early visited vertex directing to (i.e., having it as adjacent) our vertex.
*3rd column shows the distance of our visited vertex from vertex START (the root vertex); this
*distance is the length of the shortest path from START to our vertex through vertices visited
*(in BFS, this distance from START to our vertex is necessarily the globally minimal).
*4th column marks, with unit, the beginning of a new fan in the visits. A fan is a continuous
*sequence across the graph breadth, i.e., enumeration of siblings (children of the same parent
*visited earlier). New chain begins when there are no more siblings. DFS algorithm goals to maximize
*the spread of a fan.
*5th column is added if FOREST=1 (see). Here is being unit-marked a starting of a new tree.
*Argument FOREST - digit 1 or 0 (not name or expression; you may take the digit in quotes or
*apostrophes). If 0, the algorithm will limit itself traversing into breadth starting only from one
*vertex - the one specified as START. NAME will represent a single BF tree (Depth-first tree), going
*from the root START. If argument is 1, then - in case the tree from root START could not visit all
*vertices of the graph - an additional root vertex is taken (from the vertices yet unvisited), and a
*tree of yet unvisited vertices is built from it. So is repeated until lastly all vertices of the graph
*are visited. NAME in such a case will represent not one tree but a collection of trees called
*BF forest (Breadth-first forest). The start (root vertex) of a new tree of the forest is marked with
*unit in the 5th column of NAME (you must specify OUT=1, for NAME to have several columns). The
*difference of a new tree from a new fan is that a tree begins from a new start, and the distances
*in a new tree are measured from that new start, while the distances in a new fan continue to be
*measured from the old start - from the root of the tree that fan belongs to. I.e., fans are
*paths nested in trees.
*Argument ADJ - input graph. The graph can be any - undirected or directed, with or without cycles,
*with or without loops, be tree or network. It is specified via adjacency list ADJ.
*Each row in ADJ is the graph vertex, and the row contains the numbers of the vertices adjacent to it
*(with a directed graph, these are those it indicates to); the numbers must be listed left to right,
*rammed. For example, if from the 2nd vertex there go edges to vertices 1,4 и 5, the 2nd row should
*look like {1,4,5,0,...,0}, where emty spaces are padded with zeros. The order of numbers may be any:
*{4,5,1,0,...,0}, for example, is also correct. The number of columns of ADJ must be enough to carry
*all the information about the graph's edges. Adjacency list corresponds to a binary matrix in which
*unit in cell (i,j) signifies the presence of the edge (arc) from vertex i to vertex j. Adjacency list
*is another mode to represent such a matrix. If your input data are the binary matrix, you can turn it
*into the adjacency list by function /*!KO_ram*/.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_bfs(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
              /!pos= !charend('%') /!pos= !charend(')'))
!let !for= !unquote(!5) !let !out= !unquote(!6)
comp @a= !2.
comp @start= !3.
comp @g= !4.
comp @brk= 0.
comp @na= rsum(@a<>0).
comp @notd= make(nrow(@a),1,1).
!if (!for) !then
-comp @newt= make(nrow(@a)+1,1,0).
-comp @newt(1)= 1.
!ifend
!if (!out) !then
-comp !7= make(nrow(@a)+1,4,0).
-comp !7(1,4)= 1.
!else
-comp !7= make(nrow(@a)+1,1,0).
!ifend
comp @st= 1.
comp @en= 1.
comp @v= 0.
comp @v_= 0.
!if (!for) !then
-loop.
- comp @en= @st.
!ifend
- comp @notd(@start)= 0.
- comp !7(@st,1)= @start.
- loop if !7(@st,1).
-  comp @u= !7(@st,1).
-  do if @u=@g.
-   comp @en= @st.
-   comp @brk= 1.
-   break.
-  end if.
-  loop @v= 1 to @na(@u).
-   comp @v_= @a(@u,@v).
-   do if @notd(@v_).
-    comp @notd(@v_)= 0.
-    comp @en= @en+1.
     !if (!out) !then
-     comp !7(@en,1:3)= {@v_,@u,!7(@st,3)+1}.
     !else
-     comp !7(@en)= @v_.
     !ifend
-   end if.
-  end loop.
-  comp @st= @st+1.
- end loop.
!if (!for) !then
- do if any(@notd).
-  comp @start= cmax(@notd&*t(1:nrow(@notd))).
-  comp @newt(@st)= 1.
- else.
-  break.
- end if.
-end loop if @brk.
 !if (!out) !then
- comp !7= {!7,@newt}.
 !ifend
!ifend
comp !7= !7(1:@en,:).
!if (!out) !then
-do if @en>1.
-comp !7(2:@en,4)= !7(2:@en,2)<>!7(1:(@en-1),2) or !7(2:@en,2)=0.
-end if.
!ifend
release @a,@start,@u,@g,@brk,@na,@notd,@st,@en,@v,@v_ !if (!for) !then ,@newt !ifend.
!enddefine.


******** BRON-KERBOSCH MAXIMAL CLIQUES ALGORITHM (PIVOT VERSION) ********.
*/*!KO_bronkerb(adj%return%name)*/*.
*Version 1.
*This function performs Bron-Kerbosch algorithm for enumerating maximal cliques, in its pivot
*version. (Equivalent but a slower, base version of the algorithm - see function /*!KO_bronkerb2*/.)
*Input matrix ADJ is an undirected unweighted graph in the form of "adjacency list". Each row
*of ADJ is a vertex in the graph, and the row contains the vertex numbers of the vertices with which
*that vertex is adjacent (connected, neighbour). The numbers go left-to-right, rammed. For example,
*let there vertex with number 2 be connected with vertices 1, 4, 5. Then the 2-nd row of ADJ must
*look as: {1,4,5[,0...]}, where [0...] stands for padding of the row with zeros if necessary.
*The number of columns in ADJ may be any sufficient to accomodate the list of all the neighbours in
*the graph. The order of vertex numbers in a row is arbitrary: for example, that same row could be
*written as {5,1,4[,0...]}. A vertex in a graph cannot be adjacent with itself (i.e., say vertex 2
*cannot be found in the 2-nd row). Isolated vertices are also prohibited in the graph (i.e., a row
*cannot consist of just zeros - but see EXAMPLE). Adjacency list ADJ corresponds to the square
*symmetric binary adjacency matrix with zero diagonal, in which units flag the presence of a link;
*ADJ is another mode to represent the matrix. If your input data are such binary matrix, you can
*turn it into the adjacency list by function /*!KO_ram*/.
*Result name NAME - two-column matrix. Its 1-st column is the vertex numbers of the vertices of the
*maximal cliques. Its 2-nd column is the clique identifier - the clique number in the order of their
*detection by the algorithm.
*Argument RETURN - capitalized keyword (may also take in quotes or apostrophes):
*"ALL" - allow the algorithm to run to its end; NAME will enumerate all maximal cliques existing
*in the graph;
*"COUNT n" - return n maximal cliques found first, and stop; n is an integer >0, may be a digit,
*a name, or an expression (examples of specification: COUNT 10, COUNT x, COUNT abs(z)*2);
*"SIZEEQ n" - return one maximal clique consisting of n vertices, and stop; n an integer >1,
*may be a digit, a name, or an expression;
*"SIZEGE n" - return one maximal clique consisting of >=n vertices, and stop; n an integer >1,
*may be a digit, a name, or an expression.
*In the two last cases - if no cliques of the requested size were found, the result will be {0,0}.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_bronkerb(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !return= !unquote(!3)
comp @adj= !2.
!if (!head(!return)='COUNT') !then !let !cond= !conc('@cnt=',!tail(!return))
!else !if (!head(!return)='SIZEEQ') !then !let !cond= !conc('ncol(@r)=',!tail(!return))
!else !if (!head(!return)='SIZEGE') !then !let !cond= !conc('ncol(@r)>=',!tail(!return))
!ifend !ifend !ifend
comp @n= rsum(@adj>0).
comp @rs= {0,0}.
comp @ps= {1:nrow(@adj),0}.
comp @xs= {0,0}.
comp @qs= {0,0}.
comp @ri= {1,0}.
comp @pi= {nrow(@adj),0}.
comp @xi= {1,0}.
comp @qi= {1,0}.
comp !4= {0,0}.
comp @cnt= 0.
loop if ncol(@rs)>1.
 !do !w !in ('@r @p @x @q') !let !ws= !conc(!w,'s') !let !wi= !conc(!w,'i')
- comp !w= !ws(1:!wi(1)).
- comp !ws= !ws((ncol(!w)+1):ncol(!ws)).
- comp !wi= !wi(2:ncol(!wi)).
 !doend
-do if not (any(@p) or any(@x)).
- comp @cnt= @cnt+1.
- comp !4= {!4;t(@r),make(ncol(@r),1,@cnt)}.
  !if (!return<>'ALL') !then
-  do if !cond.
-   comp @cnt= -1.
-   break.
-  end if.
  !ifend
-else.
- do if not any(@x).
-  comp @#= @p.
- else if not any(@p).
-  comp @#= @x.
- else.
-  comp @#= {@p,@x}.
- end if.
- comp @n#= @n(@#).
- comp @i= mmax((@n#=mmax(@n#))&*t(1:nrow(@n#))).
- comp @q= @#(@i).
- comp @#= 0.
- do if any(@p).
-  comp @nei= @adj(@q,1:@n(@q)).
-  loop @i= 1 to ncol(@p).
-   do if not any(@nei=@p(@i)).
-    comp @#= {@#,@p(@i)}.
-   end if.
-  end loop.
-  do if ncol(@#)>1.
-   comp @#= @#(2:ncol(@#)).
-  end if.
- end if.
- do if any(@#).
-  comp @v= @#(1).
-  comp @rs= {@r,@rs}.
-  comp @ri= {ncol(@r),@ri}.
-  do if ncol(@p)=1.
-   comp @#= 0.
-  else.
-   comp @#= mmax((@p=@v)&*(1:ncol(@p))).
-   do if @#=1.
-    comp @#= @p(2:ncol(@p)).
-   else if @#= ncol(@p).
-    comp @#= @p(1:(ncol(@p)-1)).
-   else.
-    comp @#= @p({1:(@#-1),(@#+1):ncol(@p)}).
-   end if.
-  end if.
-  comp @ps= {@#,@ps}.
-  comp @pi= {ncol(@#),@pi}.
-  do if any(@x).
-   comp @#= {@x,@v}.
-  else.
-   comp @#= @v.
-  end if.
-  comp @xs= {@#,@xs}.
-  comp @xi= {ncol(@#),@xi}.
-  comp @qs= {@q,@qs}.
-  comp @qi= {1,@qi}.
-  do if any(@r).
-   comp @#= {@r,@v}.
-  else.
-   comp @#= @v.
-  end if.
-  comp @rs= {@#,@rs}.
-  comp @ri= {ncol(@#),@ri}.
-  comp @nei= @adj(@v,1:@n(@v)).
   !do !w !in('@p @x') !let !ws= !conc(!w,'s') !let !wi= !conc(!w,'i')
-   comp @#= 0.
-   do if any(!w).
-    loop @i= 1 to ncol(!w).
-     do if any(@nei=!w(@i)).
-      comp @#= {@#,!w(@i)}.
-     end if.
-    end loop.
-    do if ncol(@#)>1.
-     comp @#= @#(2:ncol(@#)).
-    end if.
-   end if.
-   comp !ws= {@#,!ws}.
-   comp !wi= {ncol(@#),!wi}.
   !doend
-  comp @qs= {@q,@qs}.
-  comp @qi= {1,@qi}.
- end if.
-end if.
end loop.
do if nrow(!4)>1.
-comp !4= !4(2:nrow(!4),:).
end if.
!if (!head(!return)='SIZEEQ' !or !head(!return)='SIZEGE') !then
-do if @cnt=-1.
- comp !4= !4((nrow(!4)-ncol(@r)+1):nrow(!4),:).
-else.
- comp !4= {0,0}.
-end if.
!ifend
release @adj,@n,@rs,@ps,@xs,@qs,@ri,@pi,@xi,@qi,@cnt,@r,@p,@x,@q,@v,@#,@nei,@i,@n#.
!enddefine.


******** BRON-KERBOSCH MAXIMAL CLIQUES ALGORITHM (BASE VERSION) ********.
*/*!KO_bronkerb2(adj%return%name)*/*.
*Version 1.
*This function performs Bron-Kerbosch algorithm for enumerating maximal cliques, in its basic
*version, and it is equivalent to function /*!KO_bronkerb*/, but is slower than it, therefore is not
*preferable - except the cases of very sparse graph, when it uses to be somewhat faster.
*The arguments are the same as in /*!KO_bronkerb*/. The order of cliques finding and the order of
*vertices inside the cliques, i.e., the order of rows in the result NAME, will usually be different
*from when /*!KO_bronkerb*/ is used, but the list and composition of the cliques under RETURN=ALL is
*always be the same as then. The vertex order in the cliques here is always one and the
*same - ascending their number.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_bronkerb2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !return= !unquote(!3)
comp @adj= !2.
!if (!head(!return)='COUNT') !then !let !cond= !conc('@cnt=',!tail(!return))
!else !if (!head(!return)='SIZEEQ') !then !let !cond= !conc('ncol(@r)=',!tail(!return))
!else !if (!head(!return)='SIZEGE') !then !let !cond= !conc('ncol(@r)>=',!tail(!return))
!ifend !ifend !ifend
comp @n= rsum(@adj>0).
comp @rs= {0,0}.
comp @ps= {1:nrow(@adj),0}.
comp @xs= {0,0}.
comp @ri= {1,0}.
comp @pi= {nrow(@adj),0}.
comp @xi= {1,0}.
comp !4= {0,0}.
comp @cnt= 0.
loop if ncol(@rs)>1.
 !do !w !in ('@r @p @x') !let !ws= !conc(!w,'s') !let !wi= !conc(!w,'i')
- comp !w= !ws(1:!wi(1)).
- comp !ws= !ws((ncol(!w)+1):ncol(!ws)).
- comp !wi= !wi(2:ncol(!wi)).
 !doend
-do if not (any(@p) or any(@x)).
- comp @cnt= @cnt+1.
- comp !4= {!4;t(@r),make(ncol(@r),1,@cnt)}.
  !if (!return<>'ALL') !then
-  do if !cond.
-   comp @cnt= -1.
-   break.
-  end if.
  !ifend
-end if.
-do if any(@p).
- comp @v= @p(1).
- comp @rs= {@r,@rs}.
- comp @ri= {ncol(@r),@ri}.
- do if ncol(@p)=1.
-  comp @#= 0.
- else.
-  comp @#= @p(2:ncol(@p)).
- end if.
- comp @ps= {@#,@ps}.
- comp @pi= {ncol(@#),@pi}.
- do if any(@x).
-  comp @#= {@x,@v}.
- else.
-  comp @#= @v.
- end if.
- comp @xs= {@#,@xs}.
- comp @xi= {ncol(@#),@xi}.
- do if any(@r).
-  comp @#= {@r,@v}.
- else.
-  comp @#= @v.
- end if.
- comp @rs= {@#,@rs}.
- comp @ri= {ncol(@#),@ri}.
- comp @nei= @adj(@v,1:@n(@v)).
  !do !w !in('@p @x') !let !ws= !conc(!w,'s') !let !wi= !conc(!w,'i')
-  comp @#= 0.
-  do if any(!w).
-   loop @i= 1 to ncol(!w).
-    do if any(@nei=!w(@i)).
-     comp @#= {@#,!w(@i)}.
-    end if.
-   end loop.
-   do if ncol(@#)>1.
-    comp @#= @#(2:ncol(@#)).
-   end if.
-  end if.
-  comp !ws= {@#,!ws}.
-  comp !wi= {ncol(@#),!wi}.
  !doend
-end if.
end loop.
do if nrow(!4)>1.
-comp !4= !4(2:nrow(!4),:).
end if.
!if (!head(!return)='SIZEEQ' !or !head(!return)='SIZEGE') !then
-do if @cnt=-1.
- comp !4= !4((nrow(!4)-ncol(@r)+1):nrow(!4),:).
-else.
- comp !4= {0,0}.
-end if.
!ifend
release @adj,@n,@rs,@ps,@xs,@ri,@pi,@xi,@cnt,@r,@p,@x,@v,@#,@nei,@i.
!enddefine.


******** CONVEX HULL: 2D JARVIS ALGORITHM ********.
*/*!KO_chjarv(data%name)*/*.
*Version 1.
*Outlines convex hull (convex polygon of all peripheral points) of point cloud DATA.
*The basic "gift-wrapping" algorithm, also called Jarvis algorithm, is designed for
*the 2-dimensional case. DATA is n x 2 data matrix, coordinates of n points (n>=3) on plane.
*Result NAME - list of point numbers constituting the convex hull. Points in NAME
*go counter-clockwise of the data.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*To outline convex hull, you may also use function /*!KO_chgrah*/.

define !KO_chjarv(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @pt= t(1:nrow(@data)).
comp @min= cmin(@data).
comp @fstpt= mmax((@data(:,1)=@min(1))&*@pt).
do if @data(@fstpt,2)=@min(2).
-comp @off= 100.
else.
-comp @off= -100.
end if.
comp @prev= {@min(1)-1E-12,@data(@fstpt,2)-@off}.
comp @orig= {@min(1),@data(@fstpt,2)}.
comp @prev= @prev-@orig.
comp @data= @data-@orig(make(1,nrow(@data),1),:).
comp !3= 0.
comp @ind= 0.
comp @ind2= 0.
loop.
-comp @sp= @data*t(@prev).
-comp @h= sqrt(rssq(@data)).
-comp @hh= @h*sqrt(rssq(@prev)).
-comp @sp= @sp/(@hh+(@hh=0)).
-comp @sp= abs(@sp-mmin(@sp))<1E-12.
-comp @nt= msum(@sp).
-do if @nt=nrow(@sp).
- comp @nxtpt= @pt.
- comp !3= {!3;@nxtpt}.
- break.
-else if @nt=1.
- comp @ind= mmax(@sp&*t(1:nrow(@sp))).
- do if @ind=1.
-  comp @ind2= 2:nrow(@sp)).
- else if @ind=nrow(@sp).
-  comp @ind2= 1:(nrow(@sp)-1).
- else.
-  comp @ind2= {1:(@ind-1),(@ind+1):nrow(@sp)}.
- end if.
-else.
- comp @sp= {@sp;1}.
- comp @ind= make(1,@nt+1,0).
- comp @ind2= 0.
- comp @s= 1.
- loop @c= 1 to @nt+1.
-  loop @i= @s to nrow(@sp).
-  end loop if @sp(@i).
-  comp @ind(@c)= @i.
-  do if @i>@s.
-   comp @ind2= {@ind2,@s:(@i-1)}.
-  end if.
-  comp @s= @i+1.
- end loop.
- release @s,@c,@i.
- comp @ind= @ind(1:(ncol(@ind)-1)).
- comp @ind2= @ind2(2:ncol(@ind2)).
- comp @ind(grade(@h(@ind)))= @ind.
-end if.
-comp @nxtpt= @pt(@ind).
-comp !3= {!3;@nxtpt}.
-comp @orig= @data(@ind(@nt),:).
-comp @data= @data(@ind2,:).
-comp @pt= @pt(@ind2).
-comp @prev= -@orig.
-comp @data= @data-@orig(make(1,nrow(@data),1),:).
end loop if any(@nxtpt=@fstpt).
comp !3= !3(2:nrow(!3)).
do if @off.
-comp !3= !3(nrow(!3):1).
end if.
release @data,@pt,@min,@fstpt,@off,@prev,@orig,@sp,@h,@hh,@nt,@ind,@ind2,@nxtpt.
!enddefine.


******** CONVEX HULL: 2D GRAHAM ALGORITHM ********.
*/*!KO_chgrah(data%name)*/*.
*Version 1.
*Outlines convex hull (convex polygon of all peripheral points) of point cloud DATA.
*Graham's algorithm is designed for the 2-dimensional case.
*DATA is n x 2 data matrix, coordinates of n points (n>=3) on plane. There must be no less than
*three non-coinciding points in the data.
*The 1-st column of NAME - the list of point numbers constituting the convex hull. If there are
*coinciding (tied) points in the convex hull, only one instance of them is displayed in NAME;
*however, there is the 2-nd column in NAME which informs of the number of instances each
*point exists. Points in NAME go clockwise of the data.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*To outline convex hull, you may also use function /*!KO_chjarv*/. These two functions are
*equivalent in result, though the algorithms are different. /*!KO_chjarv*/ shows all the points
*of the hull, including tied, in the single column of the result.

define !KO_chgrah(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @pt= t(1:nrow(@data)).
comp @1= mmin(@data(:,2)).
comp @flag= @data(:,2)=@1.
comp @2= mmin((@data(:,1)+(not @flag)*1E300)).
comp @flag= (@data(:,1)=@2)&*@flag.
comp @3= mmax(@flag&*t(1:nrow(@flag))).
comp @cnt= msum(@flag).
comp @lst= 0.
do if @cnt=1.
-do if @3=1.
- comp @ind= 2:nrow(@data).
-else if @3=nrow(@data).
- comp @ind= 1:(nrow(@data)-1).
-else.
- comp @ind= {1:(@3-1),(@3+1):nrow(@data)}.
-end if.
else.
-comp @flag= not @flag.
-comp @nnz= nrow(@flag)-@cnt.
-comp @ind= make(1,@nnz+1,0).
-comp @c= 1.
-loop @i= 1 to nrow(@flag).
- comp @ind(@c)= @i.
- comp @c= @c+@flag(@i).
-end loop.
-comp @ind= @ind(1:@nnz).
-release @nnz,@c,@i.
end if.
comp @pt= @pt(@ind).
comp @data= @data(@ind,:).
comp @data= @data-{@2,@1}(make(1,nrow(@data),1),:).
comp @sp= @data*t({1,0}).
comp @h= sqrt(rssq(@data)).
comp @sp= @sp/(@h+(@h=0)).
comp @sp= rnd(@sp*1E12).
comp @grd= grade(-@sp).
comp @sp(@grd)= @sp.
comp @h(@grd)= @h.
comp @data(@grd,:)= @data.
comp @pt(@grd)= @pt.
comp @cptfr= 0.
comp @cntfr= 0.
comp @cptlr= 0.
comp @cntlr= 0.
comp @z= 0.
do if @sp(2)=@sp(1).
-loop @lst= 2 to nrow(@sp)-1.
-end loop if @sp(@lst)<>@sp(@lst+1).
-comp @z= @h(1:@lst).
-comp @1= mmax(@z).
-do if any(@z<@1).
- comp @grd= grade(@z).
- comp @z(@grd)= @z.
- comp @pt_= @pt(1:@lst).
- comp @pt_(@grd)= @pt_.
- comp @2= 0.
- loop @i= 1 to @lst-1 if @z(@i)<@1.
-  do if @z(@i+1)<>@z(@i).
-   comp @cptfr= {@pt_(@i);@cptfr}.
-   comp @cntfr= {@i-@2,@cntfr}.
-   comp @2= @i.
-  end if.
- end loop.
- comp @cptfr= @cptfr(1:(nrow(@cptfr)-1)).
- comp @cntfr= t(@cntfr(1:(ncol(@cntfr)-1))).
- release @pt_,@i.
-end if.
end if.
do if @sp(nrow(@sp)-1)=@sp(nrow(@sp)) and @sp(nrow(@sp))<>@sp(1).
-loop @lst= nrow(@sp)-1 to 2 by -1.
-end loop if @sp(@lst)<>@sp(@lst-1).
-comp @z= @h(@lst:nrow(@sp)).
-comp @1= mmax(@z).
-do if any(@z<@1).
- comp @grd= grade(-@z).
- comp @z(@grd)= @z.
- comp @pt_= @pt(@lst:nrow(@sp)).
- comp @pt_(@grd)= @pt_.
- comp @2= nrow(@z)+1.
- loop @i= nrow(@z) to 2 by -1 if @z(@i)<@1.
-  do if @z(@i-1)<>@z(@i).
-   comp @cptlr= {@cptlr;@pt_(@i)}.
-   comp @cntlr= {@cntlr,@2-@i}.
-   comp @2= @i.
-  end if.
- end loop.
- comp @cptlr= @cptlr(2:nrow(@cptlr)).
- comp @cntlr= t(@cntlr(2:ncol(@cntlr))).
- release @pt_,@i.
-end if.
end if.
comp @flg= @sp(2:nrow(@sp))=@sp(1:(nrow(@sp)-1)).
do if any(@flg).
-comp @flg= {0;not @flg;1}.
-comp @ind= make(1,nrow(@flg)-1,0).
-comp @cnt_= make(1,ncol(@ind),1).
-comp @#= 0.
-comp @1= 1.
-loop @i= 2 to nrow(@flg).
- do if @flg(@i).
-  comp @#= @#+1.
-  comp @lst= @i-1.
-  do if @lst=@1.
-   comp @ind(@#)= @lst.
-  else.
-   comp @2= @1:@lst.
-   comp @flag= @h(@2)=mmax(@h(@2)).
-   comp @ind(@#)= mmax(@flag&*t(@2)).
-   comp @cnt_(@#)= msum(@flag).
-  end if.
-  comp @1= @i.
- end if.
-end loop.
-comp @ind= @ind(1:@#).
-comp @pt= @pt(@ind).
-comp @data= @data(@ind,:).
-comp @cnt= {@cnt,@cnt_(1:@#)}.
-release @cnt_,@#,@i.
else.
-comp @cnt= {@cnt,make(1,nrow(@data),1)}.
end if.
comp @data= {0,0;@data}.
comp @pt= {@3;@pt}.
comp !3= {2,1}.
loop @3= 3 to nrow(@pt).
-loop.
- comp @2= !3(1).
- comp @1= !3(2).
- comp @z= (@data(@2,1)-@data(@1,1))*(@data(@3,2)-@data(@1,2))-(@data(@2,2)-@data(@1,2))*(@data(@3,1)-@data(@1,1)).
- do if @z>=0.
-  break.
- else.
-  comp !3= !3(2:ncol(!3)).
- end if.
-end loop.
-comp !3= {@3,!3}.
end loop.
comp @cnt= t(@cnt(!3)).
comp !3= @pt(!3).
do if @cptfr(1).
-comp !3= {!3(1:(nrow(!3)-1));@cptfr;!3(nrow(!3))}.
-comp @cnt= {@cnt(1:(nrow(@cnt)-1));@cntfr;@cnt(nrow(@cnt))}.
end if.
do if @cptlr(1).
-comp !3= {@cptlr;!3}.
-comp @cnt= {@cntlr;@cnt}.
end if.
comp !3= {!3,@cnt}.
release @data,@pt,@1,@2,@3,@flag,@cnt,@ind,@sp,@h,@grd,@cptfr,@cntfr,@cptlr,@cntlr,@lst,@flg,@z.
!enddefine.


******** MAGNITUDES OF ELEMENTS OFFSET FROM MATRIX DIAGONAL ********.
*/*!KO_diagoff(nr%nc%diag%name)*/*.
*Version 1.
*The function marks the diagonal of NR x NC sized matrix with zero, while other elements
*are marked with the magnitude of their distance (offset) from the diagonal. Under-diagonal
*marks have (for convenience) negative sign.
*When a matrix is nonsquare, we may say it has more than one diagonal.
*Argument DIAG (capitalized word, optionally in quotes/apostrophes) allows
*to choose which diagonal is to be zero-marked and offset measured off:
*"FIRST" - first, left aka main diagonal (starts in the top-left corner of the matrix)
*"LAST" - last aka right diagonal (ends in the bottom-right corner of the matrix)
*"MIDDLE" - middle diagonal, it is in-between the first and the last;
*in case abs(NR-NC) is odd number, there are two middle diagonals.
*If the matrix is square (NR=NC), its diagonal is single, so it is unimportant then
*how DIAG is specified.

define !KO_diagoff(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @r= !2.
comp @c= !3.
!let !diag= !unquote(!4)
comp @o1= (1:@c)(make(1,@r,1),:).
comp @o2= t(1:@r)*make(1,@c,1).
comp !5= @o1-@o2.
!if (!diag<>'FIRST') !then
-do if @r<>@c.
- comp @diff= abs(@r-@c).
- comp @shift= !if (!diag='LAST') !then @diff !else trunc(@diff/2) !ifend.
- do if @r<@c.
-  comp !5= !5-@shift.
   !if (!diag='MIDDLE') !then
-   do if mod(@diff,2).
-    comp !5= !5-(!5>0).
-   end if.
   !ifend
- else.
-  comp !5= !5+@shift.
   !if (!diag='MIDDLE') !then
-   do if mod(@diff,2).
-    comp !5= !5+(!5<0).
-   end if.
   !ifend
- end if.
- release @diff,@shift.
-end if.
!ifend
release @r,@c,@o1,@o2.
!enddefine.


******** MAGNITUDES OF ELEMENTS OFFSET FROM MATRIX DIAGONAL BAND ********.
*/*!KO_diagboff(nr%nc%name)*/*.
*Version 1.
*When a matrix is nonsquare, we may say it has more than one diagonal.
*The first diagonal is left aka main diagonal: it starts from the top-right corner.
*The last diagonal or right diagonal ends in the bottom-right corner. There are abs(NR-NC)+1
*diagonals in all, they form the diagonal band.
*The function marks all the diagonal band of NR x NC sized matrix with zero, while other
*elements are marked with the magnitude of their distance (offset) from the band. Under-band
*marks have (for convenience) negative sign.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_diagboff(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @r= !2.
comp @c= !3.
comp @o1= (1:@c)(make(1,@r,1),:).
comp @o2= t(1:@r)*make(1,@c,1).
comp @diff= abs(@r-@c).
comp @i= 0.
do if @r=@c.
-comp !4= @o1-@o2.
else if @r<@c.
-loop @i= 1 to @r.
- comp @o1(@i,1:@diff+@i)= make(1,@diff,@i).
-end loop.
-comp !4= @o1-@o2.
-comp !4= !4-(!4>0)*@diff.
else if @r>@c.
-loop @i= 1 to @c.
- comp @o2(1:@diff+@i,@i)= make(@diff,1,@i).
-end loop.
-comp !4= @o1-@o2.
-comp !4= !4+(!4<0)*@diff.
end if.
release @r,@c,@o1,@o2,@diff,@i.
!enddefine.


******** READ INLINE DATA ********.
*/*!KO_read(nr nc name) data*/*.
*Version 2.
*This function is of special form of specification. It reads values offered by you within syntax
*and constructs matrix NAME of them. Use it to construct not big matrices.
*The function is equivalent to the explicit construction of a matrix of the submitted elements with
*the help of operators {,;}. Its convenience is exactly that it does not require you to write these
*operators (which takes human time). Therefore you can simply serve a set of values separated by just blanks,
*and the function will create matrix out of them.
*Arguments (note, arguments in parentheses go via blank, not via %):
*NR - number of rows in the matrix to built (positive integer: number or name, not expression)
*NC - number of columns in the matrix to built (positive integer: number or name, not expression).
*You are also in right to specify one of these two arguments by symbol ?. Then the function will itself
*calculate it, dividing the number of the input values by the magnitude of the other, specified argument
*(the division must be without the remainder).
*The inline data themselves DATA must go after the parenthesis. Don't put period between the parenthesis
*and DATA. Put period after DATA.
*Data is an array of values via blank(s) or tabulation. It doesn't matter in what number of rows and columns
*you have broken the set of values: the order of the matrix being created is defined by arguments NR, NC.
*The function reads values as free (unstructured) text, left to right, and lines top to bottom.
*The number of values must equal the product NRxNC.
*What a value being read can appear?:
*-positive, zero or negative number (examples: 2.5 -3 +4 -.001 0.15);
*-fraction or division (examples: 1/3 -3/16 3/-16);
*-variable name (example: x);
*-expression in quotes or apostrohes (examples: 'x*2' "2/3" '-ln(4)').
* Reading values, the function removes quotes/apostrophes off them (use function /*!KO_read2()*/,
* if you want to keep quotes/apostrophes, i.e. read the values as string).
*Don't use the current function to read any big dataset (thousands of elements) because it is slow
*(faster will be to take the data from syntax by command DATA LIST and then open them in MATRIX).
*This function is useful for routine work, not for programming algorithms.

define !KO_read(!pos= !enclose('(',')') /!pos= !cmdend)
!let !mx= !null !let !comma= ','
!do !val !in (!2) !let !val2= !unquote(!val)
 !if (!val2='-' !or !val2='+') !then !let !sign= !val2
 !else !if (!val2='/') !then !let !mx= !conc(!mx,!val2) !let !comma= !null
 !else !let !mx= !conc(!mx,!comma,!sign,!val2) !let !sign= !null !let !comma= ','
 !ifend !ifend
!doend
!let !nr= !head(!1) !let !nc= !head(!tail(!1)) !let !name= !tail(!tail(!1))
comp !name= {!tail(!mx)}.
!if (!nr='?') !then !let !nr= !conc('ncol(',!name,')/',!nc)
!else !if (!nc='?') !then !let !nc= !conc('ncol(',!name,')/',!nr)
!ifend !ifend
comp !name= reshape(!name,!nr,!nc).
!enddefine.


******** READ INLINE DATA (NO UNQUOTING) ********.
*/*!KO_read2(nr nc name) data*/*.
*Version 2.
*This function is identical to /*!KO_read()*/ with the only distinction that it does not take
*quotes/apostrophes off the value being read. Therefore if the value is quoted or surrounded by
*apostrophes, that value is taken as a string element into matrix NAME.

define !KO_read2(!pos= !enclose('(',')') /!pos= !cmdend)
!let !mx= !null !let !comma= ','
!do !val !in (!2)
 !if (!val='-' !or !val='+') !then !let !sign= !val
 !else !if (!val='/') !then !let !mx= !conc(!mx,!val) !let !comma= !null
 !else !let !mx= !conc(!mx,!comma,!sign,!val) !let !sign= !null !let !comma= ','
 !ifend !ifend
!doend
!let !nr= !head(!1) !let !nc= !head(!tail(!1)) !let !name= !tail(!tail(!1))
comp !name= {!tail(!mx)}.
!if (!nr='?') !then !let !nr= !conc('ncol(',!name,')/',!nc)
!else !if (!nc='?') !then !let !nc= !conc('ncol(',!name,')/',!nr)
!ifend !ifend
comp !name= reshape(!name,!nr,!nc).
!enddefine.


******** SPARSE MATRIX INTO LIST OF NONZERO VALUES ********.
*/*!KO_nzlist(mat%name1%name2)*/*.
*Version 1.
*Takes matrix (or vector) MAT of any size. Copies out from MAT all nonzero elements in one list.
*That list is the 2nd column of NAME2, and in the 1st column of NAME2 there the column numbers are listed in
*which the copied values are from. NAME2 has as many rows as the number of nonzero elements in MAT.
*Column vector NAME1 has the length as the number of rows in MAT, and there is written how many nonzero
*elements has each row of MAT.
*If MAT has no nonzero elements the function returns NAME1 and NAME2 as zero scalars.
*If you transpose input MAT the results will be different still informationally equivalent.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Function /*!KO_nzunlist*/ is opposite to this and can restore MAT from NAME1 and NAME2.

define !KO_nzlist(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @ram= @mat<>0.
comp !3= rsum(@ram).
do if any(!3).
-comp @m= @ram.
-loop @i= 2 to ncol(@m).
- comp @m(:,@i)= @m(:,@i-1)+@m(:,@i).
-end loop.
-comp @m= @m&*@ram.
-comp @ram= make(nrow(@m),ncol(@m)+1,0).
-comp @nums= 1:ncol(@m).
-comp @m= @m+(not @m)*ncol(@ram).
-loop @i= 1 to nrow(@m).
- comp @ram(@i,@m(@i,:))= @nums.
-end loop.
-comp !4= make(msum(!3),2,0).
-comp @st= 1.
-loop @i= 1 to nrow(@ram).
- comp @n= !3(@i).
- do if @n.
-  comp @en= @st+@n-1.
-  comp @rami= @ram(@i,1:@n).
-  comp !4(@st:@en,:)= {t(@rami),t(@mat(@i,@rami))}.
-  comp @st= @en+1.
- end if.
-end loop.
-release @m,@i,@nums,@st,@en,@n,@rami. 
else.
-comp !3= 0.
-comp !4= 0.
end if. 
release @mat,@ram.
!enddefine.


******** LIST INTO SPARSE MATRIX ********.
*/*!KO_nzunlist(nnz%list%name)*/*.
*Version 1.
*Creates (sparse) matrix by writing in values from a list into the empty matrix.
*Takes nonzero column vector NNZ which values must be nonnegative integers.
*These values are the numbers of elements being written in each row of the being created matrix NAME (NAME will
*have as many rows as NNZ has).
*The values to write into NAME are borrowed by the function from the 2nd column of the 2-column matrix LIST,
*and information to which column should a given value be put - is taken from the corresponding
*element of the 1st column of LIST. (Values in 1st col of LIST must be all positive integers. The values being
*written down themselves - 2nd col of LIST - may be any.)
*In 1st row of NAME there will be written NNZ(1) values of LIST, in 2nd row of NAME there will be written
*NNZ(2) next values of LIST, and so on.
*The created matrix NAME has as many columns as the maximal value in the 1st column of LIST.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Function /*!KO_nzlist*/ is opposite to this and can copy out values from NAME, creating NNZ and LIST, if
*the values to be copy are all the nonzero values existing in NAME.

define !KO_nzunlist(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @nnz= !2.
comp @list= t(!3).
comp !4= make(nrow(@nnz),rmax(@list(1,:)),0).
comp @e= 0.
loop @i= 1 to nrow(@nnz).
-do if @nnz(@i).
- comp @s= @e+1.
- comp @e= @e+@nnz(@i).
- comp !4(@i,@list(1,@s:@e))= @list(2,@s:@e).
-end if.
end loop.
release @nnz,@list,@e,@i,@s.
!enddefine.


******** RAMMING HORIZONTALLY NONZERO ELEMENTS OR THEIR INDICES ********.
*/*!KO_ram(mat%indval%lr%rand%name1%name2)*/*.
*Version 2.
*Takes matrix or vector MAT and shifts all nonzero elements in each row of MAT to left or right
*border, compacting them (no zeros will be among them).
*Argument INDVAL - capitalized keyword (may take in quotes or apostrophes):
*"VAL" - shift nonzero values themselves
*"IND" - shift indices of nonzero values (i.e., their column numbers)
*"BOTH" - do both tasks
*Argument LR - scalar. If nonpositive, the shift will be to the left; if positive, then to the right.
*Result NAME1 is the matrix with the ramming (shifting) of values or their indices done,
*sized as MAT. With INDVAL=BOTH, both results are returned, stacked: the matrix with indices joint
*below the matrix with values.
*Result NAME2 is column of length as the number of rows of MAT, showing number of nonzero elements in
*each row of MAT/NAME1.
*Argument RAND - digit (not name or expression) 0 or 1 (may put in quotes or apostrophes).
*If 1, action of randomizing will be added to the shifting: rammed values (and their indices)
*will go in random order, own in each row of NAME1.
*To manage random numbers  seed use out-of-matrix generator
*of random numbers (corresponds to commands SET MTINDEX or SET SEED).
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_ram(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
              /!pos= !charend('%') /!pos= !charend(')'))
!let !indval= !unquote(!3)
!let !rand= !unquote(!5)
comp @mat= !2.
comp @lr= !4.
comp @ind= 0.
comp @flag= @mat<>0.
comp !7= rsum(@flag).
do if cmax(!7).
-comp @m= @flag.
-loop @i= 2 to ncol(@m).
- comp @m(:,@i)= @m(:,@i-1)+@m(:,@i).
-end loop.
-do if @lr.
- comp @m= @m+ncol(@m)-!7*make(1,ncol(@m),1).
-end if.
-comp @m= @m&*@flag.
-release @flag.
-comp !6= make(nrow(@m),ncol(@m)+1,0).
 !if (!indval='BOTH') !then !let !ind= '@ind'
- comp @ind= make(nrow(@m),ncol(@m)+1,0).
 !else !let !ind= !6
 !ifend
-comp @nums= 1:ncol(@m).
-comp @m= @m+(not @m)*ncol(!6).
-loop @i= 1 to nrow(@m).
  !if (!indval='VAL' !or !indval='BOTH') !then
-  comp !6(@i,@m(@i,:))= @mat(@i,:).
  !ifend
  !if (!indval='IND' !or !indval='BOTH') !then
-  comp !ind(@i,@m(@i,:))= @nums.
  !ifend
-end loop.
-comp !6= !6(:,1:(ncol(!6)-1)).
 !if (!indval='BOTH') !then
- comp @ind= @ind(:,1:(ncol(@ind)-1)).
 !ifend
 !if (!rand) !then
- loop @i= 1 to nrow(!6).
-  do if !7(@i)>1.
-   comp @m= grade(uniform(1,!7(@i))).
-   comp @nums= 1:!7(@i).
-   do if @lr.
-    comp @nums= @nums+ncol(!6)-!7(@i).
-    comp @m= @m+ncol(!6)-!7(@i).
-   end if.
-   comp !6(@i,@m)= !6(@i,@nums).
    !if (!indval='BOTH') !then
-    comp @ind(@i,@m)= @ind(@i,@nums).
    !ifend
-  end if.
- end loop.
 !ifend
-release @m,@i,@nums.
else.
-release @flag.
-comp !6= make(nrow(@mat),ncol(@mat),0).
 !if (!indval='BOTH') !then
- comp @ind= make(nrow(@mat),ncol(@mat),0).
 !ifend
end if.
!if (!indval='BOTH') !then
-comp !6= {!6;@ind}.
!ifend
release @mat,@lr,@ind.
!enddefine.


******** RAMMING VERTICALLY NONZERO ELEMENTS OR THEIR INDICES ********.
*/*!KO_vram(mat%indval%tb%rand%name1%name2)*/*.
*Version 1.
*This function is like /*!KO_ram*/, but shifts elements vertically, not horizontally.
*Takes matrix or vector MAT and shifts all nonzero elements in each column of MAT to top or bottom
*border, compacting them (no zeros will be among them).
*Argument INDVAL - capitalized keyword (may take in quotes or apostrophes):
*"VAL" - shift nonzero values themselves
*"IND" - shift indices of nonzero values (i.e., their row numbers)
*"BOTH" - do both tasks
*Argument TB - scalar. If nonpositive, the shift will be to the top; if positive, then to the bottom.
*Result NAME1 is the matrix with the ramming (shifting) of values or their indices done,
*sized as MAT. With INDVAL=BOTH, both results are returned, harped: the matrix with indices joint
*right of the matrix with values.
*Result NAME2 is row of length as the number of columns of MAT, showing number of nonzero elements in
*each column of MAT/NAME1.
*Argument RAND - digit (not name or expression) 0 or 1 (may put in quotes or apostrophes).
*If 1, action of randomizing will be added to the shifting: rammed values (and their indices)
*will go in random order, own in each column of NAME1.
*To manage random numbers  seed use out-of-matrix generator
*of random numbers (corresponds to commands SET MTINDEX or SET SEED).
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_vram(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
               /!pos= !charend('%') /!pos= !charend(')'))
!let !indval= !unquote(!3)
!let !rand= !unquote(!5)
comp @mat= !2.
comp @tb= !4.
comp @ind= 0.
comp @flag= @mat<>0.
comp !7= csum(@flag).
do if rmax(!7).
-comp @m= @flag.
-loop @i= 2 to nrow(@m).
- comp @m(@i,:)= @m(@i-1,:)+@m(@i,:).
-end loop.
-do if @tb.
- comp @m= @m+nrow(@m)-!7(make(1,nrow(@m),1),:).
-end if.
-comp @m= @m&*@flag.
-release @flag.
-comp !6= make(nrow(@m)+1,ncol(@m),0).
 !if (!indval='BOTH') !then !let !ind= '@ind'
- comp @ind= make(nrow(@m)+1,ncol(@m),0).
 !else !let !ind= !6
 !ifend
-comp @nums= t(1:nrow(@m)).
-comp @m= @m+(not @m)*nrow(!6).
-loop @i= 1 to ncol(@m).
  !if (!indval='VAL' !or !indval='BOTH') !then
-  comp !6(@m(:,@i),@i)= @mat(:,@i).
  !ifend
  !if (!indval='IND' !or !indval='BOTH') !then
-  comp !ind(@m(:,@i),@i)= @nums.
  !ifend
-end loop.
-comp !6= !6(1:(nrow(!6)-1),:).
 !if (!indval='BOTH') !then
- comp @ind= @ind(1:(nrow(@ind)-1),:).
 !ifend
 !if (!rand) !then
- loop @i= 1 to ncol(!6).
-  do if !7(@i)>1.
-   comp @m= grade(uniform(1,!7(@i))).
-   comp @nums= 1:!7(@i).
-   do if @tb.
-    comp @nums= @nums+nrow(!6)-!7(@i).
-    comp @m= @m+nrow(!6)-!7(@i).
-   end if.
-   comp !6(@m,@i)= !6(@nums,@i).
    !if (!indval='BOTH') !then
-    comp @ind(@m,@i)= @ind(@nums,@i).
    !ifend
-  end if.
- end loop.
 !ifend
-release @m,@i,@nums.
else.
-release @flag.
-comp !6= make(nrow(@mat),ncol(@mat),0).
 !if (!indval='BOTH') !then
- comp @ind= make(nrow(@mat),ncol(@mat),0).
 !ifend
end if.
!if (!indval='BOTH') !then
-comp !6= {!6,@ind}.
!ifend
release @mat,@tb,@ind.
!enddefine.


******** PROPAGATE ROWS ********.
*/*!KO_propag(mat%times%name)*/*.
*Version 1.
*Takes matrix or vector MAT of any size. Propagates its rows in the number of copies
*specified in TIMES. In output matrix NAME the copies of each row go adjacently.
*TIMES - either positive integer, or vector lengthed as the number of rows of MAT where there is given
*the number of copies for each row separately. Values in the vector - nonnegative integers, at least one
*value greater than zero. Zero value indicates not to take that row of MAT into NAME at all.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_propag(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @times= !3.
do if nrow(@times)*ncol(@times)=1.
-comp !4= kroneker(!4,make(@times,1,1)).
else.
-comp @vec= 1:nrow(!4).
-comp @vec2= make(msum(@times),1,0).
-comp @end= 0.
-do if all(@times).
- loop @i= 1 to nrow(!4).
-  comp @start= @end+1.
-  comp @end= @start+@times(@i)-1.
-  comp @vec2(@start:@end)= make(@times(@i),1,@vec(@i)).
- end loop.
-else.
- loop @i= 1 to nrow(!4).
-  do if @times(@i).
-   comp @start= @end+1.
-   comp @end= @start+@times(@i)-1.
-   comp @vec2(@start:@end)= make(@times(@i),1,@vec(@i)).
-  end if.
- end loop.
-end if.
-comp !4= !4(@vec2,:).
-release @vec,@vec2,@start,@end,@i.
end if.
release @times.
!enddefine.


******** ADD A CONSTANT TO ELEMENTS OF A MATRIX TRIANGLE ********.
*/*!KO_tradd(mat%tr%val%name)*/*.
*Version 1.
*Adds number VAL (scalar) to elements of the upper or the lower triangle of MAT.
*Matrix MAT needs not be square, but must be at least size 2x2.
*TR - scalar; if positive, VAL is added to elements above the main diagonal;
*if nonpositive, VAL is added to elements below the main diagonal.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_tradd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @tr= !3.
comp @val= !4.
do if @tr.
-do if nrow(!5)>=ncol(!5).
- comp @m= ncol(!5)-1.
-else.
- comp @m= nrow(!5).
-end if.
-loop @i= 1 to @m.
- comp @ind= (@i+1):ncol(!5).
- comp !5(@i,@ind)= !5(@i,@ind)+@val.
-end loop.
else.
-do if ncol(!5)>=nrow(!5).
- comp @m= nrow(!5)-1.
-else.
- comp @m= ncol(!5).
-end if.
-loop @i= 1 to @m.
- comp @ind= (@i+1):nrow(!5).
- comp !5(@ind,@i)= !5(@ind,@i)+@val.
-end loop.
end if.
release @tr,@val,@m,@i,@ind.
!enddefine.


******** MULTIPLY ELEMENTS OF A MATRIX TRIANGLE BY A CONSTANT ********.
*/*!KO_trmult(mat%tr%val%name)*/*.
*Version 1.
*Multiplies elements of the upper or the lower triangle of MAT by number VAL (scalar).
*Matrix MAT needs not be square, but must be at least size 2x2.
*TR - scalar; if positive, elements above the main diagonal are multiplied by VAL;
*if nonpositive, elements below the main diagonal are multiplied by VAL.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_trmult(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @tr= !3.
comp @val= !4.
do if @tr.
-do if nrow(!5)>=ncol(!5).
- comp @m= ncol(!5)-1.
-else.
- comp @m= nrow(!5).
-end if.
-loop @i= 1 to @m.
- comp @ind= (@i+1):ncol(!5).
- comp !5(@i,@ind)= !5(@i,@ind)*@val.
-end loop.
else.
-do if ncol(!5)>=nrow(!5).
- comp @m= nrow(!5)-1.
-else.
- comp @m= ncol(!5).
-end if.
-loop @i= 1 to @m.
- comp @ind= (@i+1):nrow(!5).
- comp !5(@ind,@i)= !5(@ind,@i)*@val.
-end loop.
end if.
release @tr,@val,@m,@i,@ind.
!enddefine.


******** EXPONENTIATE ELEMENTS OF A MATRIX TRIANGLE BY A CONSTANT ********.
*/*!KO_trexp(mat%tr%power%name)*/*.
*Version 1.
*Exponentiates (raises to a power) elements of the upper or the lower triangle of MAT.
*The exponent - POWER (scalar). Limitations:
*If negative values are present in the triangle of MAT, POWER must be integer number.
*If zero values are present in the triangle of MAT, POWER must be positive number.
*Matrix MAT needs not be square, but must be at least size 2x2.
*TR - scalar; if positive, elements above the main diagonal are exponentiated;
*if nonpositive, elements below the main diagonal are exponentiated.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_trexp(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @tr= !3.
comp @val= !4.
do if @tr.
-do if nrow(!5)>=ncol(!5).
- comp @m= ncol(!5)-1.
-else.
- comp @m= nrow(!5).
-end if.
-loop @i= 1 to @m.
- comp @ind= (@i+1):ncol(!5).
- comp !5(@i,@ind)= !5(@i,@ind)&**@val.
-end loop.
else.
-do if ncol(!5)>=nrow(!5).
- comp @m= nrow(!5)-1.
-else.
- comp @m= ncol(!5).
-end if.
-loop @i= 1 to @m.
- comp @ind= (@i+1):nrow(!5).
- comp !5(@ind,@i)= !5(@ind,@i)&**@val.
-end loop.
end if.
release @tr,@val,@m,@i,@ind.
!enddefine.


******** UNFOLD THE TRIANGLES OF SQUARE MATRIX INTO VECTORS ********.
*/*!KO_unftri(mat%name1%name2%name3)*/*.
*Version 2.
*Takes matrix MAT which must be square, and returns 2-column matrix NAME1,
*where the 1st column is the unfolding of below-diagonal (lower-left) triangle of MAT
*and the 2nd  column is the unfolding of above-diagonal (upper-right) triangle of MAT.
*A pair of elements in the same row of NAME1 are symmetrical-positioned elements in MAT.
*NAME2 are the corresponding to NAME1 indices (positions) of off-diagonal elements in the
*vector that is the unfold (vectorizing) of matrix MAT. 1-st row of NAME2 pertains to
*elements of below-diagonal triangle, and 2-nd row - to elements of above-diagonal triangle.
*With the help of NAME2, you can quickly unfold triangles of any matrix sized as MAT,
*without resort a second time to this function.
*NAME3 contains memo - the numbers of rows and columns of MAT where the elements came from.
*Thus, the 1st pair of indices in NAME3 is 2 and 1, which reminds that the 1st pair of values
*of NAME1 are: the 2nd element from the 1st column and the 2nd element from the 1st row.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Function /*!KO_foltri*/ is opposite to this. It folds a pair of vectors NAME1 into the triangles
*of the square matrix MAT.

define !KO_unftri(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp @n= ncol(!3).
comp @n1= @n+1.
comp @1n= 1-@n.
comp @nn1= @n*(@n-1).
comp !3= reshape(!3,@n**2,1).
comp @uind= make(1,@nn1/2,0).
comp @lind= make(1,@nn1/2,0).
comp @ind1= make(1,@nn1/2,0).
comp @ind2= make(1,@nn1/2,0).
loop @i= 1 to @n-1.
-comp @st= @1n+@i*(@n-@i/2+.5).
-comp @ind= (@i*@n1+@1n):(@i*@n).
-comp @pos= @st:(@st+ncol(@ind)-1).
-comp @uind(@pos)= @ind.
-comp @ind= (@i*@n1):(@i+@nn1):@n.
-comp @lind(@pos)= @ind.
-comp @ind1(@pos)= make(1,ncol(@ind),@i).
-comp @ind2(@pos)= (@i+1):@n.
end loop.
comp !3= {!3(@lind),!3(@uind)}.
comp !4= {@lind;@uind}.
comp !5= {@ind2;@ind1}.
release @n,@n1,@1n,@nn1,@uind,@lind,@ind1,@ind2,@i,@st,@ind,@pos.
!enddefine.


******** FOLD TWO VECTORS INTO THE TRIANGLES OF SQUARE MATRIX ********.
*/*!KO_foltri(vecs%name1%name2%name3)*/*.
*Version 2.
*Takes matrix VECS, which must have 2 columns (the vectors), and returns square
*matrix NAME1, in which the below-diagonal (lower-left) triangle is the folded 1st vector
*and the above-diagonal (upper-right) triangle is the folded 2nd vector.
*A pair of elements in the same row of VECS become symmetrical-positioned elements in NAME1.
*The diagonal of NAME1 consists of zeros.
*NAME2 are the corresponding to NAME1 indices (positions) of off-diagonal elements in the
*vector that is the unfold (vectorizing) of matrix NAME1. 1-st row of NAME2 pertains to
*elements of below-diagonal triangle, and 2-nd row - to elements of above-diagonal triangle.
*NAME3 is a scalar, the size of matrix NAME1.
*If the number of rows in VECS does not permit to create square matrix NAME1, then NAME3
*will appear a noninteger number, and NAME1 and NAME2 will be returned as zero scalars.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Function /*!KO_unftri*/ is opposite to this. It unfolds the triangles of the square matrix NAME1
*into the pair of vectors VECS.

define !KO_foltri(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @vecs= !2.
comp !5= .5+sqrt(.25+2*nrow(@vecs)).
do if trunc(!5)=!5.
-comp @n1= !5+1.
-comp @1n= 1-!5.
-comp @nn1= !5*(!5-1).
-comp @uind= make(1,nrow(@vecs),0).
-comp @lind= make(1,nrow(@vecs),0).
-loop @i= 1 to !5-1.
- comp @st= @1n+@i*(!5-@i/2+.5).
- comp @ind= (@i*@n1+@1n):(@i*!5).
- comp @uind(@st:(@st+ncol(@ind)-1))= @ind.
- comp @ind= (@i*@n1):(@i+@nn1):!5.
- comp @lind(@st:(@st+ncol(@ind)-1))= @ind.
-end loop.
-comp !3= make(!5**2,1,0).
-comp !3(@lind)= @vecs(:,1).
-comp !3(@uind)= @vecs(:,2).
-comp !3= reshape(!3,!5,!5).
-comp !4= {@lind;@uind}.
-release @vecs,@n1,@1n,@nn1,@uind,@lind,@i,@st,@ind.
else.
-comp !3= 0.
-comp !4= 0.
end if.
!enddefine.


******** SYMMETRIZATION OF SQUARE MATRIX: REPLACE ONE TRIANGLE BY THE OTHER ********.
*/*!KO_symtri1(mat%tr%name)*/*.
*Version 1.
*Takes matrix MAT which must be square, and replaces one triangle in it by the other one;
*Returns square symmetric matrix NAME.
*Argument TR orders, which triangle to replace:
*TR non-positive scalar => below-diagonal triangle be replaced by above-diagonal triangle;
*TR positive scalar  => above-diagonal triangle be replaced by below-diagonal triangle.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_symtri1(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @tr= !3.
do if not @tr.
-comp !4= t(!4).
end if.
loop @i= 1 to ncol(!4).
-comp !4(@i,:)= t(!4(:,@i)).
end loop.
release @tr,@i.
!enddefine.


******** SYMMETRIZATION OF SQUARE MATRIX: REPLACE GREATER ELEMENT BY LESSER OR LESSER BY GREATER ********.
*/*!KO_symtri2(mat%el%name)*/*.
*Version 2.
*Takes matrix MAT which must be square, and replaces in it each off-diagonal element
*by the one position-symmetric to it, if the latter is greater (or lesser) than it;
*returns square symmetric matrix NAME.
*Argument EL orders, when to do the replacement:
*EL non-positive scalar => replace element by its symmetric, if the latter is lesser than it;
*EL positive scalar  => replace element by its symmetric, if the latter is greater than it.

define !KO_symtri2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @el= !3.
comp @t= t(!4).
do if @el.
-comp @gv= !4&*(!4>@t).
else.
-comp @gv= !4&*(!4<@t).
end if.
comp @ev= !4&*(!4=@t).
comp !4= @gv+t(@gv)+@ev.
release @el,@t,@gv,@ev.
!enddefine.


******** SORTING OF VECTOR ********.
*/*!KO_sort(vec%ord%name1%name2)*/*.
*Version 2.
*Sorts elements in vector (row or column) VEC by magnitude of  the values.
*Returns sorted vector  NAME1.
*And also vector NAME2 which is the element numbers in NAME1 arranged in the sequence
*in what they were prior the sorting (with the help of NAME2 you can restore the initial order
*of elements). NAME2 is the integer ranks of values of VEC.
*Argument ORD orders the order of sorting:
*ORD non-positive scalar => descending; ORD positive scalar => ascending.
*If you need then to reorder rows of an external array the same way the elements in VEC appeared to be
*sorted, use NAME2 as COL in /*!KO_rsortc*/ and set there ORD as 0.

define !KO_sort(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @ord= !3.
comp @ord= (@ord>0)-(@ord<=0).
comp !5= grade(@ord&*!4).
comp !4(!5)= !4.
release @ord.
!enddefine.


******** SIMPLE SORTING OF ROWS IN MATRIX ********.
*/*!KO_rsort(mat%cn%ord%name1%name2)*/*.
*Version 2.
*Sorts rows in matrix MAT by values of its column CN.
*Returns matrix NAME1.
*And also column NAME2 which is the row numbers in NAME1 arranged in the sequence in what
*the rows were prior the sorting (with the help of NAME2 you can restore the initial order
*of rows). NAME2 is the integer ranks of values of column CN (prior sorting).
*CN is the column number, positive integer scalar.
*Argument ORD orders the order of sorting:
*ORD non-positive scalar => descending; ORD positive scalar => ascending.
*If you need then to reorder rows of an external array the same way the rows in MAT appeared to be
*sorted, use NAME2 as COL in /*!KO_rsortc*/ and set there ORD as 0.

define !KO_rsort(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @cn= !3.
comp @ord= !4.
comp @ord= (@ord>0)-(@ord<=0).
comp !6= grade(@ord&*!5(:,@cn)).
comp !5(!6,:)= !5.
release @cn,@ord.
!enddefine.


******** SORTING OF ROWS IN MATRIX BY VALUES OF EXTERNAL COLUMN ********.
*/*!KO_rsortc(mat%col%ord%name1%name2)*/*.
*Version 2.
*Sorts rows in matrix MAT by values of external column COL.
*COL is a column vector with length as the number of rows in MAT.
*Returns matrix NAME1.
*And also column NAME2 which is the row numbers in NAME1 arranged in the sequence in what
*the rows were prior the sorting (with the help of NAME2 you can restore the initial order
*of rows). NAME2 is the integer ranks of values of COL.
*Argument ORD orders the order of sorting:
*ORD negative scalar => descending; ORD positive scalar => ascending.
*If ORD is 0 then COL must contain integer ranks, and the order of sorting is defined now by them.
*At that, the function spends no time at ranking. This way of specification is advantageous in the sense of
*speed, when you want to sort out rows in several matrices (with the same number of rows) in identical
*order but you don't want to joint temporarily the matrices in one for that. Then you may use
*NAME2 obtained at sorting of the first matrix as COL in sorting of the rest of the matrices.

define !KO_rsortc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @col= !3.
comp @ord= !4.
do if @ord=0.
-comp !6= @col.
else.
-comp @ord= (@ord>0)-(@ord<0).
-comp !6= grade(@ord&*@col).
end if.
-comp !5(!6,:)= !5.
release @col,@ord.
!enddefine.


******** HIERARCHICAL SORTING OF ROWS IN MATRIX ********.
*/*!KO_hiesort(mat%ord%name1%name2)*/*.
*Version 2.
*Sorts rows in matrix MAT by values of all its columns.
*Rows are sorted hierarchically: by values of the 1st column; within its identical
*values – by values of the 2nd column; within its identical values - by
*values of the 3rd column; etc. This function is like out-of-matrix SORT CASES command
*when you specify several variables in that command.
*Returns matrix NAME.
*And also column NAME2 which is the row numbers in NAME1 arranged in the sequence in what
*the rows were prior the sorting (with the help of NAME2 you can restore the initial order
*of rows).
*Argument ORD is a vector lengthed as the number of columns in MAT or is a scalar;
*it requests the order of sorting in each separate column:
*n-th element of ORD non-positive => descending sorting in n-th column;
*n-th element of ORD positive => ascending sorting in n-th column.
*Scalar ORD means same order for all the columns.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*The function is relatively slow.
*If you need then to reorder rows of an external array the same way the rows in MAT appeared to be
*sorted, use NAME2 as COL in /*!KO_rsortc*/ and set there ORD as 0.

define !KO_hiesort(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @ord= !3.
comp @ord= (@ord>0)-(@ord<=0).
do if nrow(@ord)*ncol(@ord)=1.
-comp @ord= make(1,ncol(!4),@ord).
end if.
comp !5= t(1:nrow(!4)).
comp !4= {!4,!5}.
comp !4(grade(@ord(1)*!4(:,1)),:)= !4.
do if ncol(!4)>2.
-loop @i= 2 to ncol(!4)-1.
- comp @lst= 0.
- loop. 
-  comp @fst= @lst+1.
-  loop @lst= @fst to nrow(!4)-1.
-  end loop if any(!4(@lst+1,1:(@i-1))<>!4(@lst,1:(@i-1))).
-  comp !4(grade(@ord(@i)*!4(@fst:@lst,@i))+@fst-1,:)= !4(@fst:@lst,:).
- end loop if @lst=nrow(!4).
-end loop.
-release @i,@lst,@fst.
end if.
comp !5(!4(:,ncol(!4)))= !5.
comp !4= !4(:,1:(ncol(!4)-1)).
release @ord.
!enddefine.


******** SPLITTING OF MATRIX INTO TWO ********.
*/*!KO_split(whole%dim%ind%name1%name2)*/*.
*Version 2.
*From matrix WHOLE, either its rows or its columns with indices IND are cut out,
*forming cutting matrix NAME1. The remaining rows (columns) of WHOLE form
*remnant matrix NAME2. The function returns both matrices NAME1 and NAME2.
*Argument DIM - keyword (optionally quoted or in apostrophes):
*"ROWS" - to cut out rows (word "arrays" below stands for rows);
*"COLS" - to cut out columns (word "arrays" below stands for columns).
*WHOLE must have the number of arrays more than one.
*IND is a vector (row or column, doesn't matter) with length less than there
*are arrays in WHOLE, and containing positive integers: indices of arrays in WHOLE
*that should be cut out. These numbers do not have to go in ascending order; their sequence
*defines the sequence of arrays how it will be in the cutting matrix NAME1.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Function /*!KO_merge*/ is opposite to this one and can insert cut-in NAME1 into base NAME2
*restoring matrix WHOLE.

define !KO_split(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @whole= !2.
!let !dim= !unquote(!3)
comp @ind= !4.
comp !5= @whole(!if (!dim='COLS') !then !let !nd= 'ncol' :,@ind !else !let !nd= 'nrow' @ind,: !ifend).
comp @lenind= rmax({nrow(@ind),ncol(@ind)}).
comp @notind= make(1,!nd(@whole)-@lenind,0).
comp @c= 0.
loop @i= 1 to !nd(@whole).
-do if not any(@ind=@i).
- comp @c= @c+1.
- comp @notind(@c)= @i.
-end if.
end loop.
comp !6= @whole(!if (!dim='COLS') !then :,@notind !else @notind,: !ifend).
release @whole,@ind,@lenind,@notind,@c,@i.
!enddefine.


******** MERGING OF TWO MATRICES ********.
*/*!KO_merge(base%dim%slice%ind%name)*/*.
*Version 2.
*Into base matrix BASE, all rows or columns of cut-in matrix SLICE are inserted, forming the
*united matrix NAME which is returned. Rows (columns) constituting matrix SLICE will occupy in NAME
*positions specified in IND.
*Argument DIM - keyword (optionally quoted or in apostrophes):
*"ROWS" - insert SLICE rows as rows in BASE (word "arrays" below stands for rows), number of columns
*in BASE and SLICE must be equal;
*"COLS" - insert SLICE columns as columns in BASE (word "arrays" below stands for columns), number of
*rows in BASE and SLICE must be equal.
*IND is a vector (row or column, doesn't matter) with length as the number of arrays in SLICE,
*and containing positive integers: indices of arrays in NAME which must be occupied by arrays of
*the cut-in matrix SLICE. These numbers do not have to go in ascending order; their sequence defines
*the sequence of cutting-in of arrays of SLICE between arrays of BASE.
*The maximal value in IND cannot exceed the sum of the number of arrays in BASE and SLICE.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Function /*!KO_split*/ is opposite to this one and can split NAME back into
*cutting SLICE and remnant BASE.

define !KO_merge(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @base= !2.
!let !dim= !unquote(!3)
comp @slice= !4.
comp @ind= !5.
comp !6= make(!if (!dim='COLS') !then !let !nd= 'ncol' nrow(@base),ncol(@base)+ncol(@slice) !else !let !nd= 'nrow' nrow(@base)+nrow(@slice),ncol(@base) !ifend,0).
comp !6(!if (!dim='COLS') !then :,@ind !else @ind,: !ifend)= @slice.
comp @r= 0.
loop @i= 1 to !nd(!6).
-do if not any(@ind=@i).
- comp @r= @r+1.
- comp !if (!dim='COLS') !then !6(:,@i)= @base(:,@r) !else !6(@i,:)= @base(@r,:) !ifend.
-end if.
end loop.
release @base,@slice,@ind,@r,@i.
!enddefine.


******** SPLITTING OF MATRIX HORIZONTALLY IN SEVERAL, BY GROUPS ********.
*/*!KO_msplit(mat%group%'maxcode')*/*.
*Version 1.
*Takes matrix MAT (some name, not expression) and grouping variable GROUP (sorted).
*Splits MAT horizontally in matrices corresponding to groups in GROUP. Names of the resulting
*matrices are defined automatically.
*The function is partly similar to command SPLIT FILE.
*Details:
*GROUP - case-grouping variable, column with integer group codes: 1, 2, 3,...
*(Fractional codes and values less than 1 not allowed.) Cases (rows) must be already sorted by
*the codes ascendingly.
*MAT - matrix (data) with num-of-rows as in GROUP, already sorted correspondingly with GROUP, which
*you need to break horizontally: one resultant matrix for each code of GROUP.
*The names will be concatenated from name of MAT and the group code: MAT1, MAT2,...
*'MAXCODE' - code of the last observed or of your interest group (maximal 999). It must be number
*(not variable or expression), you may put it in apostrophes or quotes.
*If some codes between 1 and maxcode are absent in GROUP the matrices for these missing groups
*will not be created.
*This function is useful for routine work, not for programming algorithms.

define !KO_msplit(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @gr= !3.
comp @ind= t(1:nrow(!2)).
comp @lst= 0.
!do !i= 1 !to !unquote(!4)
-comp @flag= (@gr=!i)&*@ind.
-comp @fst= @lst+1.
-comp @last= cmax(@flag).
-do if @last.
- comp @lst= @last.
- comp !conc(!2,!i)= !2(@fst:@lst,:).
-end if.
!doend
release @gr,@ind,@flag,@fst,@lst,@last.
!enddefine.


******** RESTRUCTURING "VARIABLES INTO CASES" ********.
*/*!KO_vartocas(data%k%ord%name1%name2)*/*.
*Version 1.
*Restructures each group of K variables (columns) of DATA into single column.
*This way each case (row) gets broken in K rows.
*Returns restructured data NAME1.
*The function is similar to SPSS command VARSTOCASES.
*DATA - matrix with number of columns multiple of K.
*K - positive integer scalar; each K columns of DATA (each "group") will produce one column in NAME1.
*Scalar ORD indicates in which order the variables producing single column are situated in DATA:
*situated adjacently (ORD positive) or situated with step number_columns_DATA/K (ORD nonpositive).
*For example, if number_columns_DATA=6 and K=2 then adjacent situation means that columns
*1 and 2 will produce column, 3 and 4 will produce column, 5 and 6 will produce column. And with situation via step 3
*columns 1 and 4 will produce column, 2 and 5 will produce column, 3 and 6 will produce column.
*Two-column matrix NAME2 reflects that order: it contains case numbers of DATA in its 1st column and
*the sequential index of a variable in "group" (1,2,...,K) of variables in its 2nd column.
*Function */*!KO_castovar*/* is opposite to this one and restructures NAME1 back in DATA.

define !KO_vartocas(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @k= !3.
comp @ord= !4.
comp !6= t({kroneker(1:nrow(!5),make(1,@k,1));kroneker(make(1,nrow(!5),1),1:@k)}).
do if @ord.
-comp !5= !5(:,kroneker(make(1,@k,1),1:ncol(!5):@k)+kroneker(0:(@k-1),make(1,ncol(!5)/@k,1))).
end if.
comp !5= reshape(!5,nrow(!5)*@k,ncol(!5)/@k).
release @k,@ord.
!enddefine.


******** RESTRUCTURING "CASES INTO VARIABLES" ********.
*/*!KO_castovar(data%k%ord%name1%name2)*/*.
*Version 1.
*Restructures each group of K cases (rows) of DATA into single row.
*This way each variable (column) gets broken in K columns.
*Returns restructured data NAME1.
*The function is similar to SPSS command CASESTOVARS.
*DATA - matrix with number of rows multiple of K.
*K - positive integer scalar; each K rows of DATA (each "group") will produce one row in NAME1.
*Rows in DATA must be already sorted so that K rows of each group appear adjacent.
*Scalar ORD specifies in which order should the variables produced of single variable of DATA be situated:
*adjacently (ORD positive) or with step number_columns_DATA (ORD nonpositive).
*For example, if number_columns_DATA=3 и K=2 then adjacent situation means
*that column 1 will produce columns 1 and 2, column 2 will produce columns 3 and 4, column 3 will produce columns 5 and 6.
*And with situation via step 3 column 1 will produce columns 1 and 4, column 2 will produce columns 2 and 5,
*column 3 will produce columns 3 and 6.
*Two-row matrix NAME2 reflects that order: it contains variable numbers of DATA
*in its 1st row and the sequential index of a case in "group" (1,2,...,K) of cases in its 2nd row.
*Function */*!KO_vartocas*/* is opposite to this one and restructures NAME1 back in DATA.

define !KO_castovar(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @k= !3.
comp @ord= !4.
comp @ind= kroneker(make(1,ncol(!5),1),1:(ncol(!5)*@k):ncol(!5))+kroneker(0:(ncol(!5)-1),make(1,@k,1)).
comp !6= {kroneker(make(1,@k,1),1:ncol(!5));kroneker(1:@k,make(1,ncol(!5),1))}.
comp !5= reshape(!5,nrow(!5)/@k,ncol(!5)*@k).
do if @ord.
-comp !6= !6(:,@ind).
-comp !5= !5(:,@ind).
end if.
release @k,@ord,@ind.
!enddefine.


******** SHIFT ROWS HORIZONTALLY ********.
*/*!KO_shift(mat%rel%displ%name)*/*.
*Version 1.
*Takes matrix MAT with no less than two rows. Returns matrix NAME where rows were
*shifted horizontally - to right or to left by the amounts (number of elements) specified in column
*vector DISPL. DISPL must be of the length 1 less than the number of rows in MAT. Its values - integers.
*Negative value is shift by this number of elements to the left, positive value is shift by this number
*of elements to the right. The 1st element is the shift for row 2, 2nd element is the shift of
*row 3, etc.
*Argument REL (scalar) explains relative what the shifts in DISPL are specified. If REL is nonpositive
*then shift of every row is the shift relative the 1st row. If REL is positive then shift of every
*row is the shift relative the previous, overlying row.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_shift(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @rel= !3.
comp @shift= !4.
do if @rel.
-loop @i= 2 to nrow(@shift).
- comp @shift(@i)= @shift(@i-1)+@shift(@i).
-end loop.
end if.
comp @marg= cmax((@shift<0)&*(-@shift)).
comp !5= make(nrow(@mat),ncol(@mat)+cmax((@shift>0)&*@shift)+@marg,0).
comp @shift= {0;@shift}.
loop @i= 1 to nrow(@shift).
-comp !5(@i,1:ncol(@mat)+@marg+@shift(@i))= @mat(@i,:).
end loop.
release @mat,@rel,@shift,@i,@marg.
!enddefine.


******** BLOCK-INCUT ********.
*/*!KO_blockic(mat1%mat2%pos%name)*/*.
*Version 1.
*Takes matrices MAT1 and MAT2 of any size and unites them, in-cutting MAT2 from bottom-right
*into MAT1, docking it to position POS in MAT1. Specifically, first, matrix NAME is created
*as block matrix: block(MAT1(1:I,1:J),MAT2). Then, if there are left cells in NAME not yet
*received values from their corresponding cells of MAT1, they receive them.
*POS - vector of length 2, containing positions I and J in MAT1. I can be from 1 to num_rows
*of MAT1. J can be from 1 to num_columns of MAT1. When I and J are both maximal, the function
*turns into function block().
*If you want to "revert" the task and in-cut MAT2 into MAT1 from top-left, then first
*change the order of rows and columns in both matrices to the opposite, and in the end change
*the order of rows and columns in NAME to the opposite.

define !KO_blockic(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @m1= !2.
comp @m2= !3.
comp @cr= !4.
comp @rs= @cr(1).
comp @cs= @cr(2).
comp !5= block(@m1(1:@rs,1:@cs),@m2).
do if @cs<ncol(@m1).
-comp @cr= mmin({ncol(@m1),ncol(!5)}).
-comp !5(1:@rs,(@cs+1):@cr)= @m1(1:@rs,(@cs+1):@cr).
end if.
do if @rs<nrow(@m1).
-comp @cr= mmin({nrow(@m1),nrow(!5)}).
-comp !5((@rs+1):@cr,1:@cs)= @m1((@rs+1):@cr,1:@cs).
end if.
release @m1,@m2,@cr,@rs,@cs.
!enddefine.


******** DYADIC SET OPERATIONS ********.
*/*!KO_setdo(set1%set2%oper%name)*/*.
*Version 1.
*Takes two sets SET1 and SET2 (arbitrary matrices) and returns their union, intersection,
*difference, or symmetric difference. Value 0 in SET1 and SET2 is reserved for emptiness, i.e.,
*it is not considered an element of a set. Any other value is an element and is accounted
*once in the resulting row vector NAME (that is, for example, intersection between {1,2,2,2,3}
*and {1,2,2,3,4} will be {1,2,3}, and not {1,2,2,3} or {1,2,2,2,3}).
*OPER - capitalized keyword (may be quoted or in apostrophes):
*"UNION" - union of two sets (values present at least in one of the sets);
*"INTER" - intersection of two sets (values present in both sets);
*"DIFF" - difference of two sets (values present in SET1 but not in SET2);
*"SYMDIFF" - symmetric difference of two sets (values present in SET1 but not in SET2, or
*in SET2 but not in SET1).
*NAME = scalar 0 means empty set at output.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_setdo(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @set1= !2.
comp @set2= !3.
!let !oper= !unquote(!4)
!if (!oper='DIFF') !then !let !n= 'not' !ifend
comp @i= 1.
comp !5= 0.
loop @i= 1 to nrow(@set1).
-loop @j= 1 to ncol(@set1).
- comp @val= @set1(@i,@j).
- do if @val<>0.
   !if (!oper='INTER' !or !oper='DIFF') !then
-   do if !n any(@set2=@val).
-    comp !5= {!5,@val}.
-    comp @set1= @set1&*(@set1<>@val).
-   end if.
   !else !if (!oper='UNION') !then
-   comp !5= {!5,@val}.
-   comp @set1= @set1&*(@set1<>@val).
-   comp @set2= @set2&*(@set2<>@val).
   !else
-   do if any(@set2=@val).
-    comp @set2= @set2&*(@set2<>@val).
-   else.
-    comp !5= {!5,@val}.
-   end if.
-   comp @set1= @set1&*(@set1<>@val).
   !ifend !ifend
- end if.
-end loop.
end loop.
!if (!oper='UNION' !or !oper='SYMDIFF') !then
-loop @i= 1 to nrow(@set2).
- loop @j= 1 to ncol(@set2).
-  comp @val= @set2(@i,@j).
-  do if @val<>0.
-   comp !5= {!5,@val}.
-   comp @set2= @set2&*(@set2<>@val).
-  end if.
- end loop.
-end loop.
!ifend
do if ncol(!5)>1.
-comp !5= !5(2:ncol(!5)).
end if.
release @set1,@set2,@i,@j,@val.
!enddefine.


******** CARTESIAN PRODUCT BETWEEN TWO SETS ********.
*/*!KO_cartp(set1%set2%ii%name)*/*.
*Version 1.
*Combines elements between two sets, creating the cartesian product. There is an option not to pair
*elements with same index numbers.
*Sets SET1 and SET2 are matrices of same or different size with >1 rows.
*Their rows are the elements in the sets.
*Result NAME is the matrix each row of which is a unique pair of elements, the 1st is from SET1,
*the 2nd is from SET2. Cartesian product is all possible paired combinations.
*Columns of NAME are the concatenation of columns of the input matrices.
*Argument II is scalar. If it is nonpositive then usual, complete cartesian product is returned.
*If II is positive then pairs representing elements having identical index numbers are excluded from the result.
*For example, if SET1 is A,B,C and SET2 is 1,2,3,4, then pairs A1, B2 and C3 will be omitted.

define !KO_cartp(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @1= !2.
comp @2= !3.
comp @ii= !4.
do if @ii and nrow(@1)>nrow(@2).
-comp @1_= @1((nrow(@2)+1):nrow(@1),:).
-comp @1= @1(1:nrow(@2),:).
-comp !5= {kroneker(@1,make(nrow(@2),1,1)),kroneker(make(nrow(@1),1,1),@2)}.
-comp @ind= kroneker(make(1,nrow(@1)-1,1),1:nrow(@2)+1)+kroneker(0:(nrow(@1)-2),make(1,nrow(@2),nrow(@2)+1)).
-comp !5= {!5(@ind,:);{kroneker(@1_,make(nrow(@2),1,1)),kroneker(make(nrow(@1_),1,1),@2)}}.
-release @1_,@ind.
else.
-comp !5= {kroneker(@1,make(nrow(@2),1,1)),kroneker(make(nrow(@1),1,1),@2)}.
-do if @ii.
- comp @ind= kroneker(make(1,nrow(@1)-1,1),1:nrow(@2)+1)+kroneker(0:(nrow(@1)-2),make(1,nrow(@2),nrow(@2)+1)).
- do if nrow(@1)<nrow(@2).
-  comp @ind= {@ind,(@ind(ncol(@ind))+2):nrow(!5)}.
- end if.
- comp !5= !5(@ind,:).
- release @ind.
-end if.
end if.
release @1,@2,@ii.
!enddefine.


******** CARTESIAN PRODUCT BETWEEN N SETS ********.
*/*!KO_ncartp(setlist%name)*/*.
*Version 1.
*Combines elements between two or more sets, creating the cartesian product.
*The sets are matrices of any size, rows of a matrix are considered the elements of that set.
*SETLIST is the list (via blanks) of the matrix names. The list may be optionally quoted/apostrophed.
*Returns matrix NAME which rows are the elements of the catresian product of those matrices, and columns
*are the concatenated columns of the matrices.

define !KO_ncartp(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
!let !sets= !unquote(!2) !let !sts= !null
!do !w !in (!sets) !let !sts= !conc(!w,' ',!sts) !doend
comp !3= !head(!sts).
!do !w !in (!tail(!sts))
-comp !3= {kroneker(!w,make(nrow(!3),1,1)),kroneker(make(nrow(!w),1,1),!3)}.
!doend
!enddefine.


******** COMBINATIONS BY K ELEMENTS ********.
*/*!KO_combk(set%k%name)*/*.
*Version 1.
*Forms combinations by k elements from a set of n elements.
*Set SET is a column or matrix with n rows: a row is an element in the set.
*K (integer scalar from 2 to n) - the needed number of elements in a combination.
*Result NAME - matrix where each row is a combination of elements.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_combk(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @v= !2.
comp @maxk= !3.
comp @n= nrow(@v).
comp @next= 1:@n+1.
comp !4= @v.
loop @k= 2 to @maxk.
-comp @nextnew= 0.
-comp @vecnew= make(1,ncol(!4)+ncol(@v),0).
-loop @i= 1 to nrow(!4).
- do if @next(@i)<=@n.
-  comp @ind= @next(@i):@n.
-  comp @vecnew= {@vecnew;{!4(@i,:)(make(1,ncol(@ind),1),:),@v(@ind,:)}}.
-  comp @nextnew= {@nextnew,@ind+1}.
- end if.
-end loop.
-comp @next= @nextnew(2:ncol(@nextnew)).
-comp !4= @vecnew(2:nrow(@vecnew),:).
end loop.
release @v,@maxk,@n,@next,@k,@nextnew,@vecnew,@i,@ind.
!enddefine.


******** ALL COMBINATIONS BY K=2,3... ELEMENTS ********.
*/*!KO_allcomb(set%maxk%prefix)*/*.
*Version 1.
*Forms, from a set of n elements, combinations by k=2, k=3, ..., k=MAXK elements.
*Returns all these combinations respectively as matrices PREFIX2, PREFIX3, etc.
*I.e. this function is like */*!KO_combk*/*, but it returns at once all combinations with different k.
*Set SET is a column or matrix with n rows: a row is an element in the set.
*MAXK - maximal number of elements in a combination. That must be integer number (not name or expression)
*from 2 to n. Optionally it can be taken in quotes or apostrophes.
*PREFIX - prefix (optionally in quotes or apostrophes) in the names of the output matrices, the sets of
*combinations.
*In the returned matrices each row is a combination of elements.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Tip: if you want to set MAXK=n but number n is not known to you in advance, specify MAXK as a number
*knowingly exceeding possible n. Matrices above PREFIXn won't be able to create and no error will occur.

define !KO_allcomb(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !maxk= !unquote(!3) !let !pref= !unquote(!4)
comp @v= !2.
comp @n= nrow(@v).
comp @next= 1:@n+1.
comp @vec= @v.
!do !k= 2 !to !maxk
-do if !k<=@n.
- comp @nextnew= 0.
- comp @vecnew= make(1,ncol(@vec)+ncol(@v),0).
- loop @i= 1 to nrow(@vec).
-  do if @next(@i)<=@n.
-   comp @ind= @next(@i):@n.
-   comp @vecnew= {@vecnew;{@vec(@i,:)(make(1,ncol(@ind),1),:),@v(@ind,:)}}.
-   comp @nextnew= {@nextnew,@ind+1}.
-  end if.
- end loop.
- comp @next= @nextnew(2:ncol(@nextnew)).
- comp @vec= @vecnew(2:nrow(@vecnew),:).
- comp !conc(!pref,!k)= @vec.
-end if.
!doend
release @v,@n,@next,@vec,@nextnew,@vecnew,@i,@ind.
!enddefine.


******** COMBINATIONS BY K ELEMENTS FROM DIFFERENT SETS ********.
*/*!KO_dscombk(sets%last%k%name)*/*.
*Version 1.
*Forms combinations by k elements belonging to different sets.
*Sets SETS is a column or matrix with n rows: a row is an element in a set.
*The sets must be gathered stacked in SETS: first go all elements (rows) of one set,
*below it all elements of a second set, and so on. There may be any number of sets and the number
*of elements in them may be different.
*LAST - vector (row or column) lengthed as the number of sets in SETS and in correspondence to them,
*containing indices of last rows in each set. For example, {3,6,11} means: the are three sets in SETS,
*with elemens occupying rows 1:3, 4:6, 7:11. 11 (last element of the last set) must equal the number
*of rows in SETS.
*K (integer scalar from 2 to nc, where nc is the number of sets, length of vector LAST) - number of
*elements in a combination.
*Result NAME - matrix where each row is a combination of k elements from different sets, no more than
*one element from a set. If K = number of sets then a combination is "by one element from all sets".
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_dscombk(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @v= !2.
comp @last= !3.
comp @maxk= !4.
comp @n= nrow(@v).
do if nrow(@last)>ncol(@last).
-comp @last= t(@last).
end if.
comp @plast= {0,@last}.
comp @explast= 0.
loop @i= 1 to ncol(@last).
-comp @val= @last(@i).
-comp @explast= {@explast,make(1,@val-@plast(@i),@val)}.
end loop.
comp @explast= @explast(2:ncol(@explast)).
comp @next= @explast+1.
comp !5= @v.
loop @k= 2 to @maxk.
-comp @nextnew= 0.
-comp @vecnew= make(1,ncol(!5)+ncol(@v),0).
-loop @i= 1 to nrow(!5).
- do if @next(@i)<=@n.
-  comp @ind= @next(@i):@n.
-  comp @vecnew= {@vecnew;{make(ncol(@ind),1,1)*!5(@i,:),@v(@ind,:)}}.
-  comp @nextnew= {@nextnew,@explast(@ind)+1}.
- end if.
-end loop.
-comp @next= @nextnew(2:ncol(@nextnew)).
-comp !5= @vecnew(2:nrow(@vecnew),:).
end loop.
release @v,@last,@maxk,@n,@next,@k,@plast,@explast,@val,@nextnew,@vecnew,@i,@ind.
!enddefine.


******** ALL COMBINATIONS BY K=2,3... ELEMENTS FROM DIFFERENT SETS ********.
*/*!KO_dsallcomb(sets%last%maxk%prefix)*/*.
*Version 1.
*Forms, from several sets, combinations by k=2, k=3, ..., k=MAXK elements belonging to different sets.
*Returns all these combinations respectively as matrices PREFIX2, PREFIX3, etc.
*I.e. this function is like */*!KO_dscombk*/*, but it returns at once all combinations with different k.
*Sets SETS is a column or matrix with n rows: a row is an element in a set.
*The sets must be gathered stacked in SETS: first go all elements (rows) of one set,
*below it all elements of a second set, and so on. There may be any number of sets and the number
*of elements in them may be different.
*LAST - vector (row or column) lengthed as the number of sets in SETS and in correspondence to them,
*containing indices of last rows in each set. For example, {3,6,11} means: the are three sets in SETS,
*with elemens occupying rows 1:3, 4:6, 7:11. 11 (last element of the last set) must equal the number
*of rows in SETS.
*MAXK - maximal number of elements in a combination. That must be integer number (not name or expression)
*from 2 to ns, where ns is the number of sets, length of vector LAST. Optionally it can be taken in
*quotes or apostrophes.
*PREFIX - prefix (optionally in quotes or apostrophes) in the names of the output matrices, the sets of
*combinations.
*In the returned matrices each row is a combination of k elements from different sets, no more than
*one element from a set.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Tip: if you want to set MAXK=ns but number ns is not known to you in advance, specify MAXK as a number
*knowingly exceeding possible ns. Matrices above PREFIXns won't be able to create and no error will occur.

define !KO_dsallcomb(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !maxk= !unquote(!4) !let !pref= !unquote(!5)
comp @v= !2.
comp @last= !3.
comp @n= nrow(@v).
do if nrow(@last)>ncol(@last).
-comp @last= t(@last).
end if.
comp @plast= {0,@last}.
comp @explast= 0.
loop @i= 1 to ncol(@last).
-comp @val= @last(@i).
-comp @explast= {@explast,make(1,@val-@plast(@i),@val)}.
end loop.
comp @explast= @explast(2:ncol(@explast)).
comp @next= @explast+1.
comp @vec= @v.
!do !k= 2 !to !maxk
-do if !k<=ncol(@last).
- comp @nextnew= 0.
- comp @vecnew= make(1,ncol(@vec)+ncol(@v),0).
- loop @i= 1 to nrow(@vec).
-  do if @next(@i)<=@n.
-   comp @ind= @next(@i):@n.
-   comp @vecnew= {@vecnew;{@vec(@i,:)(make(1,ncol(@ind),1),:),@v(@ind,:)}}.
-   comp @nextnew= {@nextnew,@explast(@ind)+1}.
-  end if.
- end loop.
- comp @next= @nextnew(2:ncol(@nextnew)).
- comp @vec= @vecnew(2:nrow(@vecnew),:).
- comp !conc(!pref,!k)= @vec.
-end if.
!doend
release @v,@last,@n,@next,@vec,@plast,@explast,@val,@nextnew,@vecnew,@i,@ind.
!enddefine.


******** HORIZONTAL ARITHMETIC OPERATIONS IN COMBINATIONS OF VARIABLES ********.
/*!KO_comboper(vars%last%maxk%return%oper%weights%alpha%prefix%prefix_res)*/*.
*Version 1.
*Columns of matrix VARS are quantitative variables (minimum 2 variables).
*The function considers all combinations of variables of K variables in a combination; K runs over
*values from 2 to MAXK. In each variable combination, quantity OPER is computed horizontally,
*i.e., within each case. Which is the result.
*Argument OPER - arithmetic action that needs to be done in each combination of variables;
*capitalized keyword (optionally in quotes/apostrophes):
*"SUM" - sum, operation of adding variables: V(1)+V(2)+...+V(K)
*"DIF" - difference, operation of subtracting variables V(1)-V(2)-...-V(K)
*"PROD" - product, operation of multiplying variables V(1)*V(2)*...*V(K)
*"QUOT" - quotient, operation of dividing variables V(1)/V(2)/.../V(K) (needs nonzero data)
*"WSUM" - weighted sum: w(1)*V(1)+w(2)*V(2)+...+w(K)*V(K) (weights w - argument WEIGHTS)
*"WDIF" - weighted difference: w(1)*V(1)-w(2)*V(2)-...-w(K)*V(K) (weights w - argument WEIGHTS)
*"EWMA" - terminal (K-th) value of exponentially weighted moving average: S(K);
* S(K)=(1-a)*S(K-1)+a*V(K), where S(K-1)=(1-a)*S(K-2)+a*V(K-1), where S(K-2)=...; and S(1)=V(1)
* (exponential weight a - argument ALPHA).
*Argument WEIGHTS - weights for WSUM and WDIF operations. Either row vector lengthed as the number
*of columns in VARS, or matrix sized as VARS; the second instance means that for different cases
*(VARS rows) weights for variables are set different. Other actions - not WSUM or WDIF - don't use
*the argument: specify any value, e.g. 0.
*Argument ALPHA - exponential weight for EWMA operation, value between 0 and 1. It is either scalar
*or column vector lengthed as the number of rows in VARS; the second instance means that for
*different cases alpha weight is set different. Other actions - not EWMA - don't use the argument:
*specify any value, e.g. 0.
*Arguments LAST, MAXK, as well as PREFIX are the same as in function /*!KO_dsallcomb*/, on which this
*function is built.
*Argument LAST - parameter to break the variables in sets. Only variables belonging to different
*sets will be combining together. It is a vector (row or column) lengthed as the number of sets
*and in corresponding to them, containing indices (ordinal numbers) of the last variables in the sets.
*For example, {3,6,11} means: there are three sets in VARS, the variable groups consisting of columns
*1:3, 4:6, 7:11. 11 (the last value in LAST) must be equal to the number of columns in VARS.
*If you don't have groups of variables, i.e., each variable is its own set, specify LAST
*as 1:ncol(VARS), i.e. the vector of numbers 1,2,...,number_of_variables.
*Argument MAXK - maximal needed to you value of K, the number of variables in a combination. That must
*be integer number (not name or expression) from 2 to ns, where ns is the number of sets, length
*of vector LAST.
*Argument RETURN - for which values of K you want to obtain results. If for all K from 2 to
*min(MAXK,ns), i.e. all results, specify RETURN as keyword ALL (capitalized).
*Otherwise, request a list of K you need, spaced integers, for example: 2 3 5. Then the function
*will compute results for those K only.
*Results: PREFIX, PREFIX_RES. Specify in place of these arguments 2 prefixes for the names of
*the matrices to return, the names consisting of these prefixes and number K as the ending.
*For each K, a pair of arrays will be output:
*- Indices of variables forming combinations - matrix with prefix PREFIX.
*  Rows in PREFIX are various combinations of VARS variables, elements in a row are indices of
*  variables in the combination.
*- Results of operation OPER - matrix with prefix PREFIX_RES. Rows in it correspond to the data cases
*  (rows of VARS), and columns correspond to the variable combinations (rows of PREFIX).
*Specifications of arguments MAXK, RETURN, PREFIX, PREFIX_RES can be optionally surrounded with
*quotes/apostrophes.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Tip: if you want to set MAXK=ns but number ns is not known to you in advance, specify MAXK as a number
*knowingly exceeding possible ns. Matrices above PREFIXns won't be able to create and no error will occur.

define !KO_comboper(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                    /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !maxk= !unquote(!4) !let !return= !unquote(!5) !let !oper= !unquote(!6) !let !comb= !unquote(!9) !let !a= !unquote(!10)
comp @vars= t(!2).
comp @last= !3.
comp @w= !7.
comp @ew= !8.
comp @v= t(1:nrow(@vars)).
do if nrow(@last)>ncol(@last).
-comp @last= t(@last).
end if.
comp @plast= {0,@last}.
comp @explast= 0.
loop @i= 1 to ncol(@last).
-comp @val= @last(@i).
-comp @explast= {@explast,make(1,@val-@plast(@i),@val)}.
end loop.
comp @explast= @explast(2:ncol(@explast)).
comp @next= @explast+1.
comp @vec= @v.
!do !k= 2 !to !maxk
-do if !k<=ncol(@last).
- comp @nextnew= 0.
- comp @vecnew= make(1,ncol(@vec)+1,0).
- loop @i= 1 to nrow(@vec).
-  do if @next(@i)<=nrow(@v).
-   comp @ind= @next(@i):nrow(@v).
-   comp @vecnew= {@vecnew;{@vec(@i,:)(make(1,ncol(@ind),1),:),@v(@ind)}}.
-   comp @nextnew= {@nextnew,@explast(@ind)+1}.
-  end if.
- end loop.
- comp @next= @nextnew(2:ncol(@nextnew)).
- comp @vec= @vecnew(2:nrow(@vecnew),:).
- comp !conc(!comb,!k)= @vec.
-end if.
!doend
release @v,@next,@vec,@plast,@explast,@val,@nextnew,@vecnew,@ind.
!if (!return='ALL') !then !let !ret='ALL' !else !let !ret= !conc(' ',!return,' ') !ifend
!if (!oper='SUM' !or !oper='WSUM') !then !let !op= '+'
!else !if (!oper='DIF' !or !oper='WDIF') !then !let !op= '-'
!else !if (!oper='PROD') !then !let !op= '&*'
!else !if (!oper='QUOT') !then !let !op= '/'
!ifend !ifend !ifend !ifend
!if (!oper='WSUM' !or !oper='WDIF') !then !let !wght= 1
-do if nrow(@w)=1.
- comp @w= t(@w)*make(1,ncol(@vars),1).
-else.
- comp @w= t(@w).
-end if.
!else !if (!oper='EWMA') !then !let !wght= 2
-do if nrow(@ew)*ncol(@ew)=1.
- comp @iscol= 0.
- comp @ewk= @ew.
-else.
- comp @iscol= 1.
-end if.
!ifend !ifend
!do !k= 2 !to !maxk !let !kk= !conc(!comb,!k) !let !intk= !conc(!a,!k)
-do if !k<=ncol(@last).
  !if (!ret='ALL' !or !index(!ret,!conc(' ',!k,' '))>0) !then
-  comp !intk= @vars(!kk(:,1),:).
   !if (!wght=1) !then
-   comp !intk= @w(!kk(:,1),:)&*!intk.
   !else !if (!wght=2) !then
-   do if @iscol.
-    comp @ewk= @ew(make(1,nrow(!kk),1),:).
-   end if.
   !ifend !ifend
-  loop @i= 2 to !k.
    !if (!wght=1) !then
-    comp !intk= !intk !op @w(!kk(:,@i),:)&*@vars(!kk(:,@i),:).
    !else !if (!wght=2) !then
-    comp !intk= (1-@ewk)&*!intk+@ewk&*@vars(!kk(:,@i),:).
-   !else
-    comp !intk= !intk !op @vars(!kk(:,@i),:).
    !ifend !ifend
-  end loop.
-  comp !intk= t(!intk).
  !else
-  release !kk.
  !ifend
-end if.
!doend
release @vars,@last,@i,@w,@ew !if (!wght=2) !then ,@ewk !ifend.
!enddefine.


******** QUANTITIES ON THE BASIS OF HORIZONTAL SUMS OR MEANS IN COMBINATIONS OF VARIABLES ********.
*/*!KO_turflike(vars%last%maxk%return%tosum%tocnt%cond%prefix%prefix_sum%prefix_cnt)*/*.
*Version 1.
*Columns of matrix VARS are quantitative variables (minimum 2 variables).
*The function considers all combinations of variables of K variables in a combination; K runs over
*values from 2 to MAXK. In each variable combination, statistics hs (see below) are computed
*horizontally, i.e., within each case. Then the function (1) sums hs across cases, returning
*result PREFIX_SUM; (2) counts cases where hs satisfies condition COND, returning result PREFIX_CNT.
*So, for each combination of variables, columns from VARS, you'll get value "column_sum(hs)"
*and "column_count(hs COND = True)".
*Arguments TOSUM and TOCNT. Here you need to specify a statistic hs, in a form of capitalized keyword.
*May request the same or different statistics in these two arguments. The choice is among the following
*statistics, to compute within each case (row of VARS):
*"SUM" - sum of data values
*"MEAN" - mean of data values
*"SUMVLD" - sum of valid data values
*"MEANVLD" - mean of valid data values
*"NVLD" - count of valid data values
*In the last three statistics, value: -999 is treated as the nonvalid value.
*If there is no valid data in a case, MEANVLD of the case evaluates as 0.
*Argument COND - the condition for counting number of cases in which the TOCNT-statistic satisfies it.
*Consists of one operator and one operand. Operators can be: GT (or >), LT (or <), EQ (or =),
*GE, LE, NE. (>=, <=, <> are not allowed to use in this function.) The operand can be a
*number or some matrix name or an expression; the negative sign may precede the operand.
*Examples of specifying COND: GT 0, EQ -4.12, LE v3, LT x(:,1).
*Operand with a name. The name can be of a scalar (number), a column vector of length =nrow(VARS),
*a row vector of length >=min(MAXK,ns) [see below of ns], or a matrix sized [nrow(VARS),>=min(MAXK,ns)].
*Column vector lets you give different values of the operand for different data cases (rows of VARS).
*Row vector lets you give different values of the operand for different K (number of variables in
*combination); K-th element in the row vector will correspond to the given K. Operand in the form of a
*matrix (of the size mentioned above) lets you give different values of the operand both for different
*cases and for different K.
*Arguments LAST, MAXK, as well as PREFIX are the same as in function /*!KO_dsallcomb*/, on which this
*function is built.
*Argument LAST - parameter to break the variables in sets. Only variables belonging to different
*sets will be combining together. It is a vector (row or column) lengthed as the number of sets
*and in corresponding to them, containing indices (ordinal numbers) of the last variables in the sets.
*For example, {3,6,11} means: there are three sets in VARS, the variable groups consisting of columns
*1:3, 4:6, 7:11. 11 (the last value in LAST) must be equal to the number of columns in VARS.
*If you don't have groups of variables, i.e., each variable is its own set, specify LAST
*as 1:ncol(VARS), i.e. the vector of numbers 1,2,...,number_of_variables.
*Argument MAXK - maximal needed to you value of K, the number of variables in a combination. That must
*be integer number (not name or expression) from 2 to ns, where ns is the number of sets, length
*of vector LAST.
*Argument RETURN - for which values of K you want to obtain results. If for all K from 2 to
*min(MAXK,ns), i.e. all results, specify RETURN as keyword ALL (capitalized).
*Otherwise, request a list of K you need, spaced integers, for example: 2 3 5. Then the function
*will compute results for those K only.
*Results: PREFIX, PREFIX_SUM, PREFIX_CNT. Specify in place of these arguments 3 prefixes for
*the names of the matrices to return, the names consisting of these prefixes and number K as the ending.
*For each K, a series of 3 arrays will be output:
*- Indices of variables forming combinations - matrix with prefix PREFIX.
*  Rows in PREFIX are various combinations of VARS variables, elements in a row are indices of
*  variables in the combination.
*- Results of summing TOSUM - column vector with prefix PREFIX_SUM. Rows in it correspond to the rows
*  in PREFIX.
*- Results of counting TOCNT - column vector with prefix PREFIX_CNT. Rows in it correspond to the rows
*  in PREFIX.
*If you don't want to have result PREFIX_SUM or PREFIX_CNT, put digit 0 in place of that agrument.
*Specifications of arguments TOSUM, TOCNT, COND, MAXK, RETURN, PREFIX, PREFIX_SUM, PREFIX_CNT can be
*optionally surrounded with quotes/apostrophes.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*Tip: if you want to set MAXK=ns but number ns is not known to you in advance, specify MAXK as a number
*knowingly exceeding possible ns. Matrices above PREFIXns won't be able to create and no error will occur.
*Particular case: if VARS are binary variables (0 vs 1), TOSUM=TOCNT="SUM", and condition COND="GT 0",
*then the function in fact yields quantities that are computed in the so-called TURF analysis.
*PREFIX_SUM will give counts of "Frequency", and PREFIX_CNT will give counts "Reach".

define !KO_turflike(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                    /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !maxk= !unquote(!4) !let !return= !unquote(!5) !let !tosum= !unquote(!6) !let !tocnt= !unquote(!7) !let !cond= !unquote(!8)
!let !comb= !unquote(!9) !let !a= !unquote(!10) !let !b= !unquote(!11)
comp @vars= t(!2).
comp @last= !3.
comp @v= t(1:nrow(@vars)).
do if nrow(@last)>ncol(@last).
-comp @last= t(@last).
end if.
comp @plast= {0,@last}.
comp @explast= 0.
loop @i= 1 to ncol(@last).
-comp @val= @last(@i).
-comp @explast= {@explast,make(1,@val-@plast(@i),@val)}.
end loop.
comp @explast= @explast(2:ncol(@explast)).
comp @next= @explast+1.
comp @vec= @v.
!do !k= 2 !to !maxk
-do if !k<=ncol(@last).
- comp @nextnew= 0.
- comp @vecnew= make(1,ncol(@vec)+1,0).
- loop @i= 1 to nrow(@vec).
-  do if @next(@i)<=nrow(@v).
-   comp @ind= @next(@i):nrow(@v).
-   comp @vecnew= {@vecnew;{@vec(@i,:)(make(1,ncol(@ind),1),:),@v(@ind)}}.
-   comp @nextnew= {@nextnew,@explast(@ind)+1}.
-  end if.
- end loop.
- comp @next= @nextnew(2:ncol(@nextnew)).
- comp @vec= @vecnew(2:nrow(@vecnew),:).
- comp !conc(!comb,!k)= @vec.
-end if.
!doend
release @v,@next,@vec,@plast,@explast,@val,@nextnew,@vecnew,@ind.
!if (!return='ALL') !then !let !ret='ALL' !else !let !ret= !conc(' ',!return,' ') !ifend
!if (!b<>0) !then !let !op= !head(!cond)
-comp @x= t(!tail(!cond)).
!ifend
!let !sum= 0 !let !sumvld= 0 !let !nvld= 0
!if (!tosum='SUM' !or !tosum='MEAN' !or !tocnt='SUM' !or !tocnt='MEAN') !then !let !sum= 1 !ifend
!if (!tosum='SUMVLD' !or !tosum='MEANVLD' !or !tocnt='SUMVLD' !or !tocnt='MEANVLD') !then !let !sumvld= 1 !ifend
!if (!tosum='NVLD' !or !tosum='MEANVLD' !or !tocnt='NVLD' !or !tocnt='MEANVLD') !then !let !nvld= 1 !ifend
!if (!nvld !or !sumvld) !then
-comp @flagvld= @vars<>-999.
 !if (!sumvld) !then
- comp @varsvld= @vars&*@flagvld.
 !ifend
!ifend
!do !k= 2 !to !maxk !let !kk= !conc(!comb,!k)
-do if !k<=ncol(@last).
  !if (!ret='ALL' !or !index(!ret,!conc(' ',!k,' '))>0) !then
   !if (!sum) !then
-   comp @wsk= @vars(!kk(:,1),:).
   !ifend
   !if (!sumvld) !then
-   comp @wsvldk= @varsvld(!kk(:,1),:).
   !ifend
   !if (!nvld) !then
-   comp @wnvldk= @flagvld(!kk(:,1),:).
   !ifend
-  loop @i= 2 to !k.
    !if (!sum) !then
-    comp @wsk= @wsk+@vars(!kk(:,@i),:).
    !ifend
    !if (!sumvld) !then
-    comp @wsvldk= @wsvldk+@varsvld(!kk(:,@i),:).
    !ifend
    !if (!nvld) !then
-    comp @wnvldk= @wnvldk+@flagvld(!kk(:,@i),:).
    !ifend
-  end loop.
   !if (!b<>0) !then
-   do if nrow(@x)>1 and ncol(@x)>1.
-    comp @x_= @x(!k,:)(make(1,nrow(!kk),1),:).
-   else if ncol(@x)>1.
-    comp @x_= @x(make(1,nrow(!kk),1),:).
-   else if nrow(@x)>1.
-    comp @x_= @x(!k).
-   else.
-    comp @x_= @x.
-   end if.
   !ifend
   !if (!a<>0) !then
    !if (!tosum='SUM') !then
-    comp !conc(!a,!k)= rsum(@wsk).
    !else !if (!tosum='MEAN') !then
-    comp !conc(!a,!k)= rsum(@wsk/!k).
    !else !if (!tosum='SUMVLD') !then
-    comp !conc(!a,!k)= rsum(@wsvldk).
    !else !if (!tosum='MEANVLD') !then
-    comp !conc(!a,!k)= rsum(@wsvldk/(@wnvldk+(@wnvldk=0))).
    !else
-    comp !conc(!a,!k)= rsum(@wnvldk).
    !ifend !ifend !ifend !ifend
   !ifend
   !if (!b<>0) !then
    !if (!tosum='SUM') !then
-    comp !conc(!b,!k)= rsum(@wsk !op @x_).
    !else !if (!tosum='MEAN') !then
-    comp !conc(!b,!k)= rsum(@wsk/!k !op @x_).
    !else !if (!tosum='SUMVLD') !then
-    comp !conc(!b,!k)= rsum(@wsvldk !op @x_).
    !else !if (!tosum='MEANVLD') !then
-    comp !conc(!b,!k)= rsum(@wsvldk/(@wnvldk+(@wnvldk=0)) !op @x_).
    !else
-    comp !conc(!b,!k)= rsum(@wnvldk !op @x_).
    !ifend !ifend !ifend !ifend
   !ifend
  !else
-  release !kk.
  !ifend
-end if.
!doend
release @vars,@last,@i !if (!nvld !or !sumvld) !then ,@flagvld !ifend !if (!sumvld) !then ,@varsvld !ifend
 !if (!sum) !then ,@wsk !ifend !if (!sumvld) !then ,@wsvldk !ifend !if (!nvld) !then ,@wnvldk !ifend
 !if (!b<>0) !then ,@x,@x_ !ifend.
!enddefine.


******** MATCHING BY HUNGARIAN ALGORITHM ********.
*/*!KO_hungar(cost%name)*/*.
*Version 1.
*Hungarian or Kuhn-Munkres algorithm matches elements one-to-one between two sets optimally so
*that the summed cost of pairs (or summed dissimilarity in pairs) become minimal.
*In graph terminology, it constructs optimal matching from a bipartite weighted or unweighted graph.
*(To note that in case of unweighted graph you can use also Hopcroft-Karp /*!KO_hopckarp*/ algorithm
*to solve the same problem.)
*The function accepts as input a n x m matrix of costs COST with nonnegative elements. The number of
*rows must be not less than the number of columns. The rows are one set of elements and the columns are the
*other set of elements. Values in matrix are taken as "cost" of a pair row-column or as "dissimilarity"
*between a row and a column. The algorithm finds for each column one matching row.
*Result NAME is 3-column matrix with number of rows as the number of columns in COST; the rows there are
*the matched pairs. The 2nd column there is the column number of the input matrix COST, the 1st column is
*the row number in COST, and the 3rd column is the value in COST on their intersection. The algorithm
*minimizes the sum in the 3rd column.
*If your input matrix represent not costs (dissimilarities) but gains (similarities), transform it to
*costs (dissimilarities) by sign-negating and then adding a constant to matrix to make all the values
*nonnegative.
*The algorithm speed is higher when the number of distinct values in COST is not great, so there is
*the reason to consider beforehand a discretization of the "cost" values.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_hungar(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @c= !2.
comp @c_= @c.
do if nrow(@c)>ncol(@c).
-comp @c= {@c,make(nrow(@c),nrow(@c)-ncol(@c),0)}.
end if.
comp @n= nrow(@c).
comp @rzero= make(1,@n,0).
comp @czero= make(@n,1,0).
comp @rnums= t(1:@n).
comp @cnums= 1:@n.
comp @c= @c-rmin(@c)*make(1,@n,1).
comp @m= (not @c)&*@cnums(make(1,@n,1),:).
loop @i= 1 to @n.
-comp @j= rmax(@m(@i,:)).
-do if @j.
- comp @m(@i,:)= @rzero.
- comp @m(:,@j)= @czero.
- comp @m(@i,@j)= 1.
-end if.
end loop.
loop.
-comp @ucr= make(@n,1,1).
-comp @ucc= not csum(@m).
-do if not any(@ucc).
- break.
-else.
- loop.
-  comp @i= 0.
-  loop @j= 1 to @n.
-   do if @ucc(@j).
-    comp @i= cmax((not @c(:,@j))&*@ucr&*@rnums).
-    do if @i.
-     break.
-    end if.
-   end if.
-  end loop.
-  do if @i.
-   comp @m(@i,@j)= -2.
-   do if any(@m(@i,:)=1).
-    comp @ucr(@i)= 0.
-    comp @ucc(rmax(@m(@i,:)&*@cnums))= 1.
-   else.
-    comp @m(@i,@j)= -3.
-    loop if any(@m(:,@j)=1).
-     comp @i2= cmax(@m(:,@j)&*@rnums).
-     comp @m(@i2,@j)= 0.
-     comp @j2= rmax((@m(@i2,:)=-2)&*@cnums).
-     comp @m(@i2,@j2)= -3.
-     comp @j= @j2.
-    end loop.
-    comp @m= @m&*(@m<>-3)+(@m=-3).
-    comp @m= @m&*(@m<>-2).
-    break.
-   end if.
-  else.
    !do !w !in ("'@ucr @rind' '@ucc @cind'")
     !let !w0= !unquote(!w) !let !w1= !head(!w0) !let !w2= !tail(!w0)
-    comp @nnz= msum(!w1).
-    do if @nnz=@n.
-     comp !w2= @cnums.
-    else if @nnz/@n>.8.
-     comp !w2= make(1,@nnz+1,0).
-     comp @cnt= 1.
-     loop @#= 1 to @n.
-      comp !w2(@cnt)= @#.
-      comp @cnt= @cnt+!w1(@#).
-     end loop.
-     comp !w2= !w2(1:@nnz).
-    else.
-     comp !w2= make(1,@nnz,0).
-     comp @cnt= 0.
-     loop @#= 1 to @n.
-      do if !w1(@#).
-       comp @cnt= @cnt+1.
-       comp !w2(@cnt)= @#.
-      end if.
-     end loop.
-    end if.
    !doend
-   comp @min= mmin(@c(@rind,@cind)).
-   comp @c= @c+(not @ucr)*make(1,@n,@min)-make(@n,1,@min)*@ucc.
-   release @nnz,@rind,@cind,@cnt,@#.
-  end if.
- end loop.
-end if.
end loop.
comp !3= make(ncol(@c_),3,0).
loop @j= 1 to ncol(@c_).
-comp @i= cmax(@m(:,@j)&*@rnums).
-comp !3(@j,:)= {@i,@j,@c_(@i,@j)}.
end loop.
release @c,@c_,@m,@n,@rzero,@czero,@rnums,@cnums,@ucr,@ucc,@i,@j.
!enddefine.


******** MATCHING BY HOPCROFT-KARP ALGORITHM ********.
*/*!KO_hopckarp(adj%name)*/*.
*Version 1.
*Hopcroft-Karp algorithm matches elements one-to-one between two sets having links between them
*so as to maximize the number of pairs, i.e., the number of links accounted for.
*In graph terminology, it constructs maximal matching from a bipartite unweighted graph.
*(To note that you can also use Hungarian /*!KO_hungar*/ algorithm to solve the same problem.)
*Input matrix ADJ is an "adjacency list" sized n x m, where n and m are the numbers of elements
*in the two sets. Each row in ADJ is an element from the "n" set, and it contains the list of element
*numbers from the "m" set, going left-to-right, rammed. For example, let there be sets "n"
*(1,2,...,n) and "m" (1,2,...,m=7), where, particularly, the 2nd element from "n" has links with the
*elements 1,4, and 5 from the "m" set. Then, the 2nd row of ADJ should look like: {1,4,5,0,0,0,0}.
*The order of the numbers plays no role: {4,5,1,0,0,0,0}, for instance, is also correct. The number of
*columns in ADJ must be m, therefore empty trails are padded with zeros up to column 7. Adjacency list
*corresponds to a binary matrix sized n x m, in which units flag the presence of a link; and it is
*another mode to represent the matrix. If your input data are the binary matrix, you can turn it into
*the adjacency list by function /*!KO_ram*/.
*Empty (zero filled) rows in ADJ are allowed - they correspond to isolated, having no links,
*elements of "n", - but you might want first to remove them, in the speed performance view.
*ADJ must bear at least one nonzero element.
*If n and m differ strongly, it is advantageous, from the speed performance point of view, to have ADJ
*wide (n<m) rather than high (n>m).
*Result NAME is 2-column matrix giving a maximal matching; the 1st column is the elements from
*the "n" set, 2nd column is the elements from the "m" set; each row is a done pairing.
*Number of rows in NAME is, therefore, the size of the matching.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_hopckarp(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @adj= !2.
comp @n= nrow(@adj).
comp @m= ncol(@adj).
comp @adj= @adj+@n*(@adj<>0).
comp @adj2= make(@m,@n,0).
comp @adj2= {make(@m,1,@n+@m+2),@adj2}.
do if ncol(@adj2)<@m.
-comp @adj= {@adj;{@adj2,make(@m,@m-ncol(@adj2),0)}}.
else if ncol(@adj2)>@m.
-comp @adj= {{@adj,make(@n,ncol(@adj2)-@m,0)};@adj2}.
-else.
-comp @adj= {@adj;@adj2}.
end if.
release @adj2.
comp @adj= {@adj;make(2,ncol(@adj),0)}.
comp @src= nrow(@adj)-1.
comp @snk= nrow(@adj).
comp @j= 0.
loop @i= 1 to @n.
-do if @adj(@i,1).
- comp @j= @j+1.
- comp @adj(@src,@j)= @i.
-end if.
end loop.
comp @nadj= rsum(@adj<>0).
comp @nums= 1:ncol(@adj).
comp !3= {-1,-1}.
comp @ginm= make(@n,1,0).
loop.
-comp @g= @adj.
-comp @ng= @nadj.
-do if nrow(!3)>1.
- comp @jj= {0;(@g(!3(2:nrow(!3),1),:)=(!3(2:nrow(!3),2)*make(1,ncol(@adj),1)))*t(@nums)}.
- comp @g(!3(2:nrow(!3),2),1)= !3(2:nrow(!3),1).
- loop @i= 2 to nrow(!3).
-  comp @grl= !3(@i,1).
-  comp @g(@grl,@jj(@i))= @g(@grl,@ng(@grl)).
-  comp @ng(@grl)= @ng(@grl)-1.
-  comp @g(@src,rmax((@g(@src,:)=@grl)&*@nums))= @g(@src,@ng(@src)).
-  comp @g(@src,@ng(@src))= 0.
-  comp @ng(@src)= @ng(@src)-1.
- end loop.
-end if.
-comp @b= make(1,@snk,1).
-comp @s= {@src,0}.
-comp @done= 1.
-loop if ncol(@s)>1.
- loop if @ng(@s(1)).
-  comp @top= @s(1).
-  comp @fst= @g(@top,@ng(@top)).
-  comp @ng(@top)= @ng(@top)-1.
-  do if @fst=@snk.
-   loop if ncol(@s)>2.
-    do if @ginm(@s(2)).
-     comp !3(@ginm(@s(2)),2)= @s(1).
-    else.
-     comp !3= {!3;{@s(2),@s(1)}}.
-     comp @ginm(@s(2))= nrow(!3).
-    end if.
-    comp @s= @s(3:ncol(@s)).
-   end loop.
-   comp @done= 0.
-  end if.
-  do if @b(@fst).
-   comp @s= {@fst,@s}.
-   comp @b(@fst)= 0.
-  end if.
- end loop.
- comp @s= @s(2:ncol(@s)).
-end loop.
end loop if @done.
comp !3= !3(2:nrow(!3),:).
comp !3(:,2)= !3(:,2)-@n.
release @adj,@n,@m,@src,@snk,@j,@nadj,@nums,@ginm,@done,@g,@ng,@jj,@i,@grl,@b,@s,@top,@fst.
!enddefine.


******** SIMPLE GREEDY MATCHING ********.
*/*!KO_greedm(mat%cg%name)*/*.
*Version 1.
*Simple greedy matching is a stepwise algorithm which pairs elements between two sets so that
*to minimize (or maximize) the sum of values within pairs. This is the same task that Hungarian
*matching (function /*!KO_hungar*/) accomplishes, but /*!KO_greedm*/ is faster and does not guarantee
*the global optimum in the sense of the final sum. The algorithm finds the least (or the largest - as
*requested) value in the matrix and records its row and column a pair (match), then cuts off the two
*from the matrix, in the remnant matrix it then again finds the least value, pairs the row and the
*column, deletes them, and so on.
*Input:
*MAT - matrix sized at least 2x2, with any numbers. If the number of columns much exceeds the number
*of rows it might have sense to transpose the matrix first, which could give win in speed.
*CG - digit (not name or expression) 0, 1, or 2 (the digit may optionally be put in quotes or
*apostrophes). If "0", the function will seek minimal values, i.e. will treat MAT as cost which
*overall must be minimized at matching. If "1", the function will seek maximal values, i.e. will
*treat MAT as gain which overall must be maximized at matching. If "2", then MAT is a sparse
*nonnegative gain matrix: with positive values as the gains and zero elements prohibiting to
*do the pairing; at least one value must be greater than 0.
*Result NAME: 3-column matrix where rows are matched pairs row-column chosen by the algorithm:
*the row number in MAT is written in the 1st NAME's column, the column number in MAT is written in
*the 2nd NAME's column, and the value on their intersection in MAT is in the 3rd NAME's column.
*The sum in this 3rd column is what the function attempts to minimize (maximize).
*In case CG="2", some last rows in NAME can be empty, because sparse MAT may not allow to pair
*all the rows and columns.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_greedm(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
!let !cg= !unquote(!3)
comp !4= make(rmin({nrow(@mat),ncol(@mat)}),3,0).
comp @rns= 1:nrow(@mat).
comp @cns= 1:ncol(@mat).
loop @i= 1 to nrow(!4).
-comp @val= !if (!cg=0) !then mmin !else mmax !ifend (@mat).
 !if (!cg=2) !then
- do if not @val.
-  break.
- end if.
 !ifend
-comp @flag= @mat=@val.
-comp @ri= mmax(@flag&*(t((1:nrow(@mat)))*make(1,ncol(@mat),1))).
-comp @ci= mmax(@flag(@ri,:)&*(1:ncol(@mat))).
-comp !4(@i,:)= {@rns(@ri),@cns(@ci),@val}.
-do if @i<nrow(!4).
- do if @ri=1.
-  comp @rind= 2:nrow(@mat).
- else if @ri=nrow(@mat).
-  comp @rind= 1:(@ri-1).
- else.
-  comp @rind= {1:(@ri-1),(@ri+1):nrow(@mat)}.
- end if.
- do if @ci=1.
-  comp @cind= 2:ncol(@mat).
- else if @ci=ncol(@mat).
-  comp @cind= 1:(@ci-1).
- else.
-  comp @cind= {1:(@ci-1),(@ci+1):ncol(@mat)}.
- end if.
- comp @mat= @mat(@rind,@cind).
- comp @rns= @rns(@rind).
- comp @cns= @cns(@cind).
-end if.
end loop.
release @mat,@rns,@cns,@i,@val,@flag,@ri,@ci,@rind,@cind.
!enddefine.


******** SIMPLE GREEDY MATCHING (DISCRETE DATA) ********.
*/*!KO_greedm2(mat%cg%name)*/*.
*Version 1.
*This function gives the same result as /*!KO_greedm*/, but it is faster in the situation when the
*number of distinct values in MAT is small relative the size of MAT. For example, when MAT is
*integers from a considerably limited range. And conversely, if there are many distinct values in MAT,
*such as, of instance, in case of continuous data, then /*!KO_greedm2*/ will be slower
*than /*!KO_greedm*/.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_greedm2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
!let !cg= !unquote(!3)
comp !4= make(rmin({nrow(@mat),ncol(@mat)}),3,0).
comp @rns= 1:nrow(@mat).
comp @cns= 1:ncol(@mat).
comp @cnt= 0.
comp @cind= 0.
loop.
-comp @val= !if (!cg=0) !then mmin !else mmax !ifend (@mat).
 !if (!cg=2) !then
- do if not @val.
-  break.
- end if.
 !ifend
-comp @rind= 0.
-comp @cnums= 1:ncol(@mat).
-loop @ri= nrow(@mat) to 1 by -1.
- comp @ci= rmax((@mat(@ri,:)=@val)&*@cnums).
- do if @ci.
-  comp @cnt= @cnt+1. 
-  comp !4(@cnt,:)= {@rns(@ri),@cns(@ci),@val}.
-  comp @cnums(@ci)= 0.
- else.
-  comp @rind= {@ri,@rind}.
- end if.
-end loop.
-do if @cnt=nrow(!4).
- break.
-end if.
-comp @rind= @rind(1:(ncol(@rind)-1)).
-comp @vec= @cnums<>0.
-comp @n= ncol(@vec).
-comp @nnz= msum(@vec).
-do if @nnz/@n>.8.
- comp @cind= make(1,@nnz+1,0).
- comp @c= 1.
- loop @i= 1 to @n.
-  comp @cind(@c)= @i.
-  comp @c= @c+@vec(@i).
- end loop.
- comp @cind= @cind(1:@nnz).
- release @c,@i.
-else.
- comp @cind= make(1,@nnz,0).
- comp @c= 0.
- loop @i= 1 to @n.
-  do if @vec(@i).
-   comp @c= @c+1.
-   comp @cind(@c)= @i.
-  end if.
- end loop.
- release @c,@i.
-end if.
-release @vec,@n,@nnz.
-comp @mat= @mat(@rind,@cind).
-comp @rns= @rns(@rind).
-comp @cns= @cns(@cind).
end loop.
release @mat,@rns,@cns,@cnt,@val,@cnums,@ri,@ci,@rind,@cind.
!enddefine.


******** FLOYD-WARSHALL ALGORITHM: SHORTEST PATHS / EASIEST PASSES ********.
*/*!KO_flowar(cost%task%path%name1%name2)*/.
*Version 1.
*Takes square matrix COST with nonnegative elements and zero diagonal;
*the matrix may be symmetric or asymmetric. Values of COST can be understood as weights in a graph
*or links in a network, and rows/columns of the matrix are the graph's vertices or net's nodes.
*TASK (keyword, capitalized, may quote or put in apostrophes) is the algorithm's goal:
*"SHWAY" - find, from each point to each point, the shortest way (Shortest path aka Shortest
*way mode of the algo). Values in COST are taken for distances between the points:
*COST(i,j) is the distance from i into j.
*From every point i to every point j, the shortest way will be found. Result NAME1 will
*contain the summative length of that shortest path from each point-row i to each point-column j.
*"EPASS" - find, from each point to each point, the path of easiest passes (Optimal routing
*aka Easiest passes mode of the algo). Values in COST are taken for difficulties to transit
*a pass from a point to a point: COST(i,j) is the difficulty from i into j.
*From every point i to every point j, the easiest way will be found:
*all the passes of which it will consist, are minimally difficult. Result NAME1 will contain
*the difficulty of the most difficult pass on that easiest way from each point-row i to each
*point-column j.
*The found (i.e., the optimal) paths, the trajectories, will be written "wrapped" in matrix NAME2.
*Element (i,j) in NAME2 is the number of the point which follows the point i in every found path
*that leads to point j as the endpoint of the path. See EXAMPLE, showing how to extract the whole
*trajectory of the path from some starting point U to some ending point V=j.
*Whether to register a trajectory or not - depends on argument PATH. Argument PATH is a digit 1 or 0
*(not name or expression, you may take the digit in quotes or apostrophes). If 0, trajectories won't
*be written down and matrix NAME2 will come out to be zero scalar.
*Missing values in COST and impossible path. If in the input matrix there should be no links between
*some points - for example, the pass from point i to point j does not exist, then set the
*cell (i,j) to number X not less than the sum of all existing values in the matrix, i.e. X>=msum(COST).
*So then, if in some cell (U,V) of NAME1 you see number X, that will signal that a path from starting
*point U to ending point V is impossible.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_flowar(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                 /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @n= ncol(!5).
!let !task= !unquote(!3)
!if (!task='EPASS') !then !let !upd= !conc('rmax({@dik,',!5,'(@k,@j)})')
!else !let !upd= !conc('@dik+',!5,'(@k,@j)')
!ifend
!let !path= !unquote(!4)
!if (!path) !then
-comp !6= (1:@n)(make(1,@n,1),:).
!else
-comp !6= 0.
!ifend
loop @k= 1 to @n.
-loop @i= 1 to @n.
- do if @i<>@k.
-  comp @dik= !5(@i,@k).
-  loop @j= 1 to @n.
-   do if !5(@i,@j)>!upd.
-    comp !5(@i,@j)= !upd.
     !if (!path) !then
-     comp !6(@i,@j)= !6(@i,@k).
     !ifend
-   end if.
-  end loop.
- end if.
-end loop.
end loop.
release @n,@k,@i,@j,@dik.
!enddefine.


******** FLOYD-WARSHALL ALGORITHM (SYMMETRIC MATRIX) ********.
*/*!KO_sflowar(cost%task%name)*/.
*Version 1.
*This function is identical to /*!KO_flowar*/, but it is only for symmetric input matrix COST.
*This function is somewhat faster than /*!KO_flowar*/.
*Argument PATH and result NAME2 are not provided.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.

define !KO_sflowar(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @n= ncol(!4).
!let !task= !unquote(!3)
!if (!task='EPASS') !then !let !upd= !conc('rmax({@dik,',!4,'(@k,@j)})')
!else !let !upd= !conc('@dik+',!4,'(@k,@j)')
!ifend
loop @k= 1 to @n.
-loop @i= 1 to @n.
- do if @i<>@k.
-  comp @dik= !4(@i,@k).
-  loop @j= @i+1 to @n.
-   do if !4(@i,@j)>!upd.
-    comp !4(@i,@j)= !upd.
-   end if.
-  end loop.
-  comp !4(:,@i)= t(!4(@i,:)).
- end if.
-end loop.
end loop.
release @n,@k,@i,@j,@dik.
!enddefine.


******** DIJKSTRA ALGORITHM: SHORTEST PATH / EASIEST PASS ********.
*/*!KO_dijkstra(cost%start%stop%task%path%name1%name2)*/.
*Version 1.
*Takes square matrix COST with nonnegative elements; the matrix may be symmetric or asymmetric.
*Values of COST can be understood as weights in a graph or links in a network, and rows/columns of
*the matrix are the graph's vertices or net's nodes. Diagonal elements in COST are ignored by the
*function (they are considered as 0).
*START - scalar; the point (vertex) number from where to find paths to the other points.
*TASK (keyword, capitalized, may quote or put in apostrophes) is the algorithm's goal:
*"SHWAY" - find, from point START to each point, the shortest way (Shortest path aka Shortest
*way mode of the algo). Values in COST are taken for distances between the points:
*COST(i,j) is the distance from i into j.
*From point START to every point j, the shortest way will be found. Result NAME1 will
*contain the summative length of that shortest path from point-row START to each point-column j.
*"EPASS" - find, from point START to each point, the path of easiest passes (Optimal routing
*aka Easiest passes mode of the algo). Values in COST are taken for difficulties to transit
*a pass from a point to a point: COST(i,j) is the difficulty from i into j.
*From point START to every point j, the easiest way will be found:
*all the passes of which it will consist, are minimally difficult. Result NAME1 will contain
*the difficulty of the most difficult pass on that easiest way from point-row START to each
*point-column j.
*The found (i.e., the optimal) paths, the trajectories, will be written "wrapped" in vector NAME2.
*The j-th element's value in NAME2 is the number of the point which precedes the point j in every
*found path that leads to point j as the endpoint of the path. See EXAMPLE, showing how to extract
*the whole trajectory of the path from the point START to some ending point V=j.
*Whether to register a trajectory or not - depends on argument PATH. Argument PATH is a digit 1 or 0
*(not name or expression, you may take the digit in quotes or apostrophes). If 0, trajectories won't
*be written down and NAME2 will come out to be zero scalar.
*Missing values in COST and impossible path. If in the input matrix there should be no links between
*some points - for example, the pass from point i to point j does not exist, then set the
*cell (i,j) to value: -1. So then, if some V-th element of NAME1 has turned out as -1, that will
*signal that a path from START to ending point V is impossible.
*Argument STOP. If you need paths from START to all the points, set STOP on value 0.
*But if you are interested in the path from START to one specific point only, indicate its number
*here. Then the algorithm will cease working upon coming to that point; you may save time through
*this. Result NAME1 will then output as scalar, not vector. STOP must be different from START.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*If you need paths from every point to every point, use function /*!KO_flowar*/.

define !KO_dijkstra(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                   /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @u= !3.
comp @stop= !4.
comp @n= ncol(@mat).
!let !task= !unquote(!5)
!let !path= !unquote(!6)
comp !7= make(1,@n,1E300).
!if (!path) !then
-comp !8= make(1,@n,0).
!else
-comp !8= 0.
!ifend
comp @q= 1:@n.
comp !7(@u)= 0.
loop if any(@q).
-comp @u=  mmax((!7=mmin(!7+(not @q)*1E300))&*@q). 
-do if @u=@stop.
- break.
-end if.
-comp @q(@u)= 0.
 !if (!task='EPASS') !then
- comp @alt= cmax({@mat(@u,:);make(1,ncol(@mat),!7(@u))}).
 !else
- comp @alt= @mat(@u,:)+!7(@u).
 !ifend
-comp @flag= @alt<!7 and @mat(@u,:)<>-1 and @q.
-comp !7= !7&*(not @flag)+@alt&*@flag.
 !if (!path) !then
- comp !8= !8&*(not @flag)+@u*@flag.
 !ifend
end loop.
do if @stop.
-comp !7= !7(@stop).
end if.
comp @q= !7=1E300.
comp !7= !7&*(not @q)-@q.
release @mat,@u,@stop,@n,@q,@alt,@flag.
!enddefine.


******** PRIM'S MINIMUM SPANNING TREE ALGORITHM ********.
*/*!KO_prim(mat%cg%start%name)*/*.
*Version 1.
*In a unipartite weighted undirected graph specified by adjacency matrix MAT, finds
*the minimum (or maximum) spanning tree, starting from vertex START (scalar, the vertex number).
*MAT - square n x n symmetric matrix. Its diagonal elements are ignored by the function.
*Argument CG sets the interpretation of MAT. CG is the digit (not name or expression) 0, 1, or 2
*(may optionally take it in quotes or apostrophes).
*If CG=0, MAT is understood as a cost or distance matrix between the vertices of a complete
*weighted graph. Result NAME is the minimum spanning tree.
*If CG=1, MAT is understood as a gain or similarity matrix between the vertices of a complete
*weighted graph. Result NAME is the maximum spanning tree. In both instances values in MAT
*may have any sign.
*If CG=2, MAT is understood as a sparse nonnegative gain matrix:
*with positive values as gains and zero elements corresponding to absent edges (incomplete graph).
*Result NAME is then the maximum spanning tree from vertex start under the condition
*of incomplete graph. If the incomplete graph is not a connected graph (i.e., not a graph where
*there exists a path from every vertex to every vertex), then the spanning tree will cover
*only its part - the subgraph containing the start vertex, - and as a result, not all n-1
*rows of NAME will be filled (some last will remain empty). Thus Prim's algorithm
*is usable with CG=2 to diagnose connectedness in a incomplete graph.
*When a graph is complete or is incomplete but connected, the spanning tree will cover all the graph
*and the size (sum of weights) of the tree does not depend on the start vertex.
*Result NAME: has three columns and n-1 rows; the rows correspond to the edges written down to the
*spanning tree. The first two elements in a row are the vertices i and j, adjacent of the edge,
*and the third value is the edge weight, i.e. the value (i,j) in MAT. Prim's algorithm
*minimizes (if MAT is cost matrix) or maximizes (if MAT is gain matrix) the sum in the 3rd column
*of NAME.
*This function may require prior setting of limit for number of cycles to enough big value
*by command SET MXLOOPS.
*If your MAT is cost matrix and the graph is incomplete (some edges nonexistent), transform it
*to the positive gain matrix by negating, and then adding a positive constant, and the absent
*edges to mark with zeros; then use CG=2.

define !KO_prim(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
!let !cg= !unquote(!3)
comp @st= !4.
comp @n= ncol(@mat).
comp !5= make(@n-1,3,0).
comp @rns= @st.
do if @st=1.
-comp @cns= 2:@n.
else if @st=@n.
-comp @cns= 1:(@n-1).
else.
-comp @cns= {1:(@st-1),(@st+1):@n}.
end if.
comp @ri= 0.
comp @ci= 0.
loop @i= 1 to nrow(!5).
-comp @m= @mat(@rns,@cns).
-comp @val= !if (!cg=0) !then mmin !else mmax !ifend (@m).
 !if (!cg=2) !then
- do if not @val.
-  break.
- end if.
 !ifend
-comp @m= @m=@val.
-comp @ci= mmax(@m&*(1:ncol(@m))(make(1,nrow(@m),1),:)).
-comp @ri= mmax(@m(:,@ci)&*t(1:nrow(@m))).
-comp !5(@i,:)= {@rns(@ri),@cns(@ci),@val}.
-comp @rns= {@rns,@cns(@ci)}.
-do if @i<nrow(!5).
- do if @ci=1.
-  comp @cns= @cns(2:ncol(@cns)).
- else if @ci=ncol(@cns).
-  comp @cns= @cns(1:(ncol(@cns)-1)).
- else.
-  comp @cns= @cns({1:(@ci-1),(@ci+1):ncol(@cns)}).
- end if.
-end if.
end loop.
release @mat,@st,@n,@rns,@cns,@i,@m,@ci,@ri.
!enddefine.
